`timescale 1 ns/10 ps
module tb;


 reg 	[63:0]		ReadData1E;
 reg 	[63:0]		ReadData2E;
 reg 	[63:0]		ReadData3E;
 reg 	[63:0]		ans;
 reg 	[2:0]		FrmE;
 wire 	[63:0]		FmaResultM;
 wire 	[4:0]	 	FmaFlagsM;

	wire 		[12:0]		aligncntE;    	// status flags
	wire 		[105:0]		rE; 				// one result of partial product sum
	wire 		[105:0]		sE; 				// other result of partial products
	wire 		[163:0]		tE;				// wire of alignment shifter	
	wire 		[8:0]		normcntE; 		// shift count for normalizer
	wire 		[12:0]		aeE; 		// multiplier expoent
	wire 					bsE;				// sticky bit of addend
	wire 					killprodE; 		// ReadData3E >> product
	wire 					prodofE; 		// ReadData1E*ReadData2E out of range
	wire					xzeroE;
	wire					yzeroE;
	wire					zzeroE;
	wire					xdenormE;
	wire					ydenormE;
	wire					zdenormE;
	wire					xinfE;
	wire					yinfE;
	wire					zinfE;
	wire					xnanE;
	wire					ynanE;
	wire					znanE;
	wire					nanE;
	wire			[8:0]		sumshiftE;
	wire					sumshiftzeroE;
    wire prodinfE;

integer fp;
reg wnan;
reg xnan;
reg ynan;
reg znan;
reg ansnan;
reg		[105:0]		s;				//	partial product 2	
reg		[51:0] 		xnorm;
reg 		[51:0] 		ynorm;


localparam period = 20;  
fma1 UUT1(.*);
fma2 UUT2(.ReadData1M(ReadData1E), .ReadData2M(ReadData2E), .ReadData3M(ReadData3E), .FrmM(FrmE),
			 .aligncntM(aligncntE), .rM(rE), .sM(sE),
			.tM(tE),	.normcntM(normcntE), .aeM(aeE), .bsM(bsE),.killprodM(killprodE),
			.xzeroM(xzeroE),	.yzeroM(yzeroE),.zzeroM(zzeroE),.xdenormM(xdenormE),.ydenormM(ydenormE),
			.zdenormM(zdenormE),.xinfM(xinfE),.yinfM(yinfE),.zinfM(zinfE),.xnanM(xnanE),.ynanM(ynanE),.znanM(znanE),
			.nanM(nanE),.sumshiftM(sumshiftE),.sumshiftzeroM(sumshiftzeroE), .prodinfM(prodinfE), .*);


initial 
    begin
    fp = $fopen("/home/kparry/riscv-wally/wally-pipelined/src/fpu/FMA/tbgen/results.dat","w");
    ReadData1E = 64'hC1F0003F7FFFFFFF;
    ReadData2E = 64'hB40DC94EA8FD6C48;
    ReadData3E = 64'h82E1FC6B349D46FE;
    ans = 64'h360DC9C4DFED9AF4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"305\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC34EFFFFFFFFFF7F;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"611\n");
    end
    ReadData1E = 64'h37E0000001010000;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h41C000FFFFFFFFEF;
    ans = 64'h41C000FFFFFFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"917\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h402020000FFFFFFE;
    ReadData3E = 64'hC49CC543222D7A93;
    ans = 64'hC49CC543222D7A93;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1223\n");
    end
    ReadData1E = 64'hC3DFC000FFFFFFFE;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1529\n");
    end
    ReadData1E = 64'hBFCFB75018356166;
    ReadData2E = 64'hB7EFFFFFF7E00000;
    ReadData3E = 64'hC3DFFFFFFFF78000;
    ans = 64'hC3DFFFFFFFF78000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1835\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'hC16FFFFFEEFFFFFF;
    ans = 64'hC16FFFFFEEFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"2141\n");
    end
    ReadData1E = 64'h47FCF548236352E4;
    ReadData2E = 64'h3F900003FDFFFFFF;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h479CF54F5D16B2B7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"2447\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"2753\n");
    end
    ReadData1E = 64'h40DFFFFFFFFEF7FF;
    ReadData2E = 64'hC80000007EFFFFFF;
    ReadData3E = 64'h4008BA14EB49A7C1;
    ans = 64'hC8F000007EFF7BFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"3059\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hB810000000600000;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h3FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"3365\n");
    end
    ReadData1E = 64'h40600000001FFF7E;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'hC00FFFFFFFFFFDFE;
    ans = 64'hC00FFFFFFFFFFDFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"3671\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hDBFFFBFFF8000000;
    ReadData3E = 64'h4A6FFC00000000FF;
    ans = 64'h4A6FFC00000000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"3977\n");
    end
    ReadData1E = 64'h434FAE5B6ECDD96F;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"4283\n");
    end
    ReadData1E = 64'h43C5385F5ABC1E16;
    ReadData2E = 64'hC9ABFFFFFFFFFEFF;
    ReadData3E = 64'h3FC007FFFFFFBFFF;
    ans = 64'hCD8291536F6499A9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"4589\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h402FFFF000000800;
    ans = 64'h402FFFF000000800;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"4895\n");
    end
    ReadData1E = 64'h3FC6C487F45FB506;
    ReadData2E = 64'h43E25F209A1547C0;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h43BA24763E787145;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"5201\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"5507\n");
    end
    ReadData1E = 64'hB48FFFFFFFFF8FFF;
    ReadData2E = 64'h7FEFC00000000800;
    ReadData3E = 64'hC34FE0000007FFFE;
    ans = 64'hF48FBFFFFFFF98DF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"5813\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h41E0040000FFFFFF;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hBFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"6119\n");
    end
    ReadData1E = 64'h41D0040000400000;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'h3FCA69118D0A7A7B;
    ans = 64'h3FCA69128D4A7A7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"6425\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3F9FEFFFC0000000;
    ReadData3E = 64'h41DFE000007FFFFF;
    ans = 64'h41DFE000007FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"6731\n");
    end
    ReadData1E = 64'h94BFEFFFFFFFFFFF;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"7037\n");
    end
    ReadData1E = 64'h388802CB750F8A0C;
    ReadData2E = 64'h7FDBCDA5B681F4BB;
    ReadData3E = 64'h4014727098BAA500;
    ans = 64'h7874DCA9E98583AE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"7343\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h43DFDFFFFFFFFE00;
    ans = 64'h43DFDFFFFFFFFE00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"7649\n");
    end
    ReadData1E = 64'h43CFFFFFFFDDFFFF;
    ReadData2E = 64'hC3FFFFFFFFFFFFFE;
    ReadData3E = 64'h0000000000000001;
    ans = 64'hC7DFFFFFFFDDFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"7955\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"8261\n");
    end
    ReadData1E = 64'h407E0003FFFFFFFE;
    ReadData2E = 64'hC05C000000000040;
    ReadData3E = 64'hEEA0000004000003;
    ans = 64'hEEA0000004000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"8567\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h6EE0000000000FDF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h3FDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"8873\n");
    end
    ReadData1E = 64'hAD7000001FFFFF00;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hC69FFFF400000000;
    ans = 64'hC69FFFF400000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"9179\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h403FC000000001FF;
    ReadData3E = 64'h403717FDCAC5D1CA;
    ans = 64'h403717FDCAC5D1CA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"9485\n");
    end
    ReadData1E = 64'h4015B627851BAB59;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h40156D89E146EAD6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"9791\n");
    end
    ReadData1E = 64'hC0301FBFFFFFFFFF;
    ReadData2E = 64'h401000000000FFC0;
    ReadData3E = 64'hDB30000000020040;
    ans = 64'hDB30000000020040;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"10097\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'h450FF03FFFFFFFFE;
    ans = 64'h450FF03FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"10403\n");
    end
    ReadData1E = 64'h3FFFF3FFFFFFFFFF;
    ReadData2E = 64'h41CC8460F8D1AAC0;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"10709\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"11015\n");
    end
    ReadData1E = 64'h40040000000FFFFE;
    ReadData2E = 64'h37EFFFBEFFFFFFFE;
    ReadData3E = 64'hB7E0000000000081;
    ans = 64'h37FFFFAEC01FFF78;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"11321\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBCA0007FFFFFFEFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"11627\n");
    end
    ReadData1E = 64'hBFB4ECBCB6018A7A;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h4000000000008FFF;
    ans = 64'h3FFF589A1A5113AA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"11933\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC0AFFFFFFFF00000;
    ReadData3E = 64'h3CAFFFF0000000FE;
    ans = 64'h3CAFFFF0000000FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"12239\n");
    end
    ReadData1E = 64'h43E01DFFFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'hC000000000000001;
    ans = 64'h43D01DFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"12545\n");
    end
    ReadData1E = 64'h434B07BEC47D5B28;
    ReadData2E = 64'hC00FFFFF7FFEFFFF;
    ReadData3E = 64'hFFEFFDFFFFE00000;
    ans = 64'hFFEFFDFFFFE00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"12851\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'hC1CFFFFFF8003FFE;
    ans = 64'hC1CFFFFFF8003FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"13157\n");
    end
    ReadData1E = 64'h96FB9A1C04CC829F;
    ReadData2E = 64'hC03FFFDFFFFFFFC0;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"13463\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"13769\n");
    end
    ReadData1E = 64'h4342F61031FB9231;
    ReadData2E = 64'hFFF1AB8857B4DC76;
    ReadData3E = 64'hBE8EFFFFFFFFFFEF;
    ans = 64'hFFF9AB8857B4DC76;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"14075\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h41C29C11EE259365;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"14381\n");
    end
    ReadData1E = 64'hC0300002000003FF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h43F2A5B9F0ADEE20;
    ans = 64'h43F2A5B9F0ADEE20;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"14687\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC1DFFFFFFE0003FE;
    ReadData3E = 64'h31800800000003FE;
    ans = 64'h31800800000003FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"14993\n");
    end
    ReadData1E = 64'h37FFFFFFFFFF87FF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"15299\n");
    end
    ReadData1E = 64'h400FFFF100000000;
    ReadData2E = 64'hC3FFFFFFFFEFFDFF;
    ReadData3E = 64'h37F00000800001FF;
    ans = 64'hC41FFFF0FFEFFE07;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"15605\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h405FFFBFFFE00000;
    ans = 64'h405FFFBFFFE00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"15911\n");
    end
    ReadData1E = 64'hFFEF7FFFFFBFFFFE;
    ReadData2E = 64'hC03EFFFDFFFFFFFF;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"16217\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"16523\n");
    end
    ReadData1E = 64'h403E7FC2EFD29C71;
    ReadData2E = 64'h40100007FFE00000;
    ReadData3E = 64'hC3DFFFFFBFF00000;
    ans = 64'hC3DFFFFFBFF00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"16829\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3FFEC05C47FE2CB6;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h8010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"17135\n");
    end
    ReadData1E = 64'hCAAFFEFFFFFC0000;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'hC50FEFFFFFC00000;
    ans = 64'hCAAFFEFFFFFC0002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"17441\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC069FFFFFFFFFFFE;
    ReadData3E = 64'h4B10000000FFFFC0;
    ans = 64'h4B10000000FFFFC0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"17747\n");
    end
    ReadData1E = 64'hC020003F7FFFFFFF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hC031003F7FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"18053\n");
    end
    ReadData1E = 64'hC0AFF7F7FFFFFFFF;
    ReadData2E = 64'hC000001FFFFFFC00;
    ReadData3E = 64'h7FE949C5A0E93D66;
    ans = 64'h7FE949C5A0E93D66;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"18359\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h801F1C5A52CE815D;
    ans = 64'h801F1C5A52CE815D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"18665\n");
    end
    ReadData1E = 64'h3EBF1FFFFFFFFFFE;
    ReadData2E = 64'h3FF000008000007F;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"18971\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"19277\n");
    end
    ReadData1E = 64'h3E120003FFFFFFFE;
    ReadData2E = 64'hC4A0000000010800;
    ReadData3E = 64'h47EFFFF7FFF80000;
    ans = 64'h47EFFFF7FFF80000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"19583\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3DCF24B20E238B01;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h000FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"19889\n");
    end
    ReadData1E = 64'h7FD0021FFFFFFFFF;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'h000FFFFFFFF8001F;
    ans = 64'h7FE0022000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"20195\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h400D52B2928D25B3;
    ReadData3E = 64'h802DFFFFFFFBFFFF;
    ans = 64'h802DFFFFFFFBFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"20501\n");
    end
    ReadData1E = 64'h47FDFFFFFFFFBFFE;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h481DFFFFFFFFBFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"20807\n");
    end
    ReadData1E = 64'hBFF00000801FFFFF;
    ReadData2E = 64'h40278FE8D20772ED;
    ReadData3E = 64'hBFDDFFFFFFFC0000;
    ans = 64'hC0287FE98EB5B94D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"21113\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'hC3AFFFA000000000;
    ans = 64'hC3AFFFA000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"21419\n");
    end
    ReadData1E = 64'h80000000000017FF;
    ReadData2E = 64'hB7FFFBFFFFFFFDFF;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h401FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"21725\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hBFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"22031\n");
    end
    ReadData1E = 64'hBCAFDFFEFFFFFFFF;
    ReadData2E = 64'h8027F6CED78A3A8C;
    ReadData3E = 64'hC7EBC49640C1DDFC;
    ans = 64'hC7EBC49640C1DDFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"22337\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h7FFFFE0000002000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"22643\n");
    end
    ReadData1E = 64'h3FFFFFFFFFFF6FFF;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'hC01000000000000D;
    ans = 64'h400FFFFFFFFEDFE4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"22949\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC7F0FFFFC0000000;
    ReadData3E = 64'h3CA000000000000F;
    ans = 64'h3CA000000000000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"23255\n");
    end
    ReadData1E = 64'hC8003BFFFFFFFFFE;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hC8203BFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"23561\n");
    end
    ReadData1E = 64'hC1DFF800000000FE;
    ReadData2E = 64'h001BACBE988157F0;
    ReadData3E = 64'h7750000FFFFFEFFE;
    ans = 64'h7750000FFFFFEFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"23867\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hC34FFFFE000FFFFF;
    ans = 64'hC34FFFFE000FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"24173\n");
    end
    ReadData1E = 64'hBF88E3F9BB250120;
    ReadData2E = 64'hC3DEAB8649B5B386;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'h4377DB24D1EDE54A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"24479\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"24785\n");
    end
    ReadData1E = 64'hC1DF000000001FFF;
    ReadData2E = 64'h403CD1841E8DF396;
    ReadData3E = 64'hB80FFFFFFDFFFFEE;
    ans = 64'hC22BEAF7FD99A0CA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"25091\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBF8FFFEBFFFFFFFE;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"25397\n");
    end
    ReadData1E = 64'hBFE4B5C1CE0FC7F3;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'hBF9000FFFFFFFDFE;
    ans = 64'hC334B5C1CE0FC7F3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"25703\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h41DFFFFFFFFFFFFF;
    ReadData3E = 64'h80100200000007FF;
    ans = 64'h80100200000007FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"26009\n");
    end
    ReadData1E = 64'h3B70010000000007;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h3EC0010000080008;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"26315\n");
    end
    ReadData1E = 64'h3C5FF7FEFFFFFFFE;
    ReadData2E = 64'h3FB000000000F7FF;
    ReadData3E = 64'h802C007FFFFFFFFF;
    ans = 64'h3C1FF7FF0001EF80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"26621\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h400E96FC00CCC222;
    ans = 64'h400E96FC00CCC222;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"26927\n");
    end
    ReadData1E = 64'hC1DFFFFFEEFFFFFE;
    ReadData2E = 64'h3E8FE007FFFFFFFF;
    ReadData3E = 64'h4000000000000000;
    ans = 64'hC07FC007EF10FBBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"27233\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"27539\n");
    end
    ReadData1E = 64'hC7EF87569D128E87;
    ReadData2E = 64'h402DFFBFFFFFFFFF;
    ReadData3E = 64'h274545C08EA0DAC8;
    ans = 64'hC82D8EA224942B78;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"27845\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h47B6991E31F48059;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"28151\n");
    end
    ReadData1E = 64'h433CCAE453CB7215;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h3FE00000401FFFFF;
    ans = 64'h469CCAE453CB7213;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"28457\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h41D8BE43AD466148;
    ReadData3E = 64'h47F1000400000000;
    ans = 64'h47F1000400000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"28763\n");
    end
    ReadData1E = 64'hC08FFFFFFFE80000;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"29069\n");
    end
    ReadData1E = 64'hB7E00000003DFFFF;
    ReadData2E = 64'h41D07FFFFFFFF800;
    ReadData3E = 64'hBFC00000000103FE;
    ans = 64'hBFC00000000103FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"29375\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h7FDFFFBFFFFFFFF7;
    ans = 64'h7FDFFFBFFFFFFFF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"29681\n");
    end
    ReadData1E = 64'h1FCFFFFFF00FFFFF;
    ReadData2E = 64'h3FD00000000047FE;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"29987\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h3CAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"30293\n");
    end
    ReadData1E = 64'hC02FFFFFFFFFFE01;
    ReadData2E = 64'hBFC0000380000000;
    ReadData3E = 64'hB810020000000080;
    ans = 64'h400000037FFFFF00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"30599\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFD007FFFFFFEFFE;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"30905\n");
    end
    ReadData1E = 64'h408000000007FBFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h453E07A9A16128DC;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"31211\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hA33FFF7FFFF7FFFE;
    ReadData3E = 64'hE19000007FF00000;
    ans = 64'hE19000007FF00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"31517\n");
    end
    ReadData1E = 64'h401001FFFFFFBFFE;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"31823\n");
    end
    ReadData1E = 64'hC09FFFFFFFFBFF00;
    ReadData2E = 64'h5FDFE08000000000;
    ReadData3E = 64'h4140000007FFFFFC;
    ans = 64'hE08FE07FFFFC02F1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"32129\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h404CD362AE8BB385;
    ans = 64'hFFF8000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"32435\n");
    end
    ReadData1E = 64'h912000000008FFFF;
    ReadData2E = 64'hC3EA190F0D3FE460;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"32741\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"33047\n");
    end
    ReadData1E = 64'hD89FF80000003FFF;
    ReadData2E = 64'h400000000003FFF7;
    ReadData3E = 64'h649FFFFFFDFFFF80;
    ans = 64'h649FFFFFFDFFFF80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"33353\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC00BA8E5DA51A56B;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"33659\n");
    end
    ReadData1E = 64'hA9B4B64A1320AA03;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h40CFFFFFFF7F7FFE;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"33965\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3177FFFFFEFFFFFF;
    ReadData3E = 64'hB80FFFFFFFEFFFFB;
    ans = 64'hB80FFFFFFFEFFFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"34271\n");
    end
    ReadData1E = 64'hC80FFE0000FFFFFF;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"34577\n");
    end
    ReadData1E = 64'h37E4E5EB09B9E630;
    ReadData2E = 64'hC0D000000002007E;
    ReadData3E = 64'h400FFFFFFFFEFFFC;
    ans = 64'h400FFFFFFFFEFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"34883\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h3FFFFFF7FDFFFFFF;
    ans = 64'h3FFFFFF7FDFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"35189\n");
    end
    ReadData1E = 64'hC01EFF7FFFFFFFFF;
    ReadData2E = 64'h41CFFFFFFFFEFFEF;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hC1FEFF80000707F3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"35495\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h0010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"35801\n");
    end
    ReadData1E = 64'h433C00000000003F;
    ReadData2E = 64'h2C453D8FBBAA6DE2;
    ReadData3E = 64'h4000400040000000;
    ans = 64'h4000400040000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"36107\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"36413\n");
    end
    ReadData1E = 64'h7FDBB7A2E8C34984;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'h40300000007FFFEF;
    ans = 64'h402C890BA3E796AE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"36719\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h41EFFFFBFFFFFFFE;
    ReadData3E = 64'h7D90080000008000;
    ans = 64'h7D90080000008000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"37025\n");
    end
    ReadData1E = 64'h47FFFFFFDFFFEFFE;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"37331\n");
    end
    ReadData1E = 64'hB80CCAF9E864CC74;
    ReadData2E = 64'hBFCFFFFFFFE000FF;
    ReadData3E = 64'h1B1FFFFF800FFFFF;
    ans = 64'h37ECCAF9E8480260;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"37637\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h43CE000000FFFFFF;
    ans = 64'h43CE000000FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"37943\n");
    end
    ReadData1E = 64'h401FFFFFF3FFFFFF;
    ReadData2E = 64'hC2A00080001FFFFF;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'hC2D0007FFA1FCFEE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"38249\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h8000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"38555\n");
    end
    ReadData1E = 64'h37EF8003FFFFFFFF;
    ReadData2E = 64'h410FFFFF7FFFFDFE;
    ReadData3E = 64'h3FF6752E3200448A;
    ans = 64'h3FF6752E3200448A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"38861\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC3FFFD9E53F69891;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"39167\n");
    end
    ReadData1E = 64'h401CFA37F61AB9DE;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'hBD48AE33FF42F0BC;
    ans = 64'hBD48AE33FF42F0BC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"39473\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC7EF912640B3BADD;
    ReadData3E = 64'hDAF00DFFFFFFFFFF;
    ans = 64'hDAF00DFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"39779\n");
    end
    ReadData1E = 64'h3FD004DE1725440E;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"40085\n");
    end
    ReadData1E = 64'hC7E0800001000000;
    ReadData2E = 64'h41F0000007FFFFDF;
    ReadData3E = 64'hBFB0000080FFFFFF;
    ans = 64'hC9E08000093FFFDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"40391\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'h4F2FE00000000003;
    ans = 64'h4F2FE00000000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"40697\n");
    end
    ReadData1E = 64'hC3D4710A80C12BF3;
    ReadData2E = 64'h0023BF0FB8CF836D;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"41003\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"41309\n");
    end
    ReadData1E = 64'hC7FFC00003FFFFFF;
    ReadData2E = 64'hB7E0003FFFFFFFFF;
    ReadData3E = 64'hFFEFDFFDFFFFFFFF;
    ans = 64'hFFEFDFFDFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"41615\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC22FFBFFFFFFFFF0;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hBFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"41921\n");
    end
    ReadData1E = 64'hB80BFFFFFFFFFFFF;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h00102000FFFFFFFE;
    ans = 64'h34BC000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"42227\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h382F7F7FFFFFFFFF;
    ReadData3E = 64'hC0124D9E33621DA3;
    ans = 64'hC0124D9E33621DA3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"42533\n");
    end
    ReadData1E = 64'h47F00FEFFFFFFFFE;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"42839\n");
    end
    ReadData1E = 64'h43E838FD0E6F07BF;
    ReadData2E = 64'h405FC00001FFFFFF;
    ReadData3E = 64'hC85FFFFFFC004000;
    ans = 64'hC85FFFFFFC004000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"43145\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h37E000001FFF7FFE;
    ans = 64'h37E000001FFF7FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"43451\n");
    end
    ReadData1E = 64'hC034B21C8FB22DE2;
    ReadData2E = 64'h001FBFFFFFFFFF7F;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h806388B85692C933;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"43757\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"44063\n");
    end
    ReadData1E = 64'h3FFFFFBFFFFFFBFF;
    ReadData2E = 64'h3F00080000000000;
    ReadData3E = 64'hC3D7F7D4202822B9;
    ans = 64'hC3D7F7D4202822B9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"44369\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC030000100004000;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"44675\n");
    end
    ReadData1E = 64'h37E00000007FFF7F;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hBFE585B80D43FB9F;
    ans = 64'hBFE585B80D43FB9F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"44981\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h8E4FFFFFBFFFEFFF;
    ReadData3E = 64'h404FFFFFFFFFE004;
    ans = 64'h404FFFFFFFFFE004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"45287\n");
    end
    ReadData1E = 64'hC0302000000000FE;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"45593\n");
    end
    ReadData1E = 64'hBEF00000037FFFFF;
    ReadData2E = 64'h46D010000000FFFF;
    ReadData3E = 64'hB8B7FFFFFFFFFBFE;
    ans = 64'hC5D0100003847FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"45899\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h46C00002000007FF;
    ans = 64'h46C00002000007FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"46205\n");
    end
    ReadData1E = 64'h456FFFFFFFFFE03F;
    ReadData2E = 64'h3803903647216A9D;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h3D83903647215733;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"46511\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"46817\n");
    end
    ReadData1E = 64'hB12FFFDFFFFFFC00;
    ReadData2E = 64'hBF40000001000040;
    ReadData3E = 64'hC3E002000FFFFFFF;
    ans = 64'hC3E002000FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"47123\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h41EF7FFFFFFFFBFE;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"47429\n");
    end
    ReadData1E = 64'h3FC482064B9A8DE5;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h371FFFFFDFFFFFF7;
    ans = 64'hBFB482064B9A8DE5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"47735\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3800400007FFFFFE;
    ReadData3E = 64'h3FF0000010000100;
    ans = 64'h3FF0000010000100;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"48041\n");
    end
    ReadData1E = 64'hFCFA1871A12F52FC;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'h7CEA1871A12F52FC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"48347\n");
    end
    ReadData1E = 64'hC347798653317A97;
    ReadData2E = 64'h3CA000000000000B;
    ReadData3E = 64'hC3CD17AC96D91948;
    ans = 64'hC3CD17AC96D91948;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"48653\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hCC3000007FFFFFFF;
    ans = 64'hCC3000007FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"48959\n");
    end
    ReadData1E = 64'h37FFFF0000400000;
    ReadData2E = 64'h47EFD7FFFFFFFFFE;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h3FFFD701403FAFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"49265\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"49571\n");
    end
    ReadData1E = 64'hBFDEB61745F36929;
    ReadData2E = 64'h37E00004000FFFFE;
    ReadData3E = 64'h434F7FFFFFFFFF7F;
    ans = 64'h434F7FFFFFFFFF7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"49877\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h40CFF7FFFFFF8000;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"50183\n");
    end
    ReadData1E = 64'hB8016E821E7114DC;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'hB7F000000001FFF7;
    ans = 64'h37F2DD043CE029BF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"50489\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC1E00FFFFFF00000;
    ReadData3E = 64'hACC10000000FFFFE;
    ans = 64'hACC10000000FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"50795\n");
    end
    ReadData1E = 64'hBFB0010000000800;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h401040040000001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"51101\n");
    end
    ReadData1E = 64'h0010003FFFFFFEFF;
    ReadData2E = 64'h43F000000040003F;
    ReadData3E = 64'h4030000007FFBFFF;
    ans = 64'h4030000007FFBFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"51407\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'hFD4003FFFFFFFFBE;
    ans = 64'hFD4003FFFFFFFFBE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"51713\n");
    end
    ReadData1E = 64'hC02000000020007E;
    ReadData2E = 64'h4801000000000003;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"52019\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"52325\n");
    end
    ReadData1E = 64'h4000004000000FFF;
    ReadData2E = 64'hB30E2B9A2B23911C;
    ReadData3E = 64'hCCD50B44486F7C9A;
    ans = 64'hCCD50B44486F7C9A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"52631\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h41F0000000000FFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"52937\n");
    end
    ReadData1E = 64'hB7FFFFFFFFFFE800;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h401F7FDFFFFFFFFF;
    ans = 64'h401F7FDFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"53243\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3DCFFFFEFFFFC000;
    ReadData3E = 64'h4100000003FFFFFE;
    ans = 64'h4100000003FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"53549\n");
    end
    ReadData1E = 64'hBB1F1FFFFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"53855\n");
    end
    ReadData1E = 64'h43DFFFFFFC7FFFFF;
    ReadData2E = 64'hFFD0FFFFFFFFFFDF;
    ReadData3E = 64'h4000000200000004;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"54161\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h47EFFF0000100000;
    ans = 64'h47EFFF0000100000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"54467\n");
    end
    ReadData1E = 64'h25EFFFFC000003FF;
    ReadData2E = 64'hBC5EEAFC4E08B871;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"54773\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"55079\n");
    end
    ReadData1E = 64'hAB4007FFEFFFFFFF;
    ReadData2E = 64'hC1E3F7FFFFFFFFFF;
    ReadData3E = 64'h408FFC0000000003;
    ans = 64'h408FFC0000000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"55385\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFCFFFFE000000FF;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h001FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"55691\n");
    end
    ReadData1E = 64'hC3D4002000000000;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h47EFFFFFFF80FFFF;
    ans = 64'h47EFFFFFFF80FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"55997\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h4EAFFFFFFFFFFFF8;
    ReadData3E = 64'hBFC630E0EEA8577C;
    ans = 64'hBFC630E0EEA8577C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"56303\n");
    end
    ReadData1E = 64'h7FD00000000FFFFE;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"56609\n");
    end
    ReadData1E = 64'h3FCFFDFFEFFFFFFF;
    ReadData2E = 64'hC3ED67F287D7D96E;
    ReadData3E = 64'hC8019F8A79446983;
    ans = 64'hC8019F8A79446983;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"56915\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h3FD800000001FFFF;
    ans = 64'h3FD800000001FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"57221\n");
    end
    ReadData1E = 64'h4040000007FFFFFF;
    ReadData2E = 64'h41100000007FFFFF;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h43500000001FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"57527\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"57833\n");
    end
    ReadData1E = 64'h7BFFFFFFFFC00080;
    ReadData2E = 64'hA9BFFFFFFFF00003;
    ReadData3E = 64'hBFC003FFFDFFFFFF;
    ans = 64'hE5CFFFFFFFB00083;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"58139\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC1EDEA3DDF65A262;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h8010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"58445\n");
    end
    ReadData1E = 64'h402FF7FFFFFFFBFF;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hC7EFFFDFFFFDFFFF;
    ans = 64'hC7EFFFDFFFFDFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"58751\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFDE20DF253846C3;
    ReadData3E = 64'hBFCA3D8764E550D3;
    ans = 64'hBFCA3D8764E550D3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"59057\n");
    end
    ReadData1E = 64'hC7700800000FFFFF;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'h47A00800000FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"59363\n");
    end
    ReadData1E = 64'h354000001FFFFFE0;
    ReadData2E = 64'hB80FFFFFFFFFFFDD;
    ReadData3E = 64'h400FFFFFFFFF7DFE;
    ans = 64'h400FFFFFFFFF7DFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"59669\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hAF2FFFFF000FFFFF;
    ans = 64'hAF2FFFFF000FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"59975\n");
    end
    ReadData1E = 64'hEC8FFFFF00400000;
    ReadData2E = 64'h3DB023FFFFFFFFFE;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hEA5023FF7F0047FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"60281\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"60587\n");
    end
    ReadData1E = 64'hC02FFFFFBFFFFFFE;
    ReadData2E = 64'h40FFCFFFFFFFFFFE;
    ReadData3E = 64'hBF8FFFFF7FFFE000;
    ans = 64'hC13FCFFFC45FFFEC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"60893\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h4893E0E3A2BD2F7E;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h0000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"61199\n");
    end
    ReadData1E = 64'h3FEC9DF9DF3C1BE6;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h345C7595B01ECD68;
    ans = 64'hC33C9DF9DF3C1BE8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"61505\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC50000000007DFFF;
    ReadData3E = 64'h3801F71074B2A70B;
    ans = 64'h3801F71074B2A70B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"61811\n");
    end
    ReadData1E = 64'hBFCFFE000000000F;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h432FFE000000000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"62117\n");
    end
    ReadData1E = 64'h403FFFFFFFF6FFFF;
    ReadData2E = 64'h40107FFF7FFFFFFF;
    ReadData3E = 64'hBB2C000002000000;
    ans = 64'h40607FFF7FFB5BFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"62423\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h43DFFFFFFFFFDFFB;
    ans = 64'h43DFFFFFFFFFDFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"62729\n");
    end
    ReadData1E = 64'h4F600004FFFFFFFF;
    ReadData2E = 64'h43DFFFFC1FFFFFFF;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h535000030FFF64FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"63035\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"63341\n");
    end
    ReadData1E = 64'h800FFFFFBFFF8000;
    ReadData2E = 64'hC00FFFFDFFFFFFFB;
    ReadData3E = 64'hC000960D221BA850;
    ans = 64'hC000960D221BA850;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"63647\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hBFDFFFFFFFFFF801;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"63953\n");
    end
    ReadData1E = 64'h3FF4F85EB86B4EA1;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'h3FC0000000E00000;
    ans = 64'hFFE4F85EB86B4EA1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"64259\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hC7EFFFFFFFEFFFDF;
    ReadData3E = 64'hBCA0000200000000;
    ans = 64'hBCA0000200000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"64565\n");
    end
    ReadData1E = 64'hC00000000010000E;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"64871\n");
    end
    ReadData1E = 64'h4BFF00000000FFFE;
    ReadData2E = 64'hBD4000001FFBFFFF;
    ReadData3E = 64'hC0FFFFFEDFFFFFFF;
    ans = 64'hC94F00003DF93FFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"65177\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h35800000000001FE;
    ans = 64'h35800000000001FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"65483\n");
    end
    ReadData1E = 64'hC0EFFC0007FFFFFE;
    ReadData2E = 64'h3F30000000040100;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC033FE000404007F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"65789\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"66095\n");
    end
    ReadData1E = 64'h7FEFFFFFFFFDFFF6;
    ReadData2E = 64'h0020200100000000;
    ReadData3E = 64'h37F00000101FFFFF;
    ans = 64'h40202000FFFEFDFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"66401\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hD6803FFFFFFF7FFF;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"66707\n");
    end
    ReadData1E = 64'h4071D1D196BCBBB4;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'hC3EAC2BB016C9CB6;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"67013\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'h3FC0400000000002;
    ReadData3E = 64'h407FFFFFFFFF01FF;
    ans = 64'h407FFFFFFFFF01FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"67319\n");
    end
    ReadData1E = 64'h41DDFBFFFFFFFFFF;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"67625\n");
    end
    ReadData1E = 64'h48A2E070EE32D8E0;
    ReadData2E = 64'hBE1CD75F7BC7FF61;
    ReadData3E = 64'h3FA2546AE9598A8E;
    ans = 64'hC6D1036F1A80D304;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"67931\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'hB7F79A72F0763484;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"68237\n");
    end
    ReadData1E = 64'h7FEFDA548DF1E93C;
    ReadData2E = 64'h401FE0000001FFFE;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"68543\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"68849\n");
    end
    ReadData1E = 64'h38AC01FFFFFFFFFF;
    ReadData2E = 64'hC70000400001FFFF;
    ReadData3E = 64'h620FFFFFFE00FFFF;
    ans = 64'h620FFFFFFE00FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"69155\n");
    end
    ReadData1E = 64'h0000000000000000;
    ReadData2E = 64'hFFE001FFFFFFFBFF;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"69461\n");
    end
    ReadData1E = 64'h400FFFB7FFFFFFFE;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'hBEC78A1A64B5276F;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"69767\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h615FFFFFFEFEFFFF;
    ReadData3E = 64'hC047FFF000000000;
    ans = 64'hC047FFF000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"70073\n");
    end
    ReadData1E = 64'hBCDFFFFFFFFFFFFA;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h801FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"70379\n");
    end
    ReadData1E = 64'h400FFFF008000000;
    ReadData2E = 64'hC0140007FFFFFFFF;
    ReadData3E = 64'h3CA0FFFFFFFFEFFF;
    ans = 64'hC033FFFE04FC01FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"70685\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h3F701FFFFEFFFFFF;
    ans = 64'h3F701FFFFEFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"70991\n");
    end
    ReadData1E = 64'h37EFFFFFEFFE0000;
    ReadData2E = 64'hBEDFFFF001FFFFFF;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"71297\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h3CA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"71603\n");
    end
    ReadData1E = 64'hBFC191453FCED168;
    ReadData2E = 64'hBFCB56D15B3927BA;
    ReadData3E = 64'h37E000000FFEFFFE;
    ans = 64'h3F9E04778DA58583;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"71909\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFEB401F3FDB027F;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"72215\n");
    end
    ReadData1E = 64'h469FDFFFFFFFFFFB;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h7FE00800000007FF;
    ans = 64'h7FE00800000007FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"72521\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3E0FFC0000007FFF;
    ReadData3E = 64'h3A4E985408339F37;
    ans = 64'h3A4E985408339F37;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"72827\n");
    end
    ReadData1E = 64'h41CF7FFFFFFBFFFE;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h01EF8000007BFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"73133\n");
    end
    ReadData1E = 64'h986FFEFFFFFFFFFC;
    ReadData2E = 64'h3F900000100000FF;
    ReadData3E = 64'h4770000000100FFF;
    ans = 64'h4770000000100FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"73439\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'hBFFFFFFFFFFFA000;
    ans = 64'hBFFFFFFFFFFFA000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"73745\n");
    end
    ReadData1E = 64'hCDFFFFFFE000003F;
    ReadData2E = 64'h395ECB367C637D45;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'hC76ECB365D984705;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"74051\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h801FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"74357\n");
    end
    ReadData1E = 64'hEB7FFDFFFFFFFFFF;
    ReadData2E = 64'hC3400000000000FC;
    ReadData3E = 64'hB81FFFFFD0000000;
    ans = 64'h6ECFFE00000001F7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"74663\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC1E04000000003FF;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h401FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"74969\n");
    end
    ReadData1E = 64'hC035623F6C0776D2;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'hBE3FFFFFFFF0007F;
    ans = 64'hBE3FFFFFFFF0007F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"75275\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC80FEFFFFFDFFFFE;
    ReadData3E = 64'h43DFFFFFFFFFBF80;
    ans = 64'h43DFFFFFFFFFBF80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"75581\n");
    end
    ReadData1E = 64'h43C0007F7FFFFFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h03F0007F7FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"75887\n");
    end
    ReadData1E = 64'hC91F80000000001F;
    ReadData2E = 64'h5944D35BDA88AE28;
    ReadData3E = 64'hC1C1000000080000;
    ans = 64'hE274800E6B1E8B84;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"76193\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hC1FFFBFFFEFFFFFF;
    ans = 64'hC1FFFBFFFEFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"76499\n");
    end
    ReadData1E = 64'h43EFACB926D1E997;
    ReadData2E = 64'hC3C9C7B965E1B911;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hC7C984A249246658;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"76805\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h000FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"77111\n");
    end
    ReadData1E = 64'hC000000000000FFA;
    ReadData2E = 64'h41F0000000020007;
    ReadData3E = 64'hC1E0000008000080;
    ans = 64'hC204000002021021;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"77417\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFECD4AA110D5FEE;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hC00FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"77723\n");
    end
    ReadData1E = 64'hC80F8001FFFFFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hC7F0000000024000;
    ans = 64'hC7F0000000024004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"78029\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC02FFFFFFFFEFFDE;
    ReadData3E = 64'h313FFFFFF000001F;
    ans = 64'h313FFFFFF000001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"78335\n");
    end
    ReadData1E = 64'h3FDFFF00003FFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"78641\n");
    end
    ReadData1E = 64'hC020000007FFFFF7;
    ReadData2E = 64'hBF23E9A5A4CF72CF;
    ReadData3E = 64'h43C003FFFDFFFFFF;
    ans = 64'h43C003FFFDFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"78947\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'h43DFFFFFFFDFBFFE;
    ans = 64'h43DFFFFFFFDFBFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"79253\n");
    end
    ReadData1E = 64'h3FC0001000000000;
    ReadData2E = 64'h001000001EFFFFFF;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"79559\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h0000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"79865\n");
    end
    ReadData1E = 64'h47EFFFFFFFFFFBFF;
    ReadData2E = 64'h3CA000000000FFC0;
    ReadData3E = 64'hBFE0000008000800;
    ans = 64'h44A000000000FDBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"80171\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h13E00001FFFFFFFE;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"80477\n");
    end
    ReadData1E = 64'h43DFE45B42DC9C12;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'h2DC0000100000008;
    ans = 64'h43BFE45B42DC9C14;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"80783\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC3DFFFFFFFFF01FF;
    ReadData3E = 64'h796FFFFDDFFFFFFF;
    ans = 64'h796FFFFDDFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"81089\n");
    end
    ReadData1E = 64'hBFD000100001FFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"81395\n");
    end
    ReadData1E = 64'h47FFFFFE7FFFFFFF;
    ReadData2E = 64'hBCA46299E234BA33;
    ReadData3E = 64'hFFD006FFFFFFFFFE;
    ans = 64'hFFD006FFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"81701\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h7FDFFFFEFFFFFEFF;
    ans = 64'h7FDFFFFEFFFFFEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"82007\n");
    end
    ReadData1E = 64'h4020000000007BFE;
    ReadData2E = 64'h800C000000001FFE;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"82313\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"82619\n");
    end
    ReadData1E = 64'h3DBFE000000000FE;
    ReadData2E = 64'h3FE000000003FFFE;
    ReadData3E = 64'h41AFACA8D40D6262;
    ans = 64'h41AFACA8D40D6262;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"82925\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC07FFFC00001FFFF;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"83231\n");
    end
    ReadData1E = 64'hBFC28E53D1BD64E5;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'hC3C8A4AE1DD1A4C3;
    ans = 64'hC3C8A4AE1DD1A4C3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"83537\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC2E18FA43B6CB543;
    ReadData3E = 64'h401000000000006E;
    ans = 64'h401000000000006E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"83843\n");
    end
    ReadData1E = 64'hFFFFFFFFFF800007;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFF800007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"84149\n");
    end
    ReadData1E = 64'h3FDFFFE1FFFFFFFE;
    ReadData2E = 64'h3B90004000000006;
    ReadData3E = 64'h381F000000000001;
    ans = 64'h3B800030FFC40005;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"84455\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h311000001FFEFFFF;
    ans = 64'h311000001FFEFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"84761\n");
    end
    ReadData1E = 64'h3FE0000010007FFF;
    ReadData2E = 64'h37D000000003EFFE;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h37C0000010046FFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"85067\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"85373\n");
    end
    ReadData1E = 64'hC7D2FD16AA96B96D;
    ReadData2E = 64'hC1D000010000003F;
    ReadData3E = 64'h37EFD7FFFFFFFFFF;
    ans = 64'h49B2FD17DA682461;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"85679\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h5F3FFE000000001F;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"85985\n");
    end
    ReadData1E = 64'hB804815C1063C1E2;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h449FFFC0000000FE;
    ans = 64'h449FFFC0000000FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"86291\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h41FFFFFF00000004;
    ReadData3E = 64'hDFB0000800000007;
    ans = 64'hDFB0000800000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"86597\n");
    end
    ReadData1E = 64'hC0F04000001FFFFF;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h433FFFFFFFFEFC00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"86903\n");
    end
    ReadData1E = 64'h37FAF5DF12024664;
    ReadData2E = 64'h4CDFFFFFFFFFC010;
    ReadData3E = 64'hBF47F140CD2482F2;
    ans = 64'h44EAF5DF12021086;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"87209\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'hC1F00000002001FF;
    ans = 64'hC1F00000002001FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"87515\n");
    end
    ReadData1E = 64'h43EFFFFFFFBDFFFF;
    ReadData2E = 64'h40164FFAE8166810;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h44164FFAE7E8631A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"87821\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"88127\n");
    end
    ReadData1E = 64'h40310000000003FE;
    ReadData2E = 64'h308100000000003F;
    ReadData3E = 64'h526FFFFFFFEF8000;
    ans = 64'h526FFFFFFFEF8000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"88433\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFCF31CDEA17BD81;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"88739\n");
    end
    ReadData1E = 64'h801647EF7B8F4A27;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'hC02F0007FFFFFFFF;
    ans = 64'hC02F0007FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"89045\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC7FE9B19A720FB35;
    ReadData3E = 64'h48000000001FFFEE;
    ans = 64'h48000000001FFFEE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"89351\n");
    end
    ReadData1E = 64'h71A7FFFFFFFFFBFF;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'hC010000000000001;
    ans = 64'h71B7FFFFFFFFFBFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"89657\n");
    end
    ReadData1E = 64'h434E071E5491DB00;
    ReadData2E = 64'hC3C233D6818376D1;
    ReadData3E = 64'h40000FFFFDFFFFFF;
    ans = 64'hC72114A5B10BB81C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"89963\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hBFEFFC000001FFFF;
    ans = 64'hBFEFFC000001FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"90269\n");
    end
    ReadData1E = 64'h45BFFFFFFFBFFF7E;
    ReadData2E = 64'h80000001FFFFFFEF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"90575\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"90881\n");
    end
    ReadData1E = 64'h3E299454FA252AF8;
    ReadData2E = 64'hC7EFFFFFFFFDFE00;
    ReadData3E = 64'h0CF00001FF000000;
    ans = 64'hC6299454FA239019;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"91187\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h0B3FFFFFFF0001FF;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"91493\n");
    end
    ReadData1E = 64'h801C00000000FFFF;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h380FFFFF0000000F;
    ans = 64'h380FFFFF0000000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"91799\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h434FFF7FFF000000;
    ReadData3E = 64'hBE90000003FF7FFE;
    ans = 64'hBE90000003FF7FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"92105\n");
    end
    ReadData1E = 64'h43F0400000007FFF;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h4410400000007FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"92411\n");
    end
    ReadData1E = 64'h3FCFFFFFFFFFFFFE;
    ReadData2E = 64'hB8BFFFFF000000FF;
    ReadData3E = 64'hFFFB36B5B1C510AD;
    ans = 64'hFFFB36B5B1C510AD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"92717\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'hFFD0207FFFFFFFFF;
    ans = 64'hFFD0207FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"93023\n");
    end
    ReadData1E = 64'h4850000040007FFE;
    ReadData2E = 64'hBD4000000000004F;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"93329\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"93635\n");
    end
    ReadData1E = 64'hBFC3FFFFEFFFFFFF;
    ReadData2E = 64'hBCD0000000100000;
    ReadData3E = 64'hBFDFFFFFFF800FFE;
    ans = 64'hBFDFFFFFFF800FFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"93941\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hE3E000000001FFFF;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"94247\n");
    end
    ReadData1E = 64'hC03FFFF7FEFFFFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hBFCF0003FFFFFFFE;
    ans = 64'hC07003DBFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"94553\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC1D00003FFFFEFFF;
    ReadData3E = 64'h380CDB22F03FB3C7;
    ans = 64'h380CDB22F03FB3C7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"94859\n");
    end
    ReadData1E = 64'hBCAE007FFFFFFFFE;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hC000000000000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"95165\n");
    end
    ReadData1E = 64'h0027CED0481828DA;
    ReadData2E = 64'hC018000000008000;
    ReadData3E = 64'h408FFF7FFFF7FFFF;
    ans = 64'h408FFF7FFFF7FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"95471\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'h427FFFFDFFFFFFFA;
    ans = 64'h427FFFFDFFFFFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"95777\n");
    end
    ReadData1E = 64'hCC1FFFFFDFFFFDFE;
    ReadData2E = 64'hC0001FFFFFBFFFFF;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"96083\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h400FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"96389\n");
    end
    ReadData1E = 64'h3F8B26434DA23461;
    ReadData2E = 64'h48A46669B857A915;
    ReadData3E = 64'h483000FFBFFFFFFE;
    ans = 64'h48494F4D99000EF8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"96695\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h402FEFFFFFFDFFFF;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h0020000000000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"97001\n");
    end
    ReadData1E = 64'h406FFF7FFFFFFFEF;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h75801FFBFFFFFFFF;
    ans = 64'h75801FFBFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"97307\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hFFF0000000080004;
    ReadData3E = 64'h7FFFFFFFFFFEDFFF;
    ans = 64'h7FFFFFFFFFFEDFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"97613\n");
    end
    ReadData1E = 64'h47FFFFFFFFFFC400;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h4B5FFFFFFFFFC3FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"97919\n");
    end
    ReadData1E = 64'h39D000007FFFFFEF;
    ReadData2E = 64'hC19A30C8DE8EA4F0;
    ReadData3E = 64'hC3FFF24324316780;
    ans = 64'hC3FFF24324316780;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"98225\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h57E00000001FFFBF;
    ans = 64'h57E00000001FFFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"98531\n");
    end
    ReadData1E = 64'h800890D34C5F8261;
    ReadData2E = 64'h43EDFFFFFFFFFBFE;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"98837\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"99143\n");
    end
    ReadData1E = 64'hA38000FFFFF7FFFF;
    ReadData2E = 64'hB7F8000003FFFFFF;
    ReadData3E = 64'hBFDFF000000001FE;
    ans = 64'hBFDFF000000001FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"99449\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC1316F6AD91EBB89;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h8010000000116F69;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"99755\n");
    end
    ReadData1E = 64'hB7FFFF9FFFFFFFFF;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'hBBC0800000001FFF;
    ans = 64'hF7EFFFA000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"100061\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h41DFFF01FFFFFFFF;
    ReadData3E = 64'hBCA1790765D39159;
    ans = 64'hBCA1790765D39159;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"100367\n");
    end
    ReadData1E = 64'h47FF7FFFEFFFFFFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"100673\n");
    end
    ReadData1E = 64'h47F0007FFFFFFFFB;
    ReadData2E = 64'h7FD000007FFDFFFF;
    ReadData3E = 64'h41CFFFF00003FFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"100979\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h4010004000100000;
    ans = 64'h4010004000100001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"101285\n");
    end
    ReadData1E = 64'h3F2FFFBFFFFC0000;
    ReadData2E = 64'hBA446C0F1185D142;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"101591\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"101897\n");
    end
    ReadData1E = 64'h40302007FFFFFFFE;
    ReadData2E = 64'h401BF920C8395318;
    ReadData3E = 64'h2C1FFFEFFFFEFFFE;
    ans = 64'h405C3121065A29D7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"102203\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC3CFBF7FFFFFFFFF;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h80AFBF7FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"102509\n");
    end
    ReadData1E = 64'h4010000020000040;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'hC5100000010000FF;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"102815\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h7FEF800002000000;
    ReadData3E = 64'h41C020001FFFFFFF;
    ans = 64'h41C020001FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"103121\n");
    end
    ReadData1E = 64'hC226A456871BA660;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"103427\n");
    end
    ReadData1E = 64'hA8FA294A615C92B1;
    ReadData2E = 64'hC0080000003FFFFF;
    ReadData3E = 64'hC000007FFFFFFEFF;
    ans = 64'hC000007FFFFFFEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"103733\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'hC7EFFFFF80FFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"104039\n");
    end
    ReadData1E = 64'hB7E0800001000000;
    ReadData2E = 64'hB9240000003FFFFE;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"104345\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"104651\n");
    end
    ReadData1E = 64'hB80F000000000008;
    ReadData2E = 64'h3AF0407FFFFFFFFF;
    ReadData3E = 64'h3FE1FC3C8689A584;
    ans = 64'h3FE1FC3C8689A584;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"104957\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h80169DE23509F708;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"105263\n");
    end
    ReadData1E = 64'h43EF7FFFFFFFF7FE;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h7FFFFFFFFFEF0000;
    ans = 64'h7FFFFFFFFFEF0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"105569\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h402FFFFFFFEFFFF8;
    ReadData3E = 64'h43D42686CD65475F;
    ans = 64'h43D42686CD65475F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"105875\n");
    end
    ReadData1E = 64'hC04FFFFFFFFC7FFF;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"106181\n");
    end
    ReadData1E = 64'hB7E00080000001FF;
    ReadData2E = 64'h41F07FFFFFFFFFFE;
    ReadData3E = 64'h3FF6DBA784E8283D;
    ans = 64'h3FF6DBA784E8283D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"106487\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'h7FFFF3FFFFFFFFFE;
    ans = 64'h7FFFF3FFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"106793\n");
    end
    ReadData1E = 64'hB800007FFDFFFFFE;
    ReadData2E = 64'h3FC0E7ABB4FEA446;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"107099\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h3FDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"107405\n");
    end
    ReadData1E = 64'hBFF15CA7EE372F75;
    ReadData2E = 64'hC1FFFFFFFFFFBFBF;
    ReadData3E = 64'hC0BFFFFE001FFFFF;
    ans = 64'h42015CA6EE371C97;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"107711\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBE00000FFFFFFFBF;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"108017\n");
    end
    ReadData1E = 64'hBFD0E4DFE770D252;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h8DC0000001FFFEFE;
    ans = 64'h8DC0000001FFFEFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"108323\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3E500200FFFFFFFF;
    ReadData3E = 64'hC1D0003FFFFFFEFE;
    ans = 64'hC1D0003FFFFFFEFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"108629\n");
    end
    ReadData1E = 64'hBFD03FFFFBFFFFFE;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"108935\n");
    end
    ReadData1E = 64'hB810011FFFFFFFFF;
    ReadData2E = 64'h3CAFFFFFF7FFFFFC;
    ReadData3E = 64'h41DFFFFDFFFFBFFE;
    ans = 64'h41DFFFFDFFFFBFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"109241\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h40008000003FFFFF;
    ans = 64'h40008000003FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"109547\n");
    end
    ReadData1E = 64'hBFF0000000022000;
    ReadData2E = 64'h42F63997A8DDC3FE;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'hC2F63997A8E0B784;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"109853\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"110159\n");
    end
    ReadData1E = 64'hC3DE554489E9FBEB;
    ReadData2E = 64'hB7FF8000000003FF;
    ReadData3E = 64'h3FFF32CBAFD63CA0;
    ans = 64'h3FFF32CBAFD63CA0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"110465\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h4C1FFC003FFFFFFE;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h434FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"110771\n");
    end
    ReadData1E = 64'hC80FFFF800000000;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'hC0098E4ADD5BCBCE;
    ans = 64'hC0098E4ADD5BCBCE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"111077\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h43DFFFF7FFC00000;
    ReadData3E = 64'hFFE00000000003DF;
    ans = 64'hFFE00000000003DF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"111383\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFC07FF;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'h8010000000000001;
    ans = 64'hB95FFFFFFFFC07FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"111689\n");
    end
    ReadData1E = 64'hC15FFFF5FFFFFFFE;
    ReadData2E = 64'hC02DFFFFFFFFFFBF;
    ReadData3E = 64'h41D08607E1470F0C;
    ans = 64'h41D266074B470F08;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"111995\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h41C386E10783B27B;
    ans = 64'h41C386E10783B27B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"112301\n");
    end
    ReadData1E = 64'hFFFB45BA74ABDCC5;
    ReadData2E = 64'h0004C737D93533CD;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hFFFB45BA74ABDCC5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"112607\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"112913\n");
    end
    ReadData1E = 64'h3FC7FFFFFFE00000;
    ReadData2E = 64'hBF7FBFFFFFFFFFFF;
    ReadData3E = 64'h37EFFFFE000FFFFF;
    ans = 64'hBF57CFFFFFE03FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"113219\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC03BFFFFFFFFFFF0;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hC340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"113525\n");
    end
    ReadData1E = 64'hC02528035E1CC78A;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h00000000FFFFBFFF;
    ans = 64'h3CE528035E1CC789;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"113831\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h40100000000003BF;
    ReadData3E = 64'h4800000001FFFFFC;
    ans = 64'h4800000001FFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"114137\n");
    end
    ReadData1E = 64'h40C00003FFFEFFFE;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'hC0A00003FFFEFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"114443\n");
    end
    ReadData1E = 64'h3815F31AA300BAC5;
    ReadData2E = 64'h00282D6384BE04FD;
    ReadData3E = 64'h40A10711683D5DEF;
    ans = 64'h40A10711683D5DEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"114749\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hB7E0000000000202;
    ans = 64'hB7E0000000000202;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"115055\n");
    end
    ReadData1E = 64'hBBCFFFFFFFFBFFFA;
    ReadData2E = 64'h95C71697A05241D8;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"115361\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h8010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"115667\n");
    end
    ReadData1E = 64'h3FC6AF6973D6EEE0;
    ReadData2E = 64'h81DFFFF000007FFF;
    ReadData3E = 64'hC02FFFFFBBFFFFFE;
    ans = 64'hC02FFFFFBBFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"115973\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h4060003C00000000;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h4010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"116279\n");
    end
    ReadData1E = 64'hBFF0000000011FFE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'hA04FFFFFFFFFEFF6;
    ans = 64'h3FE0000000011FFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"116585\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h9550001FFFFFFFFF;
    ReadData3E = 64'hBFE8C4D9D2CF8887;
    ans = 64'hBFE8C4D9D2CF8887;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"116891\n");
    end
    ReadData1E = 64'h3FB001FFFFFBFFFF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"117197\n");
    end
    ReadData1E = 64'hBDC941E1BC02969D;
    ReadData2E = 64'h3F5C6E5429567A7D;
    ReadData3E = 64'h7FF000000002003F;
    ans = 64'h7FF800000002003F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"117503\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h3D63D5CA425674DB;
    ans = 64'h3D63D5CA425674DB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"117809\n");
    end
    ReadData1E = 64'hB80C7A2531F5CBA2;
    ReadData2E = 64'hB7F00000000040FF;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"118115\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h000FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"118421\n");
    end
    ReadData1E = 64'h880FFFFFCFFFFFFE;
    ReadData2E = 64'h2D700000001FFFDE;
    ReadData3E = 64'hBCAFFC000007FFFF;
    ans = 64'hBCAFFC000007FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"118727\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC3E00007FFFFFDFF;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"119033\n");
    end
    ReadData1E = 64'hC031DFFFFFFFFFFE;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h429000000007FFFA;
    ans = 64'h429000000008477A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"119339\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h41C0000000FDFFFF;
    ReadData3E = 64'h3FDFFC0000000002;
    ans = 64'h3FDFFC0000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"119645\n");
    end
    ReadData1E = 64'hC010000008100000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"119951\n");
    end
    ReadData1E = 64'hBE67FFFF7FFFFFFF;
    ReadData2E = 64'hC3FFFFFE003FFFFF;
    ReadData3E = 64'h3E1FFFE003FFFFFE;
    ans = 64'h4277FFFE003007FD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"120257\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h403FFFEFFFF80000;
    ans = 64'h403FFFEFFFF80000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"120563\n");
    end
    ReadData1E = 64'hBFF55E36003CB592;
    ReadData2E = 64'h4913B02D75F6CF08;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'hC91A4B2928CD1BF3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"120869\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"121175\n");
    end
    ReadData1E = 64'hBF8FFFFFFFFFE01F;
    ReadData2E = 64'h43D000000000000B;
    ReadData3E = 64'h3FBF800FFFFFFFFE;
    ans = 64'hC36FFFFFFFFFE035;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"121481\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC01FFFFDFFFFDFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"121787\n");
    end
    ReadData1E = 64'hBFAFEFFFFFEFFFFE;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h3FDFFFFFFDFFF000;
    ans = 64'h3FE1FEFFFEFEF800;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"122093\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3F4001FFFFFFFFDF;
    ReadData3E = 64'hB7EFFFFFFF6FFFFF;
    ans = 64'hB7EFFFFFFF6FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"122399\n");
    end
    ReadData1E = 64'hC02047FFFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"122705\n");
    end
    ReadData1E = 64'h239000003FDFFFFF;
    ReadData2E = 64'hC69003FFFFFBFFFF;
    ReadData3E = 64'hC80FDFFFFFF80000;
    ans = 64'hC80FDFFFFFF80000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"123011\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h401FF00000000000;
    ans = 64'h401FF00000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"123317\n");
    end
    ReadData1E = 64'h5A3FFFFFFF7FFF80;
    ReadData2E = 64'h41D000007BFFFFFF;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'h5C2000007BBFFFBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"123623\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"123929\n");
    end
    ReadData1E = 64'hA3BFFFFFFEC00000;
    ReadData2E = 64'hFFFFFFFFFFFFFFAE;
    ReadData3E = 64'hC010000004080000;
    ans = 64'hFFFFFFFFFFFFFFAE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"124235\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h43E000003FFFFFFC;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"124541\n");
    end
    ReadData1E = 64'hBFE2ABEEC977610C;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'hA42FFF8FFFFFFFFE;
    ans = 64'h3FF2ABEEC977610D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"124847\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FB00FFF00000000;
    ReadData3E = 64'hC1C0000000000000;
    ans = 64'hC1C0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"125153\n");
    end
    ReadData1E = 64'h3F599261FF10D874;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"125459\n");
    end
    ReadData1E = 64'h4010000002000002;
    ReadData2E = 64'hB7EAC9B31C37E360;
    ReadData3E = 64'h401FFFFC07FFFFFE;
    ans = 64'h401FFFFC07FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"125765\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'hB18A7754689D82F3;
    ans = 64'hB18A7754689D82F3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"126071\n");
    end
    ReadData1E = 64'h381FFFFDFFFBFFFF;
    ReadData2E = 64'hC02F9C2590DFA192;
    ReadData3E = 64'h0010000000000000;
    ans = 64'hB85F9C23971954FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"126377\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"126683\n");
    end
    ReadData1E = 64'hC050001000000FFF;
    ReadData2E = 64'h30CFFFFF807FFFFF;
    ReadData3E = 64'h41F431202C87A685;
    ans = 64'h41F431202C87A685;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"126989\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC061000080000000;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"127295\n");
    end
    ReadData1E = 64'h401FFFFFFDFBFFFE;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'hC0200000200003FF;
    ans = 64'hC044000006FE00FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"127601\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h40C0000008000000;
    ReadData3E = 64'h800004001FFFFFFE;
    ans = 64'h800004001FFFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"127907\n");
    end
    ReadData1E = 64'h306000000003FBFF;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"128213\n");
    end
    ReadData1E = 64'h410FFEFFFFF7FFFF;
    ReadData2E = 64'hB0BA68A8F1E0C6BE;
    ReadData3E = 64'h0021FFFFFFFFBFFF;
    ans = 64'hB1DA67D5AC929D8D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"128519\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h4BB00020001FFFFF;
    ans = 64'h4BB00020001FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"128825\n");
    end
    ReadData1E = 64'h3FC100000FFFFFFF;
    ReadData2E = 64'h400FFFFFEFFFEFFF;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h3FE10000077FF776;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"129131\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"129437\n");
    end
    ReadData1E = 64'hC80000000037FFFF;
    ReadData2E = 64'hBF400003FFFFFFC0;
    ReadData3E = 64'h3FC00001001FFFFF;
    ans = 64'h475000040037FFCD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"129743\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FCEBFFFFFFFFFFF;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"130049\n");
    end
    ReadData1E = 64'h3F5FFFEFFFFF7FFF;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'hBCA0000000000020;
    ans = 64'hC2AFFFEFFFFF7FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"130355\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBFEE6A55223FE187;
    ReadData3E = 64'h40BF2A4F0D6A99C1;
    ans = 64'h40BF2A4F0D6A99C1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"130661\n");
    end
    ReadData1E = 64'h41DFFFE00FFFFFFF;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC52FFFE00FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"130967\n");
    end
    ReadData1E = 64'h409F8000003FFFFF;
    ReadData2E = 64'h3B6FEFFFFFFFFFFD;
    ReadData3E = 64'hC01001FFEFFFFFFF;
    ans = 64'hC01001FFEFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"131273\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h00290AEBB101790F;
    ans = 64'h001215D76202F21C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"131579\n");
    end
    ReadData1E = 64'hB9F0000400001FFF;
    ReadData2E = 64'h7FFFC7FFFFFFFFFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"131885\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"132191\n");
    end
    ReadData1E = 64'hBFA1000000000001;
    ReadData2E = 64'h7FDFFFFE003FFFFF;
    ReadData3E = 64'h7FD1000000007FFF;
    ans = 64'h7FCFE00021FCBFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"132497\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hBF2FFF8020000000;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"132803\n");
    end
    ReadData1E = 64'h3CFF8001FFFFFFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h9B5000000020003E;
    ans = 64'hC05F8001FFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"133109\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h3FC5695F385872B9;
    ReadData3E = 64'hBFEFFFFFFFFFFFE7;
    ans = 64'hBFEFFFFFFFFFFFE7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"133415\n");
    end
    ReadData1E = 64'hC1CFFFFFDFFFFFE0;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"133721\n");
    end
    ReadData1E = 64'h380C785BDC80505F;
    ReadData2E = 64'h401004000000007E;
    ReadData3E = 64'hBFC8A5BBE8CA87F4;
    ans = 64'hBFC8A5BBE8CA87F4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"134027\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hBFEEB93FEB6D9A27;
    ans = 64'hBFEEB93FEB6D9A2B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"134333\n");
    end
    ReadData1E = 64'hBE7000023FFFFFFF;
    ReadData2E = 64'hC69FBF7FFFFFFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"134639\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"134945\n");
    end
    ReadData1E = 64'h413000003FE00000;
    ReadData2E = 64'h4047D41D01C7BC9B;
    ReadData3E = 64'hC7EFEFFFFFFFDFFF;
    ans = 64'hC7EFEFFFFFFFDFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"135251\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h8010BFFFFFFFFFFF;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"135557\n");
    end
    ReadData1E = 64'hC1861137DB5A3CE3;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'hC7F04B353236C72C;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"135863\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h402FFFFFFF000040;
    ReadData3E = 64'h47EAAF592AC502DE;
    ans = 64'h47EAAF592AC502DE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"136169\n");
    end
    ReadData1E = 64'h434FFFC07FFFFFFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"136475\n");
    end
    ReadData1E = 64'hB9F94E1DAE778706;
    ReadData2E = 64'hB270000000000006;
    ReadData3E = 64'h381FC0000000FFFE;
    ans = 64'h381FC0000000FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"136781\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h3F1A7E5B60253554;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"137087\n");
    end
    ReadData1E = 64'h404B476C9FC42AB2;
    ReadData2E = 64'h35CF000003FFFFFE;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"137393\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'h4000000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"137699\n");
    end
    ReadData1E = 64'h40303FFFFFFFFEFF;
    ReadData2E = 64'h389000003FFF0000;
    ReadData3E = 64'hBFE000003FFEFFFE;
    ans = 64'hBFE000003FFEFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"138005\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h400FFFFFFFFF3FFF;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h0010000000000005;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"138311\n");
    end
    ReadData1E = 64'h297FC000FFFFFFFE;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'hBD6BDC398D105938;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"138617\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'h7FDFFBFFFFFFFEFF;
    ReadData3E = 64'h381E1EE6EF921ABD;
    ans = 64'h3CBFFBFFFFFFFEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"138923\n");
    end
    ReadData1E = 64'h735FFFF8007FFFFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"139229\n");
    end
    ReadData1E = 64'hC030000000000FE0;
    ReadData2E = 64'h0DDFFFFB00000000;
    ReadData3E = 64'h3CA0000000002400;
    ans = 64'h3CA0000000002400;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"139535\n");
    end
    ReadData1E = 64'h0000000000000001;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h43FDC6998218B001;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"139841\n");
    end
    ReadData1E = 64'h476FF80000020000;
    ReadData2E = 64'hC03000001FFBFFFE;
    ReadData3E = 64'h4340000000000001;
    ans = 64'hC7AFF8003FEA01FC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"140147\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hBFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"140453\n");
    end
    ReadData1E = 64'hC01FFFF000000001;
    ReadData2E = 64'hC0116F956B466478;
    ReadData3E = 64'hC00FFFFFFF7FFFFC;
    ans = 64'h403EDF1967075DAB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"140759\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC3EFFFFFFE03FFFE;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h840FFFFFFE03FFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"141065\n");
    end
    ReadData1E = 64'hC005937FE4462628;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'hC24FF9FFFFFFFFFE;
    ans = 64'hC24FF9FFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"141371\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC1EFFFBFFFFFFFFB;
    ReadData3E = 64'hC3566C56029D4B3B;
    ans = 64'hC3566C56029D4B3B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"141677\n");
    end
    ReadData1E = 64'h43CFF807FFFFFFFF;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"141983\n");
    end
    ReadData1E = 64'h381FFF83FFFFFFFF;
    ReadData2E = 64'hA0AC00000007FFFF;
    ReadData3E = 64'hB80F77FFFFFFFFFE;
    ans = 64'hB80F77FFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"142289\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h4072000000000002;
    ans = 64'h4072000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"142595\n");
    end
    ReadData1E = 64'hBCAFFFEFFFFFE000;
    ReadData2E = 64'h3FFFFFFEFFFFFF80;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"142901\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"143207\n");
    end
    ReadData1E = 64'h00100000403FFFFF;
    ReadData2E = 64'hC3D800000FFFFFFF;
    ReadData3E = 64'hFFF77AAF6197A823;
    ans = 64'hFFFF7AAF6197A823;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"143513\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h43D100000007FFFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"143819\n");
    end
    ReadData1E = 64'hC030000FFDFFFFFF;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h3FEFC0000000003F;
    ans = 64'h3FEFC0000000003F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"144125\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h1F2DEFFFFFFFFFFF;
    ReadData3E = 64'hC34D4DD27012E241;
    ans = 64'hC34D4DD27012E241;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"144431\n");
    end
    ReadData1E = 64'hB95FFFFFFF7FDFFE;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"144737\n");
    end
    ReadData1E = 64'h73B2AA12DD4A0E23;
    ReadData2E = 64'hFFFFFFFFFFF6FFFF;
    ReadData3E = 64'h480000000400007F;
    ans = 64'hFFFFFFFFFFF6FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"145043\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'hFFE020000000FFFF;
    ans = 64'hFFE020000000FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"145349\n");
    end
    ReadData1E = 64'hC1F001000007FFFF;
    ReadData2E = 64'h47F47FFFFFFFFFFF;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'hC9F48148000A3FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"145655\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"145961\n");
    end
    ReadData1E = 64'hC00007FFFFBFFFFF;
    ReadData2E = 64'h40000000100001FE;
    ReadData3E = 64'hC80FFFFFFFC00001;
    ans = 64'hC80FFFFFFFC00001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"146267\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h4010000008400000;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"146573\n");
    end
    ReadData1E = 64'h3800000800001000;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'h3FFC0000003FFFFF;
    ans = 64'h3FFC0000003FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"146879\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h613FFF7FFF7FFFFF;
    ReadData3E = 64'hBE28B1A0E27593C1;
    ans = 64'hBE28B1A0E27593C1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"147185\n");
    end
    ReadData1E = 64'h4800000000004001;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'h44B0000000004002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"147491\n");
    end
    ReadData1E = 64'h43EBF7FFFFFFFFFE;
    ReadData2E = 64'h0010000007FFFFC0;
    ReadData3E = 64'hC1D0007DFFFFFFFF;
    ans = 64'hC1D0007DFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"147797\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hBFA8C76ADFD94B46;
    ans = 64'hBFA8C76ADFD94B46;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"148103\n");
    end
    ReadData1E = 64'hB7F0000040000003;
    ReadData2E = 64'h43CFCFFFFFFFFFFF;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"148409\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"148715\n");
    end
    ReadData1E = 64'h3EAB991333A4F989;
    ReadData2E = 64'h7FF000001BFFFFFF;
    ReadData3E = 64'hBF800001FFFFFFFE;
    ans = 64'h7FF800001BFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"149021\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC8045EC85D798E1F;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"149327\n");
    end
    ReadData1E = 64'h9E40000800000007;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h3FF000080000001F;
    ans = 64'h3FF000080000001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"149633\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC05F80001FFFFFFF;
    ReadData3E = 64'hC7E0020000000040;
    ans = 64'hC7E0020000000040;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"149939\n");
    end
    ReadData1E = 64'hD5E03FFFFFFFFFEF;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'hD5C03FFFFFFFFFF0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"150245\n");
    end
    ReadData1E = 64'hC3C000000003FFF6;
    ReadData2E = 64'hC01FE01FFFFFFFFF;
    ReadData3E = 64'h3FDFFFFFFEF7FFFF;
    ans = 64'h43EFE0200007F7F3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"150551\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hBFC4ED8EFAC28779;
    ans = 64'hBFC4ED8EFAC28779;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"150857\n");
    end
    ReadData1E = 64'h48000010FFFFFFFE;
    ReadData2E = 64'h4070020000003FFF;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h4880021102203FFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"151163\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"151469\n");
    end
    ReadData1E = 64'h3EAFFFF800000010;
    ReadData2E = 64'hC06466011810A25A;
    ReadData3E = 64'h3FC00000000007FE;
    ans = 64'h3FBFF5CD0200C7CE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"151775\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h41C000801FFFFFFF;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"152081\n");
    end
    ReadData1E = 64'hBFF0000000000010;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h3F9FFF7FFFFFFFFA;
    ans = 64'hBFDE00080000001E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"152387\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3FF00000001000FE;
    ReadData3E = 64'h7FFFFF0000400000;
    ans = 64'h7FFFFF0000400000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"152693\n");
    end
    ReadData1E = 64'h3C4000000007FFFE;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h3C3000000007FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"152999\n");
    end
    ReadData1E = 64'h002FFF8000003FFE;
    ReadData2E = 64'hC3F00000017FFFFF;
    ReadData3E = 64'h41FFFFFFFFDFF800;
    ans = 64'h41FFFFFFFFDFF800;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"153305\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'hC1E57FCF69CF5180;
    ans = 64'hC1E57FCF69CF5180;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"153611\n");
    end
    ReadData1E = 64'hBDF00001FFFFFFFF;
    ReadData2E = 64'h508FFFFBEFFFFFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hCE8FFFFFEFFF7DFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"153917\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h0027FFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"154223\n");
    end
    ReadData1E = 64'h480000000103FFFE;
    ReadData2E = 64'hFFFFEFFFC0000000;
    ReadData3E = 64'hC1EFFFFFF8000100;
    ans = 64'hFFFFEFFFC0000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"154529\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC1D0000000007FF7;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"154835\n");
    end
    ReadData1E = 64'h408001FFFFFFE000;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'hC01DFBFFFFFFFFFF;
    ans = 64'h407F8C0FFFFFBFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"155141\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBFBFFF7FFFFFE000;
    ReadData3E = 64'hBE600000200000FE;
    ans = 64'hBE600000200000FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"155447\n");
    end
    ReadData1E = 64'h40093E4AEFD3FCC6;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h40093E4AEFD3FCC6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"155753\n");
    end
    ReadData1E = 64'hC3D1FE0000000000;
    ReadData2E = 64'hC3D07FFFFDFFFFFF;
    ReadData3E = 64'hFFDFFFFFFF00000F;
    ans = 64'hFFDFFFFFFF00000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"156059\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hC1FFFFFFFFDDFFFF;
    ans = 64'hC1FFFFFFFFDDFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"156365\n");
    end
    ReadData1E = 64'hC3FFFFEFFFFC0000;
    ReadData2E = 64'hA307FFFFFFBFFFFE;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"156671\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h8000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"156977\n");
    end
    ReadData1E = 64'h4803C1CD4177B2BE;
    ReadData2E = 64'hC80FFFFFFC200000;
    ReadData3E = 64'hC7F0001FFFFFEFFE;
    ans = 64'hD023C1CD3F133AE3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"157283\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h7FDFC00000FFFFFF;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4017F000003FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"157589\n");
    end
    ReadData1E = 64'hC04FFFFFCFFFFFFF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'h4024BEC7A4366D96;
    ans = 64'hC05D6826DB79324B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"157895\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hB1100000200000FF;
    ReadData3E = 64'h8024D4E33D003A28;
    ans = 64'h8024D4E33D003A28;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"158201\n");
    end
    ReadData1E = 64'h41BFFFFFDFFFDFFF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"158507\n");
    end
    ReadData1E = 64'h3FBF20EB2E06830F;
    ReadData2E = 64'h47C7BB68F15E5DDB;
    ReadData3E = 64'h46EFFFDFFFFFEFFE;
    ans = 64'h479719F77931B106;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"158813\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h615A2EE20A8610C4;
    ans = 64'h615A2EE20A8610C4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"159119\n");
    end
    ReadData1E = 64'h3C7680720C59CAC2;
    ReadData2E = 64'hBF80027FFFFFFFFE;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hBFD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"159425\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h0020000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"159731\n");
    end
    ReadData1E = 64'h979FF7FFFF000000;
    ReadData2E = 64'hC00800007FFFFFFF;
    ReadData3E = 64'hBF8FFFFFF8000006;
    ans = 64'hBF8FFFFFF8000006;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"160037\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3F6FFFFFFFFFFFFF;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"160343\n");
    end
    ReadData1E = 64'h36AD8E7E53A268EF;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'hBFDD8B65DCF02CB0;
    ans = 64'hBFDD8B65DCF02CB0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"160649\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h883C75A4B28F7EB9;
    ReadData3E = 64'hC7E020000FFFFFFF;
    ans = 64'hC7E020000FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"160955\n");
    end
    ReadData1E = 64'h43B003FFFFFEFFFF;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"161261\n");
    end
    ReadData1E = 64'h3FDFFE0000001FFF;
    ReadData2E = 64'hC7EF000000000020;
    ReadData3E = 64'h41DFFFFFFFFFBEFF;
    ans = 64'hC7DEFE1000001F1F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"161567\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h3F800008000007FE;
    ans = 64'h3F800008000007FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"161873\n");
    end
    ReadData1E = 64'h4480001FFFFFFFEE;
    ReadData2E = 64'h1A2FFFFFC00003FF;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h3CA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"162179\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"162485\n");
    end
    ReadData1E = 64'h7FFFBFFFFFFF7FFE;
    ReadData2E = 64'h42508003FFFFFFFE;
    ReadData3E = 64'h4010000000008FFE;
    ans = 64'h7FFFBFFFFFFF7FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"162791\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBFF0001FFFFFDFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"163097\n");
    end
    ReadData1E = 64'h4000185761105E4D;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h802FFFFFFF8007FF;
    ans = 64'h4020185761105E4E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"163403\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h37FFFFFFFFFFF000;
    ReadData3E = 64'h3F300000004FFFFF;
    ans = 64'h3F300000004FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"163709\n");
    end
    ReadData1E = 64'hC80007FFFDFFFFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"164015\n");
    end
    ReadData1E = 64'h43EFBFFFFFFFFFEE;
    ReadData2E = 64'hBF4C00001FFFFFFF;
    ReadData3E = 64'hBFDFFFFFE0200000;
    ans = 64'hC34BC8001FBFFFF0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"164321\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'hC03C2DBB3FCD85BC;
    ans = 64'hC03C2DBB3FCD85BC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"164627\n");
    end
    ReadData1E = 64'hDED0000000009000;
    ReadData2E = 64'hCFF000000080001E;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h6ED000000080901E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"164933\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"165239\n");
    end
    ReadData1E = 64'hC01F00000001FFFF;
    ReadData2E = 64'hD5F00040000007FF;
    ReadData3E = 64'h001BCC611BF190C0;
    ans = 64'h561F007C00020F85;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"165545\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hFFEFDFFFFFFFFDFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hC013EFFFFFFFFEFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"165851\n");
    end
    ReadData1E = 64'hFFF8000000000002;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'hC01FFFFF7FFFFBFE;
    ans = 64'hFFF8000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"166157\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hDAC5722C9F775777;
    ReadData3E = 64'hBCAF4D0647B4FC25;
    ans = 64'hBCAF4D0647B4FC25;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"166463\n");
    end
    ReadData1E = 64'h5FC1FFFFFFFFFFF7;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'hC340000000000001;
    ans = 64'h6311FFFFFFFFFFF8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"166769\n");
    end
    ReadData1E = 64'hC8C0000000004001;
    ReadData2E = 64'hC801E00000000000;
    ReadData3E = 64'hBFFFFFE200000000;
    ans = 64'h50D1E00000004781;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"167075\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h43D1FFFFFFFFFEFF;
    ans = 64'h43D1FFFFFFFFFEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"167381\n");
    end
    ReadData1E = 64'h3FF73F5B85DE8CFD;
    ReadData2E = 64'hCC5FF00002000000;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'hCC6733BBD98F936F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"167687\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"167993\n");
    end
    ReadData1E = 64'h28D8C4D1B6B5D02A;
    ReadData2E = 64'hC03000FBFFFFFFFF;
    ReadData3E = 64'hC3E92D9F7CD3120F;
    ans = 64'hC3E92D9F7CD3120F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"168299\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBCAEFFFFFFFFFFFD;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"168605\n");
    end
    ReadData1E = 64'h3ABFFFFFC000001F;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'hC00C30DFF54218F0;
    ans = 64'h7AAFFFFFC000001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"168911\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC261FFFDFFFFFFFF;
    ReadData3E = 64'h41FFFF00000001FF;
    ans = 64'h41FFFF00000001FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"169217\n");
    end
    ReadData1E = 64'hA0237DD94E68EFEB;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'hE0137DD94E68EFEC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"169523\n");
    end
    ReadData1E = 64'hC19C000000000001;
    ReadData2E = 64'h3FB03FFFFBFFFFFF;
    ReadData3E = 64'hB7EFFEFFFEFFFFFF;
    ans = 64'hC15C6FFFF8FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"169829\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h43CFFFFFEBFFFFFF;
    ans = 64'h43CFFFFFEBFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"170135\n");
    end
    ReadData1E = 64'hD7DC200000000000;
    ReadData2E = 64'hC80FF800007FFFFF;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h5FFC18F800707FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"170441\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC33FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"170747\n");
    end
    ReadData1E = 64'hB01A4F832C959609;
    ReadData2E = 64'h89C82BFFE0C45E6D;
    ReadData3E = 64'h8027D4762D411524;
    ans = 64'h8027D4762D411524;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"171053\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC01FFFFFFFFFFF7F;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"171359\n");
    end
    ReadData1E = 64'h3FBFFFFFFFFFFFFA;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h3801FFFFEFFFFFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"171665\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h5380000040000200;
    ReadData3E = 64'h402FFFFFFDFFFDFE;
    ans = 64'h402FFFFFFDFFFDFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"171971\n");
    end
    ReadData1E = 64'hFFEB274EF24ABE39;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"172277\n");
    end
    ReadData1E = 64'h0A1FFFFFFFFF8200;
    ReadData2E = 64'h400B6A2470F9D3B1;
    ReadData3E = 64'h3F900000000010FE;
    ans = 64'h3F900000000010FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"172583\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h5110000000000000;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"172889\n");
    end
    ReadData1E = 64'hC1FBFFFE00000000;
    ReadData2E = 64'h43FFFBFEFFFFFFFF;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"173195\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"173501\n");
    end
    ReadData1E = 64'h4020001F80000000;
    ReadData2E = 64'hC060001080000000;
    ReadData3E = 64'h3F1FFFFFFE001FFF;
    ans = 64'hC090002FE0207C02;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"173807\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h070B0825C5EC0F5A;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h3CAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"174113\n");
    end
    ReadData1E = 64'hBFCFFFC000002000;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h3FEFFFFFF8003FFF;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"174419\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hFFDA5511E91B29A4;
    ReadData3E = 64'hC0132AF4F7353936;
    ans = 64'hC019C039717C039F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"174725\n");
    end
    ReadData1E = 64'hB80C27A52E429300;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"175031\n");
    end
    ReadData1E = 64'hB7E0400000000007;
    ReadData2E = 64'h40168EE42D8C363C;
    ReadData3E = 64'hC3EACA3042143FC7;
    ans = 64'hC3EACA3042143FC7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"175337\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hC80FFFFBFFFEFFFE;
    ans = 64'hC80FFFFBFFFEFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"175643\n");
    end
    ReadData1E = 64'h3FFFFFFFE000007F;
    ReadData2E = 64'hFFD833ECC0B49AAE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FCF304D5DFD46C7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"175949\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hC00FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"176255\n");
    end
    ReadData1E = 64'h511FFFFFFFFF81FF;
    ReadData2E = 64'h2C9D720D8566E997;
    ReadData3E = 64'hD87A8F4E8B691515;
    ans = 64'hD87A8F4E8B691515;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"176561\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h7FE00000001FFFFF;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'h40000000001FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"176867\n");
    end
    ReadData1E = 64'h3FEEFFFFFFF7FFFF;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'hC80FC00000FFFFFE;
    ans = 64'hC80FC00000FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"177173\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h052FFFFFFFFEBFFE;
    ReadData3E = 64'h52325DDDFABB70E9;
    ans = 64'h52325DDDFABB70E9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"177479\n");
    end
    ReadData1E = 64'h41ED48B95654A033;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"177785\n");
    end
    ReadData1E = 64'hFFD7D8EF2EB80057;
    ReadData2E = 64'h47EC60B0CB2E134A;
    ReadData3E = 64'h40FFFFFC00000002;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"178091\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h001FFFFFC0000FFE;
    ans = 64'h001FFFFFC0000FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"178397\n");
    end
    ReadData1E = 64'hBEF0003FDFFFFFFF;
    ReadData2E = 64'h4020000000002002;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"178703\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"179009\n");
    end
    ReadData1E = 64'h43DF719CD7AE4856;
    ReadData2E = 64'h3FDFFFFFFFFEFFFD;
    ReadData3E = 64'h4B400083FFFFFFFF;
    ans = 64'h4B400083FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"179315\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC1CFFF000003FFFE;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h81EFFEFFFF83FFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"179621\n");
    end
    ReadData1E = 64'h4006D160DC3E5934;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h47FFBFFFFBFFFFFF;
    ans = 64'h47FFBFFFFBFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"179927\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h396DFC66107BEA99;
    ReadData3E = 64'hC7EC00001FFFFFFE;
    ans = 64'hC7EC00001FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"180233\n");
    end
    ReadData1E = 64'h547FFFFFCFFFFFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"180539\n");
    end
    ReadData1E = 64'hB8085FDD8B928E73;
    ReadData2E = 64'h3A7FFFFF80000020;
    ReadData3E = 64'hC7EDF1DED100B9B0;
    ans = 64'hC7EDF1DED100B9B0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"180845\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'h43CD9439FEC97506;
    ans = 64'h43CD9439FEC97506;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"181151\n");
    end
    ReadData1E = 64'h41D0000200000002;
    ReadData2E = 64'hBBD1C3DADDAEF6AB;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"181457\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"181763\n");
    end
    ReadData1E = 64'hC0A00001FFFBFFFE;
    ReadData2E = 64'hC04FFFFFFFF83FFF;
    ReadData3E = 64'h7FD000009FFFFFFE;
    ans = 64'h7FD000009FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"182069\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3FEFDFFFEFFFFFFF;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h000FEFFFF7FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"182375\n");
    end
    ReadData1E = 64'h3A3FC00000000FFF;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h3FBF800000010000;
    ans = 64'h3FBF800000010000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"182681\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h7FFFFFFFE3FFFFFF;
    ReadData3E = 64'h38100008000FFFFE;
    ans = 64'h7FFFFFFFE3FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"182987\n");
    end
    ReadData1E = 64'h4090000007FFFFEF;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFE00000000007FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"183293\n");
    end
    ReadData1E = 64'h3FEFBFFFFFFFFFFA;
    ReadData2E = 64'h4BD0200008000000;
    ReadData3E = 64'h3E03C74E5B33E858;
    ans = 64'h4BCFFF800FDFFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"183599\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'hBFE0020007FFFFFE;
    ans = 64'hBFE0020007FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"183905\n");
    end
    ReadData1E = 64'h802002000000007F;
    ReadData2E = 64'hC7F41A21FDBC75B7;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"184211\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"184517\n");
    end
    ReadData1E = 64'h403FFF00001FFFFF;
    ReadData2E = 64'hBFC0003FFFFFFEFF;
    ReadData3E = 64'h3C6FFBFFFDFFFFFE;
    ans = 64'hC00FFF7FFC1FFE7D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"184823\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC06FFFFF0003FFFE;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"185129\n");
    end
    ReadData1E = 64'h1BD0800000000004;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h41F040000000003E;
    ans = 64'h41F040000000003E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"185435\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h35E1FFFFFFFFBFFF;
    ReadData3E = 64'h3FC00000017FFFFF;
    ans = 64'h3FC00000017FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"185741\n");
    end
    ReadData1E = 64'hBB50000003FFFFBE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"186047\n");
    end
    ReadData1E = 64'hBFE2B7291B588423;
    ReadData2E = 64'h07000FDFFFFFFFFF;
    ReadData3E = 64'h4340001FFFFFFF80;
    ans = 64'h4340001FFFFFFF80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"186353\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h7FE720695AC49CE9;
    ans = 64'h7FE720695AC49CE9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"186659\n");
    end
    ReadData1E = 64'hC7E00000000000BF;
    ReadData2E = 64'hBF899D7D5F337EFB;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h47799D7D5F33802D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"186965\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"187271\n");
    end
    ReadData1E = 64'hFFDFFFFFFFFBBFFF;
    ReadData2E = 64'h7FEF000000000200;
    ReadData3E = 64'h400FF8000007FFFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"187577\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h47F0FFFFFFFFDFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"187883\n");
    end
    ReadData1E = 64'h3BDFFFFFDFFFFFFC;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h30E000000081FFFE;
    ans = 64'hBBCFFFFFDFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"188189\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hD34562EBF3B297C2;
    ReadData3E = 64'hC1B0002003FFFFFF;
    ans = 64'hC1B0002003FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"188495\n");
    end
    ReadData1E = 64'h3D0CC8372DBE587E;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCFD48372DBE5880;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"188801\n");
    end
    ReadData1E = 64'hC3EDFFF000000000;
    ReadData2E = 64'hC07F00000000001F;
    ReadData3E = 64'hC1F8C8FF1C1636A0;
    ans = 64'h447D0FF07FFFE754;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"189107\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h3FE0000020008000;
    ans = 64'h3FE0000020008000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"189413\n");
    end
    ReadData1E = 64'hBCDFFFEFFFFEFFFF;
    ReadData2E = 64'h309FFFC800000000;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"189719\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"190025\n");
    end
    ReadData1E = 64'hC0EEFFFFFF7FFFFF;
    ReadData2E = 64'hB7FFFFF9FFFFFFFF;
    ReadData3E = 64'hBFFFFFFEEFFFFFFF;
    ans = 64'hBFFFFFFEEFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"190331\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC3C000010003FFFF;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"190637\n");
    end
    ReadData1E = 64'hD83FFFFEBFFFFFFE;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h48A000007FFFBFFF;
    ans = 64'h583FFFFEBFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"190943\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hA21FFFFFDBFFFFFF;
    ReadData3E = 64'h3EC0000100000000;
    ans = 64'h3EC0000100000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"191249\n");
    end
    ReadData1E = 64'hC00FFFC000003FFF;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h400FFFC000004001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"191555\n");
    end
    ReadData1E = 64'h43F08000003FFFFF;
    ReadData2E = 64'h43F0000000010007;
    ReadData3E = 64'h3FD000008000003F;
    ans = 64'h47F0800000410806;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"191861\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h402FFFFFFFEFFFE0;
    ans = 64'h402FFFFFFFEFFFE0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"192167\n");
    end
    ReadData1E = 64'hBFCFFFFFF0001000;
    ReadData2E = 64'hBFFFFFFFE1000000;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h3FF7FFFFF4400404;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"192473\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"192779\n");
    end
    ReadData1E = 64'hB050000000000080;
    ReadData2E = 64'hC040000000023FFE;
    ReadData3E = 64'hC5F0001FFFFC0000;
    ans = 64'hC5F0001FFFFC0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"193085\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h70CFFFFFF007FFFF;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"193391\n");
    end
    ReadData1E = 64'h39E0400200000000;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h93BDEDDC1C831F20;
    ans = 64'hB9F04001FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"193697\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC004002000000000;
    ReadData3E = 64'h3F40000002007FFF;
    ans = 64'h3F40000002007FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"194003\n");
    end
    ReadData1E = 64'h401E2A166710BDF4;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'hC02E2A166710BDF4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"194309\n");
    end
    ReadData1E = 64'h47F7FFFFFFFFFE00;
    ReadData2E = 64'hBCAA2259CDEBD65C;
    ReadData3E = 64'h43E0000004000003;
    ans = 64'hC4B399435A70BF23;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"194615\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h93EFFFFFFFF800FE;
    ans = 64'h93EFFFFFFFF800FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"194921\n");
    end
    ReadData1E = 64'hAC0EEB91BA0B67AB;
    ReadData2E = 64'h344007FFFFFF7FFF;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hBFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"195227\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h801FFFFFFFFFFFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"195533\n");
    end
    ReadData1E = 64'hC021FFFFFFFFDFFF;
    ReadData2E = 64'h54B000000000002E;
    ReadData3E = 64'h77C7FFFBFFFFFFFE;
    ans = 64'h77C7FFFBFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"195839\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC02FFFFFFF7FFBFE;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"196145\n");
    end
    ReadData1E = 64'hC6CC000000007FFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'hC02FFDFFFFFFFDFF;
    ans = 64'h46EC000000007FFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"196451\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h36EE3366425C75C8;
    ReadData3E = 64'hC01E3475BEDA42B8;
    ans = 64'hC01E3475BEDA42B8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"196757\n");
    end
    ReadData1E = 64'hBF27000000000000;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h3F47000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"197063\n");
    end
    ReadData1E = 64'h43E0080007FFFFFF;
    ReadData2E = 64'hBE6000021FFFFFFF;
    ReadData3E = 64'hBE0FFF6FFFFFFFFE;
    ans = 64'hC25008022910010E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"197369\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'hBFCFFFFBFFFF7FFE;
    ans = 64'hBFCFFFFBFFFF7FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"197675\n");
    end
    ReadData1E = 64'h18B000000007EFFF;
    ReadData2E = 64'h434FFFBFFEFFFFFE;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h3FDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"197981\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h8033FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"198287\n");
    end
    ReadData1E = 64'h37E000000FFFFFF6;
    ReadData2E = 64'h33D6647C02C60C56;
    ReadData3E = 64'hB7E00800007FFFFF;
    ans = 64'hB7E00800007FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"198593\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h7FDA473ACD6FEEA1;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h401691CEB35BFBA7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"198899\n");
    end
    ReadData1E = 64'h3FDFFFFC00000000;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'hC02FFFFFFFFFFFFF;
    ans = 64'hC033FFFF7FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"199205\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h3BBFA0C90CC4D322;
    ReadData3E = 64'hC1EFFFFFFFFEF7FF;
    ans = 64'hC1EFFFFFFFFEF7FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"199511\n");
    end
    ReadData1E = 64'h3F6D51D89575BCC2;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"199817\n");
    end
    ReadData1E = 64'hB7ED478C97EEF482;
    ReadData2E = 64'hBFFBFF8000000000;
    ReadData3E = 64'hED7FFFFF00000010;
    ans = 64'hED7FFFFF00000010;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"200123\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'hC1C0000001FFFBFF;
    ans = 64'hC1C0000001FFFBFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"200429\n");
    end
    ReadData1E = 64'h3FE0001FFFFFF7FF;
    ReadData2E = 64'h434FFFFFFFFFFBF7;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'h4340001FFFFFF5FA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"200735\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h8360000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"201041\n");
    end
    ReadData1E = 64'h3815CFAC7B87B8E2;
    ReadData2E = 64'h4342C0A02E3CB0AF;
    ReadData3E = 64'hBCA00020001FFFFE;
    ans = 64'hBCA0001E671C7836;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"201347\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC3E3BFFFFFFFFFFF;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"201653\n");
    end
    ReadData1E = 64'h7FE0400000800000;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h43FFFFFDFFFFFFBF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"201959\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hB81FFFFFFF800008;
    ReadData3E = 64'h3FEFFFFFFF8FFFFF;
    ans = 64'h3FEFFFFFFF8FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"202265\n");
    end
    ReadData1E = 64'h41CFFFEFFFFFFFFD;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"202571\n");
    end
    ReadData1E = 64'hC1DFFFFFFE007FFF;
    ReadData2E = 64'h37F003FFEFFFFFFF;
    ReadData3E = 64'hBFA26A7786C5CB13;
    ans = 64'hBFA26A7786C5CB13;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"202877\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h001F000000000006;
    ans = 64'h836FFFFFFFFFFFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"203183\n");
    end
    ReadData1E = 64'hB80FEFFFFFE00000;
    ReadData2E = 64'hD3BFFC00001FFFFF;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h4BDFEC01FFFFF3FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"203489\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"203795\n");
    end
    ReadData1E = 64'h402FFFFFFFFE7FFE;
    ReadData2E = 64'hC0D5A7AD5AE857AD;
    ReadData3E = 64'hCAD5D349067B07E9;
    ans = 64'hCAD5D349067B07E9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"204101\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC06FFFFFFBBFFFFF;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"204407\n");
    end
    ReadData1E = 64'h4EF101FFFFFFFFFE;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hBFC400000FFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"204713\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'h402FFFFFF7FFFFFD;
    ReadData3E = 64'hBC0DFFFFFFFFFDFF;
    ans = 64'hBC0DFFFFFFFFFDFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"205019\n");
    end
    ReadData1E = 64'hC18A72215CF98D87;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"205325\n");
    end
    ReadData1E = 64'hC10FFFFFFFFFE000;
    ReadData2E = 64'hFFF0576627B050D5;
    ReadData3E = 64'hB80DFFFFFBFFFFFF;
    ans = 64'hFFF8576627B050D5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"205631\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'hC1D03FFFF0000000;
    ans = 64'hC1D03FFFF1000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"205937\n");
    end
    ReadData1E = 64'hBE18A0CAC8A2538B;
    ReadData2E = 64'hBF1FFFFFFFFE003F;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h3D48A0CAC8A0C9AF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"206243\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"206549\n");
    end
    ReadData1E = 64'hBF7FFC000000007F;
    ReadData2E = 64'h480BFFFFFC000000;
    ReadData3E = 64'h7FD0E7C95060F159;
    ans = 64'h7FD0E7C95060F159;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"206855\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC68000000003FFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"207161\n");
    end
    ReadData1E = 64'h53B00003FFFFFF7F;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'hC0480001FFFFFFFF;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"207467\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hC03FFFFE00000002;
    ReadData3E = 64'h40A4791D49CC2DB1;
    ans = 64'h40A4791D49CC2DB1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"207773\n");
    end
    ReadData1E = 64'hBFA00001DFFFFFFF;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"208079\n");
    end
    ReadData1E = 64'hB8000000FFFFF7FF;
    ReadData2E = 64'hBFDFFE0000000003;
    ReadData3E = 64'h41E00800000000FF;
    ans = 64'h41E00800000000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"208385\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'hBFDF7FBFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"208691\n");
    end
    ReadData1E = 64'hB80FDFFFFFFFFDFF;
    ReadData2E = 64'hC3C0000000400002;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h3BDFE000007F7E03;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"208997\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"209303\n");
    end
    ReadData1E = 64'h3FB800007FFFFFFF;
    ReadData2E = 64'h7C30000003FFFFE0;
    ReadData3E = 64'hB7EF87FFFFFFFFFF;
    ans = 64'h7BF8000085FFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"209609\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBCA39AB8CC6231C7;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"209915\n");
    end
    ReadData1E = 64'h418FE00000000000;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'hBCA007FFDFFFFFFE;
    ans = 64'hBCA007FFDFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"210221\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hDDF0001FF7FFFFFF;
    ReadData3E = 64'h40301FFFFFFFFFF8;
    ans = 64'h40301FFFFFFFFFF8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"210527\n");
    end
    ReadData1E = 64'hC09000DFFFFFFFFE;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h4010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"210833\n");
    end
    ReadData1E = 64'h41C10000000001FE;
    ReadData2E = 64'h40A2D6D7190048EE;
    ReadData3E = 64'h3FB0400FFFFFFFFF;
    ans = 64'h427404448A9050D9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"211139\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'h3F600000101FFFFE;
    ans = 64'h3F600000101FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"211445\n");
    end
    ReadData1E = 64'h403FFFFFFFFFFFF9;
    ReadData2E = 64'hC7F0001080000000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"211751\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hC340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"212057\n");
    end
    ReadData1E = 64'h47F00FFFFF7FFFFF;
    ReadData2E = 64'h3F08F8194952D2A2;
    ReadData3E = 64'h3FE6D31424C9EF4B;
    ans = 64'h4709111161D464A9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"212363\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFD00040000003FF;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hBFD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"212669\n");
    end
    ReadData1E = 64'h348EE00000000000;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'hBF2FFF7FFFFFDFFE;
    ans = 64'hBF2FFF7FFFFFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"212975\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC1D84EFBB512A72A;
    ReadData3E = 64'hBE69AF2F69623970;
    ans = 64'hBE69AF2F69623970;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"213281\n");
    end
    ReadData1E = 64'hBFFAA0994C9B21E2;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"213587\n");
    end
    ReadData1E = 64'h3A3FFFFFDFF80000;
    ReadData2E = 64'h40900000000E0000;
    ReadData3E = 64'hBFE0FFE000000000;
    ans = 64'hBFE0FFE000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"213893\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'hD8A8756C95FF9ED8;
    ans = 64'hD8A8756C95FF9ED8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"214199\n");
    end
    ReadData1E = 64'h59DFFFFFFE0007FE;
    ReadData2E = 64'hC9C000FFFFFFC000;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"214505\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h4010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"214811\n");
    end
    ReadData1E = 64'h43E256FC6CB56CC4;
    ReadData2E = 64'h7FD0001000000000;
    ReadData3E = 64'hBFF00000203FFFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"215117\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h001FFFFBFFFFF800;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h3CA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"215423\n");
    end
    ReadData1E = 64'hB7FD14550750159C;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'hC7E00007FFEFFFFF;
    ans = 64'hC7E00007FFEFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"215729\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBEBCCD02BE345203;
    ReadData3E = 64'h439FFFFFFFFFFF80;
    ans = 64'h439FFFFFFFFFFF80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"216035\n");
    end
    ReadData1E = 64'hC3D000087FFFFFFF;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'hC07FE010FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"216341\n");
    end
    ReadData1E = 64'hC04000000F7FFFFF;
    ReadData2E = 64'hC4AFFFFC00000010;
    ReadData3E = 64'hC1EFFFFFFFF7DFFE;
    ans = 64'h44FFFFFC1EFFFC1E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"216647\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h41EFE05B7227E188;
    ans = 64'h41EFE05B7227E188;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"216953\n");
    end
    ReadData1E = 64'hC26E000000000000;
    ReadData2E = 64'h002FFFF77FFFFFFF;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"217259\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"217565\n");
    end
    ReadData1E = 64'hC63FFFFFDFEFFFFF;
    ReadData2E = 64'h40100000000000FC;
    ReadData3E = 64'h47F76CA910E9A48C;
    ans = 64'h47F76CA908E9A494;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"217871\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3330000000000803;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h801FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"218177\n");
    end
    ReadData1E = 64'hB7EFFFFFFEFBFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hB7F2007FFFFFFFFF;
    ans = 64'hB7F2008000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"218483\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h37E60289AEF484C8;
    ReadData3E = 64'hC0E000000BFFFFFE;
    ans = 64'hC0E000000BFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"218789\n");
    end
    ReadData1E = 64'h3CA0007EFFFFFFFE;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"219095\n");
    end
    ReadData1E = 64'h41DF800000003FFF;
    ReadData2E = 64'h43C00000001FBFFF;
    ReadData3E = 64'h380FFFFFFFF7F800;
    ans = 64'h45AF8000003EC1FD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"219401\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'hC01962BD104E333A;
    ans = 64'hC01962BD104E333A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"219707\n");
    end
    ReadData1E = 64'h496FFFFFFF7FF800;
    ReadData2E = 64'h47E000001FFFFFFF;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'h516000001FBFFBFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"220013\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"220319\n");
    end
    ReadData1E = 64'h0ECFFFFFFFFFFFFA;
    ReadData2E = 64'hC03FFBFFFEFFFFFF;
    ReadData3E = 64'hC03FFEFBFFFFFFFE;
    ans = 64'hC03FFEFBFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"220625\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC03FFFFFFFFFFFFF;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h805EFFFFFFFFFFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"220931\n");
    end
    ReadData1E = 64'h3FF00000040000FE;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'hBCBFEFFFFFFFFFC0;
    ans = 64'h3FE00000040000FA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"221237\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC3EA4F01DB82E6F6;
    ReadData3E = 64'hC1D0000000800002;
    ans = 64'hC1D0000000800002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"221543\n");
    end
    ReadData1E = 64'hC975270E98987C51;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'hC965270E98987C50;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"221849\n");
    end
    ReadData1E = 64'hC01000000FFFFFFE;
    ReadData2E = 64'hF6EFFFFFF000FFFF;
    ReadData3E = 64'h41EFFE0080000000;
    ans = 64'h7710000008007FF6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"222155\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h400CCB5237997727;
    ans = 64'h400CCB5237997727;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"222461\n");
    end
    ReadData1E = 64'h43F0040010000000;
    ReadData2E = 64'h47A1B34990150361;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h4BA1B7B6742C5232;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"222767\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"223073\n");
    end
    ReadData1E = 64'h401FFFBFFF7FFFFF;
    ReadData2E = 64'h3E600000013FFFFF;
    ReadData3E = 64'hC1F0001040000000;
    ans = 64'hC1F0001040000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"223379\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h404FF80000000010;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h006FF8000000000C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"223685\n");
    end
    ReadData1E = 64'h2F400000000FBFFF;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'h41F606CA2848B3D9;
    ans = 64'h41F606CA2848B3D9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"223991\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC02FFFFF0001FFFF;
    ReadData3E = 64'h3E7F00003FFFFFFF;
    ans = 64'h3E7F00003FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"224297\n");
    end
    ReadData1E = 64'h3FEFFEFFE0000000;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'h3FDFFDFFBFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"224603\n");
    end
    ReadData1E = 64'h4F0A74F16F87B7F7;
    ReadData2E = 64'hC3EFDFFFFFFE0000;
    ReadData3E = 64'hC06BE163A6D740EF;
    ans = 64'hD30A5A7C7E1688F0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"224909\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'hBDA00BFFFFFFFFFF;
    ans = 64'hBDA00BFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"225215\n");
    end
    ReadData1E = 64'h480DA52CB9BE1522;
    ReadData2E = 64'hBFF0000020007FFE;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC80DA52CF5095BBB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"225521\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"225827\n");
    end
    ReadData1E = 64'h24A18A36F18AE59F;
    ReadData2E = 64'h17C00080000001FE;
    ReadData3E = 64'h356FFF8002000000;
    ans = 64'h356FFF8002000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"226133\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h41EFC6443D52B99A;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"226439\n");
    end
    ReadData1E = 64'h403BD20E5EDBB48B;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'h8C61D52AFF1B225C;
    ans = 64'h403BD20E5EDBB48D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"226745\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC07FFFFFFFFFFFDC;
    ReadData3E = 64'h41301E0000000000;
    ans = 64'h41301E0000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"227051\n");
    end
    ReadData1E = 64'h3922C21DDFDA7C7B;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"227357\n");
    end
    ReadData1E = 64'h43F00002FFFFFFFE;
    ReadData2E = 64'hBE5FFFFFFFFFFFFF;
    ReadData3E = 64'h4E4FFFFF1FFFFFFF;
    ans = 64'h4E4FFFFF1FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"227663\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'h3F2D90DCFD91775A;
    ans = 64'h3F2D90DCFD91775A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"227969\n");
    end
    ReadData1E = 64'hB667FFFFEFFFFFFF;
    ReadData2E = 64'hBE2003FFF8000000;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"228275\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"228581\n");
    end
    ReadData1E = 64'h4340020000010000;
    ReadData2E = 64'h3810000001FFFFEF;
    ReadData3E = 64'h42BB43994E714E2E;
    ans = 64'h42BB43994E714E2E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"228887\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h401FFC03FFFFFFFE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"229193\n");
    end
    ReadData1E = 64'h41FFFF7FFFFFFFE0;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h7300801FFFFFFFFE;
    ans = 64'h7300801FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"229499\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC0007FFFFFFFDFFF;
    ReadData3E = 64'h4C9FFFFDF8000000;
    ans = 64'h4C9FFFFDF8000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"229805\n");
    end
    ReadData1E = 64'h628FFFFBFC000000;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'h629FFFFBFC000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"230111\n");
    end
    ReadData1E = 64'hC01F7FFBFFFFFFFE;
    ReadData2E = 64'hC130010020000000;
    ReadData3E = 64'h3FBFF7FFFFFFDFFE;
    ans = 64'h415F81F446BDF7FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"230417\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'hBC7FFFFFFFFF07FF;
    ans = 64'hBC7FFFFFFFFF07FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"230723\n");
    end
    ReadData1E = 64'h3FF0000040000001;
    ReadData2E = 64'hC02000FFFFFF7FFF;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hC024010040038000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"231029\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"231335\n");
    end
    ReadData1E = 64'h47EFFFFFFFF7FF7F;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hC860000010003FFE;
    ans = 64'hC860000010003FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"231641\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC4DFFFFFFFFFFFC1;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"231947\n");
    end
    ReadData1E = 64'h3FFFFF7FFDFFFFFF;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'h47748A41CB17831D;
    ans = 64'h47748A41CB17831D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"232253\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hB8100000000011FF;
    ReadData3E = 64'h403FFFFFF87FFFFF;
    ans = 64'h403FFFFFF87FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"232559\n");
    end
    ReadData1E = 64'h3D7D028206CB6C42;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h3D9D028206CB6C44;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"232865\n");
    end
    ReadData1E = 64'h41407FFF80000000;
    ReadData2E = 64'hA640000020000000;
    ReadData3E = 64'hC3E200000000003E;
    ans = 64'hC3E200000000003E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"233171\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h409FFFFFFFFEFFFD;
    ans = 64'h409FFFFFFFFEFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"233477\n");
    end
    ReadData1E = 64'hC7F0004010000000;
    ReadData2E = 64'h38038FCF6BF4B472;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'hBFF7203B7D846763;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"233783\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"234089\n");
    end
    ReadData1E = 64'hC07223F0FE41D18A;
    ReadData2E = 64'hBF90001FFFFFF7FE;
    ReadData3E = 64'h3D70007FFFFF7FFF;
    ans = 64'h401224154623C8FA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"234395\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h4319A2C1C0285F37;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"234701\n");
    end
    ReadData1E = 64'hBCAFFFFFFFC01000;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'hC0EFF00FFFFFFFFF;
    ans = 64'hC0EFF00FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"235007\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h43400021FFFFFFFF;
    ReadData3E = 64'h3FCFFFFFFFE00004;
    ans = 64'h3FCFFFFFFFE00004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"235313\n");
    end
    ReadData1E = 64'hB05FFFDFFFFEFFFF;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'hB3AFFFDFFFFEFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"235619\n");
    end
    ReadData1E = 64'h0020000000000807;
    ReadData2E = 64'h43EFFFEFFFFFFFF0;
    ReadData3E = 64'h401FFFFFFCFFFFFF;
    ans = 64'h401FFFFFFCFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"235925\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h43C0000002000000;
    ans = 64'h43C0000002000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"236231\n");
    end
    ReadData1E = 64'hFFED5680A72E37A2;
    ReadData2E = 64'h41F6874673086E01;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"236537\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h036FFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"236843\n");
    end
    ReadData1E = 64'h47F0001000000000;
    ReadData2E = 64'h3D00010000000080;
    ReadData3E = 64'h380A26E8BAE640A4;
    ans = 64'h4500011001000080;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"237149\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h801B2DC8D53A6A81;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"237455\n");
    end
    ReadData1E = 64'h3FD0000000000A00;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'hE25FFFFFFF83FFFE;
    ans = 64'hE25FFFFFFF83FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"237761\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FDD188A4F128793;
    ReadData3E = 64'hBFEFFFFFFEDFFFFF;
    ans = 64'hBFEFFFFFFEDFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"238067\n");
    end
    ReadData1E = 64'h7FFFFFE0000FFFFF;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"238373\n");
    end
    ReadData1E = 64'hC3EFFFFFFFFFEE00;
    ReadData2E = 64'hC7F000000000803F;
    ReadData3E = 64'hC1CFDFBFFFFFFFFF;
    ans = 64'h4BF000000000773F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"238679\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h30F5B31E15D86A5A;
    ans = 64'h3FFFFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"238985\n");
    end
    ReadData1E = 64'hDF90800000003FFE;
    ReadData2E = 64'hD7E1FFFFFFFFFFFE;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h77829000000047FC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"239291\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"239597\n");
    end
    ReadData1E = 64'hC08FFFFF00003FFF;
    ReadData2E = 64'h3FD0080000000004;
    ReadData3E = 64'hC388EF1792D83470;
    ans = 64'hC388EF1792D83478;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"239903\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h381000000FFFFF7E;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h400FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"240209\n");
    end
    ReadData1E = 64'h403FFFFDFFFFFEFE;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'hC00FFBFDFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"240515\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hEB8FDFBFFFFFFFFF;
    ReadData3E = 64'hC032000000003FFF;
    ans = 64'hC032000000003FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"240821\n");
    end
    ReadData1E = 64'h37E00100003FFFFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"241127\n");
    end
    ReadData1E = 64'hBF8CD92404A3E8B5;
    ReadData2E = 64'hFFE000000FFF7FFF;
    ReadData3E = 64'h8000102000000000;
    ans = 64'h7F7CD924217C25EF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"241433\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h4020000010080000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"241739\n");
    end
    ReadData1E = 64'hB35FFFEFDFFFFFFE;
    ReadData2E = 64'hC0900007FFFFBFFE;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"242045\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"242351\n");
    end
    ReadData1E = 64'hBFE9E4420B1827EE;
    ReadData2E = 64'hC190000080000002;
    ReadData3E = 64'hB73C9715382BE1E6;
    ans = 64'h4189E442DA3A384A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"242657\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h41CAA0EB1CCD7918;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"242963\n");
    end
    ReadData1E = 64'hCC0FFFFFDFFFFFBF;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'hBE80080000008000;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"243269\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hB980FFFFF7FFFFFE;
    ReadData3E = 64'hA28FFFFFFF83FFFF;
    ans = 64'hA28FFFFFFF83FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"243575\n");
    end
    ReadData1E = 64'h43FFFF000000001F;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"243881\n");
    end
    ReadData1E = 64'h555FFE00000001FF;
    ReadData2E = 64'h401FFEFFFFFBFFFF;
    ReadData3E = 64'hC3E6BCB070CDA62F;
    ans = 64'h558FFD000FFC023E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"244187\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h41F0001000000001;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"244493\n");
    end
    ReadData1E = 64'h402FFFFDFFFF7FFE;
    ReadData2E = 64'h403C63BB8B2B2B77;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h407C63B9C4EF0134;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"244799\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"245105\n");
    end
    ReadData1E = 64'hC3CFF7FFFFEFFFFE;
    ReadData2E = 64'h507000000000101F;
    ReadData3E = 64'hB8000000000011FF;
    ans = 64'hD44FF7FFFFF02034;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"245411\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h000FFFFFDFFFFFF0;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h3FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"245717\n");
    end
    ReadData1E = 64'h9DF0000002FFFFFE;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hB72FFFFBFFFFFFFE;
    ans = 64'hB72FFFFBFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"246023\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hFFDDC2F0D865AC3F;
    ReadData3E = 64'hBF700000000007FF;
    ans = 64'hBFFDD2F0D865AC43;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"246329\n");
    end
    ReadData1E = 64'hBFBFFFFBFFBFFFFF;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"246635\n");
    end
    ReadData1E = 64'hB32585C41E783BBC;
    ReadData2E = 64'hC3C0001000000020;
    ReadData3E = 64'hC74FFFFFEFFFFFF8;
    ans = 64'hC74FFFFFEFFFFFF8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"246941\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h7FE285F0E49D4A7B;
    ans = 64'h7FE285F0E49D4A7B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"247247\n");
    end
    ReadData1E = 64'hC3500100000FFFFF;
    ReadData2E = 64'hBDF0E1E8C66ADBAB;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h4150E2F6E5082440;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"247553\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"247859\n");
    end
    ReadData1E = 64'h3F700007FFFEFFFE;
    ReadData2E = 64'hB7FC3FB36C2D5B84;
    ReadData3E = 64'hBFFFFBFFFFFFF7FF;
    ans = 64'hBFFFFBFFFFFFF7FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"248165\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3F97A677C36D84B9;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hBFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"248471\n");
    end
    ReadData1E = 64'h401537E560C75BAB;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h3FB0000007FFFFEF;
    ans = 64'h3FB0000007FFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"248777\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h406FFFBFFFFFF7FF;
    ReadData3E = 64'h403FE000000FFFFF;
    ans = 64'h403FE000000FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"249083\n");
    end
    ReadData1E = 64'hC00000080000000E;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"249389\n");
    end
    ReadData1E = 64'hBF9F2337AA50C4F6;
    ReadData2E = 64'h41FFFBFFFFFFFFF0;
    ReadData3E = 64'hC0BFFFFFDFFFFFBF;
    ans = 64'hC1AF1F93435B3ACE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"249695\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'hC3EE512EE6662CCD;
    ans = 64'hC3EE512EE6662CCD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"250001\n");
    end
    ReadData1E = 64'hDA1FE00000040000;
    ReadData2E = 64'h40301FFFFFFFFFDF;
    ReadData3E = 64'h0000000000000001;
    ans = 64'hDA600FE0000203DF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"250307\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"250613\n");
    end
    ReadData1E = 64'hC00FFFCFFFFFFFFF;
    ReadData2E = 64'hE97E000000800000;
    ReadData3E = 64'hBFEFFFFFF7FFF800;
    ans = 64'h699DFFD3007FFF3F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"250919\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBCAFFF7FFF800000;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h3FDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"251225\n");
    end
    ReadData1E = 64'hC02000000400001E;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hB80FFFFC0000001F;
    ans = 64'h3CD000000400001E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"251531\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC0A0004001FFFFFF;
    ReadData3E = 64'hC28F000000000100;
    ans = 64'hC28F000000000100;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"251837\n");
    end
    ReadData1E = 64'hC06292D76228137B;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000013;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"252143\n");
    end
    ReadData1E = 64'h3CA80000000000FE;
    ReadData2E = 64'h8017FB4D2CCAA271;
    ReadData3E = 64'h3F6000000010000F;
    ans = 64'h3F6000000010000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"252449\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'hBE11E49856E4D7E0;
    ans = 64'hBE11E49856E4D7E0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"252755\n");
    end
    ReadData1E = 64'hC3FBFFFEFFFFFFFE;
    ReadData2E = 64'hBFF0007FFFC00000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"253061\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"253367\n");
    end
    ReadData1E = 64'hC80EF613678AF683;
    ReadData2E = 64'hCD8172A803FEC66C;
    ReadData3E = 64'hC7EFFFFFFFFF8020;
    ans = 64'h55A0E1A983E624A4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"253673\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBF1291958811E1EE;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"253979\n");
    end
    ReadData1E = 64'hCABFFF0000003FFF;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h002FFFFFBFFFFFFC;
    ans = 64'h4A9FFF0000003FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"254285\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h6E3ABB40B7437E23;
    ReadData3E = 64'hCB5F39111E61024E;
    ans = 64'hCB5F39111E61024E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"254591\n");
    end
    ReadData1E = 64'hBCA100000000FFFE;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"254897\n");
    end
    ReadData1E = 64'hE3F00000001FBFFF;
    ReadData2E = 64'h4800FFFFFFFFFF7F;
    ReadData3E = 64'h0132E73035E25848;
    ans = 64'hEC0100000021BB7E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"255203\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'hB6B662EDDC60BE5B;
    ans = 64'hB6B662EDDC60BE5B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"255509\n");
    end
    ReadData1E = 64'hBF879A0574BE9D7E;
    ReadData2E = 64'h3FE5C6796BC3A859;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"255815\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"256121\n");
    end
    ReadData1E = 64'h3FC623839C67FBA6;
    ReadData2E = 64'h5FC855109CEE8C4A;
    ReadData3E = 64'hE8F0003FFFF00000;
    ans = 64'hE8F0003FFFF00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"256427\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h002493107A4421CE;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"256733\n");
    end
    ReadData1E = 64'hC02FFFFFE01FFFFF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h4DA0003FFFFFDFFF;
    ans = 64'h4DA0003FFFFFDFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"257039\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FEE00007FFFFFFF;
    ReadData3E = 64'h001E4E082A17FB0B;
    ans = 64'h0026A704350BFD84;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"257345\n");
    end
    ReadData1E = 64'hA240000000004000;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"257651\n");
    end
    ReadData1E = 64'h41B0000000000006;
    ReadData2E = 64'hC7EFF7FFFFFEFFFF;
    ReadData3E = 64'hB8139885504DA644;
    ans = 64'hC9AFF7FFFFFF000B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"257957\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hBDBBFFFFFFFFF7FE;
    ans = 64'hBDBBFFFFFFFFF7FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"258263\n");
    end
    ReadData1E = 64'h3FB0FFFFFFFFFEFF;
    ReadData2E = 64'h37E0000000001FFD;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"258569\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"258875\n");
    end
    ReadData1E = 64'h223E000000000007;
    ReadData2E = 64'hC020000003FFFFBE;
    ReadData3E = 64'h3F3FFFFFFE010000;
    ans = 64'h3F3FFFFFFE010000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"259181\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FFFFFFFEFF7FFFF;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h000FFFFFEFF7FFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"259487\n");
    end
    ReadData1E = 64'h403FFFFFFEE00000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h7FEFDFFFF7FFFFFF;
    ans = 64'h7FEFDFFFF7FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"259793\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h7FEA683B1EC46B7D;
    ReadData3E = 64'h37EF003FFFFFFFFF;
    ans = 64'h400A683B1EC46B7A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"260099\n");
    end
    ReadData1E = 64'h4000000000200002;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hC008000000200000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"260405\n");
    end
    ReadData1E = 64'hC03FFFC000000007;
    ReadData2E = 64'hC3C000FFFDFFFFFF;
    ReadData3E = 64'h3FE800000FFFFFFE;
    ans = 64'h441000DFFC000403;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"260711\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'hC3FE00FFFFFFFFFF;
    ans = 64'hC3FE00FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"261017\n");
    end
    ReadData1E = 64'hDC70000010020000;
    ReadData2E = 64'h40900D506FFEE146;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hDD100D50800E3360;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"261323\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"261629\n");
    end
    ReadData1E = 64'h5DA000FFFFFFFFFF;
    ReadData2E = 64'hBFE0080000000004;
    ReadData3E = 64'hBFF07FFFFFBFFFFF;
    ans = 64'hDD90090080000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"261935\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h401FFFFFF0000400;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h0041FFFFF80001FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"262241\n");
    end
    ReadData1E = 64'h412007FFFEFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h4012000040000000;
    ans = 64'hC13007FB7EFFEFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"262547\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h4380000000100003;
    ReadData3E = 64'hD43000000FFFFF7F;
    ans = 64'hD43000000FFFFF7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"262853\n");
    end
    ReadData1E = 64'hC03C5E2AE452C6D3;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h404C9E2AE452C6D1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"263159\n");
    end
    ReadData1E = 64'hC3DFAF53AFB9F351;
    ReadData2E = 64'hBFAFF8000000FFFF;
    ReadData3E = 64'h43CAD1C87055F27F;
    ans = 64'h43CEC6B56BAFD2C9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"263465\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'hC7E4008000000000;
    ans = 64'hC7E4008000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"263771\n");
    end
    ReadData1E = 64'h4033C0DD77566DA9;
    ReadData2E = 64'h47E001FFFFFFFE00;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h4823C355930555FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"264077\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hBFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"264383\n");
    end
    ReadData1E = 64'hBFC5C91E28AE3927;
    ReadData2E = 64'h800FDFFFFFFFFFFE;
    ReadData3E = 64'hC90FFF00003FFFFE;
    ans = 64'hC90FFF00003FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"264689\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h3FB000000000401F;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"264995\n");
    end
    ReadData1E = 64'h47F0007FFE000000;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h41EAF62DF12D1A41;
    ans = 64'hC800007FFE000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"265301\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFDFFFFFFE00003F;
    ReadData3E = 64'h41FFF7FFFFFC0000;
    ans = 64'h41FFF7FFFFFC0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"265607\n");
    end
    ReadData1E = 64'h2FC0000800001FFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"265913\n");
    end
    ReadData1E = 64'h380AEFDDEF1C3975;
    ReadData2E = 64'h4013173BC23790A3;
    ReadData3E = 64'h43CFC000000003FF;
    ans = 64'h43CFC000000003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"266219\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'h408000FFFFFFFFF8;
    ans = 64'h408000FFFFFFFFF8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"266525\n");
    end
    ReadData1E = 64'hB9FF000000020000;
    ReadData2E = 64'hC03DDB784BF264CB;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hC00FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"266831\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"267137\n");
    end
    ReadData1E = 64'hC1E444153B5F480B;
    ReadData2E = 64'h437CF734BD2A8438;
    ReadData3E = 64'hC88CB6C24967E7EC;
    ans = 64'hC88CB6C24967E7F5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"267443\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC1FC000000080000;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"267749\n");
    end
    ReadData1E = 64'hBFFFFFFF7FFFFF7F;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hC3F000000400FFFE;
    ans = 64'hC3F000000400FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"268055\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h48B0715E3CE4DAB7;
    ReadData3E = 64'h41E00000000000DF;
    ans = 64'h41E00000000000DF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"268361\n");
    end
    ReadData1E = 64'h4EB0FFFFFFFFFF80;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'hCEE0FFFFFFFFFF7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"268667\n");
    end
    ReadData1E = 64'h37FFFFFEEFFFFFFF;
    ReadData2E = 64'h802FFE00003FFFFE;
    ReadData3E = 64'hBFE9184238350997;
    ans = 64'hBFE9184238350997;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"268973\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h43CFCA9038906DA8;
    ans = 64'h43CFCA9038906DA8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"269279\n");
    end
    ReadData1E = 64'hBFDFFFFFF7FFE000;
    ReadData2E = 64'h5EC0000000002FFF;
    ReadData3E = 64'h4000000000000000;
    ans = 64'hDEAFFFFFF8003FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"269585\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"269891\n");
    end
    ReadData1E = 64'h43EFFFFFFFFC007E;
    ReadData2E = 64'h1B1FFBFFFFC00000;
    ReadData3E = 64'hC7E080000000003F;
    ans = 64'hC7E080000000003F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"270197\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hBFCDD72B2723C364;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"270503\n");
    end
    ReadData1E = 64'hC80FFFFEF0000000;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'hC01FFFFFFFF80007;
    ans = 64'h4B5FFFFEF0000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"270809\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h47E0000000403FFF;
    ReadData3E = 64'h4010F3DB191A1F8F;
    ans = 64'h4010F3DB191A1F8F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"271115\n");
    end
    ReadData1E = 64'h401FFFFC0007FFFF;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hC36FFFFC00080001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"271421\n");
    end
    ReadData1E = 64'h3FFFD00A729C4742;
    ReadData2E = 64'hC3CFFFFFFF1FFFFE;
    ReadData3E = 64'h3FC000000000BFFE;
    ans = 64'hC3DFD00A71BD96F7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"271727\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h43D5B0920D1BF88E;
    ans = 64'h43D5B0920D1BF88E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"272033\n");
    end
    ReadData1E = 64'h3CA27282D68BAADA;
    ReadData2E = 64'h400FFFF80003FFFF;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFEFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"272339\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h3CAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"272645\n");
    end
    ReadData1E = 64'h5F901FFFFFFFFFFE;
    ReadData2E = 64'h47EFFF000000FFFF;
    ReadData3E = 64'hBFDFFFFFF7FFFFFB;
    ans = 64'h67901F7F000080FD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"272951\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h657FFFFFFFFBFFF6;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"273257\n");
    end
    ReadData1E = 64'hC0300FFFFFFFFEFF;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'hC004003FFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"273563\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h43E1AA2F8A851184;
    ReadData3E = 64'hB8000000007FFFFD;
    ans = 64'hB8000000007FFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"273869\n");
    end
    ReadData1E = 64'h3F8FFFFDFFFFFE00;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'h0010000000000000;
    ans = 64'hFF7FFFFDFFFFFE02;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"274175\n");
    end
    ReadData1E = 64'hB7EFFF00000001FE;
    ReadData2E = 64'h77CFFFFFFFE10000;
    ReadData3E = 64'hA3A0000000007FBF;
    ans = 64'hEFCFFEFFFFE102F6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"274481\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hB93E92AC23EB1A80;
    ans = 64'hBFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"274787\n");
    end
    ReadData1E = 64'h3E2FFFFF00001FFF;
    ReadData2E = 64'h41E8C14DBD3E8F88;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h402AC14CF7343A5F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"275093\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hC00FFFFFFFFFFFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"275399\n");
    end
    ReadData1E = 64'h41C010000007FFFF;
    ReadData2E = 64'hC03FFE000000007E;
    ReadData3E = 64'hBFF0001FFFFFFFF7;
    ans = 64'hC2100EFF000BFFC6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"275705\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h1487869849373A4D;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"276011\n");
    end
    ReadData1E = 64'hC2780000000007FF;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'hC18FFC0000000004;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"276317\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h41758A7AFB86E8EF;
    ReadData3E = 64'h402C084092D67AFA;
    ans = 64'h402C084092D67AFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"276623\n");
    end
    ReadData1E = 64'h41CFFDFFFFFFFFFF;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h8000000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"276929\n");
    end
    ReadData1E = 64'hC7E14E030C337810;
    ReadData2E = 64'hB4508B55779CF496;
    ReadData3E = 64'h418FFFF00003FFFE;
    ans = 64'h418FFFF00003FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"277235\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'h78EFFFFFFF7F7FFF;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"277541\n");
    end
    ReadData1E = 64'hB8100800007FFFFF;
    ReadData2E = 64'h421BC643D44ACF51;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"277847\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h0010000000000000;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"278153\n");
    end
    ReadData1E = 64'h57F0400000000003;
    ReadData2E = 64'h3FEFF800000003FF;
    ReadData3E = 64'h3FDFFFFFFEFE0000;
    ans = 64'h57F03BF00000020A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"278459\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h4340007FFFFFFFEE;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"278765\n");
    end
    ReadData1E = 64'h33D0000003FE0000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h3FCFDFFFFFFDFFFF;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"279071\n");
    end
    ReadData1E = 64'h000FFFFFFFFFFFFE;
    ReadData2E = 64'h401000000000207F;
    ReadData3E = 64'hBFEFF7FFFFFFFFFF;
    ans = 64'hBFEFF7FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"279377\n");
    end
    ReadData1E = 64'h3EDFFF7BFFFFFFFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"279683\n");
    end
    ReadData1E = 64'hBFF007FFFFFFBFFF;
    ReadData2E = 64'hABFFFE00FFFFFFFE;
    ReadData3E = 64'h3FFFFFFFFFFFFBFB;
    ans = 64'h3FFFFFFFFFFFFBFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"279989\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h7FDFFFFFC0001000;
    ans = 64'h7FDFFFFFC0001000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"280295\n");
    end
    ReadData1E = 64'h401E7FA35DDC0D0A;
    ReadData2E = 64'hBFE0000400000040;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'hC00C7FAAFDC4E4FB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"280601\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h8000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"280907\n");
    end
    ReadData1E = 64'hD1925886906206D4;
    ReadData2E = 64'hC1E8EAED66B2863A;
    ReadData3E = 64'hC0202003FFFFFFFF;
    ans = 64'h538C9228BD9CEB27;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"281213\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC3F00000203FFFFE;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"281519\n");
    end
    ReadData1E = 64'h3FEFFFFF5FFFFFFF;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'h4029E3589C74D6C1;
    ans = 64'h4029E3589C74D6C1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"281825\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h4755B660F1AE640B;
    ReadData3E = 64'h3790000000001BFE;
    ans = 64'h3790000000001BFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"282131\n");
    end
    ReadData1E = 64'h380FF00FFFFFFFFF;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"282437\n");
    end
    ReadData1E = 64'hBFE200000000FFFE;
    ReadData2E = 64'h41FD6E4653630B63;
    ReadData3E = 64'h480FFFF008000000;
    ans = 64'h480FFFF008000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"282743\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'hC3E000000002003E;
    ans = 64'hC3E000000002003E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"283049\n");
    end
    ReadData1E = 64'hC020000400000004;
    ReadData2E = 64'h41064F83F491A335;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hC1364F898872A05F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"283355\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"283661\n");
    end
    ReadData1E = 64'hBFE5000000000000;
    ReadData2E = 64'hB49FEFFEFFFFFFFE;
    ReadData3E = 64'hC1CFDFFFFFF7FFFE;
    ans = 64'hC1CFDFFFFFF7FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"283967\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3FF1FFFFFFBFFFFF;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hBFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"284273\n");
    end
    ReadData1E = 64'hC16000000000021E;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h41C0000000000100;
    ans = 64'h41C0000000000100;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"284579\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC1D022D17DB65925;
    ReadData3E = 64'h7FDE0000FFFFFFFF;
    ans = 64'h7FDE0000FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"284885\n");
    end
    ReadData1E = 64'h3FDE0001FFFFFFFE;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"285191\n");
    end
    ReadData1E = 64'hC7FFC00000003FFF;
    ReadData2E = 64'h3D20000000021FFE;
    ReadData3E = 64'h2EEFFFFFFFFF6FFE;
    ans = 64'hC52FC0000004777B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"285497\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'hC1F0007DFFFFFFFE;
    ans = 64'hC1F0007DFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"285803\n");
    end
    ReadData1E = 64'hC173218EED2EAD60;
    ReadData2E = 64'hC620000040000080;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h47A3218F39B4E9AE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"286109\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"286415\n");
    end
    ReadData1E = 64'hC2A902A218479D05;
    ReadData2E = 64'hA0BFFFE020000000;
    ReadData3E = 64'h41DBFFFFFFFEFFFF;
    ans = 64'h41DBFFFFFFFEFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"286721\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h42F07DEEE25F8464;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"287027\n");
    end
    ReadData1E = 64'hC160000042000000;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h4013FFFFFFFFFFFF;
    ans = 64'h4013FFFFFFEFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"287333\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC16FFFDFFFFF7FFE;
    ReadData3E = 64'hC7FFFFFFF7FFFFFF;
    ans = 64'hC7FFFFFFF7FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"287639\n");
    end
    ReadData1E = 64'h7FD000000BFFFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h7C9000000BFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"287945\n");
    end
    ReadData1E = 64'hBFD0000010000008;
    ReadData2E = 64'hC1EFFFFFFBFFFFFD;
    ReadData3E = 64'h49600000001BFFFF;
    ans = 64'h49600000001BFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"288251\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hFFF003FC00000000;
    ans = 64'hFFF803FC00000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"288557\n");
    end
    ReadData1E = 64'hC06FDFBFFFFFFFFE;
    ReadData2E = 64'h002000011FFFFFFF;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h809FEFC23DBB7FFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"288863\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"289169\n");
    end
    ReadData1E = 64'h842FFFFFFFFEFFFC;
    ReadData2E = 64'hFFFFFFDFFFFFFFFD;
    ReadData3E = 64'h1E80000200000010;
    ans = 64'hFFFFFFDFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"289475\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h9590000001F7FFFF;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"289781\n");
    end
    ReadData1E = 64'h62DE00000007FFFF;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'h3FEFFFFBFFFFDFFF;
    ans = 64'h62BE000000080001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"290087\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h41E0000000200007;
    ReadData3E = 64'h414302D9B4ED0185;
    ans = 64'h414302D9B4ED0185;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"290393\n");
    end
    ReadData1E = 64'hB7F0FFFFFF7FFFFE;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"290699\n");
    end
    ReadData1E = 64'hBFFE0FFFFFFFFFFF;
    ReadData2E = 64'hBD203FFFFFFFF800;
    ReadData3E = 64'hC311550C9132B40C;
    ans = 64'hC311550C9132B40C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"291005\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'hC3DFFFF7FFFFC000;
    ans = 64'hC3DFFFF7FFFFC000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"291311\n");
    end
    ReadData1E = 64'hFFE000007FFFFFFA;
    ReadData2E = 64'h3FB3EB2F6685C313;
    ReadData3E = 64'h0000000000000000;
    ans = 64'hFFA3EB3005DF3E40;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"291617\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"291923\n");
    end
    ReadData1E = 64'hB8E701B553E06384;
    ReadData2E = 64'h001000FFFFFFFFFB;
    ReadData3E = 64'hBF9FFFFFFFFFDF80;
    ans = 64'hBF9FFFFFFFFFDF80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"292229\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3AEFFFFFFFFFBEFF;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"292535\n");
    end
    ReadData1E = 64'hC7EFF7FFFDFFFFFF;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'h45BFFFFFF7FFFE00;
    ans = 64'hC7DFF7FFFDF7FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"292841\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h41DFFFF3FFFFFFFE;
    ReadData3E = 64'hC0100000BFFFFFFF;
    ans = 64'hC0100000BFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"293147\n");
    end
    ReadData1E = 64'h319000080000000F;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"293453\n");
    end
    ReadData1E = 64'h37EFFFFFFFFF0FFF;
    ReadData2E = 64'h429FFC0003FFFFFF;
    ReadData3E = 64'h41CFFFF3FFFFFFFF;
    ans = 64'h41CFFFF3FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"293759\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'hF2600100000FFFFE;
    ans = 64'hF2600100000FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"294065\n");
    end
    ReadData1E = 64'h43403FFFFFFFDFFF;
    ReadData2E = 64'h3F900001EFFFFFFF;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"294371\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"294677\n");
    end
    ReadData1E = 64'hDFC0000000040000;
    ReadData2E = 64'h380E000000000007;
    ReadData3E = 64'h434FF7FFFEFFFFFF;
    ans = 64'hD7DE000000078007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"294983\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3FD744F5F33CC055;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"295289\n");
    end
    ReadData1E = 64'hC7E6D8109ACDBE41;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hC02DFFBFFFFFFFFF;
    ans = 64'hC7E6D8109ACDBE41;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"295595\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC020001003FFFFFF;
    ReadData3E = 64'hC050003F7FFFFFFE;
    ans = 64'hC050003F7FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"295901\n");
    end
    ReadData1E = 64'h434FFBFFFFFFFFE0;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hC000000000000000;
    ans = 64'h434FFBFFFFFFFFDF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"296207\n");
    end
    ReadData1E = 64'hC6E0001FFFFFDFFE;
    ReadData2E = 64'hC036D93D56E70AAC;
    ReadData3E = 64'h439FFFFFF0000003;
    ans = 64'h4726D96B09618AC5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"296513\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'hBFB00007FFFFFFFB;
    ans = 64'hBFB00007FFFFFFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"296819\n");
    end
    ReadData1E = 64'h4025AAF27EF72E72;
    ReadData2E = 64'h206B033642E4A6A9;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"297125\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"297431\n");
    end
    ReadData1E = 64'h938020FFFFFFFFFF;
    ReadData2E = 64'h001AFF91BAC0C01C;
    ReadData3E = 64'h3FA9ED8FA7758F35;
    ans = 64'h3FA9ED8FA7758F35;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"297737\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC000000000001080;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h8000000000002102;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"298043\n");
    end
    ReadData1E = 64'h402311AC49E54B80;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'hBEE0100000000002;
    ans = 64'h403311ABC9654B7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"298349\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFF3FFFFDFFFFFFF;
    ReadData3E = 64'hB81081FFFFFFFFFE;
    ans = 64'hB81081FFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"298655\n");
    end
    ReadData1E = 64'h43FFFFFF7FFFF000;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h440FFFFF7FFFF000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"298961\n");
    end
    ReadData1E = 64'h002632D8A1656279;
    ReadData2E = 64'hC12BFFFFDFFFFFFE;
    ReadData3E = 64'h7BF57EED210A57A2;
    ans = 64'h7BF57EED210A57A2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"299267\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'h40B1000003FFFFFE;
    ans = 64'h40B1000003FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"299573\n");
    end
    ReadData1E = 64'h7FF0200000001FFF;
    ReadData2E = 64'h400FEFFFFFFFE000;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h7FF8200000001FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"299879\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"300185\n");
    end
    ReadData1E = 64'h49A0002000000FFF;
    ReadData2E = 64'h43F0000000BFFFFF;
    ReadData3E = 64'hFFDF78B1B0067614;
    ans = 64'hFFDF78B1B0067614;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"300491\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC1C00000000001FC;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h81E00000008001FC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"300797\n");
    end
    ReadData1E = 64'hC00080000003FFFE;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'hC7EF484B79B51CC3;
    ans = 64'hC7EF484B79B51CC3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"301103\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h409FFEF7FFFFFFFF;
    ReadData3E = 64'hB7F0000000000F00;
    ans = 64'hB7F0000000000F00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"301409\n");
    end
    ReadData1E = 64'hB7FFFFFFFFDFF000;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"301715\n");
    end
    ReadData1E = 64'hC1E7B18F153C6AE9;
    ReadData2E = 64'h3FE000000040000F;
    ReadData3E = 64'hA75000FFFFFE0000;
    ans = 64'hC1D7B18F159B313C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"302021\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'hBB10000000007FBE;
    ans = 64'hBB10000000007FBE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"302327\n");
    end
    ReadData1E = 64'h7FD00000000E0000;
    ReadData2E = 64'hB33E000001FFFFFF;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hF31E0000021A3FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"302633\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"302939\n");
    end
    ReadData1E = 64'h3FCFFC0000000003;
    ReadData2E = 64'hD080E00000000000;
    ReadData3E = 64'h41ED6C31252172C3;
    ans = 64'hD060DDE400000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"303245\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFF613FC6392E9AA;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h801613FC6392E9A9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"303551\n");
    end
    ReadData1E = 64'h40407FFFFFFFBFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hC7F0000FFC000000;
    ans = 64'hC7F0000FFC000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"303857\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h416FFFFFFFFFDFF6;
    ReadData3E = 64'h40DE7B3AA73E78BE;
    ans = 64'h40DE7B3AA73E78BE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"304163\n");
    end
    ReadData1E = 64'h380FFFFEFFC00000;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"304469\n");
    end
    ReadData1E = 64'h6690003F7FFFFFFF;
    ReadData2E = 64'hFFEFFDFFDFFFFFFF;
    ReadData3E = 64'hBA300000407FFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"304775\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'hC1E0020000000200;
    ans = 64'hC1E0020000000200;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"305081\n");
    end
    ReadData1E = 64'h890BFFFFFFF80000;
    ReadData2E = 64'h7FD0001000000002;
    ReadData3E = 64'h4010000000000001;
    ans = 64'hC8EC001BFFF7FFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"305387\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"305693\n");
    end
    ReadData1E = 64'h4080000007FFFEFE;
    ReadData2E = 64'hC80FE00000000004;
    ReadData3E = 64'h001FFBBFFFFFFFFF;
    ans = 64'hC89FE0000FEFFE02;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"305999\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hB53FFFFFFFF7BFFE;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"306305\n");
    end
    ReadData1E = 64'hC1BBFFFFFFFFF800;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'hC01200007FFFFFFE;
    ans = 64'hC50BFFFFFFFFF802;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"306611\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC0E8639CEABADD60;
    ReadData3E = 64'h407FFFFFF000000E;
    ans = 64'h407FFFFFF000000E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"306917\n");
    end
    ReadData1E = 64'h403E9B00B0A9304B;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'h439E9B00B0A9304A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"307223\n");
    end
    ReadData1E = 64'h480FFFFDFFF80000;
    ReadData2E = 64'h3F41BFFFFFFFFFFF;
    ReadData3E = 64'hC1C000000000FFEF;
    ans = 64'h4761BFFEE3FB8FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"307529\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h3CE80001FFFFFFFE;
    ans = 64'h3CE80001FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"307835\n");
    end
    ReadData1E = 64'h7FD0002FFFFFFFFF;
    ReadData2E = 64'hB8700000800FFFFF;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hF85000308011802E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"308141\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h4004000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"308447\n");
    end
    ReadData1E = 64'h3FEFFFFFFFFFF807;
    ReadData2E = 64'h403000FFFFFFFFFD;
    ReadData3E = 64'h8000000000FFFFFB;
    ans = 64'h403000FFFFFFFC00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"308753\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC00BC838B3BBB5A9;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"309059\n");
    end
    ReadData1E = 64'hBFA007FFFFFFFBFF;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h3FB00000000201FF;
    ans = 64'hFF9007FFFFFFFC00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"309365\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h48B5B44331874F02;
    ReadData3E = 64'h41CFFE00FFFFFFFF;
    ans = 64'h41CFFE00FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"309671\n");
    end
    ReadData1E = 64'h000FFFFFFFFFDFBE;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h400FFFFFFFFFBF7B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"309977\n");
    end
    ReadData1E = 64'h41CF7FFFFFFFFFF0;
    ReadData2E = 64'h7FEFFDFFFEFFFFFE;
    ReadData3E = 64'h3FE96EA68E3FBB6C;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"310283\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'hBF801FFFFFFFFEFF;
    ans = 64'h400FEFE000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"310589\n");
    end
    ReadData1E = 64'h00201000007FFFFE;
    ReadData2E = 64'h480CFFB7FE0DACA9;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"310895\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'h400DFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"311201\n");
    end
    ReadData1E = 64'hCE44B2478F9C2274;
    ReadData2E = 64'h434003FDFFFFFFFF;
    ReadData3E = 64'hD79800000001FFFE;
    ans = 64'hD79800000001FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"311507\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC004F9F07751D02C;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"311813\n");
    end
    ReadData1E = 64'h2DB0041FFFFFFFFF;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h3FE960A47AF95278;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"312119\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h7FE00003FFFFFFF6;
    ReadData3E = 64'h7FE0013FFFFFFFFE;
    ans = 64'h7FE0013FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"312425\n");
    end
    ReadData1E = 64'hBFFFFE0000003FFE;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"312731\n");
    end
    ReadData1E = 64'hC02FFFFFFFBFFFFF;
    ReadData2E = 64'hB39FFFFFFFFFF808;
    ReadData3E = 64'h43FFFFF000002000;
    ans = 64'h43FFFFF000002000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"313037\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'hC03FBF7FFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"313343\n");
    end
    ReadData1E = 64'h381FFFEEFFFFFFFF;
    ReadData2E = 64'hB7EFF80000008000;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"313649\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"313955\n");
    end
    ReadData1E = 64'h3FD000003FFFFE00;
    ReadData2E = 64'h4BC000007FFFC000;
    ReadData3E = 64'hBFDFAA56060806BE;
    ans = 64'h4BA00000BFFFC000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"314261\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hFFEFFFFFFFEFFFFC;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"314567\n");
    end
    ReadData1E = 64'hC0307FEFFFFFFFFF;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'hBDD00000080FFFFF;
    ans = 64'hBDD00000080FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"314873\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC3FFFC0000FFFFFE;
    ReadData3E = 64'hBFCFFFFEFFEFFFFE;
    ans = 64'hBFCFFFFEFFEFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"315179\n");
    end
    ReadData1E = 64'h40C34F729EDBEC45;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h000FFFFFFFFFD95F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"315485\n");
    end
    ReadData1E = 64'hC03FFFEFFFFFFFDE;
    ReadData2E = 64'h8010800000004000;
    ReadData3E = 64'h41CC0000000003FF;
    ans = 64'h41CC0000000003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"315791\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h43FFFFFFEEFFFFFF;
    ans = 64'h43FFFFFFEEFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"316097\n");
    end
    ReadData1E = 64'h41C00020000000FF;
    ReadData2E = 64'hC07A332FE70D0382;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'hC24A33644D6C933E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"316403\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h801FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"316709\n");
    end
    ReadData1E = 64'h3D6FFFFE000FFFFE;
    ReadData2E = 64'hC03000407FFFFFFF;
    ReadData3E = 64'h402FFFFFF80001FE;
    ans = 64'h402FFFFFF7FFE1FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"317015\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h41D524236A0D87CB;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h401FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"317321\n");
    end
    ReadData1E = 64'h41C807C515B87415;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'hCD0080000001FFFF;
    ans = 64'hCD0080000001FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"317627\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h3F4400000000001E;
    ReadData3E = 64'hC1DFFFFFFFFBFFFA;
    ans = 64'hC1DFFFFFFFFBFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"317933\n");
    end
    ReadData1E = 64'h41CFF80000000007;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h81EFF80000000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"318239\n");
    end
    ReadData1E = 64'hC03003FFF7FFFFFE;
    ReadData2E = 64'hC80F8FFFFFFFFFFF;
    ReadData3E = 64'h3FBDFFFFFF800000;
    ans = 64'h484F97E3F037FFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"318545\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h38E0000FFFFFFF80;
    ans = 64'h38E0000FFFFFFF80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"318851\n");
    end
    ReadData1E = 64'hBFCFFFFFFC000008;
    ReadData2E = 64'h1ACFFFFFFF800010;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"319157\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h000FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"319463\n");
    end
    ReadData1E = 64'hBFC195D722242CFC;
    ReadData2E = 64'h43FA8BAAFEAC312F;
    ReadData3E = 64'h7FF44DE4CC2EAF61;
    ans = 64'h7FFC4DE4CC2EAF61;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"319769\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hFFF000000003FFFB;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hFFF800000003FFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"320075\n");
    end
    ReadData1E = 64'h43F00000007FFFFE;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'hBFBFFFF7FFEFFFFF;
    ans = 64'hBFBFFFF7FFEFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"320381\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h41EBFFFFFC000000;
    ReadData3E = 64'h41403FEFFFFFFFFE;
    ans = 64'h41403FEFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"320687\n");
    end
    ReadData1E = 64'h381FF7FFFFFFFFDF;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"320993\n");
    end
    ReadData1E = 64'h300EFFFFFFFFFFDE;
    ReadData2E = 64'h41FFF7FFDFFFFFFF;
    ReadData3E = 64'h404FFFFFFF9FFFFF;
    ans = 64'h404FFFFFFF9FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"321299\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'h43B00000007FFFF8;
    ans = 64'h43B00000007FFFF8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"321605\n");
    end
    ReadData1E = 64'h31C003FFFFFBFFFE;
    ReadData2E = 64'h3FB929CE7C20406E;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"321911\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h8000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"322217\n");
    end
    ReadData1E = 64'hB7F0000000000FFC;
    ReadData2E = 64'hC3EEE26ECB4D48FA;
    ReadData3E = 64'h41E01FFFFF7FFFFE;
    ans = 64'h41E01FFFFF7FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"322523\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h5E3FB1EF5708D61C;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"322829\n");
    end
    ReadData1E = 64'hC09000003FFFBFFF;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'hC02FFFFFFFFE01FF;
    ans = 64'hC02FFFFFFFFE017F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"323135\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h20BDA441CA99D67B;
    ReadData3E = 64'h3FFFFFFFFFFFFFFD;
    ans = 64'h3FFFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"323441\n");
    end
    ReadData1E = 64'h41C00001FE000000;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"323747\n");
    end
    ReadData1E = 64'hC02FFF0000000003;
    ReadData2E = 64'hD7F000000000007A;
    ReadData3E = 64'h316B40A4972E3E8C;
    ans = 64'h582FFF00000000F7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"324053\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hC00FFFFC00001FFF;
    ans = 64'hC00FFFFC00001FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"324359\n");
    end
    ReadData1E = 64'h401007FFDFFFFFFE;
    ReadData2E = 64'h41EFFFFFFFFFFF7E;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'h421007FFDFFFFFBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"324665\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"324971\n");
    end
    ReadData1E = 64'hB81F775CD1DFBCE5;
    ReadData2E = 64'hC5AFFFFFFFEFFFFE;
    ReadData3E = 64'h41DFFFFFF00FFFFE;
    ans = 64'h41DFFFFFF00FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"325277\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFE72B4771A92771;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"325583\n");
    end
    ReadData1E = 64'h43EFFFFFC00007FE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'hC34F000003FFFFFF;
    ans = 64'hC3E007BFE00103FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"325889\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC0DFF000003FFFFF;
    ReadData3E = 64'hA751FFFFFFFFFF7E;
    ans = 64'hA751FFFFFFFFFF7E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"326195\n");
    end
    ReadData1E = 64'h3F243DDDF2B81DCE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"326501\n");
    end
    ReadData1E = 64'hBD86FC7AA6E86ECB;
    ReadData2E = 64'hC3007FFFFFFFFFFF;
    ReadData3E = 64'hBFF00101FFFFFFFF;
    ans = 64'h4097B05E3B9FB240;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"326807\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h401FFFF800FFFFFF;
    ans = 64'h401FFFF800FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"327113\n");
    end
    ReadData1E = 64'h40F0020000000007;
    ReadData2E = 64'hBA10000003FFF000;
    ReadData3E = 64'h0010000000000001;
    ans = 64'hBB10020004007005;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"327419\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"327725\n");
    end
    ReadData1E = 64'hFFD0023FFFFFFFFF;
    ReadData2E = 64'hB80FFFFE000000FE;
    ReadData3E = 64'hBFF0000002080000;
    ans = 64'h77F0023EFFDC007E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"328031\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h402FF0000001FFFF;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"328337\n");
    end
    ReadData1E = 64'h001EFFFFFFFFFFFA;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h41EFFFFFFFFF7FFA;
    ans = 64'h41EFFFFFFFFF7FFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"328643\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h400FFFBFFFFFFFFE;
    ReadData3E = 64'hBFEF000000000006;
    ans = 64'hBFEF000000000006;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"328949\n");
    end
    ReadData1E = 64'h2CA00001DFFFFFFE;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"329255\n");
    end
    ReadData1E = 64'h3FBFFFFC03FFFFFE;
    ReadData2E = 64'h41D2003FFFFFFFFF;
    ReadData3E = 64'hB81FFF7FBFFFFFFF;
    ans = 64'h41A2003DC23807FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"329561\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'hBFC8D4B898AE6753;
    ans = 64'hBFC8D4B898AE6753;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"329867\n");
    end
    ReadData1E = 64'h37F0811ECCA520CC;
    ReadData2E = 64'hB5EC0F0DCEE7205D;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'hADECF17D3156B094;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"330173\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"330479\n");
    end
    ReadData1E = 64'h47E800000003FFFF;
    ReadData2E = 64'h002D21C92796619C;
    ReadData3E = 64'h8020000000000000;
    ans = 64'h0825D956DDB46D6D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"330785\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h380FFFFFFFFFFFFF;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"331091\n");
    end
    ReadData1E = 64'h0010000000010002;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'hBFC00000003FEFFF;
    ans = 64'hBFC00000003FEFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"331397\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC7F3FFFFFF000000;
    ReadData3E = 64'hC1AFFFFFFF7DFFFF;
    ans = 64'hC1AFFFFFFF7DFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"331703\n");
    end
    ReadData1E = 64'hC8BFFFFFFFFBFDFE;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'hC010000000000001;
    ans = 64'h48BFFFFFFFFBFE00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"332009\n");
    end
    ReadData1E = 64'hBE6F80000000007F;
    ReadData2E = 64'h3E8B72F2BC14CB1E;
    ReadData3E = 64'hC230976CFE95CEB9;
    ans = 64'hC230976CFE95CEB9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"332315\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'hC44FFFFF7FFFFF80;
    ans = 64'hC44FFFFF7FFFFF80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"332621\n");
    end
    ReadData1E = 64'hC1F01FFFFFFF0000;
    ReadData2E = 64'hF298EF2E7DB85567;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"332927\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"333233\n");
    end
    ReadData1E = 64'hBF5FD227B86CA879;
    ReadData2E = 64'h40200000003FFFFC;
    ReadData3E = 64'hD9BFFFFC1FFFFFFF;
    ans = 64'hD9BFFFFC1FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"333539\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h403D53A968AE85F4;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"333845\n");
    end
    ReadData1E = 64'h401FFFFDC0000000;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'hCF8F800007FFFFFF;
    ans = 64'hCF8F800007FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"334151\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC1D0803FFFFFFFFF;
    ReadData3E = 64'h41D9081CD2377743;
    ans = 64'h41D9081CD2377743;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"334457\n");
    end
    ReadData1E = 64'h78001E0C150FC0CA;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'hF8101E0C150FC0CB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"334763\n");
    end
    ReadData1E = 64'hC1E0000FFFFFFEFF;
    ReadData2E = 64'hBFF20000000003FF;
    ReadData3E = 64'h3CA2B9DBAAB26581;
    ans = 64'h41E20012000002DE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"335069\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h65E05E479AF5AEB5;
    ans = 64'h65E05E479AF5AEB5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"335375\n");
    end
    ReadData1E = 64'hC1F00000000400FF;
    ReadData2E = 64'h400F51AB9CFD5497;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"335681\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"335987\n");
    end
    ReadData1E = 64'h3ECFFC0000000010;
    ReadData2E = 64'hC0FFFFF8000FFFFF;
    ReadData3E = 64'hBFCFFF7FFFFBFFFF;
    ans = 64'hBFE7FDDC0086FF07;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"336293\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC0000001000001FE;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"336599\n");
    end
    ReadData1E = 64'h37EFFFFC00000004;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'hC1CFFFFFBF800000;
    ans = 64'hC1CFFFFFBF800000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"336905\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC5C679594BB014C1;
    ReadData3E = 64'h3380000203FFFFFF;
    ans = 64'h3380000203FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"337211\n");
    end
    ReadData1E = 64'h403585434B1B77D6;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hC05605434B1B77D6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"337517\n");
    end
    ReadData1E = 64'hBE3FFFFFFFFFE07F;
    ReadData2E = 64'h3F9FFFFFFFF87FFF;
    ReadData3E = 64'h3EA000000003FFBF;
    ans = 64'h3E9FFC0000080072;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"337823\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hC07238B056832559;
    ans = 64'hC07238B056832559;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"338129\n");
    end
    ReadData1E = 64'hC7FF635B2DA8AC6C;
    ReadData2E = 64'hBF300001FFFE0000;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"338435\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h400FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"338741\n");
    end
    ReadData1E = 64'h8004E9A86C24788F;
    ReadData2E = 64'h40100000FFFFFC00;
    ReadData3E = 64'h41BFDF7FFFFFFFFF;
    ans = 64'h41BFDF7FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"339047\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBF3FFF00000FFFFE;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h001FFE000FFFFEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"339353\n");
    end
    ReadData1E = 64'hB63FFFFFFFFFFDFF;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hA5BE0000000003FF;
    ans = 64'h366FFFFFFFFFFDFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"339659\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hB9F000000007FFFA;
    ReadData3E = 64'hC8061A6CA9E4273A;
    ans = 64'hC8061A6CA9E4273A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"339965\n");
    end
    ReadData1E = 64'h7FF0010000000200;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h7FF8010000000200;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"340271\n");
    end
    ReadData1E = 64'hC140400000000007;
    ReadData2E = 64'hC747571F3869690F;
    ReadData3E = 64'h43400003FFF7FFFF;
    ans = 64'h4897B47BB54B0EBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"340577\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'hC2BFBFFEFFFFFFFF;
    ans = 64'hC2BFBFFEFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"340883\n");
    end
    ReadData1E = 64'hBFFFFFFFC00003FF;
    ReadData2E = 64'hC01274DD0FBADA9C;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h4350000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"341189\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"341495\n");
    end
    ReadData1E = 64'h380F001FFFFFFFFF;
    ReadData2E = 64'hAC0FDFFFFC000000;
    ReadData3E = 64'hBFC0000000004007;
    ans = 64'hBFC0000000004007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"341801\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC0A000103FFFFFFE;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h80C002103FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"342107\n");
    end
    ReadData1E = 64'hB59FC00000000003;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'hB80FFFFFFFFFFFDA;
    ans = 64'h38FFBFC000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"342413\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBCC3CFBD727D2506;
    ReadData3E = 64'hC1C0100000001FFF;
    ans = 64'hC1C0100000001FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"342719\n");
    end
    ReadData1E = 64'hFFEE000080000000;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"343025\n");
    end
    ReadData1E = 64'h39EFFFF000000040;
    ReadData2E = 64'hBFA0000000020FFF;
    ReadData3E = 64'hBD117042903CB8AA;
    ans = 64'hBD117042903CB8AA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"343331\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'h42800000017FFFFF;
    ans = 64'h42800000017FEFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"343637\n");
    end
    ReadData1E = 64'hB9BFFFFFE0004000;
    ReadData2E = 64'hBF90200000000007;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"343943\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'hBFF0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"344249\n");
    end
    ReadData1E = 64'h3FF00080003FFFFF;
    ReadData2E = 64'h43EFFFFF7FFFFDFF;
    ReadData3E = 64'hBFFFFFFF7F7FFFFF;
    ans = 64'h43F0007FC03DFEFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"344555\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC1CFFFFFFFC0FFFE;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h81EFFFFFFFC0FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"344861\n");
    end
    ReadData1E = 64'hB02FFBFFFEFFFFFF;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h2F20000000100000;
    ans = 64'h702FFBFFFEFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"345167\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hC1FFFFFFF8000010;
    ReadData3E = 64'h0010000840000000;
    ans = 64'h821FFFFFF7F00008;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"345473\n");
    end
    ReadData1E = 64'h381007C000000000;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'hF81007BFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"345779\n");
    end
    ReadData1E = 64'hBF562488B8D64371;
    ReadData2E = 64'h92A0000017FFFFFF;
    ReadData3E = 64'h39EFDFFFFFFFDFFE;
    ans = 64'h39EFDFFFFFFFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"346085\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'h403E715430E93B27;
    ans = 64'h403A715430E93B27;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"346391\n");
    end
    ReadData1E = 64'h3A90000000BFFFFF;
    ReadData2E = 64'hC1D671BF3B5F1C48;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"346697\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"347003\n");
    end
    ReadData1E = 64'hB7DFFFFFFFFFFE00;
    ReadData2E = 64'h3F907FFFFFFFF7FE;
    ReadData3E = 64'hBF93C0A23CC11A26;
    ans = 64'hBF93C0A23CC11A26;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"347309\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'h7FF00017FFFFFFFF;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF80017FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"347615\n");
    end
    ReadData1E = 64'h3FC0000000050000;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'h4025438262DA06B0;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"347921\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hBD9100000000003F;
    ReadData3E = 64'hC01FFFFE003FFFFF;
    ans = 64'hC01FFFFE003FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"348227\n");
    end
    ReadData1E = 64'hBFD00001FFFFFE00;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"348533\n");
    end
    ReadData1E = 64'h000FFFFFFFDFFEFF;
    ReadData2E = 64'hBCA00003FFFFFFF0;
    ReadData3E = 64'h41E207FFFFFFFFFF;
    ans = 64'h41E207FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"348839\n");
    end
    ReadData1E = 64'h0010000000000000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h3FC4726CAF9F0AC5;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"349145\n");
    end
    ReadData1E = 64'hC1EC4292DAD85B10;
    ReadData2E = 64'h0FD00007FFFEFFFE;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"349451\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h3FDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"349757\n");
    end
    ReadData1E = 64'hC510000000200007;
    ReadData2E = 64'h37E8254245857C30;
    ReadData3E = 64'h215001FFFFFFF7FE;
    ans = 64'hBD08254245B5C6BF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"350063\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"350369\n");
    end
    ReadData1E = 64'hBFC000000FFFFFC0;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h40CC6452234E22D5;
    ans = 64'h40CC6452234E22D5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"350675\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h313FFFFF8000000E;
    ReadData3E = 64'h4090400000008000;
    ans = 64'h4090400000008000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"350981\n");
    end
    ReadData1E = 64'h43F5973D6929059F;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"351287\n");
    end
    ReadData1E = 64'hFA5618AB76807DB4;
    ReadData2E = 64'hA9B684B73C005DA3;
    ReadData3E = 64'h40CFFBFFFDFFFFFF;
    ans = 64'h641F19345688708F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"351593\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'hBFD40007FFFFFFFF;
    ans = 64'hBFD40007FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"351899\n");
    end
    ReadData1E = 64'hBE50000000000401;
    ReadData2E = 64'hC00FFFF03FFFFFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h4000000007FFFC10;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"352205\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"352511\n");
    end
    ReadData1E = 64'hE0B100000000003F;
    ReadData2E = 64'hC10F800000003FFE;
    ReadData3E = 64'hBFDFFFFFFFFFDFFD;
    ans = 64'h61D0BC000000223D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"352817\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FEFFFFFFFFFC002;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h434FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"353123\n");
    end
    ReadData1E = 64'h417FFF7FFFFFFF00;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h37FA53BB21817EC2;
    ans = 64'h37FA53BB21817EC2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"353429\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3DC39B8B3CE1D868;
    ReadData3E = 64'hC96FFFFFFBFFE000;
    ans = 64'hC96FFFFFFBFFE000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"353735\n");
    end
    ReadData1E = 64'h3FF00000007FFFF7;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h00000000007FFFF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"354041\n");
    end
    ReadData1E = 64'h400FFFFFFFEFF000;
    ReadData2E = 64'hB7FFFFFFFFFFF040;
    ReadData3E = 64'hC1EFFFFFFF01FFFF;
    ans = 64'hC1EFFFFFFF01FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"354347\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h312EFFFFFFFFFFFF;
    ans = 64'h312EFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"354653\n");
    end
    ReadData1E = 64'hC3F003FFFF7FFFFF;
    ReadData2E = 64'hC1FBFFFFFFFEFFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'h45FC06FFFF1EFFBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"354959\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"355265\n");
    end
    ReadData1E = 64'h4EA00000000007FE;
    ReadData2E = 64'h43EE5D7578A9E4C9;
    ReadData3E = 64'hC0108000000003FF;
    ans = 64'h529E5D7578A9F3F4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"355571\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC00000DFFFFFFFFF;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hC340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"355877\n");
    end
    ReadData1E = 64'h400FFEDFFFFFFFFF;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hC00FFFFFFBFFFBFE;
    ans = 64'hC00FFFFFFBFFFBFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"356183\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hCF22DB34B40B610A;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"356489\n");
    end
    ReadData1E = 64'h794CFFFFFFFFFFFE;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h75FD000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"356795\n");
    end
    ReadData1E = 64'hBFDFFFFFF0001000;
    ReadData2E = 64'h43CFFFC0003FFFFF;
    ReadData3E = 64'hC7F31244A4CB68E0;
    ans = 64'hC7F31244A4CB68E0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"357101\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'hC7E000200000003F;
    ans = 64'hC7E000200000003F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"357407\n");
    end
    ReadData1E = 64'hFBA7EF96594DB2B4;
    ReadData2E = 64'hC37FFFFFFFFFEBFE;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h7F37EF96594DA3BD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"357713\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h8010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"358019\n");
    end
    ReadData1E = 64'h4630000000FFFFFF;
    ReadData2E = 64'h41E07FFFFFFFBFFF;
    ReadData3E = 64'h3FBFFFAFFFFFFFFF;
    ans = 64'h482080000107BFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"358325\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3ABFFFFC40000000;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h4010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"358631\n");
    end
    ReadData1E = 64'h436F7FFDFFFFFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hC01020007FFFFFFF;
    ans = 64'h40276FFDBFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"358937\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC1E0008000002000;
    ReadData3E = 64'h7FD0000002200000;
    ans = 64'h7FD0000002200000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"359243\n");
    end
    ReadData1E = 64'hB7EFFFFFFFFFD000;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"359549\n");
    end
    ReadData1E = 64'h3FF008003FFFFFFF;
    ReadData2E = 64'hC00FFFF0007FFFFF;
    ReadData3E = 64'h80DFFC0000000003;
    ans = 64'hC01007F83C3FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"359855\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hBFB46BA6BE17B26F;
    ans = 64'hBFB46BA6BE17B26F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"360161\n");
    end
    ReadData1E = 64'h42A2C1EA3B54D9AE;
    ReadData2E = 64'h7FE00000007FFFE0;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"360467\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h0017FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"360773\n");
    end
    ReadData1E = 64'h40086C9CAE34D324;
    ReadData2E = 64'hC00FFFBFFBFFFFFF;
    ReadData3E = 64'hC020000001FFDFFF;
    ans = 64'hC0343635E9F6E191;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"361079\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3E36E176EB8922BE;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"361385\n");
    end
    ReadData1E = 64'hC0600007FFFFBFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h3FB47FFFFFFFFFFE;
    ans = 64'hC04FF5CFFFFF7FFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"361691\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC7CFFFFFF0000003;
    ReadData3E = 64'hB7EFFFFFFFFF8800;
    ans = 64'hB7EFFFFFFFFF8800;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"361997\n");
    end
    ReadData1E = 64'hBCA0807FFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"362303\n");
    end
    ReadData1E = 64'h4032001FFFFFFFFF;
    ReadData2E = 64'hE0300040003FFFFE;
    ReadData3E = 64'hC8664720EB83783B;
    ans = 64'hE072006800C8007D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"362609\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'h3FF020003FFFFFFF;
    ans = 64'h3FF020003FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"362915\n");
    end
    ReadData1E = 64'h40700000000FFFFE;
    ReadData2E = 64'h48028D94CD5609B6;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h48828D94CD689748;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"363221\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"363527\n");
    end
    ReadData1E = 64'h41E932AE9C561D16;
    ReadData2E = 64'hBFBE902517750DF2;
    ReadData3E = 64'hC1FFFFFFFFFFFE02;
    ans = 64'hC200C08827CB14EC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"363833\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FFFF80080000000;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"364139\n");
    end
    ReadData1E = 64'h4180200000200000;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h37E6BEA376F569FE;
    ans = 64'h41802000001FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"364445\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h0012DB0FA04B1CAC;
    ReadData3E = 64'h3D70000020004000;
    ans = 64'h3D70000020004000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"364751\n");
    end
    ReadData1E = 64'h41EFFFFFF800FFFE;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"365057\n");
    end
    ReadData1E = 64'hC7EFF00FFFFFFFFE;
    ReadData2E = 64'hC1FFFFFFDFFFFBFF;
    ReadData3E = 64'hBB9000400000007F;
    ans = 64'h49FFF00FE00FEBFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"365363\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h33501FFF7FFFFFFF;
    ans = 64'h33501FFF7FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"365669\n");
    end
    ReadData1E = 64'hC1EFFE000000001E;
    ReadData2E = 64'h402FFFFFFFFEF7FE;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hC22FFDFFFFFEF82D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"365975\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"366281\n");
    end
    ReadData1E = 64'hB930FFFFFFFFFBFF;
    ReadData2E = 64'h4031FFFFFFFFFF80;
    ReadData3E = 64'hBEBE982A1C07F2DE;
    ans = 64'hBEBE982A1C07F2DE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"366587\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hB800000000002FFF;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"366893\n");
    end
    ReadData1E = 64'hC015FCF3430061EA;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'hC7F98E9DA34917BB;
    ans = 64'hC7F98E9DA34917BB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"367199\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h4057FFC000000000;
    ReadData3E = 64'h3B2000000000FFF7;
    ans = 64'h3B2000000000FFF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"367505\n");
    end
    ReadData1E = 64'hFFEFF00000000200;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"367811\n");
    end
    ReadData1E = 64'hBFBAE3CE9204B8B5;
    ReadData2E = 64'h400FFFFE00000002;
    ReadData3E = 64'h403FFFFFFFFFC800;
    ans = 64'h403F9470CC70A8C2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"368117\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC1FFFFFCFFFFFFFE;
    ans = 64'hC1FFFFFCFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"368423\n");
    end
    ReadData1E = 64'h3C5FFDFFFFDFFFFE;
    ReadData2E = 64'h0004DB2EE93B91E9;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h0010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"368729\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"369035\n");
    end
    ReadData1E = 64'hC1D0000000003FBF;
    ReadData2E = 64'hBFE0000FFEFFFFFF;
    ReadData3E = 64'h46927FFFFFFFFFFF;
    ans = 64'h46927FFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"369341\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h41F0001FFFFFFFF0;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"369647\n");
    end
    ReadData1E = 64'h380FFFFFE0000100;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hBFA0000200200000;
    ans = 64'hBFA0000200200000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"369953\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC1D0008000001000;
    ReadData3E = 64'hC050000000040006;
    ans = 64'hC050000000040006;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"370259\n");
    end
    ReadData1E = 64'hC3E0000000BFFFFE;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'hC400000000BFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"370565\n");
    end
    ReadData1E = 64'hC3DFC84C1C3C21EC;
    ReadData2E = 64'hB41FFFF001FFFFFF;
    ReadData3E = 64'hC4DFFFFFFFFFFE20;
    ans = 64'hC4DFFFFFFFFFFE20;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"370871\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h24F0000000003FFD;
    ans = 64'h24F0000000003FFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"371177\n");
    end
    ReadData1E = 64'hBFD000FFFFEFFFFF;
    ReadData2E = 64'hBFBFFFFFC0000400;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h3FA000FFDFEE01FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"371483\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"371789\n");
    end
    ReadData1E = 64'h3FC0008FFFFFFFFF;
    ReadData2E = 64'hC3D000000FFFFFFC;
    ReadData3E = 64'hC02FFEFFF8000000;
    ans = 64'hC3A0009010008FFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"372095\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h1BDC6B59EFED4540;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"372401\n");
    end
    ReadData1E = 64'h4007F0F830234BBC;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'hC071FEFFFFFFFFFF;
    ans = 64'hC0713F783E7EE5A1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"372707\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hCE60000800000003;
    ReadData3E = 64'h3FD0000200000001;
    ans = 64'h3FD0000200000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"373013\n");
    end
    ReadData1E = 64'h45BFFFFBF8000000;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'hC010000000000000;
    ans = 64'h45DFFFFBF8000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"373319\n");
    end
    ReadData1E = 64'h52CFFFEFFFFFFBFE;
    ReadData2E = 64'hC3DCF5E2283E4645;
    ReadData3E = 64'h3FDFFFC004000000;
    ans = 64'hD6BCF5D3AD4D2E85;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"373625\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'h417FFDFFFFFFFFE0;
    ans = 64'h417FFDFFFFFFFFE0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"373931\n");
    end
    ReadData1E = 64'h3FC1E79A840CE4A2;
    ReadData2E = 64'hC0202000000007FE;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"374237\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"374543\n");
    end
    ReadData1E = 64'h43EFFFF9FFFFFFFF;
    ReadData2E = 64'hBB91FFFFFFF7FFFE;
    ReadData3E = 64'hBFF00000400000FF;
    ans = 64'hBFF04800327FE0FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"374849\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hB8BFFFF80000FFFE;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"375155\n");
    end
    ReadData1E = 64'hB7F781B7D9BC6FE7;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'h29EA81D8C86B6006;
    ans = 64'hBB4781B7D9BC6FE7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"375461\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBF252B3EDD51DAD4;
    ReadData3E = 64'h3F4001FFFF7FFFFF;
    ans = 64'h3F4001FFFF7FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"375767\n");
    end
    ReadData1E = 64'hA980002200000000;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"376073\n");
    end
    ReadData1E = 64'h43F2B50D2D1063E9;
    ReadData2E = 64'h3F24B8494F924E33;
    ReadData3E = 64'hC34FFD7FFFFFFFFF;
    ans = 64'hC349EF0DEE440A1B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"376379\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h37E7DAF9073C2293;
    ans = 64'h37E7DAF9073C2293;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"376685\n");
    end
    ReadData1E = 64'hFFE7E8F25A9B01D1;
    ReadData2E = 64'h8EF00000000C0000;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"376991\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"377297\n");
    end
    ReadData1E = 64'h000ABEC8547943F5;
    ReadData2E = 64'hC34FF7FFFFFFEFFE;
    ReadData3E = 64'h382001FFFF7FFFFF;
    ans = 64'h382001FFFF7FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"377603\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h77C007F7FFFFFFFE;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"377909\n");
    end
    ReadData1E = 64'h41E0001F7FFFFFFE;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h41DFFFFEFFFFFFFF;
    ans = 64'h4540001F7FFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"378215\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFEF5FFFFFFFFFFF;
    ReadData3E = 64'h3FC0000000003FFB;
    ans = 64'h3FC0000000003FFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"378521\n");
    end
    ReadData1E = 64'hC1CFF80000007FFF;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"378827\n");
    end
    ReadData1E = 64'hFFE5E5EE64C01393;
    ReadData2E = 64'hBFD00000000BFFFF;
    ReadData3E = 64'h3FC0000100001FFF;
    ans = 64'h7FC5E5EE64D08004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"379133\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h499FFF7FFFFFFFEF;
    ans = 64'h499FFF7FFFFFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"379439\n");
    end
    ReadData1E = 64'hBFF95C149819D013;
    ReadData2E = 64'h3FEFFFFFFFE0000F;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"379745\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4018000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"380051\n");
    end
    ReadData1E = 64'h98FFFF00000001FF;
    ReadData2E = 64'h76A000FBFFFFFFFF;
    ReadData3E = 64'h3CA0000000400FFE;
    ans = 64'hCFB0007BF82000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"380357\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h43CF7FFFFFFFDFFE;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h03EF7FFFFFFFE000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"380663\n");
    end
    ReadData1E = 64'hBFEEFFFDFFFFFFFE;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'hC3DF3A8CBF006383;
    ans = 64'hFFEEFFFDFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"380969\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hF80FFFFFFFC00001;
    ReadData3E = 64'hC8800000007FDFFE;
    ans = 64'hC8800000007FDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"381275\n");
    end
    ReadData1E = 64'hFFEFFFF03FFFFFFF;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"381581\n");
    end
    ReadData1E = 64'hBD848F6BEB745C8A;
    ReadData2E = 64'hBFB82E20E6DA10F1;
    ReadData3E = 64'h47E2000FFFFFFFFF;
    ans = 64'h47E2000FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"381887\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'hB430000000100080;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"382193\n");
    end
    ReadData1E = 64'hBF8BF116CF9EB336;
    ReadData2E = 64'hBFF000000200007F;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"382499\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"382805\n");
    end
    ReadData1E = 64'h002FFFFFFFFFF7FF;
    ReadData2E = 64'h3FE6BFDF70E60FB1;
    ReadData3E = 64'h41CFFE07FFFFFFFF;
    ans = 64'h41CFFE07FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"383111\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h43403FFFFFFEFFFF;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h03603FFFFFFF0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"383417\n");
    end
    ReadData1E = 64'h3800000800000003;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h4020AE1F9ECA9031;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"383723\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFCFB217BCF730CB;
    ReadData3E = 64'hBF1FF7FFE0000000;
    ans = 64'hBF1FF7FFE0000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"384029\n");
    end
    ReadData1E = 64'h41786956587A01AC;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"384335\n");
    end
    ReadData1E = 64'hC1EFFFFBFFFEFFFF;
    ReadData2E = 64'hB80FFDFFFFFFF7FF;
    ReadData3E = 64'h41EE329F77C38BD2;
    ans = 64'h41EE329F77C38BD2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"384641\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'hC005A4CE9C323304;
    ans = 64'hC005A4CE9C323304;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"384947\n");
    end
    ReadData1E = 64'hC3DD085B6A950C5E;
    ReadData2E = 64'h000800000000001F;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"385253\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"385559\n");
    end
    ReadData1E = 64'hA7571911431E5523;
    ReadData2E = 64'hC3D000000010FFFF;
    ReadData3E = 64'h3E0000000000021E;
    ans = 64'h3E0000000000021E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"385865\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h43D0000800003FFE;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"386171\n");
    end
    ReadData1E = 64'h3FCFFBFBFFFFFFFE;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'hC1E000000000FFE0;
    ans = 64'hC1E000000000FFE0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"386477\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FCF800000000004;
    ReadData3E = 64'hC3EFFFFFFFFFFD00;
    ans = 64'hC3EFFFFFFFFFFD00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"386783\n");
    end
    ReadData1E = 64'hBFF84430474AABF1;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"387089\n");
    end
    ReadData1E = 64'hBE9D7FFFFFFFFFFF;
    ReadData2E = 64'hC030007FFFFBFFFF;
    ReadData3E = 64'hC963F3DF40853CC4;
    ans = 64'hC963F3DF40853CC4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"387395\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'hD7B00000000003FB;
    ans = 64'hD7B00000000003FB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"387701\n");
    end
    ReadData1E = 64'hC1C1000000000200;
    ReadData2E = 64'hFFD000200007FFFE;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"388007\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"388313\n");
    end
    ReadData1E = 64'h9E5FF00000004000;
    ReadData2E = 64'h54E01FFFFFFFFDFF;
    ReadData3E = 64'h00099A5422D8B766;
    ans = 64'hB35017F000001E40;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"388619\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hCE80000018000000;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"388925\n");
    end
    ReadData1E = 64'hB81FC0000000003F;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h73FFFFFFBFFFFF80;
    ans = 64'h73FFFFFFBFFFFF80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"389231\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hFFD001FFFFFFFFF7;
    ReadData3E = 64'h415C71AEA4B8E680;
    ans = 64'h415C71AE64B0E680;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"389537\n");
    end
    ReadData1E = 64'h47F8A277FB0CC05C;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"389843\n");
    end
    ReadData1E = 64'h3CA7C32F64B84C77;
    ReadData2E = 64'h281FFFFEDFFFFFFF;
    ReadData3E = 64'h2E3CF9805BD36589;
    ans = 64'h2E3CF9805BD36589;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"390149\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'hC20001FFFF000000;
    ans = 64'hC20001FFFF000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"390455\n");
    end
    ReadData1E = 64'hC11FFF7FFFFEFFFF;
    ReadData2E = 64'h43FC9E1AE158F437;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC52C9DA868EC89E1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"390761\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"391067\n");
    end
    ReadData1E = 64'hBFD04000007FFFFF;
    ReadData2E = 64'hA8BFFFFFFFFF7FFA;
    ReadData3E = 64'h401338B91A01C053;
    ans = 64'h401338B91A01C053;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"391373\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC00FFFFFFFBFFFE0;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"391679\n");
    end
    ReadData1E = 64'hFFDFFFFFFEFFFEFF;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'hC7FFEFFFFFFFFBFF;
    ans = 64'h7C8FFFFFFEFFFF01;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"391985\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3FE0000000000201;
    ReadData3E = 64'hBFDA78ED1BAFDD0A;
    ans = 64'hBFDA78ED1BAFDD0A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"392291\n");
    end
    ReadData1E = 64'h401F000002000000;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'hBCDF000001FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"392597\n");
    end
    ReadData1E = 64'hC1E00001FFFFFFBE;
    ReadData2E = 64'h3FB0000400007FFF;
    ReadData3E = 64'hC030000000000008;
    ans = 64'hC1A000062000FFBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"392903\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h4DB0005FFFFFFFFF;
    ans = 64'h4DB0005FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"393209\n");
    end
    ReadData1E = 64'h403FFFFFFFFB8000;
    ReadData2E = 64'hC1D000101FFFFFFE;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'hC22000101FFBBFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"393515\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"393821\n");
    end
    ReadData1E = 64'hC07FF7FFFFF7FFFF;
    ReadData2E = 64'h7FEFFFC000FFFFFF;
    ReadData3E = 64'hBEA0400000000006;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"394127\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC012F769A2FB5162;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"394433\n");
    end
    ReadData1E = 64'hC1E000000000001D;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hC1F3EE27F9A3D71F;
    ans = 64'hC1F1EE27F9A3D71B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"394739\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h37EFFFF0FFFFFFFE;
    ReadData3E = 64'hBBB00008FFFFFFFF;
    ans = 64'hBBB00008FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"395045\n");
    end
    ReadData1E = 64'h8002000000000000;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h800F800000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"395351\n");
    end
    ReadData1E = 64'h408C252230659285;
    ReadData2E = 64'h3FEFFFE000000001;
    ReadData3E = 64'hFFEDFFFFFFFFFF7E;
    ans = 64'hFFEDFFFFFFFFFF7E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"395657\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h58CFE0007FFFFFFE;
    ans = 64'h58CFE0007FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"395963\n");
    end
    ReadData1E = 64'h233FFFFFC0000080;
    ReadData2E = 64'hC03100000001FFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"396269\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h0017FFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"396575\n");
    end
    ReadData1E = 64'hC00F7FFFFFFF0000;
    ReadData2E = 64'h41FFFFFFFFFE001F;
    ReadData3E = 64'hC3CFFFFFFFFFEC00;
    ans = 64'hC3D0000001F7F600;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"396881\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h8950000007DFFFFE;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"397187\n");
    end
    ReadData1E = 64'h43C0400000003FFE;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h001000003FFFFF7F;
    ans = 64'hC3B0400000003FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"397493\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hB2659A07DA522AC2;
    ReadData3E = 64'h4D2000FFBFFFFFFF;
    ans = 64'h4D2000FFBFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"397799\n");
    end
    ReadData1E = 64'h47EC1BAD2B3E6CDC;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h0000000000000001;
    ans = 64'hC7DC1BAD2B3E6CDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"398105\n");
    end
    ReadData1E = 64'h3FEFFFF7FFFFFFF6;
    ReadData2E = 64'hB7FFFFFFFF0003FF;
    ReadData3E = 64'h800FFFFFFFFFDDFE;
    ans = 64'hB7FFFFF7FF000435;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"398411\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hBFC000000000A000;
    ans = 64'hBFC000000000A000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"398717\n");
    end
    ReadData1E = 64'h002FF7BFFFFFFFFE;
    ReadData2E = 64'h800B7AD760176442;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"399023\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h8020000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"399329\n");
    end
    ReadData1E = 64'h5D200010000001FF;
    ReadData2E = 64'h381F000080000000;
    ReadData3E = 64'h486FFFFFFFEFFF80;
    ans = 64'h554F001F800083DE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"399635\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h4034720DB721EB5E;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"399941\n");
    end
    ReadData1E = 64'h40000000801FFFFF;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h41EC000000008000;
    ans = 64'h41EBFFFFFFC07FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"400247\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h40E0020001000000;
    ReadData3E = 64'h4016F0BC0E532E91;
    ans = 64'h4016F0BC0E532E91;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"400553\n");
    end
    ReadData1E = 64'hC24FFFFFBFFFBFFE;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"400859\n");
    end
    ReadData1E = 64'hC00816F876A3A39B;
    ReadData2E = 64'h801FFFFFFF0007FF;
    ReadData3E = 64'h3FCFFF0001FFFFFE;
    ans = 64'h3FCFFF0001FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"401165\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'hC5A0007DFFFFFFFE;
    ans = 64'hC5A0007DFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"401471\n");
    end
    ReadData1E = 64'h3FFFFFFFBFEFFFFF;
    ReadData2E = 64'hC0A52E0A869D4883;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hC0B52E4A5C369C70;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"401777\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h8020000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"402083\n");
    end
    ReadData1E = 64'hA350000021FFFFFF;
    ReadData2E = 64'hA66FF8000000000F;
    ReadData3E = 64'hC3D0080000001000;
    ans = 64'hC3D0080000001000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"402389\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hB7E0022000000000;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"402695\n");
    end
    ReadData1E = 64'h043FFFFBFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h278F80000000003F;
    ans = 64'h278F80000000003F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"403001\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h414AC3E82996F770;
    ReadData3E = 64'h3F10000003FFFF7E;
    ans = 64'h3F10000003FFFF7E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"403307\n");
    end
    ReadData1E = 64'hC3D013FFFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"403613\n");
    end
    ReadData1E = 64'hB7F9E772425B7128;
    ReadData2E = 64'hBFAFFDC000000000;
    ReadData3E = 64'h380FFFFFFFFBFFFD;
    ans = 64'h381067967FEF4B19;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"403919\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'hFFFFBEFFFFFFFFFF;
    ans = 64'hFFFFBEFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"404225\n");
    end
    ReadData1E = 64'hB7E37A866D226CAF;
    ReadData2E = 64'h3FD49B0AB002569D;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h3CA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"404531\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"404837\n");
    end
    ReadData1E = 64'h43FA57E9C5624B8A;
    ReadData2E = 64'hC800000FFFFF7FFE;
    ReadData3E = 64'h4342599B7B2573CC;
    ans = 64'hCC0A58041D4B3E2A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"405143\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBF800000000407FF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"405449\n");
    end
    ReadData1E = 64'hC3E0000008FFFFFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h47403FFFDFFFFFFF;
    ans = 64'h47403FFFE0000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"405755\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3E50FFFFFFFFFFF6;
    ReadData3E = 64'h467001F7FFFFFFFF;
    ans = 64'h467001F7FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"406061\n");
    end
    ReadData1E = 64'h7FD000000FFFFFFA;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'hFFE000001FFFFFF2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"406367\n");
    end
    ReadData1E = 64'h7FECAFD59CB7B74C;
    ReadData2E = 64'h41DBCBEB843B4CF3;
    ReadData3E = 64'h41F8A2957B307164;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"406673\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'hC340000004080000;
    ans = 64'hC340000004080000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"406979\n");
    end
    ReadData1E = 64'hB81FF80000000001;
    ReadData2E = 64'hC1C71750D077617C;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h39F7118AFC4343A4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"407285\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"407591\n");
    end
    ReadData1E = 64'h802FFFFFFFFFFA00;
    ReadData2E = 64'h37E6AE1D90D70705;
    ReadData3E = 64'hBFDFFF77FFFFFFFE;
    ans = 64'hBFDFFF77FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"407897\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC1C00203FFFFFFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hBFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"408203\n");
    end
    ReadData1E = 64'h43EFFBFFFFFFDFFF;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'hC036CD7A7F9E1F02;
    ans = 64'hC41FFBFFFFFFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"408509\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFE280673AD588A9;
    ReadData3E = 64'h4025DE3CDA0994EF;
    ans = 64'h4025DE3CDA0994EF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"408815\n");
    end
    ReadData1E = 64'hC3EFFFF0000FFFFF;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'hC340000000000001;
    ans = 64'h441FFF70000FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"409121\n");
    end
    ReadData1E = 64'h4032E061B83BDA94;
    ReadData2E = 64'hC8F000000001FBFF;
    ReadData3E = 64'hBF4ED64192CA3998;
    ans = 64'hC932E061B83E31E7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"409427\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hE7C002000FFFFFFF;
    ans = 64'hE7C002000FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"409733\n");
    end
    ReadData1E = 64'hC1E7545A88785004;
    ReadData2E = 64'hC120007FFF800000;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h431755152A91F0F2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"410039\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"410345\n");
    end
    ReadData1E = 64'hC2D91B69773500AA;
    ReadData2E = 64'h40B1000000000006;
    ReadData3E = 64'hFFDFFFFFFFE0007F;
    ans = 64'hFFDFFFFFFFE0007F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"410651\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h4340000FFFDFFFFF;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"410957\n");
    end
    ReadData1E = 64'hBCAE912E62741E61;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'hBCAC430028149BD0;
    ans = 64'h3FFE912E62741E62;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"411263\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC3EFFC0000000003;
    ReadData3E = 64'h4031FFFFFFFDFFFF;
    ans = 64'h4031FFFFFFFDFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"411569\n");
    end
    ReadData1E = 64'hC1C000000027FFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h452000000027FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"411875\n");
    end
    ReadData1E = 64'hC0CFFF7FFBFFFFFF;
    ReadData2E = 64'h7E000000FFFFFFFB;
    ReadData3E = 64'h800FFFFC0003FFFE;
    ans = 64'hFEDFFF81FBF7FFB5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"412181\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h401FFFFFFFFE003E;
    ans = 64'h401FFFFFFFFE003E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"412487\n");
    end
    ReadData1E = 64'hB8900000080007FF;
    ReadData2E = 64'h433FE0000FFFFFFE;
    ReadData3E = 64'h8000000000000000;
    ans = 64'hBBDFE0001FF00FF4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"412793\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"413099\n");
    end
    ReadData1E = 64'hC174AF337885BA36;
    ReadData2E = 64'h801FFFFFFFEFFFDF;
    ReadData3E = 64'h41FBFFFFFFFFFFFC;
    ans = 64'h41FBFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"413405\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h001153E88EFDD9B6;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"413711\n");
    end
    ReadData1E = 64'hBFD2000000FFFFFF;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hC16FFFFFFFFF7FEF;
    ans = 64'h7FC2000001000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"414017\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h84D00001000000FE;
    ReadData3E = 64'hC0500020000000FF;
    ans = 64'hC0500020000000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"414323\n");
    end
    ReadData1E = 64'hDE54CE8EEB63015D;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"414629\n");
    end
    ReadData1E = 64'hB2200027FFFFFFFF;
    ReadData2E = 64'h47E1F36BF0B15ECA;
    ReadData3E = 64'h407FFFFFFFFFF007;
    ans = 64'h407FFFFFFFFFF007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"414935\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h43D00FFFFFFFFEFF;
    ans = 64'h43D00FFFFFFFFEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"415241\n");
    end
    ReadData1E = 64'hC1EFBFF7FFFFFFFF;
    ReadData2E = 64'hC032BA8DF73A392C;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"415547\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"415853\n");
    end
    ReadData1E = 64'h00000000000C0000;
    ReadData2E = 64'hBE80000FFFEFFFFE;
    ReadData3E = 64'h763FFFFC07FFFFFE;
    ans = 64'h763FFFFC07FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"416159\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hCF531BF125DB15C1;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h3CAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"416465\n");
    end
    ReadData1E = 64'hC1CFFFFEEFFFFFFF;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h474E6EDD61B287DA;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"416771\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hBFF0004000000040;
    ReadData3E = 64'h480FFDFFFFFFFFDE;
    ans = 64'h480FFDFFFFFFFFDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"417077\n");
    end
    ReadData1E = 64'h3CA0000100000100;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'h4000000000000000;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"417383\n");
    end
    ReadData1E = 64'h8017C536DDFF8DDC;
    ReadData2E = 64'h40B0003FFFFFFFDF;
    ReadData3E = 64'h43E1000000000007;
    ans = 64'h43E1000000000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"417689\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h434100000000000F;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"417995\n");
    end
    ReadData1E = 64'hC03FFC0010000000;
    ReadData2E = 64'h354FFFFFFDFFFFFE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"418301\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"418607\n");
    end
    ReadData1E = 64'hED20034307B6E182;
    ReadData2E = 64'h3FDFFFFC000000FF;
    ReadData3E = 64'hC1DDBCE05D5025DF;
    ans = 64'hED100341074E810B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"418913\n");
    end
    ReadData1E = 64'h0010000000000001;
    ReadData2E = 64'h3CA42E8DA0B707E4;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'h801FFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"419219\n");
    end
    ReadData1E = 64'hC0CF800000800000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'hC0000002007FFFFF;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"419525\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBF00001003FFFFFF;
    ReadData3E = 64'hC34FFC07FFFFFFFE;
    ans = 64'hC34FFC07FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"419831\n");
    end
    ReadData1E = 64'hBF91F142D19F36E5;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"420137\n");
    end
    ReadData1E = 64'h7FFFFF00000007FF;
    ReadData2E = 64'h3F0FFFFFFDFFFEFE;
    ReadData3E = 64'hBFE4E89A16B1F0B7;
    ans = 64'h7FFFFF00000007FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"420443\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h4010001FFFFFFFFB;
    ans = 64'h4010001FFFFFFFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"420749\n");
    end
    ReadData1E = 64'h4030000207FFFFFF;
    ReadData2E = 64'h3FB0000001000001;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"421055\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"421361\n");
    end
    ReadData1E = 64'h000D9AB04F057759;
    ReadData2E = 64'h8020000000007FFF;
    ReadData3E = 64'h43BFFFFFFFFFFFFF;
    ans = 64'h43BFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"421667\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBDAEFFFFFFFFEFFF;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h000FFFFFFFFE1000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"421973\n");
    end
    ReadData1E = 64'h3D04003FFFFFFFFF;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h8010000FFFFFF000;
    ans = 64'h8010000FFFFFEFD8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"422279\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC03FEFFFFFFFFF7F;
    ReadData3E = 64'h40000010000001FE;
    ans = 64'h40000010000001FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"422585\n");
    end
    ReadData1E = 64'hA1CFFFC0000FFFFF;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"422891\n");
    end
    ReadData1E = 64'hBFE0000000010200;
    ReadData2E = 64'hBFDDE00000000000;
    ReadData3E = 64'hC04F72F0FC0052B0;
    ans = 64'hC04F5510FC0050CE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"423197\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h40420003FFFFFFFF;
    ans = 64'h40420003FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"423503\n");
    end
    ReadData1E = 64'h1FD0000000001010;
    ReadData2E = 64'hF7E201FFFFFFFFFF;
    ReadData3E = 64'h4340000000000000;
    ans = 64'hD7C2020000001213;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"423809\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"424115\n");
    end
    ReadData1E = 64'h806E1FFFFFFFFFFF;
    ReadData2E = 64'h3FEFFFFFFEFBFFFF;
    ReadData3E = 64'hFFFFFFFFFFFB8000;
    ans = 64'hFFFFFFFFFFFB8000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"424421\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h7FF03FFFFFF7FFFF;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h7FF83FFFFFF7FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"424727\n");
    end
    ReadData1E = 64'hB21FFDFFFEFFFFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'hBFCFBFFFFFFFFFDF;
    ans = 64'hBFCFBFFFFFFFFFDF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"425033\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC1B603569C15ABE9;
    ReadData3E = 64'hC270000008000006;
    ans = 64'hC270000008000006;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"425339\n");
    end
    ReadData1E = 64'hBF8007FFFFFFBFFE;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"425645\n");
    end
    ReadData1E = 64'hB8795CFD10295802;
    ReadData2E = 64'h3940200000000FFF;
    ReadData3E = 64'hDBC07FFFFFFFF800;
    ans = 64'hDBC07FFFFFFFF800;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"425951\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hBF9250BCC49C6F75;
    ans = 64'hBF9250BCC49C6F75;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"426257\n");
    end
    ReadData1E = 64'hBF3FFFFFFFFFFFFF;
    ReadData2E = 64'h3FEFFFFE000003FF;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010007FFFF80001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"426563\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"426869\n");
    end
    ReadData1E = 64'hC7E77FD7CFDA2EB7;
    ReadData2E = 64'hC7EBFFFFFFFEFFFE;
    ReadData3E = 64'h7FF512B59DBA0209;
    ans = 64'h7FFD12B59DBA0209;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"427175\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3F9122AE3B5AC8CD;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"427481\n");
    end
    ReadData1E = 64'hC03FFFFFFFFFF00F;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'h002FFFFFFFFF5FFF;
    ans = 64'hBCFFFFFFFFFFF00E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"427787\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC1F0000003FFFE00;
    ReadData3E = 64'h41F00200000001FE;
    ans = 64'h41F00200000001FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"428093\n");
    end
    ReadData1E = 64'hFFE20000000FFFFE;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'hFCA20000000FFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"428399\n");
    end
    ReadData1E = 64'h3C92000000000006;
    ReadData2E = 64'h8023D9D6A90F4495;
    ReadData3E = 64'hC1E0201FFFFFFFFF;
    ans = 64'hC1E0201FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"428705\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hBF2CF30ABF2C898A;
    ans = 64'hBF2CF30ABF2C898A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"429011\n");
    end
    ReadData1E = 64'hBFF110E2C0331245;
    ReadData2E = 64'h47E0000010000FFF;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'hC7E110E2D1440615;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"429317\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"429623\n");
    end
    ReadData1E = 64'hC80FFFFFC0FFFFFE;
    ReadData2E = 64'h9CE636531A22D396;
    ReadData3E = 64'hCC000000403FFFFF;
    ans = 64'hCC000000403FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"429929\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC3FFFFF80000007F;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"430235\n");
    end
    ReadData1E = 64'hB7E040000003FFFF;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hC03FFFFFFFDFFBFF;
    ans = 64'hC03FFFFFFFDFFBFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"430541\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h000E000000000040;
    ReadData3E = 64'h3EF000000000010F;
    ans = 64'h3EF000000000010F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"430847\n");
    end
    ReadData1E = 64'h802FFFFFFFF07FFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"431153\n");
    end
    ReadData1E = 64'h41C1008000000000;
    ReadData2E = 64'h43EB1E5DE79D6E0D;
    ReadData3E = 64'hBEB007FFFFFF8000;
    ans = 64'h45BCD11CB90681D9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"431459\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'hB7E002FFFFFFFFFF;
    ans = 64'hB7E002FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"431765\n");
    end
    ReadData1E = 64'hC020000040000001;
    ReadData2E = 64'hBFB000000000077F;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFF7FFFFDFFFFC3E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"432071\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"432377\n");
    end
    ReadData1E = 64'h40600000043FFFFE;
    ReadData2E = 64'h4AB80191FEDD0220;
    ReadData3E = 64'hC2520000000003FF;
    ans = 64'h4B280192053D6CE5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"432683\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h40207FFFFFBFFFFF;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"432989\n");
    end
    ReadData1E = 64'h3FCB7B8B687AC30A;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'hCA50000000000408;
    ans = 64'hCA50000000000408;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"433295\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFC1FFFFFF;
    ReadData3E = 64'hC7FAFFFFFFFFFFFF;
    ans = 64'hC7FAFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"433601\n");
    end
    ReadData1E = 64'h51900080000007FF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h51900080000007FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"433907\n");
    end
    ReadData1E = 64'h80047FFFFFFFFFFF;
    ReadData2E = 64'h3FF0000001008000;
    ReadData3E = 64'h3FB5BEBC9BBB5778;
    ans = 64'h3FB5BEBC9BBB5778;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"434213\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h1CA0000000000070;
    ans = 64'h1CA0000000000070;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"434519\n");
    end
    ReadData1E = 64'h472578883E29B548;
    ReadData2E = 64'h475A91AD5F59CD8E;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h4E91D3AD2F7635F8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"434825\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"435131\n");
    end
    ReadData1E = 64'h3BE7FFFFFFFFFFFD;
    ReadData2E = 64'h480FFFFC003FFFFF;
    ReadData3E = 64'hC7AFFFFF7FFFFFDE;
    ans = 64'hC7AFFFFF7FFFFFDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"435437\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hAA300001FFEFFFFE;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"435743\n");
    end
    ReadData1E = 64'h4FC000000001FFEF;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h3F800011FFFFFFFF;
    ans = 64'h4FC000000001FFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"436049\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBC5FBD681C0884C4;
    ReadData3E = 64'h3D3FFC000000FFFF;
    ans = 64'h3D3FFC000000FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"436355\n");
    end
    ReadData1E = 64'h000F0000001FFFFE;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h8000FFFFFFDFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"436661\n");
    end
    ReadData1E = 64'h339FFFFFFFF9FFFE;
    ReadData2E = 64'h659F200000000000;
    ReadData3E = 64'hC3E000003FFFDFFF;
    ans = 64'h594F1FFFFFFA29FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"436967\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'h13B7D8137EFDC31C;
    ans = 64'h13B7D8137EFDC31C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"437273\n");
    end
    ReadData1E = 64'h7FF00000040001FF;
    ReadData2E = 64'hBFFFFFFFFFFFEFFE;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'h7FF80000040001FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"437579\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h0037FFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"437885\n");
    end
    ReadData1E = 64'hC03FFFFFFFFBFFBF;
    ReadData2E = 64'hBFFFE0003FFFFFFF;
    ReadData3E = 64'h41E6E39FBC98F40E;
    ans = 64'h41E6E39FC490F41E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"438191\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hB7F0000000100003;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"438497\n");
    end
    ReadData1E = 64'h40FB93DAC8FEC06B;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h40100000007FFF00;
    ans = 64'h410B93FAC8FEC16B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"438803\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3D9FFFFFFF800100;
    ReadData3E = 64'hBCAFFFFFFFB7FFFE;
    ans = 64'hBCAFFFFFFFB7FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"439109\n");
    end
    ReadData1E = 64'hBFDFC6D013CB57F9;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'h0000000000000000;
    ans = 64'hBFEFC6D013CB57FB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"439415\n");
    end
    ReadData1E = 64'h3CAFFFFFFE3FFFFF;
    ReadData2E = 64'h3FFFFFFFFFF7FFF7;
    ReadData3E = 64'hED7EFFFF7FFFFFFE;
    ans = 64'hED7EFFFF7FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"439721\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hC0219DFC89EECA7C;
    ans = 64'hC0219DFC89EECA7C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"440027\n");
    end
    ReadData1E = 64'hBD20000000200FFF;
    ReadData2E = 64'hBFEB900DFBE9BCDF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h3FE00000000000DC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"440333\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h003BFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"440639\n");
    end
    ReadData1E = 64'hDFD0000000010800;
    ReadData2E = 64'h00020000000FFFFF;
    ReadData3E = 64'hBFC0007FFFFFFFBE;
    ans = 64'hBFC0007FFFFFFFBE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"440945\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h8010000000801FFF;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"441251\n");
    end
    ReadData1E = 64'h41CA78579544CC5A;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h3187FFFFFFFEFFFF;
    ans = 64'h41EA78579544CC58;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"441557\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h54AFFFFC40000000;
    ReadData3E = 64'h76E00800000000FF;
    ans = 64'h76E00800000000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"441863\n");
    end
    ReadData1E = 64'hBF1BFFFFEFFFFFFF;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"442169\n");
    end
    ReadData1E = 64'h43CFFFFFF803FFFE;
    ReadData2E = 64'h40D00004FFFFFFFF;
    ReadData3E = 64'hCF20000000008800;
    ans = 64'hCF20000000008800;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"442475\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h3F00080400000000;
    ans = 64'h3F00080400000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"442781\n");
    end
    ReadData1E = 64'h8020B3E20AEAF03E;
    ReadData2E = 64'hC34F8007FFFFFFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"443087\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h004FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"443393\n");
    end
    ReadData1E = 64'h3FDFFEFBFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFF7FFFF7F;
    ReadData3E = 64'hC04ADED407B3CC8E;
    ans = 64'hC04B5ECFF793CD90;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"443699\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FD0000010000007;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"444005\n");
    end
    ReadData1E = 64'hB81FFFC000000020;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h0019444730A7E4F3;
    ans = 64'hB84FFFC00000001E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"444311\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FD0000041FFFFFF;
    ReadData3E = 64'h400100007FFFFFFF;
    ans = 64'h400100007FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"444617\n");
    end
    ReadData1E = 64'h47E0800001FFFFFE;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"444923\n");
    end
    ReadData1E = 64'h3CA0007FFDFFFFFE;
    ReadData2E = 64'h4D40000081FFFFFF;
    ReadData3E = 64'h400806D221155416;
    ans = 64'h49F0008080040FED;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"445229\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'h40FFEBFFFFFFFFFE;
    ans = 64'h40FFEBFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"445535\n");
    end
    ReadData1E = 64'h38100000100FFFFE;
    ReadData2E = 64'h402FFFF0000003FE;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"445841\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"446147\n");
    end
    ReadData1E = 64'h3FF1FFFFFEFFFFFE;
    ReadData2E = 64'h4159EFFAF3E842A2;
    ReadData3E = 64'hBFD0000100003FFE;
    ans = 64'h415D2DFA40C64A44;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"446453\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h401FFFFFBFFFFEFE;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"446759\n");
    end
    ReadData1E = 64'h403F000000040000;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h400F9880943E2802;
    ans = 64'h439F00000003FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"447065\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBF7F360561360260;
    ReadData3E = 64'h44C9E2E2A78E50F2;
    ans = 64'h44C9E2E2A78E50F2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"447371\n");
    end
    ReadData1E = 64'hC0500001F8000000;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'hC3AF8003EFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"447677\n");
    end
    ReadData1E = 64'h3E11000000000020;
    ReadData2E = 64'h3FD00007FFFFFFFD;
    ReadData3E = 64'hC8654A4F1E1F4FDC;
    ans = 64'hC8654A4F1E1F4FDC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"447983\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h3810000010020000;
    ans = 64'h400FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"448289\n");
    end
    ReadData1E = 64'h43D0000000080007;
    ReadData2E = 64'h2ACFFFEFFFBFFFFF;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h2EAFFFEFFFD00005;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"448595\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"448901\n");
    end
    ReadData1E = 64'h002FFFFFFF000000;
    ReadData2E = 64'h54EFFFFFFE3FFFFF;
    ReadData3E = 64'hFFC003FFFFFFFFFF;
    ans = 64'hFFC003FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"449207\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC1FFFFFFF800001F;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"449513\n");
    end
    ReadData1E = 64'h3EF0000407FFFFFE;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'hBFBB0E2E40869182;
    ans = 64'h7EF0000407FFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"449819\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FB2E318F3B70CE6;
    ReadData3E = 64'hFF432ECC5E786035;
    ans = 64'hFF432ECC5E786035;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"450125\n");
    end
    ReadData1E = 64'h3FFFFFFF81FFFFFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'hC340000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"450431\n");
    end
    ReadData1E = 64'hC05FFFFFFFFFFF7E;
    ReadData2E = 64'hC80FFFBFFFFFFF7E;
    ReadData3E = 64'h480C1D6E12BFFBDE;
    ans = 64'h48801BFD6E12BF7A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"450737\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h4B9000000007DFFF;
    ans = 64'h4B9000000007DFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"451043\n");
    end
    ReadData1E = 64'h47F0FBFFFFFFFFFF;
    ReadData2E = 64'h48007FFFFFFFFDFF;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h500183DFFFFFFDDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"451349\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"451655\n");
    end
    ReadData1E = 64'hC0674B0ABB8C523C;
    ReadData2E = 64'h43E003FFF8000000;
    ReadData3E = 64'h9D500000010FFFFF;
    ans = 64'hC45750DD7295AFF3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"451961\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h7FD4BFB4A018A1E6;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h4008BFB4A018A1E5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"452267\n");
    end
    ReadData1E = 64'hBFD6729F998C9D9E;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'hC730000001FFF7FF;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"452573\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBFDFFFFFFF3FFFFE;
    ReadData3E = 64'h381FFFFFFBFFFFFE;
    ans = 64'h381FFFFFFBFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"452879\n");
    end
    ReadData1E = 64'h427000001FFFFC00;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"453185\n");
    end
    ReadData1E = 64'hC800001000000003;
    ReadData2E = 64'h3F1039D67DF32534;
    ReadData3E = 64'hBF700000000007FF;
    ans = 64'hC72039E6B7C9A32A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"453491\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h456A3212BAB43622;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"453797\n");
    end
    ReadData1E = 64'h403001000000001F;
    ReadData2E = 64'h40220FFFFFFFFFFF;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"454103\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC340000000000000;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"454409\n");
    end
    ReadData1E = 64'hBF7FE8F6D20EF31D;
    ReadData2E = 64'h401E00000000003E;
    ReadData3E = 64'h3FEFFFFFFFF01FFF;
    ans = 64'h3FEE215989A13FBC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"454715\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hB7FFFFFFFFF7FFDF;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hBFD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"455021\n");
    end
    ReadData1E = 64'h002FFFFFFFFFFFFE;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h7FDFC000003FFFFF;
    ans = 64'h7FDFC000003FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"455327\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h2243DFFFFFFFFFFE;
    ReadData3E = 64'hF575672B472E7543;
    ans = 64'hF575672B472E7543;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"455633\n");
    end
    ReadData1E = 64'hC34FFF7FFFFFBFFF;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"455939\n");
    end
    ReadData1E = 64'h5BE07FFFFFFFFFFF;
    ReadData2E = 64'hC1C8DB69C5C29889;
    ReadData3E = 64'h3F1FFFFFFFC01000;
    ans = 64'hDDB9A24513F0AD4C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"456245\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'hC1F90BEF36AF6276;
    ans = 64'hC1F90BEF36AF6276;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"456551\n");
    end
    ReadData1E = 64'hC32D08A5B4CB2446;
    ReadData2E = 64'hC300000001007FFF;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"456857\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h4010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"457163\n");
    end
    ReadData1E = 64'h380FFDFFFFFFFFFD;
    ReadData2E = 64'hC7E0000000040200;
    ReadData3E = 64'hFFD1FFFFFFFFFFFE;
    ans = 64'hFFD1FFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"457469\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3F8FFFC000000001;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h3CA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"457775\n");
    end
    ReadData1E = 64'hBCA2F22027F558C5;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'hC86FFF2000000000;
    ans = 64'hC86FFF2000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"458081\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hCD2000000FFFFEFF;
    ReadData3E = 64'h41CB8762E572139A;
    ans = 64'h41CB8762E572139A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"458387\n");
    end
    ReadData1E = 64'hBFD20FFFFFFFFFFF;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"458693\n");
    end
    ReadData1E = 64'hB803BB596AD832F9;
    ReadData2E = 64'h41E02D7CF46C5A88;
    ReadData3E = 64'hB1F00000007FBFFE;
    ans = 64'hB9F3F3726FB862F2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"458999\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'hBFE68188CE9F46C1;
    ans = 64'hBFE68188CE9F46C1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"459305\n");
    end
    ReadData1E = 64'hBE407FFFFFFFFFFE;
    ReadData2E = 64'h3FD84D410DEC4BE3;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"459611\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"459917\n");
    end
    ReadData1E = 64'hC3E0000000000440;
    ReadData2E = 64'h3C2FFF000000007F;
    ReadData3E = 64'hA3E000001FFE0000;
    ans = 64'hC01FFF00000008FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"460223\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hB820000040020000;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h801FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"460529\n");
    end
    ReadData1E = 64'hC800000FFFF7FFFE;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'hC3C0100010000000;
    ans = 64'hC3C0100010000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"460835\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h0011FFFFFF7FFFFE;
    ReadData3E = 64'hC1DFFFFEFFF7FFFF;
    ans = 64'hC1DFFFFEFFF7FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"461141\n");
    end
    ReadData1E = 64'h2CB0800000000800;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"461447\n");
    end
    ReadData1E = 64'h37EF00000000007F;
    ReadData2E = 64'h9DE0000000000FFC;
    ReadData3E = 64'hBF0FFFFFFFFFFDF6;
    ans = 64'hBF0FFFFFFFFFFDF6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"461753\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h409FFFFFEFFFFFBF;
    ans = 64'h409FFFFFEFFFFFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"462059\n");
    end
    ReadData1E = 64'h8027B36FC882BDE2;
    ReadData2E = 64'h002FFFFFFFFFFC03;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"462365\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"462671\n");
    end
    ReadData1E = 64'h403FFFFFFF00001F;
    ReadData2E = 64'h380FFFFFFF000800;
    ReadData3E = 64'h40300000000F7FFF;
    ans = 64'h40300000000F7FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"462977\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC39FFFFFFC000008;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h83CFFFFFFC000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"463283\n");
    end
    ReadData1E = 64'h80120DDC76612AE5;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h46AAF8F1DA03076B;
    ans = 64'h46AAF8F1DA03076B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"463589\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC03770390354A8D4;
    ReadData3E = 64'h3E9000FFFFFFF000;
    ans = 64'h3E9000FFFFFFF000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"463895\n");
    end
    ReadData1E = 64'hA4A1D2744FEB9D7F;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"464201\n");
    end
    ReadData1E = 64'h4031B1054AFF739B;
    ReadData2E = 64'hFFD0000080000004;
    ReadData3E = 64'hBFE000010FFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"464507\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h4830000FFFC00000;
    ans = 64'h4830000FFFC00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"464813\n");
    end
    ReadData1E = 64'hFFD0000000000FFF;
    ReadData2E = 64'h4800400040000000;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"465119\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"465425\n");
    end
    ReadData1E = 64'hC02FFFFFC00003FF;
    ReadData2E = 64'hC34FFFF8000003FE;
    ReadData3E = 64'h845FFFF7FFFBFFFE;
    ans = 64'h438FFFF7C00017FD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"465731\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3E50000000003FEF;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h0000000008000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"466037\n");
    end
    ReadData1E = 64'h407FFFF0000001FF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h306B9067D3C8EFCE;
    ans = 64'hC06FFFF0000001FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"466343\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC570400000000FFE;
    ReadData3E = 64'h47EFBFFBFFFFFFFF;
    ans = 64'h47EFBFFBFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"466649\n");
    end
    ReadData1E = 64'h3F4F80001FFFFFFE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFE003F00003FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"466955\n");
    end
    ReadData1E = 64'h1FE000000020001F;
    ReadData2E = 64'h8027BF4A25DCD90E;
    ReadData3E = 64'hBD9199B5B792B56D;
    ans = 64'hBD9199B5B792B56D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"467261\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'hBFD000001000001F;
    ans = 64'hBFD000001000001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"467567\n");
    end
    ReadData1E = 64'h37E0400000000FFF;
    ReadData2E = 64'hBF6FFFFFFFDF7FFE;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"467873\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"468179\n");
    end
    ReadData1E = 64'h3EF0000080000100;
    ReadData2E = 64'h3D76E46D9FFCED73;
    ReadData3E = 64'hA8965757C3E2E3B5;
    ans = 64'h3C76E46E57205BE1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"468485\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h400FFFFC0000001E;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"468791\n");
    end
    ReadData1E = 64'hBFDFC00400000000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'hC1F31AF02D2008C2;
    ans = 64'hC1F31AF02D1818C1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"469097\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h8B90000040000001;
    ReadData3E = 64'hC34FBA3A82A10FA5;
    ans = 64'hC34FBA3A82A10FA5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"469403\n");
    end
    ReadData1E = 64'hC1E0006FFFFFFFFF;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h41E0006FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"469709\n");
    end
    ReadData1E = 64'h7AC002003FFFFFFF;
    ReadData2E = 64'h40D000000003FFF0;
    ReadData3E = 64'hC04FFC000003FFFE;
    ans = 64'h7BA002004004006F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"470015\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h404FFFFFFFFE0080;
    ans = 64'h404FFFFFFFFE0080;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"470321\n");
    end
    ReadData1E = 64'h7FDE1D66FE274D5D;
    ReadData2E = 64'h401FFF00000000FE;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"470627\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"470933\n");
    end
    ReadData1E = 64'h3FD0040080000000;
    ReadData2E = 64'h3FBFFFFFEFFFFBFF;
    ReadData3E = 64'h357FFFFFFFFF8001;
    ans = 64'h3FA0040077FDFDBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"471239\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h002FFF80000003FE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"471545\n");
    end
    ReadData1E = 64'hC18000FFBFFFFFFF;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h0EF6C97A58CFBD2E;
    ans = 64'h418000FFC0000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"471851\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC7AAF283BA042806;
    ReadData3E = 64'h3FBA9508F09BCF9D;
    ans = 64'h3FBA9508F09BCF9D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"472157\n");
    end
    ReadData1E = 64'h450F37C9621767D3;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hC51F37C9621767D2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"472463\n");
    end
    ReadData1E = 64'hBFF007B2DC33D3AC;
    ReadData2E = 64'hBFE02000003FFFFE;
    ReadData3E = 64'h47E0000400010000;
    ans = 64'h47E0000400010000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"472769\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h3CABEF444A1B7558;
    ans = 64'h3CABEF444A1B7558;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"473075\n");
    end
    ReadData1E = 64'hFFE0000007FFFFFF;
    ReadData2E = 64'hC3E000801FFFFFFF;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"473381\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"473687\n");
    end
    ReadData1E = 64'hC3C001FE00000000;
    ReadData2E = 64'h41C822E65EA006B2;
    ReadData3E = 64'h5B90000000080100;
    ans = 64'h5B90000000080100;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"473993\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFF200;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"474299\n");
    end
    ReadData1E = 64'h3FBF8000001FFFFE;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h37E00000007FEFFE;
    ans = 64'hBFCF800000200000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"474605\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC08FFFFFE01FFFFF;
    ReadData3E = 64'hC02FFFFFFFFBFF7E;
    ans = 64'hC02FFFFFFFFBFF7E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"474911\n");
    end
    ReadData1E = 64'h3FC004FFFFFFFFFE;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h0010000000000001;
    ans = 64'hBFE004FFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"475217\n");
    end
    ReadData1E = 64'h3E035583CD466A30;
    ReadData2E = 64'h002000000000BFFE;
    ReadData3E = 64'hCC41000001FFFFFE;
    ans = 64'hCC41000001FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"475523\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h4010000000000820;
    ans = 64'h4010000000000820;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"475829\n");
    end
    ReadData1E = 64'h40500000000006FE;
    ReadData2E = 64'hBF3078FF64F622E6;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEF7C3804D84EAD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"476135\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"476441\n");
    end
    ReadData1E = 64'hCC7FFFFFFFFFFFFE;
    ReadData2E = 64'hBFE3FFFFFFEFFFFF;
    ReadData3E = 64'h49DFFF7DFFFFFFFE;
    ans = 64'h4C73FFFFFFF007FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"476747\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBFF94B59D0203802;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"477053\n");
    end
    ReadData1E = 64'h801FFFF0000FFFFF;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h3F04D4E9A95B7E9A;
    ans = 64'h3F04D4E9A95B7E9A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"477359\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hAB8FFFC00000001E;
    ReadData3E = 64'h43E07FFBFFFFFFFF;
    ans = 64'h43E07FFBFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"477665\n");
    end
    ReadData1E = 64'h37EFFDFFFFFFFBFE;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'hB80FFDFFFFFFFC00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"477971\n");
    end
    ReadData1E = 64'hC34020000000001F;
    ReadData2E = 64'hC34FFFFFFFFFFFF0;
    ReadData3E = 64'h7FE277DFE7545D62;
    ans = 64'h7FE277DFE7545D62;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"478277\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'hC034585A9C396FF4;
    ans = 64'hC034585A9C396FF4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"478583\n");
    end
    ReadData1E = 64'h3FD0001FDFFFFFFE;
    ReadData2E = 64'hB59FFFFFFFFDFFFA;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"478889\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h804DFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"479195\n");
    end
    ReadData1E = 64'h3FD1B49966115E41;
    ReadData2E = 64'hBFDBFFFFFF7FFFFF;
    ReadData3E = 64'hC3C3FFFFFFFFFFFE;
    ans = 64'hC3C3FFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"479501\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3E96DFBD778A9099;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"479807\n");
    end
    ReadData1E = 64'h40A38217A755983F;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'h28D00000000040FF;
    ans = 64'hC3F38217A755983F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"480113\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hFFF1AFE5F2357440;
    ReadData3E = 64'h39733A3911627A6D;
    ans = 64'hFFF9AFE5F2357440;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"480419\n");
    end
    ReadData1E = 64'hC3C0804000000000;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"480725\n");
    end
    ReadData1E = 64'hBFFFFFFFFFFDFFFA;
    ReadData2E = 64'hC1FFFFFFFFE0003F;
    ReadData3E = 64'h3F7F675AC649130D;
    ans = 64'h420FFFFFFFDE0FED;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"481031\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h3FEFFFFFFE004000;
    ans = 64'h3FEFFFFFFE004000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"481337\n");
    end
    ReadData1E = 64'h37E0000080000FFF;
    ReadData2E = 64'hB926FE9CADDEFE5D;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"481643\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h837FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"481949\n");
    end
    ReadData1E = 64'hC120000000007FFE;
    ReadData2E = 64'h3FF0B4D0ECAA7A9E;
    ReadData3E = 64'h414002000007FFFF;
    ans = 64'h4137A99789BA7FDD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"482255\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h040FFFFFFF800FFF;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h400FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"482561\n");
    end
    ReadData1E = 64'hBFE000000010FFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'hC180010000200000;
    ans = 64'h433FFFFFFE21DFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"482867\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h400A0615FD54A801;
    ReadData3E = 64'hBFD20F1FD86CA453;
    ans = 64'hBFD20F1FD86CA453;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"483173\n");
    end
    ReadData1E = 64'h801FFF00001FFFFF;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"483479\n");
    end
    ReadData1E = 64'hC1C000000003FFFE;
    ReadData2E = 64'hBFF001000000003E;
    ReadData3E = 64'h001003FFF7FFFFFF;
    ans = 64'h41C001000004007C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"483785\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hBFEFFFFDFFBFFFFE;
    ans = 64'hC013FFFFBFF80000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"484091\n");
    end
    ReadData1E = 64'h4D2FBFFFFFF80000;
    ReadData2E = 64'hC0300FFFFF000000;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hCD6FDFBFFDFBF800;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"484397\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"484703\n");
    end
    ReadData1E = 64'hBFC5A1020AA218E6;
    ReadData2E = 64'hC3E5495908A19130;
    ReadData3E = 64'h402A691B151F3561;
    ans = 64'h43BCC679A28B88D3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"485009\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h000D1C68E3B85FA9;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"485315\n");
    end
    ReadData1E = 64'hD58000FFFFFFFFEF;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'h4150008000000010;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"485621\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hBF940956532022AE;
    ReadData3E = 64'hB806C023758F4DA2;
    ans = 64'hB806C023758F4DA2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"485927\n");
    end
    ReadData1E = 64'hC3E0000000040040;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"486233\n");
    end
    ReadData1E = 64'h3F84200000000000;
    ReadData2E = 64'h4302931B6A82A168;
    ReadData3E = 64'hBE2FDFFFFFFFDFFF;
    ans = 64'h42975D087BF84F05;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"486539\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h43EE227816B0B1FF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"486845\n");
    end
    ReadData1E = 64'hBEFFFFFFFF7FFFE0;
    ReadData2E = 64'h41EFFFFFFF5FFFFF;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'hC0FFFFFFFEDFFFDF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"487151\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"487457\n");
    end
    ReadData1E = 64'h3F9FFFFFF0400000;
    ReadData2E = 64'h3C1001DFFFFFFFFE;
    ReadData3E = 64'h47B7CC518010305D;
    ans = 64'h47B7CC518010305D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"487763\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'hC1DFFFFE40000000;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h3FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"488069\n");
    end
    ReadData1E = 64'hC2AFFFFF80080000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h381FFFFC0007FFFE;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"488375\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFF;
    ReadData2E = 64'h3F94D7DCFF72CCDB;
    ReadData3E = 64'h000FFFFFFFFFDF7F;
    ans = 64'h0010A6BEE7FB75E6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"488681\n");
    end
    ReadData1E = 64'hB7E8D5DDDB5169D9;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"488987\n");
    end
    ReadData1E = 64'h3FEFFFFFFFDFFDFF;
    ReadData2E = 64'hC1C939221EF1CD9B;
    ReadData3E = 64'h4091FFFFFF000000;
    ans = 64'hC1C9391FDED89305;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"489293\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h41C0FFFFFFFFFDFF;
    ans = 64'h41C0FFFFFFFFFDFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"489599\n");
    end
    ReadData1E = 64'h802FFFFEFFFFFFFF;
    ReadData2E = 64'h43F00FFFFFFBFFFF;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h84300FFF7F7BFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"489905\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"490211\n");
    end
    ReadData1E = 64'h400FFFFFF000001F;
    ReadData2E = 64'h3104707296E48252;
    ReadData3E = 64'h901FFFF0003FFFFF;
    ans = 64'h312470728CAC491A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"490517\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h0F8000000FFFDFFE;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hBFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"490823\n");
    end
    ReadData1E = 64'h3F700000004FFFFF;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'hFFF000400000000E;
    ans = 64'hFFF800400000000E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"491129\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hFFF0080000FFFFFE;
    ReadData3E = 64'hC00FFFFFFFFFFBBE;
    ans = 64'hFFF8080000FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"491435\n");
    end
    ReadData1E = 64'h3FD408ECF56E2FA7;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"491741\n");
    end
    ReadData1E = 64'h7FFFFFFFE003FFFE;
    ReadData2E = 64'h4043FFFFFFDFFFFF;
    ReadData3E = 64'hBFB017FFFFFFFFFF;
    ans = 64'h7FFFFFFFE003FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"492047\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h801FFFFBFFDFFFFE;
    ans = 64'h801FFFFBFFDFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"492353\n");
    end
    ReadData1E = 64'h7FDFFFFFF0FFFFFF;
    ReadData2E = 64'hC3F0000000007FFC;
    ReadData3E = 64'h0000000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"492659\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"492965\n");
    end
    ReadData1E = 64'hC3C8000000008000;
    ReadData2E = 64'hCC2000003FFFFF7E;
    ReadData3E = 64'h800000100000000F;
    ans = 64'h4FF8000060007F3D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"493271\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h4030800000000100;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h3FDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"493577\n");
    end
    ReadData1E = 64'h8010000000400000;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'hBF200000000037FE;
    ans = 64'hBF200000000037FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"493883\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBFEE952F8893D136;
    ReadData3E = 64'h4FFFFE03FFFFFFFE;
    ans = 64'h4FFFFE03FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"494189\n");
    end
    ReadData1E = 64'hC7E007FFFFFFFFF7;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"494495\n");
    end
    ReadData1E = 64'h412FF7FFFFFFEFFE;
    ReadData2E = 64'h3FD785891B76C59A;
    ReadData3E = 64'h3A200007FFFF7FFF;
    ans = 64'h41177FA7B92FDC24;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"494801\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h47FFFFF7FFFFFEFE;
    ans = 64'h47FFFFF7FFFFFEFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"495107\n");
    end
    ReadData1E = 64'hBFF8335B11D329F9;
    ReadData2E = 64'hD5BFFFFFFE000200;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"495413\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"495719\n");
    end
    ReadData1E = 64'h2F90000001EFFFFF;
    ReadData2E = 64'h3CA0000000000090;
    ReadData3E = 64'h43C63BA6AC9ACFB3;
    ans = 64'h43C63BA6AC9ACFB3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"496025\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3F7FF0000000FFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"496331\n");
    end
    ReadData1E = 64'h4030FFFFFFFDFFFF;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h3FBFFE000000FFFE;
    ans = 64'h3FBFFE0000010086;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"496637\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3FFB5677EF9DA389;
    ReadData3E = 64'h801FF80007FFFFFF;
    ans = 64'h0016B4EFD73B4710;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"496943\n");
    end
    ReadData1E = 64'hC1CFFFFF7FFFFDFE;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC00000000FFFFFC1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"497249\n");
    end
    ReadData1E = 64'h3FB0000000000402;
    ReadData2E = 64'hBDF71B09ABA2E778;
    ReadData3E = 64'h381312450C546CDA;
    ans = 64'hBDB71B09ABA2ED42;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"497555\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hBFB0000800800000;
    ans = 64'hBFB0000800800000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"497861\n");
    end
    ReadData1E = 64'h0020007FFFFFFE00;
    ReadData2E = 64'h3BBC144DCA2CAA95;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"498167\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"498473\n");
    end
    ReadData1E = 64'hBF7FFFFFFFBFF7FF;
    ReadData2E = 64'hBFC00001007FFFFF;
    ReadData3E = 64'h37F521633D893DFC;
    ans = 64'h3F500001005FFBFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"498779\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h7FFFFFFFF0007FFE;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h7FFFFFFFF0007FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"499085\n");
    end
    ReadData1E = 64'hC03001FFFFFBFFFF;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h10EFFFF77FFFFFFF;
    ans = 64'hC01001FFFFFBFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"499391\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h37FE000000000008;
    ReadData3E = 64'h001000083FFFFFFF;
    ans = 64'h001000083FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"499697\n");
    end
    ReadData1E = 64'hC86000800007FFFF;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'hC840008000080000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"500003\n");
    end
    ReadData1E = 64'hC1C07FFFFFDFFFFF;
    ReadData2E = 64'hBFE5A99ACE18D898;
    ReadData3E = 64'h401E22E2A005217B;
    ans = 64'h41B656E7ABE704CE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"500309\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h43EFDFFFFF800000;
    ans = 64'h43EFDFFFFF800000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"500615\n");
    end
    ReadData1E = 64'h4010000100001FFF;
    ReadData2E = 64'h40107FFFFC000000;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"500921\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"501227\n");
    end
    ReadData1E = 64'h3FE010003FFFFFFF;
    ReadData2E = 64'h002473AD742511FA;
    ReadData3E = 64'h3F33A5421650E578;
    ans = 64'h3F33A5421650E578;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"501533\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h41580000000000FE;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h0187FFFFE00000FC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"501839\n");
    end
    ReadData1E = 64'h801FFF6FFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h40FF7FFE00000000;
    ans = 64'h40FF7FFE00000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"502145\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h434FFFFFF7FBFFFE;
    ReadData3E = 64'h3FA0003FFFF80000;
    ans = 64'h3FA0003FFFF80000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"502451\n");
    end
    ReadData1E = 64'h4345049737D2616D;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'h4335049737D2616C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"502757\n");
    end
    ReadData1E = 64'hB7FFDFFFFFFFFFFF;
    ReadData2E = 64'hC32FFE0020000000;
    ReadData3E = 64'hCF1FF7FFFFF7FFFF;
    ans = 64'hCF1FF7FFFFF7FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"503063\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'h018FFFFFFFFFFBF6;
    ans = 64'h019000000FFFFDFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"503369\n");
    end
    ReadData1E = 64'h43F5E49AFA9FC2DA;
    ReadData2E = 64'hB810000000000BFE;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"503675\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"503981\n");
    end
    ReadData1E = 64'h3FAB32343B9F34D5;
    ReadData2E = 64'h3F800000F7FFFFFF;
    ReadData3E = 64'hC149FFFFFFFFFFFF;
    ans = 64'hC149FFFFFFF266E4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"504287\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hA9CFFFBFFFFFFFFB;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h000FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"504593\n");
    end
    ReadData1E = 64'h3F4000100000003F;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h422FF00000007FFF;
    ans = 64'h422FF0000000803F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"504899\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC02FFFF8001FFFFE;
    ReadData3E = 64'hB67A3E22C04C77F7;
    ans = 64'hB67A3E22C04C77F7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"505205\n");
    end
    ReadData1E = 64'hB74BFFFFFFF7FFFE;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"505511\n");
    end
    ReadData1E = 64'hBF1FFFFBFFFFFDFF;
    ReadData2E = 64'h41CA02C43C1F6969;
    ReadData3E = 64'h3FEEDE240FD09F22;
    ans = 64'hC0FA02B18CB4D85C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"505817\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hC3FFFBF7FFFFFFFF;
    ans = 64'hC3FFFBF7FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"506123\n");
    end
    ReadData1E = 64'hC013FFBFFFFFFFFF;
    ReadData2E = 64'h3FE80000FFFFFFFF;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h4011002F60020000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"506429\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hBFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"506735\n");
    end
    ReadData1E = 64'hBFE000001FFFFC00;
    ReadData2E = 64'h43DFF7FFFF000000;
    ReadData3E = 64'hC900000001FFEFFF;
    ans = 64'hC900000001FFEFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"507041\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h37EFFF800007FFFF;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"507347\n");
    end
    ReadData1E = 64'hBF8C03FFFFFFFFFE;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'hC3FFFFFFFFFF7FFC;
    ans = 64'hC3FFFFFFFFFF7FFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"507653\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC20FBFDFFFFFFFFF;
    ReadData3E = 64'hC3F000007F7FFFFE;
    ans = 64'hC3F000007F7FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"507959\n");
    end
    ReadData1E = 64'hFFFF7C5CCE30AC8C;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hFFFF7C5CCE30AC8C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"508265\n");
    end
    ReadData1E = 64'hC8F001FFFFFFFDFF;
    ReadData2E = 64'h37F0001FFE000000;
    ReadData3E = 64'h3B7CC989A4C18870;
    ans = 64'hC0F0022001FFBDFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"508571\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'hE20A7354F18E1999;
    ans = 64'hE20A7354F18E1999;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"508877\n");
    end
    ReadData1E = 64'hC03DFFFFFFFFFFF7;
    ReadData2E = 64'hBFE38A75AE2155C3;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'h402CA39CA67E80C3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"509183\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"509489\n");
    end
    ReadData1E = 64'h402080000001FFFF;
    ReadData2E = 64'h3FBFFFFFFFFFFFCF;
    ReadData3E = 64'h43E00FFFFFFEFFFE;
    ans = 64'h43E00FFFFFFEFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"509795\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hE40008000000000F;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"510101\n");
    end
    ReadData1E = 64'h43600000000040FE;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'h3F323FFFFFFFFFFE;
    ans = 64'h43800000000040FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"510407\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h8010F897F57CA997;
    ReadData3E = 64'h3540000100000200;
    ans = 64'h3540000100000200;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"510713\n");
    end
    ReadData1E = 64'h43C800000000001F;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h43E800000000001D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"511019\n");
    end
    ReadData1E = 64'hC1761E69B4B257B1;
    ReadData2E = 64'h3F9FFFFFFEFFFE00;
    ReadData3E = 64'hC0000001BFFFFFFF;
    ans = 64'hC1261E6DB401D301;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"511325\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'hC3CFFFFFFFF7FF7F;
    ans = 64'hC3CFFFFFFFF7FF7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"511631\n");
    end
    ReadData1E = 64'h000C9B46C7FC9346;
    ReadData2E = 64'h3FB0003FFFFFFFFA;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"511937\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"512243\n");
    end
    ReadData1E = 64'h3F14AB7C91DA31FD;
    ReadData2E = 64'h401000000007FF7E;
    ReadData3E = 64'h9DA3FFFFFFC00000;
    ans = 64'h3F34AB7C91E48713;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"512549\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h403FF0000000003E;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"512855\n");
    end
    ReadData1E = 64'h83EC91CEF3FA9AAA;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h401FC003FFFFFFFE;
    ans = 64'h401FC003FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"513161\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h4006FEED39B40EA1;
    ReadData3E = 64'h37FA9D329C356AF4;
    ans = 64'h37FA9D329C356AF4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"513467\n");
    end
    ReadData1E = 64'hC0BD0366D6818E16;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hC0ED0366D6818E15;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"513773\n");
    end
    ReadData1E = 64'hC03E12E5D71D50F1;
    ReadData2E = 64'hB7EF7FFF00000000;
    ReadData3E = 64'hDD800000007BFFFF;
    ans = 64'hDD800000007BFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"514079\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'hBEB0408E721B8F27;
    ans = 64'hBEB0408E721B8F27;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"514385\n");
    end
    ReadData1E = 64'h800FFFFFFFFFFFF8;
    ReadData2E = 64'h3DBC9CA6A89A9686;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"514691\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h3CAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"514997\n");
    end
    ReadData1E = 64'h1CA378B1AF531F03;
    ReadData2E = 64'hC0B0000000007FDF;
    ReadData3E = 64'h3FF00003FFFFFFEE;
    ans = 64'h3FF00003FFFFFFEE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"515303\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h475040001FFFFFFF;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"515609\n");
    end
    ReadData1E = 64'h47E0000000000000;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h40187746E35FBECD;
    ans = 64'h4B30000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"515915\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h400000007FFFFFBE;
    ReadData3E = 64'hC020000000C00000;
    ans = 64'hC020000000C00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"516221\n");
    end
    ReadData1E = 64'hA4B00000001FFFFA;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h0010000000000000;
    ans = 64'hA8100000001FFFF9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"516527\n");
    end
    ReadData1E = 64'h381FEFFFFFFFFF00;
    ReadData2E = 64'h47E000002000000F;
    ReadData3E = 64'h7FDFFC0000020000;
    ans = 64'h7FDFFC0000020000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"516833\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h7FE00000003FFEFF;
    ans = 64'h7FE00000003FFEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"517139\n");
    end
    ReadData1E = 64'h43FFFFC000000004;
    ReadData2E = 64'hC7E0001FFFFDFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'hCBEFFFFFFF7C0008;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"517445\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'h037FFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"517751\n");
    end
    ReadData1E = 64'hBFFFDFFFFDFFFFFE;
    ReadData2E = 64'hBFFEB4C9BBCE1AC1;
    ReadData3E = 64'h3F8FFFFFF0002000;
    ans = 64'h400EB614F0170029;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"518057\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h405FDFFFFFFFFFF8;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"518363\n");
    end
    ReadData1E = 64'hCF4FFFFFEFFFFEFF;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h80182D1FFA57A89A;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"518669\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC22FFF77FFFFFFFF;
    ReadData3E = 64'hC3CFFE07FFFFFFFF;
    ans = 64'hC3CFFE07FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"518975\n");
    end
    ReadData1E = 64'h3FFC000000020000;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h7FEC000000020002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"519281\n");
    end
    ReadData1E = 64'hBD000000006FFFFE;
    ReadData2E = 64'h41A85FEB28D8E1AF;
    ReadData3E = 64'h3800400400000000;
    ans = 64'hBEB85FEB2983811A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"519587\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h2F5FFFFFFFBFFFF0;
    ans = 64'h401FFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"519893\n");
    end
    ReadData1E = 64'h381FFFFF7FFFFEFE;
    ReadData2E = 64'hD92FFEFFE0000000;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hD15FFEFF6003FF7E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"520199\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h401FFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"520505\n");
    end
    ReadData1E = 64'h72AC000000FFFFFF;
    ReadData2E = 64'h37E6E3CCCDE32795;
    ReadData3E = 64'hBFB000004000003E;
    ans = 64'h6AA4075334DDE108;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"520811\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3FFF7A6C08BB70BE;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"521117\n");
    end
    ReadData1E = 64'h47F00010003FFFFE;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'hC7FFE00000100000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"521423\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h4017D5BA5D98B0EF;
    ReadData3E = 64'h400FFFFFFBFFFFBF;
    ans = 64'h400FFFFFFBFFFFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"521729\n");
    end
    ReadData1E = 64'hC003BFFFFFFFFFFF;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"522035\n");
    end
    ReadData1E = 64'hBFAEF1261ABA05FC;
    ReadData2E = 64'hC1ED08B6DD3A362B;
    ReadData3E = 64'h43E00020007FFFFF;
    ans = 64'h43E000200081C12E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"522341\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h480FF3FFFFFFFFFE;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"522647\n");
    end
    ReadData1E = 64'hC3DB82C6E85F00F3;
    ReadData2E = 64'h333000001000003F;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"522953\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"523259\n");
    end
    ReadData1E = 64'hBFBFFFFFFFFFF003;
    ReadData2E = 64'hB7EFFBFFFFC00000;
    ReadData3E = 64'h41E000FE00000000;
    ans = 64'h41E000FE00000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"523565\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC7EFFFF7FFFFFDFF;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"523871\n");
    end
    ReadData1E = 64'hC1DFFFFFFFDE0000;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC710000003FFFFDF;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"524177\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hB2F79EF3767E3212;
    ReadData3E = 64'hF4F3FFFFFF800000;
    ans = 64'hF4F3FFFFFF800000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"524483\n");
    end
    ReadData1E = 64'hBF7800000000003F;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"524789\n");
    end
    ReadData1E = 64'hC012751F6F8812B9;
    ReadData2E = 64'h41D0000000003FFE;
    ReadData3E = 64'h7FFFFFFBFFFFFFEE;
    ans = 64'h7FFFFFFBFFFFFFEE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"525095\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hC0E000002000FFFE;
    ans = 64'hC0E000002000FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"525401\n");
    end
    ReadData1E = 64'h37EFC00000040000;
    ReadData2E = 64'hC3C007FFFEFFFFFF;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0003F9FBFFC11;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"525707\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"526013\n");
    end
    ReadData1E = 64'h43CFFFFFFFFEFFFC;
    ReadData2E = 64'hC040003EFFFFFFFF;
    ReadData3E = 64'h3FE07FFBFFFFFFFF;
    ans = 64'hC420003EFFFF7FFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"526319\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hCFEA28B4A04A46E4;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hBFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"526625\n");
    end
    ReadData1E = 64'hC20FFE00000003FF;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h3FCFFFF000200000;
    ans = 64'h3FCFFFF000200000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"526931\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h4CB00001FFFFFFDE;
    ReadData3E = 64'h4030000002100000;
    ans = 64'h4030000002100000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"527237\n");
    end
    ReadData1E = 64'h38101FFFC0000000;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"527543\n");
    end
    ReadData1E = 64'h886FFFFFFBF7FFFF;
    ReadData2E = 64'h4003A7ECDA013DBE;
    ReadData3E = 64'h402FFF6FFFFFFFFF;
    ans = 64'h402FFF6FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"527849\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'hC00FFFF000001FFE;
    ans = 64'hC00FFFF000001FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"528155\n");
    end
    ReadData1E = 64'hBFCFFFFEFFFDFFFF;
    ReadData2E = 64'h3FC0000004007FFF;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'hBF9FFFFF07FEFFBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"528461\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"528767\n");
    end
    ReadData1E = 64'h3A6FFFFFFFFF7DFF;
    ReadData2E = 64'hBFE2000007FFFFFF;
    ReadData3E = 64'h3FCE1E88FE3E7FCE;
    ans = 64'h3FCE1E88FE3E7FCE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"529073\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC7E1400000000000;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"529379\n");
    end
    ReadData1E = 64'hBE30000003FFF7FE;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h001E000000001FFF;
    ans = 64'h001E000002001FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"529685\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h43477310794BA290;
    ReadData3E = 64'hC010D7F981A2B919;
    ans = 64'hC010D7F981A2B919;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"529991\n");
    end
    ReadData1E = 64'h43D00000047FFFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"530297\n");
    end
    ReadData1E = 64'h407000000000083E;
    ReadData2E = 64'h480C00003FFFFFFF;
    ReadData3E = 64'h7FF551A8F234313E;
    ans = 64'h7FFD51A8F234313E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"530603\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hBF7FFFFF80008000;
    ans = 64'hBF7FFFFF80008000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"530909\n");
    end
    ReadData1E = 64'hC3D3DFFFFFFFFFFE;
    ReadData2E = 64'hC7ECF815284B7A7C;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h4BD1FE152406E111;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"531215\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"531521\n");
    end
    ReadData1E = 64'hC01F7FFFFFFFFEFF;
    ReadData2E = 64'h001FFFFFEFFFFFFF;
    ReadData3E = 64'h4032000000000040;
    ans = 64'h4032000000000040;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"531827\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h0B7FF7FFFC000000;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"532133\n");
    end
    ReadData1E = 64'h43D526C2A80AF822;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'hC7F39734008F12C8;
    ans = 64'hC7F39734008F12C8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"532439\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hA5B000007FFF7FFF;
    ReadData3E = 64'h3FEFC5EE1ADC8D49;
    ans = 64'h3FEFC5EE1ADC8D49;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"532745\n");
    end
    ReadData1E = 64'h4036000000000000;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC020000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"533051\n");
    end
    ReadData1E = 64'hC13FFFFFFE0003FF;
    ReadData2E = 64'hBFC000003FFFFBFF;
    ReadData3E = 64'hC020000041FFFFFF;
    ans = 64'h410FFFC07DFEF3F5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"533357\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'hC3FF80000000001F;
    ans = 64'hC3FF80000000001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"533663\n");
    end
    ReadData1E = 64'h484FFFDFFFFFBFFF;
    ReadData2E = 64'h40001007FFFFFFFF;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h48600FF7EFF7DFDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"533969\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"534275\n");
    end
    ReadData1E = 64'hC03EFFFFFFFFFEFF;
    ReadData2E = 64'h491BFFDFFFFFFFFE;
    ReadData3E = 64'hA3D003FFFFFFFFFE;
    ans = 64'hC96B1FE0FFFFFF1D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"534581\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h3FF192ACE36CC33A;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"534887\n");
    end
    ReadData1E = 64'hBCA010001FFFFFFE;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h480FFE000000000F;
    ans = 64'h480FFE000000000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"535193\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC05FFFFFF8000080;
    ReadData3E = 64'hBD700000200001FF;
    ans = 64'hBD700000200001FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"535499\n");
    end
    ReadData1E = 64'h4012E71496D6009C;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h3FFA31D6D253FEC5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"535805\n");
    end
    ReadData1E = 64'hFFE000001C000000;
    ReadData2E = 64'hBF4FFFFFFF80001F;
    ReadData3E = 64'hC0002DB970BBFE4F;
    ans = 64'h7F4000001BC0000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"536111\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h5A373ED6A8B003D5;
    ans = 64'h5A373ED6A8B003D5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"536417\n");
    end
    ReadData1E = 64'h801FFF7FFFFFDFFF;
    ReadData2E = 64'h401117AFD8D51332;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"536723\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"537029\n");
    end
    ReadData1E = 64'h182FFFFFF7FFFBFE;
    ReadData2E = 64'h464003FF7FFFFFFF;
    ReadData3E = 64'hC7F00004FFFFFFFE;
    ans = 64'hC7F00004FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"537335\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC0300000000801FE;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"537641\n");
    end
    ReadData1E = 64'h000000002FFFFFFF;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hC02000000000007E;
    ans = 64'hC02000000000007E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"537947\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC1FFF81FFFFFFFFE;
    ReadData3E = 64'h3FDFFFFFFFF77FFF;
    ans = 64'h3FDFFFFFFFF77FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"538253\n");
    end
    ReadData1E = 64'hC7FFFFFFFFFFFBF0;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hC000000000000000;
    ans = 64'h47EFFFFFFFFFFBF2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"538559\n");
    end
    ReadData1E = 64'h43EFFFFFFF7FFEFE;
    ReadData2E = 64'hC1E0007FFE000000;
    ReadData3E = 64'hC1EE098A9F9738A5;
    ans = 64'hC5E0007FFDBFFD7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"538865\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h402A0113351DAFC1;
    ans = 64'h402A0113351DAFC1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"539171\n");
    end
    ReadData1E = 64'h3EC00080000001FE;
    ReadData2E = 64'h088007FFFFFFF800;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"539477\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"539783\n");
    end
    ReadData1E = 64'h030C000000020000;
    ReadData2E = 64'h3F72EFB08F55AC92;
    ReadData3E = 64'h40DF76F22692A59F;
    ans = 64'h40DF76F22692A59F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"540089\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hB80FFFFFFFC00001;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h001FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"540395\n");
    end
    ReadData1E = 64'h808E8D71D4EB26E2;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h605B57A010AEBBA0;
    ans = 64'h605B57A010AEBBA0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"540701\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hCAAE000000FFFFFF;
    ReadData3E = 64'hFA00847923480522;
    ans = 64'hFA00847923480522;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"541007\n");
    end
    ReadData1E = 64'hFFE000007FBFFFFE;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h7FE000007FBFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"541313\n");
    end
    ReadData1E = 64'h402FFFFFFFFDFFFE;
    ReadData2E = 64'h3FDC00000001FFFF;
    ReadData3E = 64'h3FFFC00000400000;
    ans = 64'h4021F80000081FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"541619\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h41BE6A4FDAD1844F;
    ans = 64'h41BE6A4FDAD1844F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"541925\n");
    end
    ReadData1E = 64'hC6A0000000080007;
    ReadData2E = 64'hFFD0000000088000;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"542231\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"542537\n");
    end
    ReadData1E = 64'h4604C11700626B2B;
    ReadData2E = 64'h3CAA520793793965;
    ReadData3E = 64'h3FCFFFFF00001FFF;
    ans = 64'h42C1121665239C7D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"542843\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h7FFFFFFFFFFFFA00;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h7FFFFFFFFFFFFA00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"543149\n");
    end
    ReadData1E = 64'hC03FF7FFFFFFFF7F;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'hC3FFFFF8000003FF;
    ans = 64'hC3FFFFF8000003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"543455\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC3F0000003F7FFFF;
    ReadData3E = 64'h34CC2F99040AE502;
    ans = 64'h34CC2F99040AE502;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"543761\n");
    end
    ReadData1E = 64'h3F3FE00000000000;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFF003FC00000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"544067\n");
    end
    ReadData1E = 64'h37F000000DFFFFFF;
    ReadData2E = 64'h47F00000000043FF;
    ReadData3E = 64'hC07FFFFFFEFFFFBE;
    ans = 64'hC07FEFFFFEF1FF7A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"544373\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'hBF8961BE48C37E82;
    ans = 64'hBF8961BE48C37E82;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"544679\n");
    end
    ReadData1E = 64'hBFE2083CCB3D47F8;
    ReadData2E = 64'h43D0443E6399E6DF;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hC3C265261DC90B18;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"544985\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"545291\n");
    end
    ReadData1E = 64'h434F8000FFFFFFFF;
    ReadData2E = 64'h406FFFFC007FFFFF;
    ReadData3E = 64'h4010007FFFFFC000;
    ans = 64'h43CF7FFD107DE002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"545597\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC3DFFC0003FFFFFE;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h840FFC0003FFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"545903\n");
    end
    ReadData1E = 64'h2F30001FFEFFFFFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'hC1062003748F8F29;
    ans = 64'hC1062003748F8F29;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"546209\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC07FF7FFFFFFFF7F;
    ReadData3E = 64'h3F90000000201FFE;
    ans = 64'h3F90000000201FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"546515\n");
    end
    ReadData1E = 64'h801FFFFFEE000000;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"546821\n");
    end
    ReadData1E = 64'h3FCFFFFFEEFFFFFF;
    ReadData2E = 64'hB7FFFFFFF7FFFFEF;
    ReadData3E = 64'hB89FFF00000FFFFF;
    ans = 64'hB8A00080000737FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"547127\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'hC0B0000000081FFF;
    ans = 64'hC0B0000000081FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"547433\n");
    end
    ReadData1E = 64'h3C4FF00040000000;
    ReadData2E = 64'hBFCFFFFFFE00007E;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h4010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"547739\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"548045\n");
    end
    ReadData1E = 64'hC020000000040007;
    ReadData2E = 64'h3469DC294133AA1B;
    ReadData3E = 64'hC3DFE000000007FE;
    ans = 64'hC3DFE000000007FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"548351\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hBCCCC98E4AA02611;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"548657\n");
    end
    ReadData1E = 64'h3D46FFFFFFFFFFFF;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h3FC31D5708AC8D44;
    ans = 64'h3FC31D5708ABD544;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"548963\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC70000007F7FFFFF;
    ReadData3E = 64'hFFD000047FFFFFFF;
    ans = 64'hFFD000047FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"549269\n");
    end
    ReadData1E = 64'h3A5FFFF007FFFFFE;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hBFD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"549575\n");
    end
    ReadData1E = 64'hC040000003FFFFFB;
    ReadData2E = 64'hBDA000FFFBFFFFFF;
    ReadData3E = 64'hC3F1B104FF39AF76;
    ans = 64'hC3F1B104FF39AF76;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"549881\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'hB7FFE0000001FFFE;
    ans = 64'hB7FFE0000001FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"550187\n");
    end
    ReadData1E = 64'h7FFF8000000FFFFE;
    ReadData2E = 64'hB7EDCF2899E0A641;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'h7FFF8000000FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"550493\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"550799\n");
    end
    ReadData1E = 64'h405F100000000000;
    ReadData2E = 64'h1EA0001FF7FFFFFF;
    ReadData3E = 64'hC34FFFE0001FFFFF;
    ans = 64'hC34FFFE0001FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"551105\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h41D55E87D7361155;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"551411\n");
    end
    ReadData1E = 64'h43C0FFFFFFFFFFF0;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h027FFFFFFBFFFFFC;
    ans = 64'hC720FFFFFFFFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"551717\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC1A000000000FEFF;
    ReadData3E = 64'hBFD0FFFDFFFFFFFF;
    ans = 64'hBFD0FFFDFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"552023\n");
    end
    ReadData1E = 64'hFFF84054BAD6D8D0;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'hFFF84054BAD6D8D0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"552329\n");
    end
    ReadData1E = 64'h001F800000000004;
    ReadData2E = 64'h40CFFFFFFFF7FEFE;
    ReadData3E = 64'hC3DFFFBBFFFFFFFE;
    ans = 64'hC3DFFFBBFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"552635\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'hBFD000000800000F;
    ans = 64'hBFD000000800000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"552941\n");
    end
    ReadData1E = 64'hBFBFFFBFFFFDFFFF;
    ReadData2E = 64'h80200FFFFFFFFFFB;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"553247\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hC010FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"553553\n");
    end
    ReadData1E = 64'h2EE0003FFFFC0000;
    ReadData2E = 64'hC3C00000FFF7FFFF;
    ReadData3E = 64'h3E8FFFFFFFFFFFEF;
    ans = 64'h3E8FFFFFFFFFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"553859\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h431C03FFFFFFFFFF;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"554165\n");
    end
    ReadData1E = 64'hC030000000000FBF;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'h3FDEE29310989E27;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"554471\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h44D000003FFDFFFE;
    ReadData3E = 64'h3D2000000000FFBF;
    ans = 64'h3D2000000000FFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"554777\n");
    end
    ReadData1E = 64'h401FF7BFFFFFFFFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"555083\n");
    end
    ReadData1E = 64'h401FFFBFFFFFFFFC;
    ReadData2E = 64'hC0000000000100FF;
    ReadData3E = 64'h3FDB0C7F1AA6DCB7;
    ans = 64'hC02F275C072CCB10;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"555389\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'h7FDFC00003FFFFFF;
    ans = 64'h7FDFC00003FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"555695\n");
    end
    ReadData1E = 64'hC3EFFFFFEFFFDFFE;
    ReadData2E = 64'hB7F0000000FFEFFF;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"556001\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"556307\n");
    end
    ReadData1E = 64'h380FFFBFFE000000;
    ReadData2E = 64'h47F1301F5B57A021;
    ReadData3E = 64'hC7FFFF800000000F;
    ans = 64'hC7FFFF800000000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"556613\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h002FFFFFFFFFFEF6;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"556919\n");
    end
    ReadData1E = 64'h73D9FCECC51EB691;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'hC1C0000000003FBE;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"557225\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'h403000001FFFFFFD;
    ReadData3E = 64'h480007FFEFFFFFFF;
    ans = 64'h480007FFEFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"557531\n");
    end
    ReadData1E = 64'h40D0007FFFE00000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"557837\n");
    end
    ReadData1E = 64'hC03243999D53ABF6;
    ReadData2E = 64'h345000002001FFFF;
    ReadData3E = 64'hB8900403FFFFFFFF;
    ans = 64'hB8900403FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"558143\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'hBFCA7F53316EAB07;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"558449\n");
    end
    ReadData1E = 64'h4020080000000FFE;
    ReadData2E = 64'h755FFFFFFFFF5FFF;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h759007FFFFFFBFD5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"558755\n");
    end
    ReadData1E = 64'h001FFFFFFFFFFFFE;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"559061\n");
    end
    ReadData1E = 64'h3E37118D48FFBADF;
    ReadData2E = 64'h3CA0000000800000;
    ReadData3E = 64'h401FFFF7FFFFFFC0;
    ans = 64'h401FFFF7FFFFFFC0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"559367\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hDB63FFF7FFFFFFFF;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'hD813FFF7FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"559673\n");
    end
    ReadData1E = 64'h434000000007FFEF;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h43EFFFFFFFFFFFFF;
    ans = 64'h43EFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"559979\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h4BB2000000004000;
    ReadData3E = 64'h43C5EE619EB6C788;
    ans = 64'h4862000000004000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"560285\n");
    end
    ReadData1E = 64'h4007000000000000;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h0000000000000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"560591\n");
    end
    ReadData1E = 64'h002200000000003F;
    ReadData2E = 64'h3FCFFFFFFF0000FE;
    ReadData3E = 64'hC3CB85B401C44430;
    ans = 64'hC3CB85B401C44430;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"560897\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'hAEC0000020000000;
    ans = 64'hAEC0000020000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"561203\n");
    end
    ReadData1E = 64'hBFC004000001FFFF;
    ReadData2E = 64'hCD600000003C0000;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'h4D300400003E0EFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"561509\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h000FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"561815\n");
    end
    ReadData1E = 64'h37FFFFBFFFFFFBFE;
    ReadData2E = 64'hBCAFFFFF7FFFFFBE;
    ReadData3E = 64'hB28400000000001F;
    ans = 64'hB4BFFFBF80037BBC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"562121\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hB800000001000FFE;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hC00FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"562427\n");
    end
    ReadData1E = 64'h3FD4FEF9A261B85F;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'hC1FFFFEFFFFFFFF6;
    ans = 64'hC1FFFFEFFFFFFFF6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"562733\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h41EFFFFFFFFFF00F;
    ReadData3E = 64'h7FFFFFFFFFC0001F;
    ans = 64'h7FFFFFFFFFC0001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"563039\n");
    end
    ReadData1E = 64'hC3C0000000003FFD;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"563345\n");
    end
    ReadData1E = 64'h7FF4760C8E5F3973;
    ReadData2E = 64'hCC9FFFF87FFFFFFE;
    ReadData3E = 64'hB80FFE007FFFFFFE;
    ans = 64'h7FFC760C8E5F3973;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"563651\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h3FB0010000000007;
    ans = 64'h3FB0010000000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"563957\n");
    end
    ReadData1E = 64'h403A7713E8818439;
    ReadData2E = 64'h800FFFFFFFFFFFCF;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"564263\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h0000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"564569\n");
    end
    ReadData1E = 64'hC0144254342743BB;
    ReadData2E = 64'hB1F20FFFFFFFFFFF;
    ReadData3E = 64'hBE1FFFFFFFFFFFC1;
    ans = 64'hBE1FFFFFFFFFFFC1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"564875\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h4340080000000002;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4008040000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"565181\n");
    end
    ReadData1E = 64'hBDA000000001FF7F;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h41C0003FFFFFFFFE;
    ans = 64'h41C0003FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"565487\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC1EFFFFE0003FFFF;
    ReadData3E = 64'h4CC0807FFFFFFFFE;
    ans = 64'h4CC0807FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"565793\n");
    end
    ReadData1E = 64'h6911007FFFFFFFFE;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"566099\n");
    end
    ReadData1E = 64'hC25FFFFFFFFF6FFF;
    ReadData2E = 64'h408FFFFFFFFC000F;
    ReadData3E = 64'h41DFFFFFFFC000FF;
    ans = 64'hC2FFFFF7FFFB701E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"566405\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h405CB2EDB604D8DE;
    ans = 64'h405CB2EDB604D8DE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"566711\n");
    end
    ReadData1E = 64'h3ED003FF7FFFFFFF;
    ReadData2E = 64'h3A0D68E6205F2D62;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"567017\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"567323\n");
    end
    ReadData1E = 64'h002FFFFFFF00000F;
    ReadData2E = 64'h380000FFFFFFC000;
    ReadData3E = 64'h800B8F38D5581BB9;
    ans = 64'h800B8F38D5581BB9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"567629\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFB7317BC451054B;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"567935\n");
    end
    ReadData1E = 64'h401FE1EDC4F937A1;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hD8EBE3EE3DA505F1;
    ans = 64'hD8EBE3EE3DA505F1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"568241\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC3E2000000004000;
    ReadData3E = 64'h41EFFFFFFFFF1FFE;
    ans = 64'h41EFFFFF6FFF1FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"568547\n");
    end
    ReadData1E = 64'hCEAFFFFC7FFFFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hCB6FFFFC7FFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"568853\n");
    end
    ReadData1E = 64'h5150200000000003;
    ReadData2E = 64'h41EA317A7FF8829C;
    ReadData3E = 64'h3FC3FFFFFFFFFFFC;
    ans = 64'h534A65DD74F873A6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"569159\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'hC38FFFF7FFFFBFFE;
    ans = 64'hC38FFFF7FFFFBFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"569465\n");
    end
    ReadData1E = 64'h68000003FFFFC000;
    ReadData2E = 64'hB807FFFFFFFFDFFE;
    ReadData3E = 64'h0010000000000001;
    ans = 64'hE0180005FFFF7FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"569771\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"570077\n");
    end
    ReadData1E = 64'h41EFC450BF744BAF;
    ReadData2E = 64'h327C8450FAFD9B42;
    ReadData3E = 64'hC3E07FDFFFFFFFFF;
    ans = 64'hC3E07FDFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"570383\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFE0000000FBFFFF;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"570689\n");
    end
    ReadData1E = 64'h3CA0000000000017;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h41FFFFFFE0000007;
    ans = 64'h41FFFFFFE0000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"570995\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h43D000FFF7FFFFFF;
    ReadData3E = 64'h400125CBE8496822;
    ans = 64'h40801225C3E84967;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"571301\n");
    end
    ReadData1E = 64'h40CFFFF3FFFFFFFE;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000001000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"571607\n");
    end
    ReadData1E = 64'h3E40010003FFFFFE;
    ReadData2E = 64'hA40FFFFFFDFDFFFE;
    ReadData3E = 64'h406000FFFFFFFBFF;
    ans = 64'h406000FFFFFFFBFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"571913\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'h7FE0000000200001;
    ans = 64'h7FE0000000200001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"572219\n");
    end
    ReadData1E = 64'h3C3FFFFFFFF8FFFF;
    ReadData2E = 64'h43D2002000000000;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h4022001FFFFC0FF8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"572525\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"572831\n");
    end
    ReadData1E = 64'hC7F0000000000000;
    ReadData2E = 64'hBF9FFFDDFFFFFFFE;
    ReadData3E = 64'h40DE38BD2F586C5E;
    ans = 64'h479FFFDDFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"573137\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC0DFFFF7FEFFFFFF;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFE0000000007FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"573443\n");
    end
    ReadData1E = 64'h43C0007FFFBFFFFE;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'hC789B6254B55142E;
    ans = 64'hC789B6254B55142E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"573749\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3F7FFFFFFFFFDF7E;
    ReadData3E = 64'hFFE2010000000000;
    ans = 64'hFFE2010000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"574055\n");
    end
    ReadData1E = 64'h8810000000000000;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"574361\n");
    end
    ReadData1E = 64'hC3D00037FFFFFFFE;
    ReadData2E = 64'h800021D5197DD041;
    ReadData3E = 64'hFFEFFFFFFF800001;
    ans = 64'hFFEFFFFFFF800001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"574667\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h402ABB9F5C413B98;
    ans = 64'h402ABB9F5C413B98;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"574973\n");
    end
    ReadData1E = 64'hBFC000003FFFFFFF;
    ReadData2E = 64'h3F503FFFFFEFFFFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"575279\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"575585\n");
    end
    ReadData1E = 64'hC01DFFFFEFFFFFFE;
    ReadData2E = 64'h7FD07FFFFFFFFFFD;
    ReadData3E = 64'hC3420FFFFFFFFFFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"575891\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hACD00000002FFFFE;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"576197\n");
    end
    ReadData1E = 64'hBE9FE20000000000;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'h381FFFFFFFFFF7BE;
    ans = 64'hBE9FE20000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"576503\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC3F00001FEFFFFFF;
    ReadData3E = 64'hC1D0FFFFFFDFFFFF;
    ans = 64'hC1D10001FFE03FDF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"576809\n");
    end
    ReadData1E = 64'hC1FCE7CFD0B6C050;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'hC20CE7CFD096C04F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"577115\n");
    end
    ReadData1E = 64'h3F93E91553500BEA;
    ReadData2E = 64'h41C0003FFFFFFDFE;
    ReadData3E = 64'hBCC5DA60B1BB0238;
    ans = 64'h4163E964F7A556AB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"577421\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h403FFFFFFFFFF807;
    ans = 64'h403FFFFFFFFFF807;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"577727\n");
    end
    ReadData1E = 64'hC03BFFFFFDFFFFFF;
    ReadData2E = 64'h40CF7FFFBFFFFFFE;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"578033\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"578339\n");
    end
    ReadData1E = 64'h4020000200000000;
    ReadData2E = 64'hDFEFFEFFFFFFFDFF;
    ReadData3E = 64'hC02003FFFFFFF7FE;
    ans = 64'hE01FFF03FFDFFDFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"578645\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBCAFFFEFFFFFFFF0;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"578951\n");
    end
    ReadData1E = 64'h7FD0800200000000;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h801FF00000000000;
    ans = 64'h7FE0800200000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"579257\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h001000000003FDFF;
    ReadData3E = 64'hBF40007FFFFFC000;
    ans = 64'hBF40007FFFFFC000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"579563\n");
    end
    ReadData1E = 64'h001014EBF46D200B;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"579869\n");
    end
    ReadData1E = 64'h3FDFFFFF00000020;
    ReadData2E = 64'h2B63DAD3E09AF819;
    ReadData3E = 64'hBFEB93086E12EDF7;
    ans = 64'hBFEB93086E12EDF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"580175\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'h3FD00FDFFFFFFFFF;
    ans = 64'h3FD00FE000000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"580481\n");
    end
    ReadData1E = 64'h37F7FFFFFFF00000;
    ReadData2E = 64'h001FFFFFE00000FE;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"580787\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"581093\n");
    end
    ReadData1E = 64'h801FF00000000000;
    ReadData2E = 64'h6BC000009FFFFFFF;
    ReadData3E = 64'hB7FEAFA4CE7A7EFB;
    ans = 64'hB7FEAFA4CE7A7EFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"581399\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3F20080000000020;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h3BD0080000000020;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"581705\n");
    end
    ReadData1E = 64'h9E93DC7553D65AE9;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'hBFFFFFFEFFFFFEFF;
    ans = 64'hBFFFFFFEFFFFFEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"582011\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hFFFFFFF7F0000000;
    ReadData3E = 64'h41CFFFFBFDFFFFFF;
    ans = 64'hFFFFFFF7F0000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"582317\n");
    end
    ReadData1E = 64'hB6BFFFFFEFFFFFBF;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h3FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"582623\n");
    end
    ReadData1E = 64'h3C9FFFF7FFFF7FFF;
    ReadData2E = 64'h41D437651490379B;
    ReadData3E = 64'h3DD0000FFFFFFEFF;
    ans = 64'h3E84396008B6A199;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"582929\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'h80008007FFFFFFFF;
    ans = 64'h3CCFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"583235\n");
    end
    ReadData1E = 64'hBF783D0BB5184AF6;
    ReadData2E = 64'hC06FFFFFFE00001E;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h4350000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"583541\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"583847\n");
    end
    ReadData1E = 64'hC7FFFFFFFFFE7FFF;
    ReadData2E = 64'hC010000000000FFF;
    ReadData3E = 64'h43C00002FFFFFFFF;
    ans = 64'h481FFFFFFFFE9FFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"584153\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hCA00001FFFFFEFFE;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'hC6B0001FFFFFEFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"584459\n");
    end
    ReadData1E = 64'h37E00007FFFFEFFE;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h000B49FE64045BF1;
    ans = 64'h38100007FFFFEFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"584765\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC16C0000001FFFFE;
    ReadData3E = 64'h404FFFFBFFFFFFEF;
    ans = 64'h404FFFFBFFFC7FEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"585071\n");
    end
    ReadData1E = 64'hBC5FFFFFFF900000;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hBFE1FFFFFFF90001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"585377\n");
    end
    ReadData1E = 64'h381FFFFFE01FFFFF;
    ReadData2E = 64'hC34FFFFF007FFFFF;
    ReadData3E = 64'hBFBBFFF7FFFFFFFF;
    ans = 64'hBFBBFFF7FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"585683\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h3F86422413F8A4BC;
    ans = 64'h3FF02C844827F14A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"585989\n");
    end
    ReadData1E = 64'h3E6FFFFFFFFF0000;
    ReadData2E = 64'h426118431D11305B;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'h40E117431D10A799;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"586295\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h4008000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"586601\n");
    end
    ReadData1E = 64'hD890000000000050;
    ReadData2E = 64'hBF6FFFFBFFFFFEFF;
    ReadData3E = 64'h47EFDFFFFFFFFFFD;
    ans = 64'h580FFFFBFFFFFF9F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"586907\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h4000040000000000;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h3CB0040000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"587213\n");
    end
    ReadData1E = 64'hFFF7FFFFFFFFFDFE;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'hBCA0007FF0000000;
    ans = 64'hFFFFFFFFFFFFFDFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"587519\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FDFFFFFC0200000;
    ReadData3E = 64'hBFCF31C105642E15;
    ans = 64'hBFCF31C105642E13;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"587825\n");
    end
    ReadData1E = 64'hBD80000000FFFFF6;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'hFD70000000FFFFF6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"588131\n");
    end
    ReadData1E = 64'h3FEC96403447C6AD;
    ReadData2E = 64'h0002000000000008;
    ReadData3E = 64'hBFE00006FFFFFFFF;
    ans = 64'hBFE00006FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"588437\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h3F2FFFC1FFFFFFFF;
    ans = 64'h7C90000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"588743\n");
    end
    ReadData1E = 64'hB4E1E8C057F4DF36;
    ReadData2E = 64'h4039B6117BCEDD18;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"589049\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'h7C90000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"589355\n");
    end
    ReadData1E = 64'hC346816BDCBFF1FE;
    ReadData2E = 64'h30C000000401FFFF;
    ReadData3E = 64'hB7FFC3FFFFFFFFFF;
    ans = 64'hB7FFC3FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"589661\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC7FFFFFFC7FFFFFF;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"589967\n");
    end
    ReadData1E = 64'hBDA000010000000E;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h4E3FFFFFFFFFFBFC;
    ans = 64'hFDA000010000000D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"590273\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FE0003BFFFFFFFF;
    ReadData3E = 64'h3F1FFFFFFFFFF87F;
    ans = 64'h3F20000000000440;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"590579\n");
    end
    ReadData1E = 64'hC2D00001FFFFF7FE;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"590885\n");
    end
    ReadData1E = 64'hCDE3FFFFFFF7FFFE;
    ReadData2E = 64'hC00111153EA8B7A3;
    ReadData3E = 64'hCC7FF7FFFEFFFFFF;
    ans = 64'h4DF5555A6E525D00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"591191\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'hC3CE07315524DA3F;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"591497\n");
    end
    ReadData1E = 64'hBFC0000FFFFFE000;
    ReadData2E = 64'h948FFFFF5FFFFFFF;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"591803\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"592109\n");
    end
    ReadData1E = 64'h88DAAC772E18D0E4;
    ReadData2E = 64'h37EC230E5E47728C;
    ReadData3E = 64'h43F1000000400000;
    ans = 64'h43F1000000400000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"592415\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h402CF1C09BD3F810;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"592721\n");
    end
    ReadData1E = 64'hC37FFFFFFF00000F;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h3FBFFBFFFFFF7FFE;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"593027\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hDB1FFFFFFFF00080;
    ReadData3E = 64'hBE52000000000040;
    ans = 64'hD7CFFFFFFFF00080;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"593333\n");
    end
    ReadData1E = 64'h3560000000408000;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"593639\n");
    end
    ReadData1E = 64'hBF1FFE07FFFFFFFF;
    ReadData2E = 64'hC1D0000004000FFE;
    ReadData3E = 64'hB6CFFFFFFF040000;
    ans = 64'h40FFFE0807FFA1F9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"593945\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h4BEFFFFFA0000000;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"594251\n");
    end
    ReadData1E = 64'hFFEEC7E41D48D9E4;
    ReadData2E = 64'h403E000000000008;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"594557\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"594863\n");
    end
    ReadData1E = 64'hC7FFFFFF00000004;
    ReadData2E = 64'hBFDEFFEFFFFFFFFF;
    ReadData3E = 64'hC1CFFFFE20000000;
    ans = 64'h47EEFFEF08008003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"595169\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC04FFFFFFFFBFFFE;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h434FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"595475\n");
    end
    ReadData1E = 64'hC1C0008007FFFFFF;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hC6B4F3502E84A07B;
    ans = 64'hC6B4F3502E84A07B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"595781\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FEBFFFFFFFFFBFF;
    ReadData3E = 64'hB80A388A566C4856;
    ans = 64'h3C9BFFFFFFFFFBFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"596087\n");
    end
    ReadData1E = 64'hDFEFC007FFFFFFFF;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h200FC007FFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"596393\n");
    end
    ReadData1E = 64'h3FB00000003DFFFF;
    ReadData2E = 64'hC022000000000FFE;
    ReadData3E = 64'h7FD04FFFFFFFFFFF;
    ans = 64'h7FD04FFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"596699\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h40000000000807FE;
    ans = 64'h40000000000807FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"597005\n");
    end
    ReadData1E = 64'h00200000003FFFF7;
    ReadData2E = 64'h03BB6A461A67FAB0;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hBFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"597311\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"597617\n");
    end
    ReadData1E = 64'h5720000080001FFE;
    ReadData2E = 64'h3FDFFFFFF00007FE;
    ReadData3E = 64'hB80FF80003FFFFFF;
    ans = 64'h57100000780023BD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"597923\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hCBBFFFFFFFFF07FF;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hC86FFFFFFFFF07FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"598229\n");
    end
    ReadData1E = 64'h800FFFEFFFFFFFF6;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'hC3D20000003FFFFF;
    ans = 64'hC3D20000003FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"598535\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC00000001FFFF800;
    ReadData3E = 64'h5388000000000007;
    ans = 64'h5388000000000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"598841\n");
    end
    ReadData1E = 64'hFFDFDFFFFFFFFFFA;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h400FDFFFFFFFFFF9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"599147\n");
    end
    ReadData1E = 64'h38152BAF87CCD661;
    ReadData2E = 64'hC80FFFFFFFFFE000;
    ReadData3E = 64'h3816D0CC103DB538;
    ans = 64'hC0352BAF87CCC135;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"599453\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h41C0040000000100;
    ans = 64'h41C0040000000100;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"599759\n");
    end
    ReadData1E = 64'hC7FFFF8100000000;
    ReadData2E = 64'hC060000000001FFC;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h486FFF8100003FF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"600065\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h8010000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"600371\n");
    end
    ReadData1E = 64'hC00F0000000003FF;
    ReadData2E = 64'h3FA07FFFFFFFFEFF;
    ReadData3E = 64'h41FA412950FB1DC1;
    ans = 64'h41FA412950F91E41;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"600677\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h47FC3AA54E792590;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h44AC3AA54E792590;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"600983\n");
    end
    ReadData1E = 64'hC7E000FFFFFFBFFF;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'h3FE744AF1E7842FB;
    ans = 64'h449000FFFFFFBFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"601289\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC96FFFFFEFFFFFFF;
    ReadData3E = 64'hC1FB8000D97D3B15;
    ans = 64'hC61FFFFFEFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"601595\n");
    end
    ReadData1E = 64'hC34DDFFFFFFFFFFF;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"601901\n");
    end
    ReadData1E = 64'h4080000000000FEF;
    ReadData2E = 64'hC1E3A3F229A0185E;
    ReadData3E = 64'hC3CFFFFFFFFFFFF5;
    ans = 64'hC3D000004E8FC8A1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"602207\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'hC801FFFFFFFFFF7F;
    ans = 64'hC801FFFFFFFFFF7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"602513\n");
    end
    ReadData1E = 64'hBFCFFBFFFFFFF000;
    ReadData2E = 64'hBFC00FFFFBFFFFFF;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFCBFC8080FFE204;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"602819\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'hB95FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"603125\n");
    end
    ReadData1E = 64'h40DFFFFFFFFFBDFF;
    ReadData2E = 64'hC1E000000047FFFE;
    ReadData3E = 64'hC000000000000202;
    ans = 64'hC2D000000047DF7D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"603431\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h40D00000000003FF;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00FFFFFFFFFEFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"603737\n");
    end
    ReadData1E = 64'hC0310003FFFFFFFF;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hBB8FFC000000007F;
    ans = 64'h40110003FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"604043\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h43D0000004000003;
    ReadData3E = 64'hBDF0010000200000;
    ans = 64'h4080000003FFF802;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"604349\n");
    end
    ReadData1E = 64'h4B40000000000408;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"604655\n");
    end
    ReadData1E = 64'h3800002007FFFFFE;
    ReadData2E = 64'h41EFFF7F7FFFFFFF;
    ReadData3E = 64'h403000000000007E;
    ans = 64'h403000000000007E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"604961\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h402FFFB7FFFFFFFF;
    ans = 64'h402FFFB7FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"605267\n");
    end
    ReadData1E = 64'h3FCF03FFFFFFFFFF;
    ReadData2E = 64'hBFEFFFFFDFFBFFFE;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'hBFCF03FFE0F81F75;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"605573\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"605879\n");
    end
    ReadData1E = 64'hC01000000000FFFE;
    ReadData2E = 64'h380FFFFFF7FFFFFE;
    ReadData3E = 64'h41F00FFFFFFFFFFE;
    ans = 64'h41F00FFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"606185\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3AEFDFFBFFFFFFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"606491\n");
    end
    ReadData1E = 64'hBFFFFFFFFDFEFFFF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'hC2A0000008000200;
    ans = 64'hC2A0000008000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"606797\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h41EFFFFE000003FF;
    ReadData3E = 64'h48A0080020000000;
    ans = 64'h48A0080020000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"607103\n");
    end
    ReadData1E = 64'h40700000000FFF7F;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"607409\n");
    end
    ReadData1E = 64'hC00FFFC0007FFFFF;
    ReadData2E = 64'hC3F0040000000008;
    ReadData3E = 64'hB659C33D767BD014;
    ans = 64'h441003DFF8401007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"607715\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hC1F2000000000FFE;
    ans = 64'hC1F2000000000FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"608021\n");
    end
    ReadData1E = 64'h402008FFFFFFFFFF;
    ReadData2E = 64'h3242C59B272E2C9F;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'h3272D02A4E743697;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"608327\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"608633\n");
    end
    ReadData1E = 64'h800FFFE00000FFFE;
    ReadData2E = 64'h3FA0001001FFFFFF;
    ReadData3E = 64'h43C003FFFFFFFFEF;
    ans = 64'h43C003FFFFFFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"608939\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h002000000801FFFF;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"609245\n");
    end
    ReadData1E = 64'hBFA5D7C957A5866C;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h055FFFFFFFFFEFF7;
    ans = 64'h3FA5D7C957A5866B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"609551\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FBFFC0001FFFFFF;
    ReadData3E = 64'h47ECC30A16FAE247;
    ans = 64'h47ECC30A16FAE247;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"609857\n");
    end
    ReadData1E = 64'h4031FFFFFFFFFFFD;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC350000000000004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"610163\n");
    end
    ReadData1E = 64'h3FFE33B9FA55737D;
    ReadData2E = 64'h3F40800000002000;
    ReadData3E = 64'h88800007FDFFFFFF;
    ans = 64'h3F4F2557CA285B80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"610469\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'hC905F36EE24991F7;
    ans = 64'hC905F36EE24991F7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"610775\n");
    end
    ReadData1E = 64'hC1D827F19B2DAC3D;
    ReadData2E = 64'h3F6F6A8CE25E0151;
    ReadData3E = 64'h0010000000000000;
    ans = 64'hC157B720B8485DBA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"611081\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"611387\n");
    end
    ReadData1E = 64'h43F213A13D95B43D;
    ReadData2E = 64'hC1DFFFFFF0020000;
    ReadData3E = 64'h3EEFFF0000000040;
    ans = 64'hC5E213A1348D04D8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"611693\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hFFF0000000010001;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'hFFF8000000010001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"611999\n");
    end
    ReadData1E = 64'h37E3FFFFFFFFFFFD;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h3FEFDFFFFBFFFFFE;
    ans = 64'h3FEFDFFFFBFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"612305\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h55BFFFFFFF87FFFF;
    ReadData3E = 64'hC18000000000FBFE;
    ans = 64'h526FFFFFFF87FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"612611\n");
    end
    ReadData1E = 64'h434DFEFFFFFFFFFE;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'hC35DFEFFFFFFFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"612917\n");
    end
    ReadData1E = 64'h7FE80001FFFFFFFF;
    ReadData2E = 64'hD48FFFFFFFFFBFBF;
    ReadData3E = 64'h001FFFF001000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"613223\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h5CADFFFFFFFEFFFF;
    ans = 64'h5CADFFFFFFFEFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"613529\n");
    end
    ReadData1E = 64'hB79C1C39DCEBED2B;
    ReadData2E = 64'h41E053AEA45FFB4B;
    ReadData3E = 64'h8000000000000001;
    ans = 64'hB98CAF3F1CD529BE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"613835\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"614141\n");
    end
    ReadData1E = 64'hB35FFEFFFFFFFFEE;
    ReadData2E = 64'h43CFFFFFFFF80000;
    ReadData3E = 64'h41EFFF7FFBFFFFFF;
    ans = 64'h41EFFF7FFBFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"614447\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFC0002000000007;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"614753\n");
    end
    ReadData1E = 64'hC3D242D889644CD5;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h408FFFFFFFFFFFFF;
    ans = 64'h43F242D889644CD5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"615059\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC02000017FFFFFFE;
    ReadData3E = 64'h4170000FFFFFFFC0;
    ans = 64'h4170000FFFFFFFC0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"615365\n");
    end
    ReadData1E = 64'h7FE0FFFFF8000000;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"615671\n");
    end
    ReadData1E = 64'h428821EC16CB5777;
    ReadData2E = 64'hA0DFFF7FFFFFFBFE;
    ReadData3E = 64'h402E008000000000;
    ans = 64'h402E008000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"615977\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'h39F00000000201FF;
    ans = 64'hBCBFFFFFFFFFFEFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"616283\n");
    end
    ReadData1E = 64'hBFFFFDFFFFFFFBFF;
    ReadData2E = 64'hB06800000000000F;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"616589\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"616895\n");
    end
    ReadData1E = 64'h402BFFFFFFFFFFDF;
    ReadData2E = 64'h3CA000010003FFFE;
    ReadData3E = 64'h3F8FFFFF03FFFFFF;
    ans = 64'h3F8FFFFF0400037F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"617201\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC060000300000000;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'hBD0F800600000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"617507\n");
    end
    ReadData1E = 64'hC03FFFFFFFF00000;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hF9FFFBE000000000;
    ans = 64'hF9FFFBE000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"617813\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBFFFFFFFFFEFFF80;
    ReadData3E = 64'h3FEFFC1FFFFFFFFF;
    ans = 64'h3FEFFC1FFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"618119\n");
    end
    ReadData1E = 64'h4F98000000000010;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h4000000000000001;
    ans = 64'hCFC800000000000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"618425\n");
    end
    ReadData1E = 64'hC05FFFFFFE000001;
    ReadData2E = 64'h4CDFFDFFFFFFC000;
    ReadData3E = 64'h38000000000001FF;
    ans = 64'hCD4FFDFFFDFFE001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"618731\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hB7E07FFFEFFFFFFF;
    ans = 64'hBCCFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"619037\n");
    end
    ReadData1E = 64'h41D05FFFFFFFFFFF;
    ReadData2E = 64'hB7E0400001FFFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"619343\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"619649\n");
    end
    ReadData1E = 64'h80400000003FFEFE;
    ReadData2E = 64'hA979AC1965929EB1;
    ReadData3E = 64'hAD5FFF8000040000;
    ans = 64'hAD5FFF8000040000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"619955\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h802FF8007FFFFFFF;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"620261\n");
    end
    ReadData1E = 64'hAFBFFFF77FFFFFFF;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'h3D700100000000FF;
    ans = 64'h3D700100000000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"620567\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h84E6EACDC3A87267;
    ReadData3E = 64'hC1C400FFFFFFFFFF;
    ans = 64'hC1C400FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"620873\n");
    end
    ReadData1E = 64'hC3CFF800000000FE;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'h471FF80000000100;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"621179\n");
    end
    ReadData1E = 64'hC13FC000007FFFFF;
    ReadData2E = 64'hFFE2000FFFFFFFFF;
    ReadData3E = 64'h002A87E3D45EB225;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"621485\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h37FFFE00007FFFFF;
    ans = 64'hBFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"621791\n");
    end
    ReadData1E = 64'hC7E0000000000FFF;
    ReadData2E = 64'h047FFFF7FF7FFFFE;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"622097\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h3CD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"622403\n");
    end
    ReadData1E = 64'h40D02007FFFFFFFF;
    ReadData2E = 64'hC2D0000010000000;
    ReadData3E = 64'h3F8FF800000000FF;
    ans = 64'hC3B02008102007FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"622709\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h002007FFF7FFFFFE;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h0010000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"623015\n");
    end
    ReadData1E = 64'h079F657DBB247EA0;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'hC25FFFFFBFFFFFDF;
    ans = 64'hC78F657DBB247EA0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"623321\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hBEEFFFFFEFFFFEFE;
    ReadData3E = 64'hC3C9C7B1F71A5415;
    ans = 64'hC3C9C7B1F71A5415;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"623627\n");
    end
    ReadData1E = 64'h4030003FFFFFE000;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"623933\n");
    end
    ReadData1E = 64'h381FFDFFFFFFFFF7;
    ReadData2E = 64'h8013948FB121E71F;
    ReadData3E = 64'h411CAEB412B45C02;
    ans = 64'h411CAEB412B45C02;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"624239\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'h403007E000000000;
    ans = 64'hFC90000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"624545\n");
    end
    ReadData1E = 64'hBFEEFFFFFFFEFFFF;
    ReadData2E = 64'h69A22FBA6F273FE0;
    ReadData3E = 64'h4340000000000001;
    ans = 64'hE9A19E3C9BAD7463;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"624851\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hFC9FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"625157\n");
    end
    ReadData1E = 64'hBFF0400002000000;
    ReadData2E = 64'hBB10000000020007;
    ReadData3E = 64'hC1D3FFFF7FFFFFFF;
    ans = 64'hC1D3FFFF7FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"625463\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h37EFFFEFFFFDFFFF;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h349FFFEFFFFDFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"625769\n");
    end
    ReadData1E = 64'hBE01000000100000;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'hBFD79D33FB7309A1;
    ans = 64'h7E010000000FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"626075\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h3FF0008000007FFE;
    ReadData3E = 64'hC00D7B1E10F4ADFA;
    ans = 64'hC00D7B1E10F4ADFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"626381\n");
    end
    ReadData1E = 64'h434000000011FFFF;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"626687\n");
    end
    ReadData1E = 64'h3F26F62E6A714B72;
    ReadData2E = 64'hBFC0803FFFFFFFFF;
    ReadData3E = 64'h7FD72CEED46BAEAA;
    ans = 64'h7FD72CEED46BAEAA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"626993\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'hDBAEFFFFF0000000;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"627299\n");
    end
    ReadData1E = 64'h8332000010000000;
    ReadData2E = 64'h4010001FFFF00000;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"627605\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"627911\n");
    end
    ReadData1E = 64'h510000200FFFFFFF;
    ReadData2E = 64'hC1E000003FFFFFF7;
    ReadData3E = 64'hB80FFFFFFE000002;
    ans = 64'hD2F0002050008036;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"628217\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h480FFFFDFFFE0000;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"628523\n");
    end
    ReadData1E = 64'h380FFFFDE0000000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'hC01FBFFFFFFF7FFF;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"628829\n");
    end
    ReadData1E = 64'h3CA0000000000000;
    ReadData2E = 64'h00201FFFFFFFFFFC;
    ReadData3E = 64'h47F17FFFFFFFFFFE;
    ans = 64'h47F17FFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"629135\n");
    end
    ReadData1E = 64'h436F040000000000;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"629441\n");
    end
    ReadData1E = 64'h421FFFEFFBFFFFFF;
    ReadData2E = 64'h8013FFFFFFFEFFFF;
    ReadData3E = 64'h207100007FFFFFFF;
    ans = 64'h207100007FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"629747\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h43FFFFFFFFFC003E;
    ans = 64'h43FFFFFFFFFC003E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"630053\n");
    end
    ReadData1E = 64'hFFEFF7FFFF800000;
    ReadData2E = 64'h3FDE5AB5F7C3CD81;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'hFFDE531F49CC71B6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"630359\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"630665\n");
    end
    ReadData1E = 64'h400FFE000000003F;
    ReadData2E = 64'hBFDDE4D607B230AB;
    ReadData3E = 64'hC0235239DEB3F143;
    ans = 64'hC0270E98D5FE27FB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"630971\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h38F581177512E12D;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"631277\n");
    end
    ReadData1E = 64'hC03FFFFF7FFFFC00;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'hBE40003FFFFFFFFB;
    ans = 64'hBE40003FFFFFFFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"631583\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h43D35E20BE8A9BED;
    ReadData3E = 64'h43DFF80000007FFF;
    ans = 64'h43DFF80000008000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"631889\n");
    end
    ReadData1E = 64'h026BEB43DDA12286;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"632195\n");
    end
    ReadData1E = 64'hD52FFDFFFFBFFFFF;
    ReadData2E = 64'hC3F3CE1F9D970A2F;
    ReadData3E = 64'h43FF8007FFFFFFFF;
    ans = 64'h5933CCE2BB75947F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"632501\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h47E000007FF7FFFF;
    ans = 64'h47E000007FF7FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"632807\n");
    end
    ReadData1E = 64'hBDDE688351B4049B;
    ReadData2E = 64'h7FEFBFE000000000;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hFDDE2B93E28D4ADE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"633113\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"633419\n");
    end
    ReadData1E = 64'h3FDFC00000001FFE;
    ReadData2E = 64'hC1324744F5C38C98;
    ReadData3E = 64'hC0000000FFF7FFFF;
    ans = 64'hC12222BA6BD857C3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"633725\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBF0FE29A47D0D921;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"634031\n");
    end
    ReadData1E = 64'hF0BFFFBFFFBFFFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h3E06C8F7F58C7A61;
    ans = 64'h3E06C8F7F58C7A61;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"634337\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3ED8926415D98F05;
    ReadData3E = 64'h3FD100000007FFFE;
    ans = 64'h3FD100000007FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"634643\n");
    end
    ReadData1E = 64'hBFF0000000400001;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h8000000000800002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"634949\n");
    end
    ReadData1E = 64'h480FFFFBFFFFFFF7;
    ReadData2E = 64'h480FFFFFFFFFFFFE;
    ReadData3E = 64'h381A32D57211B378;
    ans = 64'h502FFFFBFFFFFFF5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"635255\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h3D2692C6C6F6F414;
    ans = 64'h3D2692C6C6F6F414;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"635561\n");
    end
    ReadData1E = 64'h7FFFFFBFFFFFFFFE;
    ReadData2E = 64'h37FFDFFFFF800000;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h7FFFFFBFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"635867\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"636173\n");
    end
    ReadData1E = 64'hBA50000FFFFFDFFE;
    ReadData2E = 64'h37FFFFEFFDFFFFFE;
    ReadData3E = 64'h3FB4000000000007;
    ans = 64'h3FB4000000000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"636479\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hB06D5172C00F57DE;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"636785\n");
    end
    ReadData1E = 64'hFFFFFFF0003FFFFF;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'hFFEFFBFFFF000000;
    ans = 64'hFFFFFFF0003FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"637091\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFF0100000003FFF;
    ReadData3E = 64'h402FFFEFFF7FFFFF;
    ans = 64'h402FFFEFFF7FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"637397\n");
    end
    ReadData1E = 64'hC7F93F6288E6DA64;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'h8010000000000000;
    ans = 64'hC4B93F6288E6DA63;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"637703\n");
    end
    ReadData1E = 64'hC3FFFFFFFFFFFCFF;
    ReadData2E = 64'hBFDFF3DD4FECC275;
    ReadData3E = 64'hFFDFF801FFFFFFFF;
    ans = 64'hFFDFF801FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"638009\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'h43DFFC1FFFFFFFFF;
    ans = 64'h43DFFC1FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"638315\n");
    end
    ReadData1E = 64'hC1CFFFFFFBFF7FFF;
    ReadData2E = 64'h4060000207FFFFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hC240000206003FBC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"638621\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h3C80000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"638927\n");
    end
    ReadData1E = 64'hFFD00001FFF00000;
    ReadData2E = 64'hC340003FFFEFFFFF;
    ReadData3E = 64'hBFEFFBFFFFFFF7FE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"639233\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h7F60000000000DFF;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'h7C10000000000E00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"639539\n");
    end
    ReadData1E = 64'h41DDFFFFF7FFFFFE;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hBD8FFFFFBFFFFFFC;
    ans = 64'h41BDFFFFF8000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"639845\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h404BFFFFFFFFFF7F;
    ReadData3E = 64'hBE20050000000000;
    ans = 64'hBE2004FC80000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"640151\n");
    end
    ReadData1E = 64'hBFD00000080FFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h0000000000000001;
    ans = 64'hBFC00000080FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"640457\n");
    end
    ReadData1E = 64'h3DF0000200000200;
    ReadData2E = 64'h4CD000004003FFFF;
    ReadData3E = 64'hC1C0008000FFFFFE;
    ans = 64'h4AD0000240040A00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"640763\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h014FFFFF7FFFBFFF;
    ans = 64'h3C90000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"641069\n");
    end
    ReadData1E = 64'h406FFFFFFF0000FF;
    ReadData2E = 64'h41D020000000007F;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h42501FFFFF7F20FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"641375\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h3C90000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"641681\n");
    end
    ReadData1E = 64'h3BA211127EC45600;
    ReadData2E = 64'h358FFFFF80004000;
    ReadData3E = 64'h3FCFFFFFFFFFDFF0;
    ans = 64'h3FCFFFFFFFFFDFF0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"641987\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFB2000000000003;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"642293\n");
    end
    ReadData1E = 64'hC3CFFBFFDFFFFFFF;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'hEE6C9467FCA47D87;
    ans = 64'hEE6C9467FCA47D87;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"642599\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h5EB01FF7FFFFFFFF;
    ReadData3E = 64'h37E0037FFFFFFFFE;
    ans = 64'h5B601FF800000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"642905\n");
    end
    ReadData1E = 64'h47FC4B89B376A36D;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"643211\n");
    end
    ReadData1E = 64'h3FFBFBFFFFFFFFFE;
    ReadData2E = 64'hC00F00000FFFFFFF;
    ReadData3E = 64'hB819596CB6C487DA;
    ans = 64'hC01B1C200DFDFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"643517\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h3FF0000000003EFF;
    ans = 64'h3FF0000000003F00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"643823\n");
    end
    ReadData1E = 64'hC05F7FFFEFFFFFFF;
    ReadData2E = 64'h8000100000000003;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hBFD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"644129\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"644435\n");
    end
    ReadData1E = 64'hB7FFFFF800000008;
    ReadData2E = 64'h3FFF6C3812CF83F7;
    ReadData3E = 64'h811FFFFFFFFF7C00;
    ans = 64'hB80F6C3037C17F4B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"644741\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFCFC00000000010;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"645047\n");
    end
    ReadData1E = 64'hC03FFFFF000FFFFE;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hB81FBFFFFFFFF000;
    ans = 64'hC03FFFFF000FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"645353\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBF3FFFFF00000004;
    ReadData3E = 64'h401000007DFFFFFF;
    ans = 64'h401000007DFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"645659\n");
    end
    ReadData1E = 64'hC1F00000000001FF;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"645965\n");
    end
    ReadData1E = 64'h40154B404B0501A2;
    ReadData2E = 64'h380000000003FFFE;
    ReadData3E = 64'hBFEFFEFFFDFFFFFF;
    ans = 64'hBFEFFEFFFDFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"646271\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'hCDC3FFEFFFFFFFFF;
    ans = 64'hCDC3FFEFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"646577\n");
    end
    ReadData1E = 64'h403000083FFFFFFE;
    ReadData2E = 64'h43FFFFEFFEFFFFFF;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h444000003F7BDFBC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"646883\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"647189\n");
    end
    ReadData1E = 64'h41E0000020FFFFFE;
    ReadData2E = 64'hBD4FFFFFFFDFEFFF;
    ReadData3E = 64'h332FB80000000000;
    ans = 64'hBF40000020EFF7FD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"647495\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h43D0000103FFFFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h4080100104000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"647801\n");
    end
    ReadData1E = 64'h42A01FFFFFFFFFFF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h43E00FFF7FFFFFFF;
    ans = 64'h43E0100183FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"648107\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC3C7CB413AF78371;
    ReadData3E = 64'hC02000004001FFFE;
    ans = 64'hC0784B413CF79372;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"648413\n");
    end
    ReadData1E = 64'hC1E0000001FBFFFF;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"648719\n");
    end
    ReadData1E = 64'hC00FFC0000001FFF;
    ReadData2E = 64'h3F3FFC000007FFFE;
    ReadData3E = 64'h801352A02834F823;
    ans = 64'hBF5FF80080081EF9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"649025\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'h402DFA51AAFBA1CF;
    ans = 64'h402DFA51AAFBA1CF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"649331\n");
    end
    ReadData1E = 64'h402D6AD405E9D0E2;
    ReadData2E = 64'h4006699A3858A969;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h40449A8611ACF7D4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"649637\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"649943\n");
    end
    ReadData1E = 64'h801FFFF77FFFFFFF;
    ReadData2E = 64'hC01040000000000F;
    ReadData3E = 64'h2C4000FFFDFFFFFF;
    ans = 64'h2C4000FFFDFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"650249\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h38003FFFF7FFFFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hBFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"650555\n");
    end
    ReadData1E = 64'h3B10000100000007;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h47F0000200000003;
    ans = 64'h47F0000200000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"650861\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC679CC0F7B01DBF2;
    ReadData3E = 64'hC7F000000FFFDFFE;
    ans = 64'hC7F000000FFFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"651167\n");
    end
    ReadData1E = 64'hBCA403A0FE29F2F7;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"651473\n");
    end
    ReadData1E = 64'hC7FFC603B5F552D5;
    ReadData2E = 64'hC01000000800001E;
    ReadData3E = 64'h40E00007FFFEFFFF;
    ans = 64'h481FC603C5D854EC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"651779\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'hC0300000000077FE;
    ans = 64'hC0300000000077FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"652085\n");
    end
    ReadData1E = 64'h3CA7F3ED2FDBC84E;
    ReadData2E = 64'hC806121D06A51CCD;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'hC4C08542005A4C7C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"652391\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"652697\n");
    end
    ReadData1E = 64'h3FAFFFFFFFFFFBBE;
    ReadData2E = 64'h428FFF800001FFFF;
    ReadData3E = 64'h43D000080000001F;
    ans = 64'h43D000080FFFC01F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"653003\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h7FD4BB6592591753;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h7C84BB6592591754;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"653309\n");
    end
    ReadData1E = 64'h41EFFFFF7FFFFFDF;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hC3CFFFFFFBFFEFFF;
    ans = 64'hC3CFFFFFF7FFF00F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"653615\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h4160000200000020;
    ReadData3E = 64'h3FD0000001FFFFFF;
    ans = 64'h3FD000000300001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"653921\n");
    end
    ReadData1E = 64'hBAEC7B3DC3F4513A;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h401FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"654227\n");
    end
    ReadData1E = 64'hC7DFFFFFFF7FBFFE;
    ReadData2E = 64'h001FFBFFEFFFFFFF;
    ReadData3E = 64'hBFF5AE5B27D1523E;
    ans = 64'hBFF5AE5B27D1523E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"654533\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'hC330001000000FFF;
    ans = 64'hC330001000000FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"654839\n");
    end
    ReadData1E = 64'h41D0000000007FEF;
    ReadData2E = 64'hA5E040000000FFFE;
    ReadData3E = 64'h8000000000000000;
    ans = 64'hA7C04000000181ED;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"655145\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"655451\n");
    end
    ReadData1E = 64'h7FE000007F7FFFFE;
    ReadData2E = 64'hBFFFFFFFFFC00400;
    ReadData3E = 64'hBFCFFFFFFFFFFCFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"655757\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC3C000800000007E;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hC07004800000007F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"656063\n");
    end
    ReadData1E = 64'hC5B00000000FFFFD;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h47DDFFFFFFFFDFFF;
    ans = 64'hC90FFFFC401FFFF9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"656369\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC7E1FFFFFFFFFFBF;
    ReadData3E = 64'h7FEFFFFBFFFFFFF0;
    ans = 64'h7FEFFFFBFFFFFFF0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"656675\n");
    end
    ReadData1E = 64'hC0CFFEFFFC000000;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hC42FFEFFFBFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"656981\n");
    end
    ReadData1E = 64'h40054BEC3E7261FE;
    ReadData2E = 64'h4433FF2ED149427C;
    ReadData3E = 64'h41E7FFFFFFFDFFFE;
    ans = 64'h444A9DD0E023E7B4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"657287\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h66DFFEFFBFFFFFFE;
    ans = 64'h66DFFEFFBFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"657593\n");
    end
    ReadData1E = 64'hBFEF7FFF7FFFFFFE;
    ReadData2E = 64'h37E0000000FFFDFE;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"657899\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h7C90000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"658205\n");
    end
    ReadData1E = 64'hBF6FFFFFBFFFFC00;
    ReadData2E = 64'hC1CFFFFFDFFFFFFF;
    ReadData3E = 64'h401000000FFFFFDE;
    ans = 64'h41500000CFFFFF1F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"658511\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3ED50A68C526C672;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h3CB00002A14D18A4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"658817\n");
    end
    ReadData1E = 64'hBFDFF7FFFFFFFFBF;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'hBFCEC55F3EC9F46F;
    ans = 64'hFFCFF7FFFFFFFFC1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"659123\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC1272DC69C306421;
    ReadData3E = 64'hB80FFBFFFFFFFFEE;
    ans = 64'hBDD72DC69C306422;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"659429\n");
    end
    ReadData1E = 64'hBFD078B42033B059;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h4000000000000000;
    ans = 64'hFFD078B42033B058;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"659735\n");
    end
    ReadData1E = 64'h43C8703ECCDE6D0D;
    ReadData2E = 64'h3DAFFFFFFFFF7FDF;
    ReadData3E = 64'h8881DECC22925CB8;
    ans = 64'h4188703ECCDE0B33;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"660041\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h5760003FFF000000;
    ans = 64'h7CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"660347\n");
    end
    ReadData1E = 64'h3F04000000FFFFFF;
    ReadData2E = 64'h7FD7FFFFFFFFFFEF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"660653\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'h7CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"660959\n");
    end
    ReadData1E = 64'h41CFFC000000003E;
    ReadData2E = 64'hDFA002000007FFFF;
    ReadData3E = 64'h43EFFFFFFFFFDF80;
    ans = 64'hE17FFFFF800FFE3C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"661265\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FF002000000001E;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'h3CA002000000001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"661571\n");
    end
    ReadData1E = 64'h78303FFFFFFFFFDE;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h405000000000403E;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"661877\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC1DFFFFFFFFFFFDE;
    ReadData3E = 64'h800BFFFFFFFFFFFF;
    ans = 64'hBE8FFFFFFFFFFFE0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"662183\n");
    end
    ReadData1E = 64'hC340000000003EFF;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"662489\n");
    end
    ReadData1E = 64'hC3E0000040000FFE;
    ReadData2E = 64'h4270000000003FFE;
    ReadData3E = 64'h3120BFFFFFFFFFFE;
    ans = 64'hC660000040004FFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"662795\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h3E50000000001800;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"663101\n");
    end
    ReadData1E = 64'hFFFFFF7FFDFFFFFF;
    ReadData2E = 64'hBF8392D6EC4E306A;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hFFFFFF7FFDFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"663407\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"663713\n");
    end
    ReadData1E = 64'hB6A0000008FFFFFF;
    ReadData2E = 64'h001FFFFDFFFC0000;
    ReadData3E = 64'h41EFFEFFFFFFFFC0;
    ans = 64'h41EFFEFFFFFFFFC0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"664019\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFD000000400001F;
    ReadData3E = 64'h0010000000000000;
    ans = 64'hBC80000004000020;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"664325\n");
    end
    ReadData1E = 64'hBA0FFFFFFFFFFFFE;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h380FFFFFFFF60000;
    ans = 64'h380FFFFFFFF60000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"664631\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBB50007FFFFFFFFD;
    ReadData3E = 64'hC1CFFFFFFDFFFFEF;
    ans = 64'hC1CFFFFFFDFFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"664937\n");
    end
    ReadData1E = 64'h40207FFFFFFBFFFF;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"665243\n");
    end
    ReadData1E = 64'hCF2FFBFFFF000000;
    ReadData2E = 64'h271FFFE00000000E;
    ReadData3E = 64'h40C000007FFFFC00;
    ans = 64'h40C000007FFFFC00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"665549\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hBC4000200000007E;
    ans = 64'hBC4000200000007E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"665855\n");
    end
    ReadData1E = 64'h7FFF7FFF7FFFFFFF;
    ReadData2E = 64'hC3E3FFFF80000000;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h7FFF7FFF7FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"666161\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"666467\n");
    end
    ReadData1E = 64'h3E2BFD12145D6BEC;
    ReadData2E = 64'h3FD77C5DA6D45E1A;
    ReadData3E = 64'hC3D7FFFFFFC00000;
    ans = 64'hC3D7FFFFFFC00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"666773\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC90FFFFFFFFFFFFF;
    ReadData3E = 64'h8000000000000001;
    ans = 64'hC5C0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"667079\n");
    end
    ReadData1E = 64'hB80FFFFFF7FDFFFE;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'hBFC22F02C76C27CA;
    ans = 64'hBFC22F02C76C27CA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"667385\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC1D3484EB2F79C31;
    ReadData3E = 64'hBBB0000001FF7FFF;
    ans = 64'hBE83484EB2F79CB2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"667691\n");
    end
    ReadData1E = 64'h406FFFFFC000000F;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"667997\n");
    end
    ReadData1E = 64'h4014A782EDFCB5DA;
    ReadData2E = 64'hC01000001FFFFFF8;
    ReadData3E = 64'hBFDFF800001FFFFE;
    ans = 64'hC0352763174C3BAC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"668303\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'hC2F00080003FFFFF;
    ans = 64'hC2F00080003FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"668609\n");
    end
    ReadData1E = 64'hBFCB16A4DDFFFF66;
    ReadData2E = 64'h480E26AB973442D3;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC7E985F6B4E7B914;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"668915\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"669221\n");
    end
    ReadData1E = 64'hC01FFFFFFFFFF3FE;
    ReadData2E = 64'h37E000009FFFFFFF;
    ReadData3E = 64'hBFD001000001FFFE;
    ans = 64'hBFD001000001FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"669527\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FE3BE9E4638F42A;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"669833\n");
    end
    ReadData1E = 64'hBD80000FF8000000;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'h79EFFFFFFFFFFFFF;
    ans = 64'h79EFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"670139\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC1F0000000401FFF;
    ReadData3E = 64'h3FE1000000001FFF;
    ans = 64'h3FE0FFFF00001FFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"670445\n");
    end
    ReadData1E = 64'hBFC539959DBB1BD9;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"670751\n");
    end
    ReadData1E = 64'hABA000FFFF800000;
    ReadData2E = 64'hC090000000FFFFFE;
    ReadData3E = 64'h3E1FFFFFFFFFE080;
    ans = 64'h3E1FFFFFFFFFE080;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"671057\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hC1F00001FBFFFFFF;
    ans = 64'hC1F00001FBFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"671363\n");
    end
    ReadData1E = 64'h3FE0000003FF7FFF;
    ReadData2E = 64'hC1D00000000807FF;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'hC1C00000020787FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"671669\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"671975\n");
    end
    ReadData1E = 64'hC1FC34B48347777F;
    ReadData2E = 64'h41F0000000000808;
    ReadData3E = 64'hC34E4805E11D8622;
    ans = 64'hC3FC387D8403A958;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"672281\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3F6E674F9B9487B2;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"672587\n");
    end
    ReadData1E = 64'hC1BB7F6EF09742E6;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'hBFF830AAB3C2C890;
    ans = 64'hBFF830AA9843599F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"672893\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBF22A961D8D5A41D;
    ReadData3E = 64'hC1FFDFFFFFFFF7FF;
    ans = 64'hC1FFDFFFFFFFF7FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"673199\n");
    end
    ReadData1E = 64'h3FF319FB01ADB4BC;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCBB19FB01ADB4BB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"673505\n");
    end
    ReadData1E = 64'hBFD346D8CF582601;
    ReadData2E = 64'hFFD0000000000DFF;
    ReadData3E = 64'hC800208000000000;
    ans = 64'h7FB346D8CF5836DE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"673811\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h434FFFFFEFEFFFFF;
    ans = 64'h434FFFFFEFEFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"674117\n");
    end
    ReadData1E = 64'h000FFFFFFFF7FFF6;
    ReadData2E = 64'hC3E0000002200000;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"674423\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FCFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"674729\n");
    end
    ReadData1E = 64'hBFEBB35F44DCC45E;
    ReadData2E = 64'hB88EA4683179AA7F;
    ReadData3E = 64'hBFDA67B4157D037F;
    ans = 64'hBFDA67B4157D037F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"675035\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC1EFFBFFFFFF7FFF;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"675341\n");
    end
    ReadData1E = 64'hCDE000000007FFFF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'hB8107F68833FB81B;
    ans = 64'h4DD000000007FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"675647\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBD1F52A155869D56;
    ReadData3E = 64'hC00000000800FFFF;
    ans = 64'hC00000000800FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"675953\n");
    end
    ReadData1E = 64'h47FFFFFDFFFFFF7E;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'hC7EFFFFDFFFFFF7C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"676259\n");
    end
    ReadData1E = 64'hC15001FFFFFEFFFF;
    ReadData2E = 64'h0010000C00000000;
    ReadData3E = 64'h58700000037FFFFE;
    ans = 64'h58700000037FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"676565\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'hA8B18549F87BF9E4;
    ans = 64'hBC90000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"676871\n");
    end
    ReadData1E = 64'h43FFFBFFFEFFFFFF;
    ReadData2E = 64'h683FFFFFF8000800;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h6C4FFBFFF70107FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"677177\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA8000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"677483\n");
    end
    ReadData1E = 64'h3FF002000000000F;
    ReadData2E = 64'h401FFFFF80000002;
    ReadData3E = 64'h3F339027CC991E83;
    ans = 64'h40200226E0479942;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"677789\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h40A003FFFFFF0000;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"678095\n");
    end
    ReadData1E = 64'h3FDFFFF8007FFFFF;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h4030009000000000;
    ans = 64'h402F81201FFE0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"678401\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h4025C56E3B070359;
    ReadData3E = 64'h4530020000000000;
    ans = 64'h4530020000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"678707\n");
    end
    ReadData1E = 64'hC01FC0001FFFFFFE;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h401FC0001FFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"679013\n");
    end
    ReadData1E = 64'h43F00000000000FD;
    ReadData2E = 64'h7FE001FE00000000;
    ReadData3E = 64'hC340001800000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"679319\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h7FDFFFFF7FFFFF00;
    ans = 64'h7FDFFFFF7FFFFF00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"679625\n");
    end
    ReadData1E = 64'hC6D0100FFFFFFFFF;
    ReadData2E = 64'hC800000000800006;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'h4EE0101000808085;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"679931\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'hBCA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"680237\n");
    end
    ReadData1E = 64'h801FFFFFFFFFFFDD;
    ReadData2E = 64'h400FDFFFFFEFFFFE;
    ReadData3E = 64'hB81FFC0008000000;
    ans = 64'hB81FFC0008000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"680543\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hB7E1F816D741E8FF;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"680849\n");
    end
    ReadData1E = 64'hC7E200000007FFFF;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'hBFC0000020020000;
    ans = 64'h47E2000000080000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"681155\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FB081FFFFFFFFFE;
    ReadData3E = 64'hBCA001FFFFFFFFE0;
    ans = 64'hBC9DF3BFFFFFFFC0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"681461\n");
    end
    ReadData1E = 64'h40200FBFFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h0000000000000000;
    ans = 64'hC0300FBFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"681767\n");
    end
    ReadData1E = 64'h3DBD04283127119A;
    ReadData2E = 64'hBFBFFF7FFFFFFEFF;
    ReadData3E = 64'hC80FFBFFFFFFFFFA;
    ans = 64'hC80FFBFFFFFFFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"682073\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'hB7F800000003FFFF;
    ans = 64'hBCB0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"682379\n");
    end
    ReadData1E = 64'hED3FFFFFFDFFDFFE;
    ReadData2E = 64'hC1EFF7FFFFDFFFFF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h6F3FF7FFFDE06005;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"682685\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'hBCB0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"682991\n");
    end
    ReadData1E = 64'hBEAFFFFFFC080000;
    ReadData2E = 64'h403F7FFFFFFFFDFF;
    ReadData3E = 64'h3FC0000000000807;
    ans = 64'h3FBFFE0800004E90;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"683297\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h9052000000000000;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"683603\n");
    end
    ReadData1E = 64'h458FFFFC0000007F;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'hBFB0007FFEFFFFFE;
    ans = 64'hC59FFFFC0000007F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"683909\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FDFFFFFFF07FFFE;
    ReadData3E = 64'hBF4FFFFFFFEFFEFF;
    ans = 64'hBF4FFFFFFFEFFCFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"684215\n");
    end
    ReadData1E = 64'h3E90001001000000;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"684521\n");
    end
    ReadData1E = 64'hB7E0100000000000;
    ReadData2E = 64'hC00F7FFDFFFFFFFF;
    ReadData3E = 64'h3FD0000001000040;
    ans = 64'h3FD0000001000040;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"684827\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'hBFCFFFFF00000100;
    ans = 64'hBFCFFFFF00000110;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"685133\n");
    end
    ReadData1E = 64'h3E7D7FFFFFFFFFFE;
    ReadData2E = 64'hBEA0010000001000;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBD2DA1D800001D7E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"685439\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'h0000000000000000;
    ans = 64'hBCC0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"685745\n");
    end
    ReadData1E = 64'hB80FFEFFFFFFF000;
    ReadData2E = 64'hB80000000FFF8000;
    ReadData3E = 64'hBFDFFFFFFFBC0000;
    ans = 64'hBFDFFFFFFFBC0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"686051\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h6AB32A11A3CA63BC;
    ReadData3E = 64'hC000000000000000;
    ans = 64'h67632A11A3CA63BD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"686357\n");
    end
    ReadData1E = 64'hC2E44ABF8AECE4F4;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h3A000FFE00000000;
    ans = 64'h43044ABF8AECE4F4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"686663\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h7FDE000003FFFFFF;
    ReadData3E = 64'hFFD5FFFFFFFFFFFF;
    ans = 64'hFFD5FFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"686969\n");
    end
    ReadData1E = 64'h82C0000180000000;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"687275\n");
    end
    ReadData1E = 64'hFFDFAEFD5C6B7ABA;
    ReadData2E = 64'h00220E29C3D865C9;
    ReadData3E = 64'h89013D056D7CB36A;
    ans = 64'hC011E0746CF21C59;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"687581\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'h12EFFFFFFFF007FF;
    ans = 64'hBCC0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"687887\n");
    end
    ReadData1E = 64'hFFF0400000000100;
    ReadData2E = 64'h3F4FFBFFFFFFEFFF;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'hFFF8400000000100;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"688193\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"688499\n");
    end
    ReadData1E = 64'h381FFFFFFD7FFFFF;
    ReadData2E = 64'h403FFFE000000001;
    ReadData3E = 64'hC0107FFF7FFFFFFE;
    ans = 64'hC0107FFF7FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"688805\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h3FE0080000000008;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"689111\n");
    end
    ReadData1E = 64'h37E000007FFFFFFD;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hBF5DF62C415BD684;
    ans = 64'hBF5DF62C415BD684;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"689417\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC3FFFFFFFFFFFFFE;
    ReadData3E = 64'h3F10000002FFFFFF;
    ans = 64'hC0AFFFFFF7FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"689723\n");
    end
    ReadData1E = 64'hCC82000000001FFF;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h4FD2000000001FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"690029\n");
    end
    ReadData1E = 64'h4035714CDD36D597;
    ReadData2E = 64'hD99008000001FFFF;
    ReadData3E = 64'h7FDE800000000000;
    ans = 64'h7FDE800000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"690335\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h477FFE0000200000;
    ans = 64'h477FFE0000200000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"690641\n");
    end
    ReadData1E = 64'hDA2001FFFFFFF000;
    ReadData2E = 64'hCE77EAC3142B09A3;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h68A7EDC06C8D7719;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"690947\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"691253\n");
    end
    ReadData1E = 64'hC2EFFFFFFFE0007E;
    ReadData2E = 64'hBFE745C4C8FE6D0F;
    ReadData3E = 64'h402000008000007E;
    ans = 64'h42E745C4C8E728A6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"691559\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC00FFFEFFFFF7FFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"691865\n");
    end
    ReadData1E = 64'h41300010000FFFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h3FD1000000001FFE;
    ans = 64'hC4900010000FFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"692171\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hBFBFFFFDEFFFFFFE;
    ReadData3E = 64'hBF00000000000BFF;
    ans = 64'hBF000000000013FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"692477\n");
    end
    ReadData1E = 64'hBCA4DA12D074B723;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hC33FFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"692783\n");
    end
    ReadData1E = 64'hC0F000000000013F;
    ReadData2E = 64'hC3F00080000000FF;
    ReadData3E = 64'hC3E80A604241B1B5;
    ans = 64'h44F00073FACFE11D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"693089\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'hA25FEFFFFFF00000;
    ans = 64'hFC90000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"693395\n");
    end
    ReadData1E = 64'hC1B000040FFFFFFF;
    ReadData2E = 64'h429003FFFFC00000;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'hC450040410C3FFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"693701\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'hFC90000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"694007\n");
    end
    ReadData1E = 64'h400000000000807F;
    ReadData2E = 64'hBFB00000807FFFFF;
    ReadData3E = 64'hC1CFFFFFFFF003FF;
    ans = 64'hC1D0000000000200;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"694313\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hB81FFFFFDC000000;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"694619\n");
    end
    ReadData1E = 64'hC00FFFFFDFF7FFFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'hC01000010000007F;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"694925\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC3D8C4C50B88973B;
    ReadData3E = 64'h3FF801FFFFFFFFFE;
    ans = 64'hC088B8C40B88973D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"695231\n");
    end
    ReadData1E = 64'h41F00000FFFFFFFD;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'h4010000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"695537\n");
    end
    ReadData1E = 64'h4000000001FFFFFF;
    ReadData2E = 64'hC03D3AEEE26F78DA;
    ReadData3E = 64'h4070200001000000;
    ans = 64'h4068F144487A4A53;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"695843\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h38000000008000FF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"696149\n");
    end
    ReadData1E = 64'h801FFFFFFFFFFFFE;
    ReadData2E = 64'h47F0080200000000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"696455\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"696761\n");
    end
    ReadData1E = 64'hB80FC10000000000;
    ReadData2E = 64'hC3FFFDFFFFFFFEFF;
    ReadData3E = 64'h491C0000001FFFFF;
    ans = 64'h491C0000001FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"697067\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hC0327E3DBCE6EF3A;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hBFD0000000000025;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"697373\n");
    end
    ReadData1E = 64'hC0153BC492FC310A;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h8000001FFEFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"697679\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'h37F139BA63EE665E;
    ReadData3E = 64'hB7E0000000002040;
    ans = 64'hB7E000000000203F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"697985\n");
    end
    ReadData1E = 64'h47F4000000007FFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"698291\n");
    end
    ReadData1E = 64'h00000001FFFFFF00;
    ReadData2E = 64'h3FCFFF9FFFFFFFFE;
    ReadData3E = 64'h381FFFFFFFFFEFBF;
    ans = 64'h381FFFFFFFFFEFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"698597\n");
    end
    ReadData1E = 64'h3CA0000000000001;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'hFFFF5FFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"698903\n");
    end
    ReadData1E = 64'hBC505D00DA08A9C0;
    ReadData2E = 64'hC6A24FCBC7903757;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"699209\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h4010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"699515\n");
    end
    ReadData1E = 64'h3FFAC71E6E51C317;
    ReadData2E = 64'h91B0000000003FEF;
    ReadData3E = 64'h43E80957D015CBB2;
    ans = 64'h43E80957D015CBB2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"699821\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hFFE000013FFFFFFF;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'hFCA000013FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"700127\n");
    end
    ReadData1E = 64'hC1D21AEBDD830161;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'hC03FFFFEFFFFFFFE;
    ans = 64'hC03FFFFEFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"700433\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h4FC1D94A43C02150;
    ReadData3E = 64'hB7E001FDFFFFFFFF;
    ans = 64'h4C81D94A43C0214F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"700739\n");
    end
    ReadData1E = 64'h4070B5D5F5B8F37E;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"701045\n");
    end
    ReadData1E = 64'h2A20000010000800;
    ReadData2E = 64'h3EDFFFFFDBFFFFFF;
    ReadData3E = 64'hC01C22E3E5DB9124;
    ans = 64'hC01C22E3E5DB9124;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"701351\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h4B2FFFFFFFB7FFFF;
    ans = 64'h4B2FFFFFFFB7FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"701657\n");
    end
    ReadData1E = 64'hC3F3FEFFFFFFFFFF;
    ReadData2E = 64'h7FE000000002007F;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"701963\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"702269\n");
    end
    ReadData1E = 64'h3F60000001FFFFEE;
    ReadData2E = 64'h0000000007FFFFFD;
    ReadData3E = 64'hC1CFFFFFFEBFFFFF;
    ans = 64'hC1CFFFFFFEBFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"702575\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h401FFFFFFFFC07FF;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h3CDFFFFFFFFC07FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"702881\n");
    end
    ReadData1E = 64'hC2FFFFFE000007FF;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h835003FFF7FFFFFF;
    ans = 64'h835203FFD800007F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"703187\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h0B700000004003FF;
    ReadData3E = 64'h002000FFFFFFEFFF;
    ans = 64'h08300000004003FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"703493\n");
    end
    ReadData1E = 64'h3FA001FFE0000000;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"703799\n");
    end
    ReadData1E = 64'h4040800200000000;
    ReadData2E = 64'h3FC7D33A1738766D;
    ReadData3E = 64'h8000000007FFFF7F;
    ans = 64'h401891D6E2597D07;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"704105\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h8018A1B4427BBEB8;
    ans = 64'h8018A1B4427BBEB6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"704411\n");
    end
    ReadData1E = 64'h37FFFFF3FFFFFFFE;
    ReadData2E = 64'hBFB0100000000002;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"704717\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"705023\n");
    end
    ReadData1E = 64'hC0A000000000FFFE;
    ReadData2E = 64'hC3FFFFFFFFFFFFB8;
    ReadData3E = 64'hB7E01FFFFFFFFF80;
    ans = 64'h44B000000000FFDA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"705329\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC34F0AC72AA75605;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'hC00F0AC72AA75604;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"705635\n");
    end
    ReadData1E = 64'hB86FFFFFFF820000;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hB8D00002000FFFFF;
    ans = 64'hB8D00002000FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"705941\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h41DFFFFFE00001FF;
    ReadData3E = 64'h41C3C27FCF1953FB;
    ans = 64'h41C3C27FCF1953FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"706247\n");
    end
    ReadData1E = 64'hBF3003F7FFFFFFFF;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"706553\n");
    end
    ReadData1E = 64'h40AFFFFFFFBEFFFF;
    ReadData2E = 64'hC1F5BB80746EB9E1;
    ReadData3E = 64'h7FD7FFFFFBFFFFFE;
    ans = 64'h7FD7FFFFFBFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"706859\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h3F8FFFFEFFFFFFBF;
    ans = 64'h3F8FFFFEFFFFFFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"707165\n");
    end
    ReadData1E = 64'hC7D0007FEFFFFFFF;
    ReadData2E = 64'hC18CE14D9BCE0B02;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h496CE23489599BD5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"707471\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"707777\n");
    end
    ReadData1E = 64'hC1F2000FFFFFFFFF;
    ReadData2E = 64'h4020000000080008;
    ReadData3E = 64'h4030003FFFFF7FFF;
    ans = 64'hC222000FFFE8FF90;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"708083\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h5D100000003FFFFE;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h59D00000003FFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"708389\n");
    end
    ReadData1E = 64'hDD90003FFFFFEFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hBA1FF7FFFFFFE000;
    ans = 64'hDA50003FFFFFEFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"708695\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h47EFF0001FFFFFFE;
    ReadData3E = 64'hC1E0001FFFFFFFFF;
    ans = 64'h44AFF0001FFFFEFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"709001\n");
    end
    ReadData1E = 64'hB81FFFFFF7F7FFFF;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"709307\n");
    end
    ReadData1E = 64'h5CBD7EB455888645;
    ReadData2E = 64'hBCA010000000007F;
    ReadData3E = 64'h40E3FFFFFFFFFFFE;
    ans = 64'hD96D9C3309DE0FB5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"709613\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hBF51E1746AB411A1;
    ans = 64'hBF51E1746AB410A1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"709919\n");
    end
    ReadData1E = 64'h3CAFFFFFE3FFFFFF;
    ReadData2E = 64'hC0200000FEFFFFFF;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC010000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"710225\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"710531\n");
    end
    ReadData1E = 64'h802DB8D352BF1533;
    ReadData2E = 64'h3ED78B4045674233;
    ReadData3E = 64'hFFDFFFBF7FFFFFFF;
    ans = 64'hFFDFFFBF7FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"710837\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h43C000000000021E;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"711143\n");
    end
    ReadData1E = 64'hBF9FFFFFFF000007;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'hC090001FFFFFFFFF;
    ans = 64'hC090002FFFFFFF7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"711449\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBFA0000000007FC0;
    ReadData3E = 64'h802F6FEF6C4FB6E9;
    ans = 64'hBC60000000007FBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"711755\n");
    end
    ReadData1E = 64'h3F3BFBFFFFFFFFFF;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3F2BFC0000001FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"712061\n");
    end
    ReadData1E = 64'h4142CAC584401A28;
    ReadData2E = 64'hFFDBAC0C29D4EE17;
    ReadData3E = 64'hBFB003EAEA289C1E;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"712367\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'hBDF82D162E8BAF54;
    ans = 64'hBDF82D15AE8BAF54;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"712673\n");
    end
    ReadData1E = 64'h40FFFF000003FFFF;
    ReadData2E = 64'hC0088FAE85CFB9E8;
    ReadData3E = 64'h4000000000000001;
    ans = 64'hC1188EE2085E9D5F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"712979\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"713285\n");
    end
    ReadData1E = 64'h9ECF000000004000;
    ReadData2E = 64'h7FEFFFFBFFFFFF80;
    ReadData3E = 64'hC1DFFFFF80001FFE;
    ans = 64'hDECEFFFC20003F84;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"713591\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC7F5FAB540D0DED9;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"713897\n");
    end
    ReadData1E = 64'hBFDFFFFFFC3FFFFE;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'hBF6E00000000003F;
    ans = 64'hBFE01DFFFE1FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"714203\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h406FFFFFEFFFFDFF;
    ReadData3E = 64'hB211FFFD53E700AE;
    ans = 64'h3D2FFFFFEFFFFDFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"714509\n");
    end
    ReadData1E = 64'h002FFFFDFFFEFFFE;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"714815\n");
    end
    ReadData1E = 64'hC7F00000020003FF;
    ReadData2E = 64'hBFE00021FFFFFFFF;
    ReadData3E = 64'hC0AFF800003FFFFF;
    ans = 64'h47E000220200083E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"715121\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hB81D1619C3444AAD;
    ans = 64'h3CAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"715427\n");
    end
    ReadData1E = 64'hC3C8F81F205C84C5;
    ReadData2E = 64'hBD9000007FFFFFF7;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'h4168F81FA81D7DBA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"715733\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CBFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"716039\n");
    end
    ReadData1E = 64'hBFBFFFFFFC3FFFFF;
    ReadData2E = 64'h3FBFFF0000007FFF;
    ReadData3E = 64'h3F8FFFFFFFFFFE03;
    ans = 64'h3EC0003BF6005000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"716345\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC04FDFFBFFFFFFFE;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"716651\n");
    end
    ReadData1E = 64'h43CFFFFFFFFF3FFF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'hBFCBFECEE81F8C27;
    ans = 64'h43DFFFFFFFFF3FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"716957\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FE0000000000036;
    ReadData3E = 64'hBFC0080000200000;
    ans = 64'hBFC00800001FFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"717263\n");
    end
    ReadData1E = 64'h41EA89148D512055;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h41FA89148D512053;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"717569\n");
    end
    ReadData1E = 64'h390FFFFEFFFF7FFE;
    ReadData2E = 64'h40B1FFFFFFFF0000;
    ReadData3E = 64'h224FFFFFF8000006;
    ans = 64'h39D1FFFF6FFEB7FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"717875\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h37F000003BFFFFFF;
    ans = 64'h3CBFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"718181\n");
    end
    ReadData1E = 64'h71EFFFF7FFFFFDFF;
    ReadData2E = 64'h8008BDF18FA06A55;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"718487\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'h3CB7FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"718793\n");
    end
    ReadData1E = 64'h3FCFFFFFFF800007;
    ReadData2E = 64'h7FD3FFFFFFFFF000;
    ReadData3E = 64'h380E3FFFFFFFFFFF;
    ans = 64'h7FB3FFFFFFAFF004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"719099\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h41EFFFFCFFFFFFFF;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"719405\n");
    end
    ReadData1E = 64'h3FF00000000000EE;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'h15D0000002000FFF;
    ans = 64'h40000000000000EF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"719711\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBF13670077E329FC;
    ReadData3E = 64'h66EFFFEFFFFFFFFC;
    ans = 64'h66EFFFEFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"720017\n");
    end
    ReadData1E = 64'hBF00005FFFFFFFFF;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'hBF20005FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"720323\n");
    end
    ReadData1E = 64'h483FFFFFFFFF7FBE;
    ReadData2E = 64'h51300400000003FE;
    ReadData3E = 64'h37E0008000400000;
    ans = 64'h598003FFFFFFC3CD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"720629\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'hC01000000007FFFB;
    ans = 64'hC01000000007FFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"720935\n");
    end
    ReadData1E = 64'h802FFFFFFF000080;
    ReadData2E = 64'h3F00000000FFFFE0;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"721241\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h3CCFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"721547\n");
    end
    ReadData1E = 64'h3F100007FFFFF7FF;
    ReadData2E = 64'h5AEFFFFFFFDFFFFD;
    ReadData3E = 64'hBE7FFFFFFFF7FFFB;
    ans = 64'h5A100007FFEFF7F5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"721853\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC0E00007FFF7FFFF;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000002001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"722159\n");
    end
    ReadData1E = 64'h37FDFDFFFFFFFFFF;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h2B9BC6B387D1ACCE;
    ans = 64'h381DFE0000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"722465\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBFCFC000000000FE;
    ReadData3E = 64'hD65E3F9A7F0FAF49;
    ans = 64'hD65E3F9A7F0FAF49;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"722771\n");
    end
    ReadData1E = 64'hB800000800001000;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"723077\n");
    end
    ReadData1E = 64'h3FBFF80004000000;
    ReadData2E = 64'hC06FFFBFFFFFFFFC;
    ReadData3E = 64'h40300000000007BF;
    ans = 64'hC02FEF8027FFE07A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"723383\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'h4C40000000000000;
    ans = 64'h4C40000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"723689\n");
    end
    ReadData1E = 64'h3F30001000000007;
    ReadData2E = 64'h800000000000007E;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"723995\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h3CDFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"724301\n");
    end
    ReadData1E = 64'hC01FFFFFFFFDFFFB;
    ReadData2E = 64'h3FE15EA250BE553F;
    ReadData3E = 64'h3FBFFFFFFFFFFFED;
    ans = 64'hC010DEA250BD3F52;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"724607\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBF79C78656F15D19;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h400FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"724913\n");
    end
    ReadData1E = 64'h41E540947D9D3B75;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'h95FFFBFFFFF7FFFF;
    ans = 64'h453540947D9D3B75;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"725219\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h400FFDFFFFFFBFFF;
    ReadData3E = 64'h472FFFFFFFFF81FF;
    ans = 64'h472FFFFFFFFF81FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"725525\n");
    end
    ReadData1E = 64'hD0FFFEFFFFFFFFC0;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"725831\n");
    end
    ReadData1E = 64'h401989593420B76C;
    ReadData2E = 64'hC1E4F4FFF9EF9D25;
    ReadData3E = 64'h05800FFFFFFFFFC0;
    ans = 64'hC210B95B4EB4B8D6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"726137\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h40322892719C4A7F;
    ans = 64'h40362892719C4A7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"726443\n");
    end
    ReadData1E = 64'hBF0998BC3CFD5996;
    ReadData2E = 64'hEEE00007FFF7FFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'h6DF998C9094EABB5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"726749\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"727055\n");
    end
    ReadData1E = 64'h37FCCB1A46C03634;
    ReadData2E = 64'h40A000400001FFFF;
    ReadData3E = 64'hB7FFFFFFC00003FF;
    ans = 64'h38ACC78D7334EA96;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"727361\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h40EF9E572874A507;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFF030B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"727667\n");
    end
    ReadData1E = 64'hC1F001FFFFFFFFFC;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'hDC492B5C56F6CEC6;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"727973\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h801FFFFFFDFFFEFF;
    ReadData3E = 64'h43E9488598C128E1;
    ans = 64'h43E9488598C128E1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"728279\n");
    end
    ReadData1E = 64'hC030000000000037;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"728585\n");
    end
    ReadData1E = 64'h97200000040001FF;
    ReadData2E = 64'h7FD00000000003FF;
    ReadData3E = 64'h00200001EFFFFFFF;
    ans = 64'hD7000000040005FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"728891\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h8CCEC95BB8DB66DD;
    ans = 64'h7CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"729197\n");
    end
    ReadData1E = 64'h417010000FFFFFFF;
    ReadData2E = 64'hC03FFC0000000000;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'hC1C00DFE0FFDFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"729503\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"729809\n");
    end
    ReadData1E = 64'hBDBBFF0000000000;
    ReadData2E = 64'h802FFFFFFFFC003F;
    ReadData3E = 64'hC010000006FFFFFF;
    ans = 64'hC010000006FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"730115\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hB7E2CD542725A7DF;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h3FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"730421\n");
    end
    ReadData1E = 64'hC03000000000021F;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h43DE0340FC13DE79;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"730727\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FB0000000002000;
    ReadData3E = 64'hC4F000000003FFF7;
    ans = 64'hC4F000000003FFF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"731033\n");
    end
    ReadData1E = 64'h41D3EC019940A24C;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"731339\n");
    end
    ReadData1E = 64'hFFE00001FFFFFFF7;
    ReadData2E = 64'hC01B84F8B54B3138;
    ReadData3E = 64'hBFC00FFFFFFFFFBF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"731645\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h8930007FFFFFFFEF;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"731951\n");
    end
    ReadData1E = 64'hBFBFFFFFFF1FFFFE;
    ReadData2E = 64'h41DFDFC000000000;
    ReadData3E = 64'h8010000000000000;
    ans = 64'hC1AFDFBFFF20E1BE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"732257\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"732563\n");
    end
    ReadData1E = 64'hBF9FFFFE00000000;
    ReadData2E = 64'h402FFFFFFFFFF80F;
    ReadData3E = 64'hC3C2F72E444107D4;
    ans = 64'hC3C2F72E444107D4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"732869\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC06101FFFFFFFFFE;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hBFE0000000000111;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"733175\n");
    end
    ReadData1E = 64'hD28FFFFFFFF800FE;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC3F000000041FFFF;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"733481\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h4010400001FFFFFE;
    ReadData3E = 64'hBF447AB4C8297DFA;
    ans = 64'hBF447AB4C8295D7A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"733787\n");
    end
    ReadData1E = 64'hC3C32F485BF17E30;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"734093\n");
    end
    ReadData1E = 64'hBFB0000400008000;
    ReadData2E = 64'h480219B474B1BFFE;
    ReadData3E = 64'h37EFFFFFFFF7FF7E;
    ans = 64'hC7C219B8FB1F6DF8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"734399\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h43EFFFFFFFFFFFF3;
    ans = 64'h43EFFFFFFFFFFFF3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"734705\n");
    end
    ReadData1E = 64'h511FFFFF00FFFFFE;
    ReadData2E = 64'h00000007FFFFFFBF;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h102FFFFEFFFC0006;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"735011\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"735317\n");
    end
    ReadData1E = 64'h3F8FFBFFFFFFF7FF;
    ReadData2E = 64'h374FFFF7FFFFBFFE;
    ReadData3E = 64'h37E01FFFFFFFFFFE;
    ans = 64'h37E0201FFBF800FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"735623\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h47FFFFFFFFF8001F;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h44BFFFFFFFF8001E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"735929\n");
    end
    ReadData1E = 64'hC800000800000040;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"736235\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h8000000000000028;
    ReadData3E = 64'h6A28080000000000;
    ans = 64'h6A28080000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"736541\n");
    end
    ReadData1E = 64'hBC6007FFFFFEFFFF;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"736847\n");
    end
    ReadData1E = 64'h8000000110000000;
    ReadData2E = 64'hC2F0000080200000;
    ReadData3E = 64'h400FBFFFFFFFDFFF;
    ans = 64'h400FBFFFFFFFDFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"737153\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'hBFBFFDFFFFFFFFBE;
    ans = 64'hBFBFFDFFFFFFFFBE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"737459\n");
    end
    ReadData1E = 64'hC7FFFFFFEFF7FFFF;
    ReadData2E = 64'h3E8FFFFFBFFFFFFC;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"737765\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"738071\n");
    end
    ReadData1E = 64'hAD200000000047FE;
    ReadData2E = 64'h404F2BAADD169721;
    ReadData3E = 64'hC0C001FEFFFFFFFF;
    ans = 64'hC0C001FEFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"738377\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hB813FDFFFFFFFFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"738683\n");
    end
    ReadData1E = 64'h456000000007BFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h4E7FFFFF70000000;
    ans = 64'h4E7FFFFF70000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"738989\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC1FA8B35CF63049E;
    ReadData3E = 64'hBFDFF7FDFFFFFFFE;
    ans = 64'hBFDFF804A2CD73D7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"739295\n");
    end
    ReadData1E = 64'h41CD859AA5735B17;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"739601\n");
    end
    ReadData1E = 64'h330FFFFFFBFF7FFE;
    ReadData2E = 64'hC00F7FFFFDFFFFFF;
    ReadData3E = 64'h3FC00003FFF7FFFE;
    ans = 64'h3FC00003FFF7FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"739907\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'hC1F007FFFBFFFFFE;
    ans = 64'hC1F007FFFBFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"740213\n");
    end
    ReadData1E = 64'h407FFFFFFF00001E;
    ReadData2E = 64'h3FF00000000001FB;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"740519\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"740825\n");
    end
    ReadData1E = 64'hC0207FFFFFFFF7FE;
    ReadData2E = 64'hC3FFBFFFEFFFFFFF;
    ReadData3E = 64'h3FE03FFFFFFFF000;
    ans = 64'h44305EFFF7BFF80E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"741131\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FFFFFFF80000007;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CC3FFFFC0000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"741437\n");
    end
    ReadData1E = 64'hC1F02007FFFFFFFF;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'hBFD36778D0F4A90A;
    ans = 64'hBFD36776CCF3A90A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"741743\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h08BFFFFFBFFBFFFE;
    ReadData3E = 64'hBFD9B3973EE16671;
    ans = 64'hBFD9B3973EE16671;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"742049\n");
    end
    ReadData1E = 64'h36DFFFDFFFFFFF7E;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"742355\n");
    end
    ReadData1E = 64'h401EFFFFFFFF0000;
    ReadData2E = 64'h41E00000021FFFFF;
    ReadData3E = 64'h7FF0000001000004;
    ans = 64'h7FF8000001000004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"742661\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'hC01FEFFFFFE00000;
    ans = 64'hC01FEFFFFFE00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"742967\n");
    end
    ReadData1E = 64'hC7800047FFFFFFFF;
    ReadData2E = 64'h3818718D1A7C2293;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"743273\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"743579\n");
    end
    ReadData1E = 64'hC24FFFFF80020000;
    ReadData2E = 64'h37FFFFF400000000;
    ReadData3E = 64'hC1DFFFFFFFFF0001;
    ans = 64'hC1DFFFFFFFFF0001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"743885\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3DDFFFFFFFFF000F;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h3A9FFFFFFFFF000E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"744191\n");
    end
    ReadData1E = 64'hC0C0041FFFFFFFFF;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hB81000000FFFFFDF;
    ans = 64'h40A0041FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"744497\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC2500003FFFFFFF8;
    ReadData3E = 64'hC75FFFF7FFFFFFC0;
    ans = 64'hC75FFFF7FFFFFFC0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"744803\n");
    end
    ReadData1E = 64'hDF2007FFFFFEFFFF;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'h5F0007FFFFFF0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"745109\n");
    end
    ReadData1E = 64'h3FB00002007FFFFE;
    ReadData2E = 64'h47FE08445C9805CA;
    ReadData3E = 64'hFFDE800000000000;
    ans = 64'hFFDE800000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"745415\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'hBFFFFF8000002000;
    ans = 64'hBFFFFF8000002000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"745721\n");
    end
    ReadData1E = 64'hC03FFFFFFFC7FFFF;
    ReadData2E = 64'h403000000000040E;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000101;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"746027\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"746333\n");
    end
    ReadData1E = 64'hFFD0000023FFFFFE;
    ReadData2E = 64'hC3DFFFE07FFFFFFF;
    ReadData3E = 64'hB80FFDFFFFFFFEFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"746639\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC03B1DEB7D342DFF;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'hBCFB1DEB7D342DFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"746945\n");
    end
    ReadData1E = 64'h37FFC3CD8026EDA0;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'hBEAB43550AFC66D5;
    ans = 64'hBEAB43550AFC66D5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"747251\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h43C00003FDFFFFFF;
    ReadData3E = 64'hBEBDFFFFFEFFFFFE;
    ans = 64'h40800003FD0FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"747557\n");
    end
    ReadData1E = 64'hECA939F14EB859D3;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h6C9939F14EB859D5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"747863\n");
    end
    ReadData1E = 64'hC7BDA4A9027C7292;
    ReadData2E = 64'hBFFFF80000000003;
    ReadData3E = 64'h3EC0FFFFFFFBFFFF;
    ans = 64'h47CD9D3FD83BD378;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"748169\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h4030080FFFFFFFFF;
    ans = 64'h4030080FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"748475\n");
    end
    ReadData1E = 64'h434FFFE000000000;
    ReadData2E = 64'h43E0007FFFFFFFFD;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h4740006FFF7FFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"748781\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"749087\n");
    end
    ReadData1E = 64'h801BFBE73AF53F7E;
    ReadData2E = 64'hBFFFC0007FFFFFFF;
    ReadData3E = 64'h480FFFFFF7FDFFFE;
    ans = 64'h480FFFFFF7FDFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"749393\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3E4000FFF8000000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"749699\n");
    end
    ReadData1E = 64'hBF6FE01000000000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'hBE7636E818A167A2;
    ans = 64'h3F6FDFE3922FCEBB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"750005\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hB32832CD4DA733A1;
    ReadData3E = 64'hB7F0010000100000;
    ans = 64'hB7F0010000100000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"750311\n");
    end
    ReadData1E = 64'h3CAAA8F7EF2045EC;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"750617\n");
    end
    ReadData1E = 64'hC0A00FFFFFFF0000;
    ReadData2E = 64'hC016EE0768D982C6;
    ReadData3E = 64'hBFEFFC0000000FFF;
    ans = 64'h40C704758040ED68;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"750923\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'hBEC0000001F7FFFF;
    ans = 64'hBEC0000001FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"751229\n");
    end
    ReadData1E = 64'hB7F2000000000FFF;
    ReadData2E = 64'h7FDFFFDFFBFFFFFF;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hF7E1FFEDFDC00FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"751535\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FDFFFFFFFFFFFF8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"751841\n");
    end
    ReadData1E = 64'hB780000003FFFFDF;
    ReadData2E = 64'h3FC1FFFFFFBFFFFF;
    ReadData3E = 64'hBFB6BDB75B174AF1;
    ans = 64'hBFB6BDB75B174AF1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"752147\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC01FFE0000020000;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"752453\n");
    end
    ReadData1E = 64'h402FFFFFFFFFFFFE;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h5EFB355B02575E41;
    ans = 64'h5EFB355B02575E41;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"752759\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC01000007EFFFFFF;
    ReadData3E = 64'hC0300100003FFFFF;
    ans = 64'hC0300100003FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"753065\n");
    end
    ReadData1E = 64'h3F400BFFFFFFFFFF;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'hBF500BFFFFFFFBFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"753371\n");
    end
    ReadData1E = 64'hC018F904A86762F5;
    ReadData2E = 64'hC0ABCA88E32554C2;
    ReadData3E = 64'hC2248DCD904C6353;
    ans = 64'hC2248DCCE2CB0DDA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"753677\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h37EFFFFFFFFC01FF;
    ans = 64'hBCBFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"753983\n");
    end
    ReadData1E = 64'hC7FFF7FFFFFFDFFE;
    ReadData2E = 64'h0010000080000FFF;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"754289\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000009;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"754595\n");
    end
    ReadData1E = 64'hFFD79DB1F2CD62F9;
    ReadData2E = 64'hBABFFFFBFFFFFFFA;
    ReadData3E = 64'h40B00001FFFFFFFF;
    ans = 64'h7AA79DAEFF17249B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"754901\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h7FDE00000001FFFF;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"755207\n");
    end
    ReadData1E = 64'h47FFFC0000000FFE;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h43F0000000200002;
    ans = 64'hC81FFC0000000FFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"755513\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC02FFFFFDFFFBFFE;
    ReadData3E = 64'h7FD00000000FFFF6;
    ans = 64'h7FD00000000FFFF6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"755819\n");
    end
    ReadData1E = 64'h39B00020000FFFFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hB9D00020000FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"756125\n");
    end
    ReadData1E = 64'hB588000800000000;
    ReadData2E = 64'h000BBA2331B70CC7;
    ReadData3E = 64'hADC0FFFFFFFBFFFF;
    ans = 64'hADC0FFFFFFFBFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"756431\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'hC3FFFFC002000000;
    ans = 64'hC3FFFFC002000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"756737\n");
    end
    ReadData1E = 64'h418000000007FFE0;
    ReadData2E = 64'hBFD00003FFFBFFFE;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hC16000042003FFE0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"757043\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'hBCC8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"757349\n");
    end
    ReadData1E = 64'hBCAC000000000003;
    ReadData2E = 64'h41F0000000001E00;
    ReadData3E = 64'h403000000000000A;
    ans = 64'h402FFFFFE4000014;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"757655\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hB67629A54B42C449;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"757961\n");
    end
    ReadData1E = 64'hF7F75C323BE19773;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h3CAFFFFFFFFFFFC1;
    ans = 64'h78275C323BE19772;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"758267\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h37EFFFFFEFFFDFFE;
    ReadData3E = 64'h4F000004000003FF;
    ans = 64'h4F000004000003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"758573\n");
    end
    ReadData1E = 64'h43DC3CB2A36EEDF9;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'h0010000000000000;
    ans = 64'hC40C3CB2A36EEDF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"758879\n");
    end
    ReadData1E = 64'hC030007FFFFFFBFF;
    ReadData2E = 64'h0011000001FFFFFE;
    ReadData3E = 64'hC1FFFFDBFFFFFFFE;
    ans = 64'hC1FFFFDBFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"759185\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'h47FFFFFFFFFFAFFF;
    ans = 64'h47FFFFFFFFFFAFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"759491\n");
    end
    ReadData1E = 64'hBCAFF81FFFFFFFFE;
    ReadData2E = 64'hFFDCF736D2BEB917;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h7C9CF015FC40DC26;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"759797\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hBFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"760103\n");
    end
    ReadData1E = 64'hC0600FFFFF7FFFFF;
    ReadData2E = 64'h43D0000000007800;
    ReadData3E = 64'h9326AC824E53D1EB;
    ans = 64'hC4400FFFFF807877;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"760409\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3A523E7FF4DA7D44;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"760715\n");
    end
    ReadData1E = 64'h37E35FEF92C012CE;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'hB7FF7FFFFFF80000;
    ans = 64'hBB335FEF92C012D1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"761021\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h40706E21CD654F01;
    ReadData3E = 64'hBCA000000000083E;
    ans = 64'h3D306621CD654EFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"761327\n");
    end
    ReadData1E = 64'h00279DCB6C5826C9;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h83879DCB6C5826C8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"761633\n");
    end
    ReadData1E = 64'hC00000000FFFFFF7;
    ReadData2E = 64'h3FEFFF87FFFFFFFE;
    ReadData3E = 64'h3F68E540420AA576;
    ans = 64'hBFFFF3157FDE8299;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"761939\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'hC1DFFFFFF0004000;
    ans = 64'hC1DFFFFFF1004000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"762245\n");
    end
    ReadData1E = 64'hBEC7FE99C03ACAB9;
    ReadData2E = 64'h33D00000FFFFDFFE;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"762551\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'h0010000000000000;
    ans = 64'hFC9FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"762857\n");
    end
    ReadData1E = 64'hC03FFEFFDFFFFFFF;
    ReadData2E = 64'hC001000004000000;
    ReadData3E = 64'h7FDFFFFFFFFFDFFE;
    ans = 64'h7FDFFFFFFFFFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"763163\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h7FEFEF7FFFFFFFFF;
    ReadData3E = 64'hC010000000000000;
    ans = 64'h7CAFEF7FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"763469\n");
    end
    ReadData1E = 64'h5A3FEB6457A93869;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hC1CE7828510813A0;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"763775\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hBFDFFFF7FFF7FFFE;
    ReadData3E = 64'hC3C00000BFFFFFFF;
    ans = 64'hC3C00000BFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"764081\n");
    end
    ReadData1E = 64'hC01F80000000000E;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"764387\n");
    end
    ReadData1E = 64'h4D4FEFFFFFFFF7FF;
    ReadData2E = 64'hFFFFEFFFEFFFFFFE;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'hFFFFEFFFEFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"764693\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h8C60003FFFC00000;
    ans = 64'hFCAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"764999\n");
    end
    ReadData1E = 64'hC01FFE0000000FFE;
    ReadData2E = 64'hA3000007FFFFFFFD;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"765305\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'h8000000000000001;
    ans = 64'hFCAFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"765611\n");
    end
    ReadData1E = 64'h40F000000FFE0000;
    ReadData2E = 64'hBFBFBFEFFFFFFFFF;
    ReadData3E = 64'h402904B537BDEEA5;
    ans = 64'hC0BFB36DC520190A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"765917\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h434FFFFFC00003FF;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4017FFFFE0000200;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"766223\n");
    end
    ReadData1E = 64'hBFBD6817418AD7FF;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'hAAD0400001FFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"766529\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h3FE00800007FFFFF;
    ReadData3E = 64'hC1C0000038000000;
    ans = 64'hC1C0000038000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"766835\n");
    end
    ReadData1E = 64'h434216727B125476;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"767141\n");
    end
    ReadData1E = 64'h8C0FFEFFFFFFFE00;
    ReadData2E = 64'h4360000000FFFFEF;
    ReadData3E = 64'h43F000003FFFFE00;
    ans = 64'h43F000003FFFFE00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"767447\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h403FFFFFFFBDFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"767753\n");
    end
    ReadData1E = 64'h43CDFC0000000000;
    ReadData2E = 64'h3800FE000D1AEA44;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBC9FFE027FF6770F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"768059\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"768365\n");
    end
    ReadData1E = 64'h3FEDFFC000000000;
    ReadData2E = 64'hB680007C00000000;
    ReadData3E = 64'hC1E001000FFFFFFF;
    ans = 64'hC1E001000FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"768671\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFF;
    ReadData2E = 64'h380CE2920341D715;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hBFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"768977\n");
    end
    ReadData1E = 64'hC2201001FFFFFFFF;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'hD32FFFFFFF803FFE;
    ans = 64'hD32FFFFFFF803FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"769283\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBF2400FFFFFFFFFE;
    ReadData3E = 64'h3810008000100000;
    ans = 64'hBBE400FFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"769589\n");
    end
    ReadData1E = 64'hBCA20FFFFFFFFFFE;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"769895\n");
    end
    ReadData1E = 64'hB810040000000400;
    ReadData2E = 64'h3FD0004000000080;
    ReadData3E = 64'h50C07FFFFFFBFFFF;
    ans = 64'h50C07FFFFFFBFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"770201\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'hBF0FFFDFBFFFFFFE;
    ans = 64'hBF0FFFDFBFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"770507\n");
    end
    ReadData1E = 64'h400000201FFFFFFF;
    ReadData2E = 64'hC65FFBFEFFFFFFFE;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'hC66FFC3F37F5FDFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"770813\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"771119\n");
    end
    ReadData1E = 64'h400F00007FFFFFFF;
    ReadData2E = 64'h41D000000004FFFF;
    ReadData3E = 64'h381FFBFFFFBFFFFF;
    ans = 64'h41EF00008009AFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"771425\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBFCFFFEFEFFFFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"771731\n");
    end
    ReadData1E = 64'h4020000007FFFC00;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'hC00DBEC4D86AAF0A;
    ans = 64'hC00DBEC4D86AAF0A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"772037\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBF3FFFFFFFEFFDFF;
    ReadData3E = 64'hBFDFFFFFF00003FF;
    ans = 64'hBFDFFFFFF00003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"772343\n");
    end
    ReadData1E = 64'h801FFFFFEBFFFFFE;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"772649\n");
    end
    ReadData1E = 64'hDC1FFBFBFFFFFFFE;
    ReadData2E = 64'h47FFFFFFF4000000;
    ReadData3E = 64'hBF3FFDFFFFFFFFFF;
    ans = 64'hE42FFBFBF401817E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"772955\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'hC1E63CFF38491175;
    ans = 64'hC1E63CFF38491175;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"773261\n");
    end
    ReadData1E = 64'h3800A00000000000;
    ReadData2E = 64'hBFAFFFFF003FFFFF;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'hB7C09FFF7B213FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"773567\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"773873\n");
    end
    ReadData1E = 64'h7D00000001FFFBFE;
    ReadData2E = 64'hC3C0000001000004;
    ReadData3E = 64'hBFC0000000083FFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"774179\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC0156543FCA323DC;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE000000000000B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"774485\n");
    end
    ReadData1E = 64'hBF200001001FFFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'hBD501BFFFFFFFFFF;
    ans = 64'hBD501BFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"774791\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h47EFFF80007FFFFF;
    ReadData3E = 64'h3FEFFFFFEFFF0000;
    ans = 64'h44AFFF80007FFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"775097\n");
    end
    ReadData1E = 64'hC02FFFFFFFF7FFEF;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"775403\n");
    end
    ReadData1E = 64'hBFEC0000000007FF;
    ReadData2E = 64'h000F9D0DF8DF20DF;
    ReadData3E = 64'h32AFFFFF6FFFFFFF;
    ans = 64'h32AFFFFF6FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"775709\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hBFD367D06C6CD192;
    ans = 64'hBFD367D06C6CD192;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"776015\n");
    end
    ReadData1E = 64'hB2EFFFFFFFBFFFFF;
    ReadData2E = 64'h41E0007FC0000000;
    ReadData3E = 64'h0000000000000000;
    ans = 64'hB4E0007FBFDFFF00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"776321\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"776627\n");
    end
    ReadData1E = 64'hE160010000000080;
    ReadData2E = 64'h20E1FFFFFFFFFFFF;
    ReadData3E = 64'h293DF7FFFFFFFFFE;
    ans = 64'hC25201200000008F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"776933\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hFFE4F8119C381900;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'hFCA4F8119C3818FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"777239\n");
    end
    ReadData1E = 64'h40192DFAA3A08E50;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hC7E000800003FFFF;
    ans = 64'hC7E000800003FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"777545\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FFFF80000FFFFFF;
    ReadData3E = 64'hC790000000000006;
    ans = 64'hC790000000000006;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"777851\n");
    end
    ReadData1E = 64'h43CFBFFFFFFBFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h408FDFFFFFFBFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"778157\n");
    end
    ReadData1E = 64'h43EFFFF7FFFF0000;
    ReadData2E = 64'h002000008000001F;
    ReadData3E = 64'h409FFFF7BFFFFFFF;
    ans = 64'h409FFFF7BFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"778463\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'hBCAFFFFFFFE00000;
    ans = 64'hBCA7FFFFFFE00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"778769\n");
    end
    ReadData1E = 64'h480040000000001F;
    ReadData2E = 64'hC3EFFFFFFFFFF03F;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"779075\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"779381\n");
    end
    ReadData1E = 64'hC5B81A8A748009FF;
    ReadData2E = 64'h40100000000000C0;
    ReadData3E = 64'hC0F0001FFFFFFEFE;
    ans = 64'hC5D81A8A74800B20;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"779687\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h43CEFFFFFC000000;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'h408EFFFFFBFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"779993\n");
    end
    ReadData1E = 64'hE337B88BE495EEA7;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h400FFFFFF7EFFFFF;
    ans = 64'hE327B88BE495EEA6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"780299\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h16C929AB0692347E;
    ReadData3E = 64'hC02382DEFC30F29B;
    ans = 64'hC02382DEFC30F29B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"780605\n");
    end
    ReadData1E = 64'h40AFFFBFFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'hC000000000000000;
    ans = 64'h409FF7BFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"780911\n");
    end
    ReadData1E = 64'hBF0FFFF8000000FF;
    ReadData2E = 64'hC1C040000007FFFF;
    ReadData3E = 64'h40F01FFFFFFFFF7F;
    ans = 64'h40F83FFDF803FFBE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"781217\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'hC1FEFFFFFFF7FFFE;
    ans = 64'hC1FEFFFFFFF7FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"781523\n");
    end
    ReadData1E = 64'hC10FFF8000000000;
    ReadData2E = 64'h2D300000003FFFC0;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"781829\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"782135\n");
    end
    ReadData1E = 64'h3EDFFFFE0000000E;
    ReadData2E = 64'h3FE0002FFFFFFFFF;
    ReadData3E = 64'h41D01FFEFFFFFFFE;
    ans = 64'h41D01FFF0000000E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"782441\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC3FFFFF100000000;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'hC0BFFFF0FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"782747\n");
    end
    ReadData1E = 64'hB810000000041000;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'h49660E2E11B09A23;
    ans = 64'h49660E2E11B09A23;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"783053\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h43D010000000007F;
    ReadData3E = 64'hE2C0000010000FFF;
    ans = 64'hE2C0000010000FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"783359\n");
    end
    ReadData1E = 64'hBFC1FEFFFFFFFFFF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FEB804000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"783665\n");
    end
    ReadData1E = 64'hC15FFFFFDFFFFFBF;
    ReadData2E = 64'h37E003FFFF7FFFFE;
    ReadData3E = 64'h3FA000000BFFFFFE;
    ans = 64'h3FA000000BFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"783971\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h3FCFFFF7FEFFFFFF;
    ans = 64'h3FCFFFF7FF000007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"784277\n");
    end
    ReadData1E = 64'hC7EE008000000000;
    ReadData2E = 64'h403FFFFFFFF7BFFF;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'hC83E007FFFF843DE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"784583\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hBFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"784889\n");
    end
    ReadData1E = 64'hB7FC6D6EAD2BD6AB;
    ReadData2E = 64'hC7EFFFFFFFFF87FE;
    ReadData3E = 64'h438FFFEE00000000;
    ans = 64'h438FFFEE00000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"785195\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3800FFFFFFFFFF7E;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h34C0FFFFFFFFFF7D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"785501\n");
    end
    ReadData1E = 64'hFFE00000200000FF;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hC362D940E64F79A5;
    ans = 64'hFFE00000200000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"785807\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h001F7FFFFFFEFFFF;
    ReadData3E = 64'hFFF8EFD2FE79C24A;
    ans = 64'hFFF8EFD2FE79C24A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"786113\n");
    end
    ReadData1E = 64'h40810000001FFFFE;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'h4080F800001FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"786419\n");
    end
    ReadData1E = 64'hC007FFFFFFFFC000;
    ReadData2E = 64'h401FC00000080000;
    ReadData3E = 64'hC3C01FFFFFFFFFFF;
    ans = 64'hC3C01FFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"786725\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'h43EFFFFFFFEFDFFE;
    ans = 64'h43EFFFFFFFEFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"787031\n");
    end
    ReadData1E = 64'h3FF0800000000003;
    ReadData2E = 64'h47FA61529E587B8B;
    ReadData3E = 64'hC340000000000000;
    ans = 64'h47FB345D334B3F6C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"787337\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FF0000000000003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"787643\n");
    end
    ReadData1E = 64'h3E4124E67B52FD98;
    ReadData2E = 64'hC00FFF0000000FFE;
    ReadData3E = 64'h8727490D6F398EA1;
    ans = 64'hBE61245D541F2B91;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"787949\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h804FFFF7FFFFFFEE;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h800000000000000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"788255\n");
    end
    ReadData1E = 64'hBF9526E7BD8A1F46;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h480FFFDFFFFFFFDF;
    ans = 64'h480FFFDFFFFFFFDF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"788561\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h381020000007FFFE;
    ReadData3E = 64'h37E00003FFF7FFFF;
    ans = 64'h37E00003FFF80007;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"788867\n");
    end
    ReadData1E = 64'hFFDFF7FFFFFFFBFF;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'hFFEFF7FFFFFFFC01;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"789173\n");
    end
    ReadData1E = 64'h003C97950DCA0015;
    ReadData2E = 64'hEFACB38DCC424DA6;
    ReadData3E = 64'h278003FFFFEFFFFE;
    ans = 64'hAFF9A51108403B01;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"789479\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hC95F800100000000;
    ans = 64'hC95F800100000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"789785\n");
    end
    ReadData1E = 64'h7FF0EFBB42DBF4E3;
    ReadData2E = 64'h2255FFFFFFFFFFFF;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h7FF8EFBB42DBF4E3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"790091\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFEFFFFFFFFFFFF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"790397\n");
    end
    ReadData1E = 64'h46F342BC1CEBE027;
    ReadData2E = 64'hC3D6C59A9D585605;
    ReadData3E = 64'h802DFFFFFFFFBFFE;
    ans = 64'hCADB69A273A42F3F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"790703\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC00FFDFFFFFFFFFF;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"791009\n");
    end
    ReadData1E = 64'h30B00000010FFFFF;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'hC7F903C3D334F340;
    ans = 64'hC7F903C3D334F340;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"791315\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3CAFC89A50DE63CA;
    ReadData3E = 64'hC04C000007FFFFFF;
    ans = 64'hC04C000007FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"791621\n");
    end
    ReadData1E = 64'hC3C0000000BFFFFE;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hC3E0000000BFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"791927\n");
    end
    ReadData1E = 64'hBF8F80000000007F;
    ReadData2E = 64'h000000013FFFFFFE;
    ReadData3E = 64'hC3DFFFF803FFFFFF;
    ans = 64'hC3DFFFF803FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"792233\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'hB80FFFF00007FFFF;
    ans = 64'h3CD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"792539\n");
    end
    ReadData1E = 64'hC00000000001003E;
    ReadData2E = 64'h402CDC4C5AFCC35F;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC0406E262D7F48CA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"792845\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FE000000000000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"793151\n");
    end
    ReadData1E = 64'h43CFFFFC0003FFFF;
    ReadData2E = 64'h8020FFFFFFFFFFFB;
    ReadData3E = 64'h467691F27D3EA690;
    ans = 64'h467691F27D3EA690;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"793457\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC96004000003FFFF;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"793763\n");
    end
    ReadData1E = 64'hFFFFC7FFFFFFFFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h7FEFFFFFFF7FFFE0;
    ans = 64'hFFFFC7FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"794069\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hB7E1000FFFFFFFFE;
    ReadData3E = 64'hC34FFFFFFDFFFFFE;
    ans = 64'hC34FFFFFFDFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"794375\n");
    end
    ReadData1E = 64'h406FFFFFF7FFF7FE;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h43BFFFFFF7FFF7FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"794681\n");
    end
    ReadData1E = 64'hFFD0002000000040;
    ReadData2E = 64'hC03C96F971286B1E;
    ReadData3E = 64'h41EFFC0200000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"794987\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'h420000040000FFFF;
    ans = 64'h420000040010FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"795293\n");
    end
    ReadData1E = 64'h8020010001FFFFFE;
    ReadData2E = 64'h525FFFFEFFFFF000;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"795599\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'h3FFC000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"795905\n");
    end
    ReadData1E = 64'h7FFFFC03FFFFFFFF;
    ReadData2E = 64'hB7EFFFFFEEFFFFFF;
    ReadData3E = 64'h00100003FFFF7FFF;
    ans = 64'h7FFFFC03FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"796211\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hCD7FFFFFFFFDFFF6;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"796517\n");
    end
    ReadData1E = 64'hDB40010002000000;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h3FC35D556174AFCD;
    ans = 64'hDEA0010001FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"796823\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBFD7BFFFFFFFFFFF;
    ReadData3E = 64'h806852766CBEC085;
    ans = 64'hBC97BFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"797129\n");
    end
    ReadData1E = 64'h3FDFFFFDFFEFFFFF;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h433FFFFDFFEFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"797435\n");
    end
    ReadData1E = 64'hB4B6E07873102932;
    ReadData2E = 64'h434FF87FFFFFFFFF;
    ReadData3E = 64'h47E34D80C103EAA3;
    ans = 64'h47E34D80C103EAA3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"797741\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h43400FFFFFF80000;
    ans = 64'h7C9FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"798047\n");
    end
    ReadData1E = 64'h41300010000FFFFF;
    ReadData2E = 64'hBFAFFFFFDFFFFF80;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hC0F0001FF00FEFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"798353\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h7CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"798659\n");
    end
    ReadData1E = 64'hC1E85FEE97A05DE1;
    ReadData2E = 64'h43C4000000080000;
    ReadData3E = 64'hC1CFF0B2863E46AD;
    ans = 64'hC5BE77EA3D94A551;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"798965\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h4000001FFFFFFFFF;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"799271\n");
    end
    ReadData1E = 64'h44BC001000000000;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h49E9A8F935ADEB66;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"799577\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h416AA0B165976F71;
    ReadData3E = 64'hBF0010000003FFFE;
    ans = 64'hBF000F957D3E69A0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"799883\n");
    end
    ReadData1E = 64'h80003FFFFFE00000;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'hBFAFFFFFEFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"800189\n");
    end
    ReadData1E = 64'hC1DFFEFFFFFFFEFF;
    ReadData2E = 64'h43C003FEFFFFFFFF;
    ReadData3E = 64'hFB4FFFFF800000FF;
    ans = 64'hFB4FFFFF800000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"800495\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'hAD5080B9606E81BA;
    ans = 64'h7CAFFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"800801\n");
    end
    ReadData1E = 64'h37F000000000DFFF;
    ReadData2E = 64'h3F8A7CA782E53728;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"801107\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"801413\n");
    end
    ReadData1E = 64'hB5FFFFFFF0000003;
    ReadData2E = 64'hBFB0000000000900;
    ReadData3E = 64'h40B0000000000F7F;
    ans = 64'h40B0000000000F7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"801719\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hA81FFFF00000007F;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h401FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"802025\n");
    end
    ReadData1E = 64'hD5A00009FFFFFFFF;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h364FFFFFFC3FFFFF;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"802331\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FAFFFFF003FFFFF;
    ReadData3E = 64'h41DE7402306C1129;
    ans = 64'h41DE7402306C1129;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"802637\n");
    end
    ReadData1E = 64'h2A3FFFFFF8000000;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"802943\n");
    end
    ReadData1E = 64'hC00007FFFFFFFFFF;
    ReadData2E = 64'hBFE00077FFFFFFFE;
    ReadData3E = 64'h400E0000007FFFFE;
    ans = 64'h4013021E0F3FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"803249\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'hBE8000003FFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"803555\n");
    end
    ReadData1E = 64'hBFBFFFE00007FFFF;
    ReadData2E = 64'hBFEFFFFFFEBFFFFE;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFD80008004DFFB0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"803861\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h000FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"804167\n");
    end
    ReadData1E = 64'hC96000000FFFFFDE;
    ReadData2E = 64'h31D03FFFFFFF8000;
    ReadData3E = 64'hC1C87E0A0FF31A23;
    ans = 64'hC1C87E0A0FF31A23;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"804473\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3345D50CBBB891A1;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hC00FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"804779\n");
    end
    ReadData1E = 64'hC7F0B2F9A4CB116D;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h9C019C696C0DD16A;
    ans = 64'h9C019C696C0DD16A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"805085\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hB7F00000007FFE00;
    ReadData3E = 64'hBEAFF7FFF0000000;
    ans = 64'hBEAFF7FFF0000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"805391\n");
    end
    ReadData1E = 64'hCDA521B1B27F8608;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"805697\n");
    end
    ReadData1E = 64'h41D3F2F4802615C5;
    ReadData2E = 64'hC00FFFF7FFDFFFFE;
    ReadData3E = 64'h40AFFFFF80000020;
    ans = 64'hC1F3F2EE835506C6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"806003\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'hC07FFFFFFF7FFFFE;
    ans = 64'hC07FFFFFFF7FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"806309\n");
    end
    ReadData1E = 64'h4051FFFFFFFFF7FF;
    ReadData2E = 64'h3FF80A0A70A9D9F4;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h405B0B4BBEBF092C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"806615\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h8000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"806921\n");
    end
    ReadData1E = 64'h40EFFFFFCFFFFFFF;
    ReadData2E = 64'hBF46B20A5AF687D4;
    ReadData3E = 64'h4010403FFFFFFFFE;
    ans = 64'hC044AA0238EB784B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"807227\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h4B0F8001FFFFFFFF;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h47CF8001FFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"807533\n");
    end
    ReadData1E = 64'h405000FFFFFF7FFE;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h3FBFE00100000000;
    ans = 64'h3FBFE00100000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"807839\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h402E000001FFFFFF;
    ReadData3E = 64'hC1DFFFFBFFFFBFFF;
    ans = 64'hC1DFFFFBFFFFBFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"808145\n");
    end
    ReadData1E = 64'hC0F0001FFFFFFEFF;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"808451\n");
    end
    ReadData1E = 64'hC02409224A572D98;
    ReadData2E = 64'hC06FFFFFE0020000;
    ReadData3E = 64'h3FFFFFFFF007FFFF;
    ans = 64'h40A40D22364D4CE0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"808757\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h4B87A4A7FDD82AE3;
    ans = 64'h4B87A4A7FDD82AE3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"809063\n");
    end
    ReadData1E = 64'hC0F0001FFFFFFFFB;
    ReadData2E = 64'hC089DB2B2EDEB592;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'h4189DB5EE5351347;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"809369\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"809675\n");
    end
    ReadData1E = 64'h4A6040000000000F;
    ReadData2E = 64'hC0300100000003FF;
    ReadData3E = 64'hC3E0000000000000;
    ans = 64'hCAA041040000041E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"809981\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h7FE000000FFFFFFF;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'h7CA000000FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"810287\n");
    end
    ReadData1E = 64'h41FFF1070335C8D9;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hBFA0DC2898783816;
    ans = 64'hBFA0DC48897F3B4C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"810593\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3D93E99A04CEC718;
    ReadData3E = 64'hC3DFFDFFDFFFFFFF;
    ans = 64'hC3DFFDFFDFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"810899\n");
    end
    ReadData1E = 64'hC1F14340A63773D7;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"811205\n");
    end
    ReadData1E = 64'hA32C85A44ACBA581;
    ReadData2E = 64'h41F7FFFFFFFFFFDE;
    ReadData3E = 64'h43CFFFFFBFF00000;
    ans = 64'h43CFFFFFBFF00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"811511\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'hBC92546CCE634083;
    ans = 64'hBC92546CCE634085;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"811817\n");
    end
    ReadData1E = 64'h9DD03FFFFFFFFF7F;
    ReadData2E = 64'hBFE5AE33346526BE;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h1DC604EC0136BAAA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"812123\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"812429\n");
    end
    ReadData1E = 64'hA4BA947C9B590FEF;
    ReadData2E = 64'hC7FBC71D8AAF5885;
    ReadData3E = 64'hB7E9F7E971FFC3F6;
    ans = 64'hB7E9F7E971FFC3F6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"812735\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC3FFFDFFF7FFFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'hC0BFFCFFF7FFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"813041\n");
    end
    ReadData1E = 64'h402100000FFFFFFE;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'hBF6007FFFFFFFFBE;
    ans = 64'hBF600800000010BE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"813347\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h48000FFFEFFFFFFF;
    ReadData3E = 64'hBF80EAC8590DC991;
    ans = 64'h44C00FFFEFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"813653\n");
    end
    ReadData1E = 64'h4150000040001FFF;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h433FFFFFFFF00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"813959\n");
    end
    ReadData1E = 64'h37EFFFFFF8000006;
    ReadData2E = 64'h401000000F7FFFFE;
    ReadData3E = 64'h41D0000020001FFF;
    ans = 64'h41D0000020001FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"814265\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h4641C8183EC0208B;
    ans = 64'h4641C8183EC0208B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"814571\n");
    end
    ReadData1E = 64'h40CDFFFFFF800000;
    ReadData2E = 64'h3FB000000000000A;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h408DFFFFFF800013;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"814877\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"815183\n");
    end
    ReadData1E = 64'h3FEC00000000001F;
    ReadData2E = 64'h47EFFFFFEFFFFFEE;
    ReadData3E = 64'h4046D38AE2B5B7C9;
    ans = 64'h47EBFFFFF200000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"815489\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC1EFEFFFFFFFFE00;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFE00001FEFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"815795\n");
    end
    ReadData1E = 64'h3F9FFFFFFFFFF080;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h4380000100000004;
    ans = 64'h4380000100000004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"816101\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h50AE6549B7B9413E;
    ReadData3E = 64'hC0700000FFFFFF00;
    ans = 64'h4D6E6549B7B9413C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"816407\n");
    end
    ReadData1E = 64'hC3F1FFFFFFFFFFF7;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'hC010000000000001;
    ans = 64'h43E1FFFFFFFFFFF6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"816713\n");
    end
    ReadData1E = 64'hC3CF000008000000;
    ReadData2E = 64'h00000001F7FFFFFF;
    ReadData3E = 64'hADC7AEB97D7AAA63;
    ans = 64'hADC7AEB97D7AAA63;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"817019\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h00107FFFEFFFFFFE;
    ans = 64'hBC9FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"817325\n");
    end
    ReadData1E = 64'h400C00000FFFFFFE;
    ReadData2E = 64'h3F803FFFFFC00000;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"817631\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"817937\n");
    end
    ReadData1E = 64'h43FEFFFFFFDFFFFE;
    ReadData2E = 64'h3BE000003FFFFFF7;
    ReadData3E = 64'h3F1F7FDFFFFFFFFE;
    ans = 64'h3FEF00FC7ADFFFEC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"818243\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h40380000000000FE;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000060;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"818549\n");
    end
    ReadData1E = 64'h4310010040000000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h373650909086FED3;
    ans = 64'hC31001003FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"818855\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hCA64000000000FFF;
    ReadData3E = 64'hC3E800000000FFFF;
    ans = 64'hC724000000000FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"819161\n");
    end
    ReadData1E = 64'h447FFFEFFFFFFFFC;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'hC47FFFEFFFFFFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"819467\n");
    end
    ReadData1E = 64'hB800000400000000;
    ReadData2E = 64'h400FFBFFFFDFFFFE;
    ReadData3E = 64'hC9B000000000401E;
    ans = 64'hC9B000000000401E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"819773\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h3FF000000000DFFF;
    ans = 64'h3FF000000000DFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"820079\n");
    end
    ReadData1E = 64'hBE40000020008000;
    ReadData2E = 64'hC7ECCDDA5176EED4;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"820385\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"820691\n");
    end
    ReadData1E = 64'hD57E91FB06496C86;
    ReadData2E = 64'h381B4E4682CE5ADB;
    ReadData3E = 64'hDC6BFFFFFEFFFFFE;
    ans = 64'hDC6BFFFFFEFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"820997\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBF110000000001FF;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0008800000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"821303\n");
    end
    ReadData1E = 64'hC0B0200000000003;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'hBFC1D73DADDDF9B6;
    ans = 64'h40B01FDC5184A448;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"821609\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC009137C731CDB82;
    ReadData3E = 64'h44500000001FFFDF;
    ans = 64'h44500000001FFFDF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"821915\n");
    end
    ReadData1E = 64'h3F0FFBFFFFFFFFFA;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hC000003FF7FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"822221\n");
    end
    ReadData1E = 64'hBFDF3226C82519BB;
    ReadData2E = 64'h093FFFFFFFFC0002;
    ReadData3E = 64'hBBEF000000010000;
    ans = 64'hBBEF000000010000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"822527\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h800FFFFFFFF00000;
    ans = 64'hBCBFFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"822833\n");
    end
    ReadData1E = 64'h40300FFDFFFFFFFF;
    ReadData2E = 64'h411FFFFFFFFDFFFF;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"823139\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"823445\n");
    end
    ReadData1E = 64'h52B0200000FFFFFE;
    ReadData2E = 64'hC7FDC39DD92F7A18;
    ReadData3E = 64'hC940002001FFFFFF;
    ans = 64'hDABDFF2516BE12E6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"823751\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3FBF800001FFFFFF;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h3C7F800001FFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"824057\n");
    end
    ReadData1E = 64'h41C0010007FFFFFE;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'hA39000000403FFFF;
    ans = 64'hC1D0010007FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"824363\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC7EBCC5F21936D2F;
    ReadData3E = 64'hD65B21AA7893B120;
    ans = 64'hD65B21AA7893B120;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"824669\n");
    end
    ReadData1E = 64'hC010400000008000;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h4031400000007FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"824975\n");
    end
    ReadData1E = 64'hBFD000003FFFFFFF;
    ReadData2E = 64'h4060000020FFFFFE;
    ReadData3E = 64'hC7FFE0000000003E;
    ans = 64'hC7FFE0000000003E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"825281\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'hB9EFFFFFFFF60000;
    ans = 64'hBCD000000000003E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"825587\n");
    end
    ReadData1E = 64'h41D00000FFFFFFF7;
    ReadData2E = 64'h0E8000000000027F;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h434FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"825893\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"826199\n");
    end
    ReadData1E = 64'hC03FFFFFEFFFFE00;
    ReadData2E = 64'hC1B69575AD8E4938;
    ReadData3E = 64'hC03EF1A51B0A9C79;
    ans = 64'h42069575A14BFFCF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"826505\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h401FEFFFFFFFBFFE;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h3CDFEFFFFFFFBFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"826811\n");
    end
    ReadData1E = 64'hB8008E2BDAE7849A;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h4D700008000003FF;
    ans = 64'h4D700008000003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"827117\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h802F800003FFFFFE;
    ReadData3E = 64'h3CA21EADA7E7EABE;
    ans = 64'h3CA21EADA7E7EABE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"827423\n");
    end
    ReadData1E = 64'h40454B061F63C997;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hC0655B061F63C998;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"827729\n");
    end
    ReadData1E = 64'hC3DFFDFF00000000;
    ReadData2E = 64'hC34000FFFFFFFFEF;
    ReadData3E = 64'h381FF7FFF8000000;
    ans = 64'h472FFFFEDFEFFFDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"828035\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'hFFE2D34305DB32DD;
    ans = 64'hFFE2D34305DB32DD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"828341\n");
    end
    ReadData1E = 64'hC3D4C997322DB688;
    ReadData2E = 64'hC7EFFFFFE7FFFFFF;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'h4BD4C99722968522;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"828647\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h3FEFFFFFFFFFFFF0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"828953\n");
    end
    ReadData1E = 64'h43401FFFFFFFFFEF;
    ReadData2E = 64'hBFC6B4955343E8C0;
    ReadData3E = 64'h3FACA027F5F02257;
    ans = 64'hC316E1FE7DEA7079;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"829259\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC0717FFFFFFFFFFE;
    ReadData3E = 64'h0000000000000000;
    ans = 64'hBD317FFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"829565\n");
    end
    ReadData1E = 64'h3F30200000007FFE;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'hBFEDFFF7FFFFFFFE;
    ans = 64'hC28020000000877E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"829871\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h480FFF07FFFFFFFE;
    ReadData3E = 64'hFFE0FFFFFFEFFFFF;
    ans = 64'hFFE0FFFFFFEFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"830177\n");
    end
    ReadData1E = 64'h4340000010400000;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'hC690000010400001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"830483\n");
    end
    ReadData1E = 64'h801138665FBF5823;
    ReadData2E = 64'h41C88AD4452DDA53;
    ReadData3E = 64'h837330BE9EFFCD8A;
    ans = 64'h837330BEAC34CFC6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"830789\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'hBE314AE1C1E96403;
    ans = 64'hC0000000008A570E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"831095\n");
    end
    ReadData1E = 64'h000FFFB800000000;
    ReadData2E = 64'hBFDFFFFFFE020000;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"831401\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hC011FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"831707\n");
    end
    ReadData1E = 64'hBE5F0A7A7523EF50;
    ReadData2E = 64'hBFFFFFFFFFFC0010;
    ReadData3E = 64'h3F800403FFFFFFFF;
    ans = 64'h3F80040BC29E9D47;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"832013\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h3F7FFFC00000001E;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"832319\n");
    end
    ReadData1E = 64'h93B00004000FFFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h3FAF4698DF3288FF;
    ans = 64'h3FAF4698DF3288FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"832625\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC0CFFF7FFFFEFFFF;
    ReadData3E = 64'h5175F45600193651;
    ans = 64'h5175F45600193651;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"832931\n");
    end
    ReadData1E = 64'h41CDFFFFFFFFF000;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"833237\n");
    end
    ReadData1E = 64'h413FFFFFE0001000;
    ReadData2E = 64'h43400000FFF7FFFF;
    ReadData3E = 64'hD649F219711FB93B;
    ans = 64'hD649F219711FB93B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"833543\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hB812005F57A6E0BC;
    ans = 64'hFCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"833849\n");
    end
    ReadData1E = 64'hC91FFEFFFFFFFFF0;
    ReadData2E = 64'hDB00080000FFFFFF;
    ReadData3E = 64'hC000000000000001;
    ans = 64'h6430077FC0FFF7F7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"834155\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'hFCAFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"834461\n");
    end
    ReadData1E = 64'h403F4AF7E517A201;
    ReadData2E = 64'h41CFFC000000001F;
    ReadData3E = 64'hCC6000000001E000;
    ans = 64'hCC6000000001E000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"834767\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h801FFFFFDFEFFFFE;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"835073\n");
    end
    ReadData1E = 64'h37F000000000000F;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'hDE9FF7FFFFFFFF7E;
    ans = 64'hF7F000000000000E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"835379\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hC00007FFFFFFFFF8;
    ReadData3E = 64'h3FDD260F6EE2590B;
    ans = 64'h3FDD260F6EE25903;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"835685\n");
    end
    ReadData1E = 64'h3FFC0000003FFFFF;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"835991\n");
    end
    ReadData1E = 64'hC1EF000400000000;
    ReadData2E = 64'h800000000FFFFFFE;
    ReadData3E = 64'hBCA00800000000FF;
    ans = 64'hBCA00800000000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"836297\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h4BEFFFFFFFFFFFF5;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"836603\n");
    end
    ReadData1E = 64'h41EFFFBF7FFFFFFE;
    ReadData2E = 64'h4000400000000010;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h42003FDF3F10000F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"836909\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"837215\n");
    end
    ReadData1E = 64'h800000000FFFFBFE;
    ReadData2E = 64'h3FFF168ED14CE3D1;
    ReadData3E = 64'hC0301FFF7FFFFFFE;
    ans = 64'hC0301FFF7FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"837521\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'h2200000000000420;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h434FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"837827\n");
    end
    ReadData1E = 64'hC1FF0000000001FF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h000FFF0008000000;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"838133\n");
    end
    ReadData1E = 64'h3CAFFFFFFFFFFFFE;
    ReadData2E = 64'hBF5FFFFFFFDFFFFF;
    ReadData3E = 64'h34B00FFFFFFFFFBE;
    ans = 64'hBC1FFFFFFFDFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"838439\n");
    end
    ReadData1E = 64'hC1FFFFFFBFFFFFFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h8010000000000001;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"838745\n");
    end
    ReadData1E = 64'hB7F0000040000020;
    ReadData2E = 64'h37FFFFFF00000002;
    ReadData3E = 64'h4057FFFFFFFF7FFF;
    ans = 64'h4057FFFFFFFF7FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"839051\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h37E000081FFFFFFF;
    ans = 64'h37E000081FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"839357\n");
    end
    ReadData1E = 64'h41408003FFFFFFFF;
    ReadData2E = 64'h3FC07FFFFFFFFEFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'h411104001FFFFEF5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"839663\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"839969\n");
    end
    ReadData1E = 64'h43EBF2B862C4592E;
    ReadData2E = 64'hC09F6B9863FDA598;
    ReadData3E = 64'h40100DFFFFFFFFFE;
    ans = 64'hC49B711B5078BE27;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"840275\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h47F00000007EFFFF;
    ReadData3E = 64'hC340000000000000;
    ans = 64'h47D00000007EFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"840581\n");
    end
    ReadData1E = 64'hBF9FFC0000001FFE;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'hA48FFBFFFFFFFF7E;
    ans = 64'hA48FFBFFFFFFFF7E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"840887\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC021FFFFFFFFF7FE;
    ReadData3E = 64'hCC7FFFFFDFFFFDFF;
    ans = 64'hCC7FFFFFDFFFFDFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"841193\n");
    end
    ReadData1E = 64'h4017A9EC462C22C8;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h003BA9EC462C22C5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"841499\n");
    end
    ReadData1E = 64'hB7F0000000000086;
    ReadData2E = 64'h4BFFFFFFE000FFFE;
    ReadData3E = 64'hBF70000002000000;
    ans = 64'hC3FFFFFFE001010A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"841805\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h80027DBC297E85BC;
    ans = 64'h00018243D6817A44;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"842111\n");
    end
    ReadData1E = 64'h8540020000000FFF;
    ReadData2E = 64'hBFFF3B0F41DCC6B0;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"842417\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h800C000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"842723\n");
    end
    ReadData1E = 64'h47F003FFFFFFFDFF;
    ReadData2E = 64'hC3DFFFEFFFFFC000;
    ReadData3E = 64'h43F400000000000F;
    ans = 64'hCBE003F7FDFFDDF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"843029\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h00000003FFFFFFFC;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h4010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"843335\n");
    end
    ReadData1E = 64'h402FF0000007FFFE;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h3EEFFFFFFFFC4000;
    ans = 64'h3EEFFFFFFFFC4000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"843641\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hB7EFFFFFBFFFFF7E;
    ReadData3E = 64'hBFCE000000100000;
    ans = 64'hBFCE000000100000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"843947\n");
    end
    ReadData1E = 64'hEBB4000FFFFFFFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"844253\n");
    end
    ReadData1E = 64'h001CB8573191EF15;
    ReadData2E = 64'hC19000800003FFFF;
    ReadData3E = 64'h41EFFFFFE03FFFFF;
    ans = 64'h41EFFFFFE03FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"844559\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'hBFC00003FFFF7FFE;
    ans = 64'hBFC00003FFFF7FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"844865\n");
    end
    ReadData1E = 64'h3334BE2D7996C148;
    ReadData2E = 64'h3EC03F7FFFFFFFFE;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hBFD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"845171\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h3C80000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"845477\n");
    end
    ReadData1E = 64'h800F800000080000;
    ReadData2E = 64'h7FD00000000037FF;
    ReadData3E = 64'h483FFFFFDFF7FFFF;
    ans = 64'h483FFFFFDFF7FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"845783\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FF000000FBFFFFE;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00DFFFFFE07FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"846089\n");
    end
    ReadData1E = 64'hA96FFFFFEFFBFFFF;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'hBFFBFFFFDFFFFFFF;
    ans = 64'hBFFBFFFFDFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"846395\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h62BBFFFFEFFFFFFF;
    ReadData3E = 64'h40300000403FFFFF;
    ans = 64'h629BFFFFEFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"846701\n");
    end
    ReadData1E = 64'h41CAB753F5C8C6D1;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"847007\n");
    end
    ReadData1E = 64'hFFF16C91BA884B3B;
    ReadData2E = 64'h52E000000001E000;
    ReadData3E = 64'h5A8000000FFEFFFF;
    ans = 64'hFFF96C91BA884B3B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"847313\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'h400EC0D680F9B3E5;
    ans = 64'h400EC0D680F9B3E5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"847619\n");
    end
    ReadData1E = 64'h3FB7999C7249E9A2;
    ReadData2E = 64'hC1C0040FFFFFFFFE;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'hC1879F9A7302EE63;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"847925\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"848231\n");
    end
    ReadData1E = 64'hBFE0000000000000;
    ReadData2E = 64'h43C87A46755279AC;
    ReadData3E = 64'hC00EAE236666A2DC;
    ans = 64'hC3B87A46755279AC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"848537\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h400FFF7BFFFFFFFF;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h4007FFDEFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"848843\n");
    end
    ReadData1E = 64'h3F6FFFFFC0000010;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h3F70000001FEFFFF;
    ans = 64'h3F73FFFFF9FF0001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"849149\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3F8F73F460C4C31D;
    ReadData3E = 64'h7FD542F883C0A5EF;
    ans = 64'h7FD542F883C0A5EF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"849455\n");
    end
    ReadData1E = 64'hBE70002003FFFFFE;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"849761\n");
    end
    ReadData1E = 64'hC1C197CF2A7E3F8F;
    ReadData2E = 64'hBFEFFFFFFFFF7FBE;
    ReadData3E = 64'hC3CFFC00000FFFFF;
    ans = 64'hC3CFFBFFFFFE6830;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"850067\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'hBFD0080000003FFE;
    ans = 64'hBFC0100000007FFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"850373\n");
    end
    ReadData1E = 64'h3FC00001FFFFDFFF;
    ReadData2E = 64'hAF80000004100000;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hAF500002040FE081;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"850679\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"850985\n");
    end
    ReadData1E = 64'h417007FFDFFFFFFF;
    ReadData2E = 64'h41F000FFFFFC0000;
    ReadData3E = 64'hC0BFFFFFFFFFF880;
    ans = 64'h437009005FF9FBFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"851291\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFF011FFFFFFFFFF;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF4048000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"851597\n");
    end
    ReadData1E = 64'hBEE04000007FFFFF;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'h4396138DF7EB9A59;
    ans = 64'h4396138DF7EB9A59;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"851903\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBE1FFFBFFBFFFFFF;
    ReadData3E = 64'hC7F00F8000000000;
    ans = 64'hC7F00F8000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"852209\n");
    end
    ReadData1E = 64'hB44FFFFFF000000F;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"852515\n");
    end
    ReadData1E = 64'h0005396BB2A6A133;
    ReadData2E = 64'h80214BC0C3457605;
    ReadData3E = 64'hC02C6B4A0B5C3DD6;
    ans = 64'hC02C6B4A0B5C3DD6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"852821\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'h7FD25208760C3106;
    ans = 64'h7FD25208760C3106;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"853127\n");
    end
    ReadData1E = 64'h34F00010003FFFFF;
    ReadData2E = 64'h000335BE3408EDB3;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h0010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"853433\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"853739\n");
    end
    ReadData1E = 64'h3FF000FFFFFFFFFC;
    ReadData2E = 64'hB807FFFFFFEFFFFE;
    ReadData3E = 64'hC1F0000100000006;
    ans = 64'hC1F0000100000006;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"854045\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h42BA5BD75A09BAF5;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h429A5BD75A09BCF5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"854351\n");
    end
    ReadData1E = 64'hC347ACE083F9E32E;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h3F8020000FFFFFFF;
    ans = 64'hC347ACE083F9E32D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"854657\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hB87850D0E3E0A2DD;
    ReadData3E = 64'hBFD0000000FFFFDE;
    ans = 64'hBFD0000000FFFFDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"854963\n");
    end
    ReadData1E = 64'hB80FF8000003FFFE;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"855269\n");
    end
    ReadData1E = 64'h3FE000000000000C;
    ReadData2E = 64'hC3F54C10FC0D1720;
    ReadData3E = 64'hBFFFFFEFFF7FFFFF;
    ans = 64'hC3E54C10FC0D1730;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"855575\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'h43FFFFFFFFF80040;
    ans = 64'h43FFFFFFFFF80040;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"855881\n");
    end
    ReadData1E = 64'h41C2D222AA9A0638;
    ReadData2E = 64'h43EFFFEFFFEFFFFF;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h45C2D219417F47D9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"856187\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"856493\n");
    end
    ReadData1E = 64'hBFE0800000001FFF;
    ReadData2E = 64'h3FDFFFFFFE0001FF;
    ReadData3E = 64'hBD9FFFE00000000F;
    ans = 64'hBFD07FFFFEFA2104;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"856799\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hB810080010000000;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"857105\n");
    end
    ReadData1E = 64'h402FFFFFFFFBFFDF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h405BFFFFBFFFFFFF;
    ans = 64'h4061FFFFDFFF7FFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"857411\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC1CFFF7FFFFEFFFF;
    ReadData3E = 64'h79E000FFFFFFFFFD;
    ans = 64'h79E000FFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"857717\n");
    end
    ReadData1E = 64'h3FDF8E1FB8E7AC41;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC0081C7811C614F0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"858023\n");
    end
    ReadData1E = 64'hF4D0700000000000;
    ReadData2E = 64'h4061AE8809F49782;
    ReadData3E = 64'hC3EFFFFFF7FFFF7F;
    ans = 64'hF5422A4DC23A47A7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"858329\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h7FD00000000003EF;
    ans = 64'h7FD00000000003EF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"858635\n");
    end
    ReadData1E = 64'hB7F0000FBFFFFFFF;
    ReadData2E = 64'h3FC4294C9A3326B1;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"858941\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h4020FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"859247\n");
    end
    ReadData1E = 64'h37F000000FFFDFFE;
    ReadData2E = 64'h0B900040000007FF;
    ReadData3E = 64'hC61000407FFFFFFF;
    ans = 64'hC61000407FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"859553\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC7EFFFFFFFBFFFFA;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'hC7CFFFFFFFBFFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"859859\n");
    end
    ReadData1E = 64'h408FC007FFFFFFFE;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'hC0000000080FFFFF;
    ans = 64'h40AFBC07FFFDFBFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"860165\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC7E007FFFFFFBFFE;
    ReadData3E = 64'h400000001FFFEFFF;
    ans = 64'hC7C007FFFFFFBFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"860471\n");
    end
    ReadData1E = 64'h3FBFF80000000040;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4003FF0000000009;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"860777\n");
    end
    ReadData1E = 64'hC342000FFFFFFFFF;
    ReadData2E = 64'hBFF00000000001FD;
    ReadData3E = 64'h43FF0000000007FE;
    ans = 64'h43FF0240020007FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"861083\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'hC3E4000000200000;
    ans = 64'hC3E4000000200000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"861389\n");
    end
    ReadData1E = 64'h464010003FFFFFFE;
    ReadData2E = 64'h42300000800FFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"861695\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC008000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"862001\n");
    end
    ReadData1E = 64'h404000020000001F;
    ReadData2E = 64'hFA9FFFC000001FFF;
    ReadData3E = 64'hC68FFFFFFFFFFB80;
    ans = 64'hFAEFFFC3FFF8203D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"862307\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h9950008000000003;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBCA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"862613\n");
    end
    ReadData1E = 64'hBFFFFF0000100000;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h40C5214A8EE75583;
    ans = 64'h40C51D4AAEE75383;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"862919\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h5153FFFFFFFFBFFF;
    ReadData3E = 64'hC35001FFFF000000;
    ans = 64'h5133FFFFFFFFBFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"863225\n");
    end
    ReadData1E = 64'hFFDFFFFFFFFFFEBE;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"863531\n");
    end
    ReadData1E = 64'hFFF8B44D2CF18037;
    ReadData2E = 64'h3D200000043FFFFF;
    ReadData3E = 64'h43E7FBFFFFFFFFFF;
    ans = 64'hFFF8B44D2CF18037;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"863837\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h41E000800000001E;
    ans = 64'h41E000800040001E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"864143\n");
    end
    ReadData1E = 64'h3FB0800000000003;
    ReadData2E = 64'hB9600400007FFFFE;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"864449\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4320000000000004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"864755\n");
    end
    ReadData1E = 64'h3DFFFFFFEFDFFFFF;
    ReadData2E = 64'h43D0000FFFBFFFFF;
    ReadData3E = 64'h802FEFFDFFFFFFFE;
    ans = 64'h41E0000FF7AFF7EF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"865061\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3E6ABE5E424DBD3F;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h3E4ABE5E424DBD3F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"865367\n");
    end
    ReadData1E = 64'hBFB909D69E02A222;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h380EFFFFFFFFFFEF;
    ans = 64'hC30909D69E02A224;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"865673\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h38100000007EFFFF;
    ReadData3E = 64'hC39FFFFFFE1FFFFF;
    ans = 64'hC39FFFFFFE1FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"865979\n");
    end
    ReadData1E = 64'h002FFFFFFF000400;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"866285\n");
    end
    ReadData1E = 64'h453FFFFFF0000003;
    ReadData2E = 64'h47F00000047FFFFF;
    ReadData3E = 64'hC0238DB0F8672290;
    ans = 64'h4D3FFFFFF8FFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"866591\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'hC1E7E7CD6D9B2D7E;
    ans = 64'h432FFFFE81832925;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"866897\n");
    end
    ReadData1E = 64'h3FCC803E813F7834;
    ReadData2E = 64'hC3E010000001FFFE;
    ReadData3E = 64'h4340000000000001;
    ans = 64'hC3BC7CBEBFC447B0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"867203\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'h432FFFFFFFFFFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"867509\n");
    end
    ReadData1E = 64'h052FF00004000000;
    ReadData2E = 64'h47E007FFFE000000;
    ReadData3E = 64'hDBE2000000007FFE;
    ans = 64'hDBE2000000007FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"867815\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFE00FC000000000;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'hBFC00FC000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"868121\n");
    end
    ReadData1E = 64'hC3DC1BC00754990F;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'hD87E57C40FADB738;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"868427\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3E10010007FFFFFE;
    ReadData3E = 64'h413000000007FFFF;
    ans = 64'h4130000000080000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"868733\n");
    end
    ReadData1E = 64'h3EA1DF1B96B50BDF;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'h7E91DF1B96B50BE0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"869039\n");
    end
    ReadData1E = 64'h9B50000000000FEE;
    ReadData2E = 64'hBFFFE00000080000;
    ReadData3E = 64'hFFF02B3B3E780E07;
    ans = 64'hFFF82B3B3E780E07;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"869345\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h401C08C927D263F5;
    ans = 64'h7FCFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"869651\n");
    end
    ReadData1E = 64'h37EEFBFFFFFFFFFF;
    ReadData2E = 64'hDFB0002001000000;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hD7AEFC3DF9EFBFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"869957\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h7FCFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"870263\n");
    end
    ReadData1E = 64'h3ECFFFFC000000FF;
    ReadData2E = 64'hC1FFFE3FFFFFFFFE;
    ReadData3E = 64'hFFFFFEFFFFFFF7FF;
    ans = 64'hFFFFFEFFFFFFF7FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"870569\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBF0FEF8000000000;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"870875\n");
    end
    ReadData1E = 64'h2550003FFFFFFFEF;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'hAAA103DECE05BF0F;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"871181\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hFFFF0000000001FE;
    ReadData3E = 64'h2D80000000001FF7;
    ans = 64'hFFFF0000000001FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"871487\n");
    end
    ReadData1E = 64'h41D7748AB1A3FE5C;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"871793\n");
    end
    ReadData1E = 64'h800D0D06EE913F9A;
    ReadData2E = 64'h3DFFE0007FFFFFFE;
    ReadData3E = 64'h3FCFDFFFFFFFDFFE;
    ans = 64'h3FCFDFFFFFFFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"872099\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'hC000523B717EC726;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"872405\n");
    end
    ReadData1E = 64'hC1F4B64F0908F803;
    ReadData2E = 64'h2EF00008000001FF;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"872711\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"873017\n");
    end
    ReadData1E = 64'h47EFFFFC00000FFF;
    ReadData2E = 64'hC7F0000000A00000;
    ReadData3E = 64'hBF8FFFFFFFF803FF;
    ans = 64'hCFEFFFFC01400FD7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"873323\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h41D81517C3598347;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"873629\n");
    end
    ReadData1E = 64'h47F55D8D580F7615;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC3BF17E3E4C41B97;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"873935\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBCA1CA7B0D7924D0;
    ReadData3E = 64'h002A7B153AF12BAD;
    ans = 64'hBC81CA7B0D7924D0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"874241\n");
    end
    ReadData1E = 64'hC7163FA8199608CB;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"874547\n");
    end
    ReadData1E = 64'hBFC00001FDFFFFFF;
    ReadData2E = 64'hC4D0000020000007;
    ReadData3E = 64'hE2EFFFFBFFFFFFFE;
    ans = 64'hE2EFFFFBFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"874853\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'hC417FFFFFFFFFFBF;
    ans = 64'hC417FFFFFFFFFFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"875159\n");
    end
    ReadData1E = 64'h801FFFFFFDFFFF7F;
    ReadData2E = 64'h7FD0000000008FFE;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC00FFFFFFF008FBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"875465\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"875771\n");
    end
    ReadData1E = 64'h3CA5C07DE83266F8;
    ReadData2E = 64'h800FFFFFFFFFEF00;
    ReadData3E = 64'h5770000010080000;
    ans = 64'h5770000010080000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"876077\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h4DD000001FFFFFFA;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"876383\n");
    end
    ReadData1E = 64'hCEF4C8D3D8B55D06;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'hA51F200000000000;
    ans = 64'hA51F200000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"876689\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC0300000080003FF;
    ReadData3E = 64'h43DD48A0CE4511A9;
    ans = 64'h43DD48A0CE4511A9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"876995\n");
    end
    ReadData1E = 64'h524B89CF496B0D8A;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h926B89CF496B0D8A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"877301\n");
    end
    ReadData1E = 64'h47F00002000000FF;
    ReadData2E = 64'hB85000000001FFF7;
    ReadData3E = 64'h41E0800000000000;
    ans = 64'h41E07FFFF7FFFF00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"877607\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h4030EE949384D389;
    ans = 64'h4030EE949384D389;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"877913\n");
    end
    ReadData1E = 64'hBF00000001001FFE;
    ReadData2E = 64'h415001FFFFFFFBFF;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'hC05FC40002007802;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"878219\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hBCAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"878525\n");
    end
    ReadData1E = 64'hC01E32185C15D267;
    ReadData2E = 64'h40184C07F358ED80;
    ReadData3E = 64'hBE69FFFFFFFFFFFF;
    ans = 64'hC046ED50BFE8B8A0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"878831\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hCD200000000003BF;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'hCD000000000003BF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"879137\n");
    end
    ReadData1E = 64'hC7F5F2DFDDF5DD2D;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h20C00000007BFFFE;
    ans = 64'h20C00000007BFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"879443\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hB810000000000FFC;
    ReadData3E = 64'h400FBFFFFFFF7FFF;
    ans = 64'h400FBFFFFFFF7FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"879749\n");
    end
    ReadData1E = 64'h7FFFFFFFBFFFFF7F;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h7FFFFFFFBFFFFF7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"880055\n");
    end
    ReadData1E = 64'hBFDFFFFFFEFFFFFF;
    ReadData2E = 64'hC044AF12B6A74C7E;
    ReadData3E = 64'h3C251824E2D48BA5;
    ans = 64'h4034AF12B601D3E8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"880361\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hB7E0000040000007;
    ans = 64'hBC80000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"880667\n");
    end
    ReadData1E = 64'h3FB00000100003FE;
    ReadData2E = 64'hC01FFFFF80001FFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFF7FFFFE80009DE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"880973\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'hBC80000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"881279\n");
    end
    ReadData1E = 64'h401D507E92C5E594;
    ReadData2E = 64'h000F620D896F063D;
    ReadData3E = 64'hBE2FFFFEFFFFBFFF;
    ans = 64'hBE2FFFFEFFFFBFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"881585\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFEFFFFFFFFFBDFF;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC0207FFFFFFFFEF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"881891\n");
    end
    ReadData1E = 64'hC3D0002000010000;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h002FFFFFFFBFFFFE;
    ans = 64'h409000200000FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"882197\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h41ED56A3C3AB3F70;
    ReadData3E = 64'hC3F0003FFFFFFFF7;
    ans = 64'hC3F0003FFFFC5523;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"882503\n");
    end
    ReadData1E = 64'h43EFFE000000007F;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h0000000000000001;
    ans = 64'hC0AFFE000000007D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"882809\n");
    end
    ReadData1E = 64'h434FFFFF00000002;
    ReadData2E = 64'hC1DBCB06C86758B7;
    ReadData3E = 64'h406FE00000001000;
    ans = 64'hC53BCB05EA0F2276;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"883115\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h381F8003FFFFFFFF;
    ans = 64'hBC8FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"883421\n");
    end
    ReadData1E = 64'h4303C48ADE74C3B5;
    ReadData2E = 64'h00200027FFFFFFFE;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h3FDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"883727\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h8010000000000000;
    ans = 64'hBFB0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"884033\n");
    end
    ReadData1E = 64'hDE90020000007FFF;
    ReadData2E = 64'h3FD0000400000010;
    ReadData3E = 64'hC3EE5C580A6A45A4;
    ans = 64'hDE7002040080800F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"884339\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h37E454181AF8A93E;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"884645\n");
    end
    ReadData1E = 64'hFFE000200003FFFF;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h0020CEB3FC1309E7;
    ans = 64'h7FC0002000040000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"884951\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FB007FFFFFFFFBF;
    ReadData3E = 64'hD00FFFFFFFFE0002;
    ans = 64'hD00FFFFFFFFE0002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"885257\n");
    end
    ReadData1E = 64'hB7E0001001FFFFFF;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"885563\n");
    end
    ReadData1E = 64'hBFF0000000000803;
    ReadData2E = 64'hEADE000000400000;
    ReadData3E = 64'h401FFBFFFFFFFFF0;
    ans = 64'h6ADE000000400F06;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"885869\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'hC17FFFFFFFF7FFFD;
    ans = 64'hC180000000FBFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"886175\n");
    end
    ReadData1E = 64'h4010005FFFFFFFFF;
    ReadData2E = 64'h43E0000100400000;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'h440000610046017F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"886481\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h0000000000000001;
    ans = 64'hBFC0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"886787\n");
    end
    ReadData1E = 64'hC21DFFFFDFFFFFFF;
    ReadData2E = 64'h41EF7FFFFFFFFFFD;
    ReadData3E = 64'h3FC2C86C00258B86;
    ans = 64'hC41D87FFE07FFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"887093\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FE0000081FFFFFF;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hBFFDFFFFEFC00002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"887399\n");
    end
    ReadData1E = 64'hBFB000000007C000;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hBF5E0000003FFFFF;
    ans = 64'h3F9E2000000B8002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"887705\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hB8000000000005FF;
    ReadData3E = 64'h3FE00002001FFFFE;
    ans = 64'h3FE00002001FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"888011\n");
    end
    ReadData1E = 64'h3E3000100007FFFE;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"888317\n");
    end
    ReadData1E = 64'hC3F0FFFFFF7FFFFF;
    ReadData2E = 64'h41FFFEFFFFFFFFFB;
    ReadData3E = 64'hBD9FFFFFFFFFFFDD;
    ans = 64'hC600FF77FF8003FC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"888623\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'hD6DFFFFDFFFFDFFE;
    ans = 64'hD6DFFFFDFFFFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"888929\n");
    end
    ReadData1E = 64'h3CAFFFF7FFEFFFFF;
    ReadData2E = 64'h3FFDB1E7398F283A;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h3CC2D8EFE6834071;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"889235\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"889541\n");
    end
    ReadData1E = 64'hC02FFFFF80003FFF;
    ReadData2E = 64'h05B041FFFFFFFFFF;
    ReadData3E = 64'hBFDFFEFFFFFFFFBE;
    ans = 64'hBFDFFEFFFFFFFFBE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"889847\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3FF6779EE2033B07;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h4002CEF3DC406760;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"890153\n");
    end
    ReadData1E = 64'h3FF044898012431F;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h3EC774E18486AEBC;
    ans = 64'hBFF044869176128E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"890459\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFFFFFF000000000;
    ReadData3E = 64'h4010000402000000;
    ans = 64'h400C000A04000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"890765\n");
    end
    ReadData1E = 64'hC2A0000008002000;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"891071\n");
    end
    ReadData1E = 64'hBFA2EDC3DB985385;
    ReadData2E = 64'h4043EFFFFFFFFFFE;
    ReadData3E = 64'hC1FFFFFFFFFFFFFF;
    ans = 64'hC2000000000BCB23;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"891377\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h408000000000007E;
    ans = 64'h407FF800000000FC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"891683\n");
    end
    ReadData1E = 64'h0012D90EB735AC29;
    ReadData2E = 64'h41E00000000006FF;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h0202D90EB6F5B467;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"891989\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"892295\n");
    end
    ReadData1E = 64'h6C26B0BFD2FBCAAE;
    ReadData2E = 64'hC300000005000000;
    ReadData3E = 64'h2380021000000000;
    ans = 64'hEF36B0BFDA1306A0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"892601\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hF4100400007FFFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hF3F00400007FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"892907\n");
    end
    ReadData1E = 64'hC02FFFFFFFFFFE1F;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h7FDFFFFFFFFF8004;
    ans = 64'h7FDFFFFFFFFF8004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"893213\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hBFFFFFFFFBFFEFFE;
    ReadData3E = 64'h43F00040000000FF;
    ans = 64'h43F00040000000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"893519\n");
    end
    ReadData1E = 64'h41EBFFFFFFFFFF80;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC3400000E0000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"893825\n");
    end
    ReadData1E = 64'h3C9B86D50A168045;
    ReadData2E = 64'hFFDF7FFFFE000000;
    ReadData3E = 64'hFFFFFFFFFF7FFFDF;
    ans = 64'hFFFFFFFFFF7FFFDF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"894131\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h3800003FFFFFEFFF;
    ans = 64'hBFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"894437\n");
    end
    ReadData1E = 64'h4023805779AA7BA6;
    ReadData2E = 64'h41F0000000000000;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'h4223805779AA7BA6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"894743\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"895049\n");
    end
    ReadData1E = 64'hB751FFFBFFFFFFFF;
    ReadData2E = 64'hC3C9ED9349BBCC6D;
    ReadData3E = 64'h38100000FFFFFFC0;
    ans = 64'h3B2D2B3F378E7392;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"895355\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h7E1FFFFFFFFFDFF0;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h7DFFFFFFFFFFDFF0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"895661\n");
    end
    ReadData1E = 64'hBAE0000000003FFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'hBFB35446159C93A3;
    ans = 64'hBFB35446159C93A3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"895967\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h3A9DFFFFFFFFFEFF;
    ReadData3E = 64'hB8000000FFFFFDFE;
    ans = 64'h3A7DFFFFFFFFDEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"896273\n");
    end
    ReadData1E = 64'hBFCF52E5F0E06610;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h4021F52E5F0E0660;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"896579\n");
    end
    ReadData1E = 64'h802E487204168AE2;
    ReadData2E = 64'h3FD0002080000000;
    ReadData3E = 64'h7FDFFFFC01FFFFFF;
    ans = 64'h7FDFFFFC01FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"896885\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hC1F3FFFFFFFEFFFF;
    ans = 64'hC1F40000000EFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"897191\n");
    end
    ReadData1E = 64'hC01F000001000000;
    ReadData2E = 64'hBFE000200001FFFF;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h400F003E0103E1FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"897497\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"897803\n");
    end
    ReadData1E = 64'hB7F0220000000000;
    ReadData2E = 64'h4804298279080EA6;
    ReadData3E = 64'h43E0000FFFFFFFBE;
    ans = 64'h43E0000FFFFFFFBE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"898109\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC90000040000FFFE;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hC8E000040000FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"898415\n");
    end
    ReadData1E = 64'h3FF1896CDCA1FBD1;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'h802400000FFFFFFF;
    ans = 64'hC021896CDCA1FBD0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"898721\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC01FFFBC00000000;
    ReadData3E = 64'hB7E52B803F675515;
    ans = 64'hBFFFFFBC00000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"899027\n");
    end
    ReadData1E = 64'h41D3FD5D0A16B1AB;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hC203FD5D0A36B1AA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"899333\n");
    end
    ReadData1E = 64'hB7F0200000000004;
    ReadData2E = 64'hBFBFFFFFFFFF00FF;
    ReadData3E = 64'hDFB3322F1716D84C;
    ans = 64'hDFB3322F1716D84C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"899639\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'hC7E0040000000000;
    ans = 64'hC7E0040000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"899945\n");
    end
    ReadData1E = 64'h43F020000003FFFF;
    ReadData2E = 64'h0DF007FFFFFFC000;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"900251\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'hC31FFFFFFFFFFFE2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"900557\n");
    end
    ReadData1E = 64'hC80BFFFFFFFF8000;
    ReadData2E = 64'hC36D1CF2F5E2A67E;
    ReadData3E = 64'hC1FFFFFFFF007FFE;
    ans = 64'h4B8979549725DD3A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"900863\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hFFFB6A3A2432DE41;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'hFFFB6A3A2432DE41;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"901169\n");
    end
    ReadData1E = 64'hC3E000000000407F;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h3CAFFF9000000000;
    ans = 64'h474000000000407E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"901475\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h2768D5A43EE64697;
    ReadData3E = 64'h1AA00000FFFFBFFF;
    ans = 64'h2748D5A43EE64697;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"901781\n");
    end
    ReadData1E = 64'h380FFFFFFFFFFFFF;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"902087\n");
    end
    ReadData1E = 64'hC13F3F19214DF452;
    ReadData2E = 64'h3FF0001FFFF00000;
    ReadData3E = 64'h402FFFFFFFFFC7FF;
    ans = 64'hC13F3F479F60F7D5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"902393\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'h3C5020007FFFFFFF;
    ans = 64'hFFC0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"902699\n");
    end
    ReadData1E = 64'h3FEFFFFFF9FFFFFF;
    ReadData2E = 64'h3F10000000008004;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"903005\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hFFC0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"903311\n");
    end
    ReadData1E = 64'h3FCFFFFFFC000010;
    ReadData2E = 64'hBCA00000003FFBFE;
    ReadData3E = 64'h5C6D7C81CE014EF9;
    ans = 64'h5C6D7C81CE014EF9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"903617\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hC1FD2AEBDB57C7AA;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hC1DD2AEBDB57C7AA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"903923\n");
    end
    ReadData1E = 64'hC0121AB24E47D412;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'h403FC00000010000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"904229\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h1FB0002FFFFFFFFF;
    ReadData3E = 64'h380FFFEFFFE00000;
    ans = 64'h380FFFEFFFE00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"904535\n");
    end
    ReadData1E = 64'h800F795795E454AD;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'h4006F2AF2BC8A959;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"904841\n");
    end
    ReadData1E = 64'h47EFFFFFFFFEFDFF;
    ReadData2E = 64'hC080000780000000;
    ReadData3E = 64'hBFE0000FFFFFDFFE;
    ans = 64'hC88000077FFF7EFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"905147\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'hFFD6C1C477639F3D;
    ans = 64'hFFE360E23BB1CF9E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"905453\n");
    end
    ReadData1E = 64'hB80FFC0000004000;
    ReadData2E = 64'h3FDC7ACB3D8E0D56;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"905759\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h4000000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"906065\n");
    end
    ReadData1E = 64'h405FFF01FFFFFFFE;
    ReadData2E = 64'h471E00FFFFFFFFFF;
    ReadData3E = 64'h41D000000FFF7FFF;
    ans = 64'h478E0011D80FFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"906371\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'h8EB00000017FFFFF;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h8E900000017FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"906677\n");
    end
    ReadData1E = 64'hFFD40000000001FF;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'hBFDFFFFFFC008000;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"906983\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hFFFFFFFFE0000200;
    ReadData3E = 64'h38120EB5CFFF927A;
    ans = 64'hFFFFFFFFE0000200;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"907289\n");
    end
    ReadData1E = 64'hD960000000DFFFFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"907595\n");
    end
    ReadData1E = 64'h40DFFFFFFC040000;
    ReadData2E = 64'h7FE000000020FFFF;
    ReadData3E = 64'hFFF00000007FFEFE;
    ans = 64'hFFF80000007FFEFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"907901\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h43EBFFFFFF7FFFFE;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"908207\n");
    end
    ReadData1E = 64'h381000003FFFFFFF;
    ReadData2E = 64'hBF1886CB0D330AA3;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"908513\n");
    end
    ReadData1E = 64'h3FD0000000000000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"908819\n");
    end
    ReadData1E = 64'h407ED04D6C9499F6;
    ReadData2E = 64'h380FFFFEFFFBFFFE;
    ReadData3E = 64'h3810000007FFFFE0;
    ans = 64'h389EE04C76165486;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"909125\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hB02083FFFFFFFFFE;
    ReadData3E = 64'h8000000000000001;
    ans = 64'hB00083FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"909431\n");
    end
    ReadData1E = 64'hB450000003DFFFFF;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'hC02FFFFFE0100000;
    ans = 64'hC02FFFFFE0100000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"909737\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC3EFFFFF00003FFF;
    ReadData3E = 64'hB7FF000000000002;
    ans = 64'hC3CFFFFF00004001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"910043\n");
    end
    ReadData1E = 64'hB3DCBBCBF4FC7B3C;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"910349\n");
    end
    ReadData1E = 64'hC01FE7FFFFFFFFFF;
    ReadData2E = 64'h43DFD4A9159E1E89;
    ReadData3E = 64'h43C000000F7FFFFF;
    ans = 64'hC40EBCC995D5E7F1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"910655\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'h59FFFE00000003FE;
    ans = 64'h59FFFE00000003FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"910961\n");
    end
    ReadData1E = 64'hC7FE000000001FFF;
    ReadData2E = 64'hC3CFFFFFFFFC0100;
    ReadData3E = 64'hC010000000000001;
    ans = 64'h4BDDFFFFFFFC60EF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"911267\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"911573\n");
    end
    ReadData1E = 64'h43D7DDC6B07E1242;
    ReadData2E = 64'h7FD000000013FFFF;
    ReadData3E = 64'hC34FFEFFFFFFFFE0;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"911879\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h21E1000000001FFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"912185\n");
    end
    ReadData1E = 64'h3BE00000008007FF;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h45013FFFFFFFFFFF;
    ans = 64'h45013FFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"912491\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h56DC000000000080;
    ReadData3E = 64'hFFE0000000020010;
    ans = 64'hFFE0000000020010;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"912797\n");
    end
    ReadData1E = 64'h5D9EF3F25478434E;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"913103\n");
    end
    ReadData1E = 64'hC01F00000000003E;
    ReadData2E = 64'h72A001FFFFFFEFFF;
    ReadData3E = 64'hBCE000007FFBFFFF;
    ans = 64'hF2CF03DFFFFFE13C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"913409\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h3CAFC07FFFFFFFFE;
    ans = 64'h3CAFC07FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"913715\n");
    end
    ReadData1E = 64'h429E54A46244D94F;
    ReadData2E = 64'h41FEFFFFFFFFDFFE;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h44AD61FF3F32942E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"914021\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"914327\n");
    end
    ReadData1E = 64'hC3E00007FE000000;
    ReadData2E = 64'h949F7FFFFDFFFFFF;
    ReadData3E = 64'hC1E000003FFFF7FE;
    ans = 64'hC1E000003FFFF7FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"914633\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBCA42A241576C81A;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"914939\n");
    end
    ReadData1E = 64'h41FFFF8000000010;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h3FF048A52C15DD23;
    ans = 64'h3FF048A52C15DD23;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"915245\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFD9E0724E2AFED0;
    ReadData3E = 64'hC3DEACACA3A5320B;
    ans = 64'hC3DEACACA3A5320B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"915551\n");
    end
    ReadData1E = 64'hC0FFFE00001FFFFF;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBDAFFE10001FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"915857\n");
    end
    ReadData1E = 64'hB80FFFF7FFFEFFFF;
    ReadData2E = 64'h3F40000007FFFEFF;
    ReadData3E = 64'h41EFFFFFEFBFFFFE;
    ans = 64'h41EFFFFFEFBFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"916163\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h43440000000003FF;
    ans = 64'h43440000000003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"916469\n");
    end
    ReadData1E = 64'h3FC199DB34D4E268;
    ReadData2E = 64'h402FF800000003FE;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'h3FC9574BE07AF628;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"916775\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"917081\n");
    end
    ReadData1E = 64'hFFD0003FBFFFFFFE;
    ReadData2E = 64'hC00FFFFCFFFFFFFF;
    ReadData3E = 64'h380FFFE000080000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"917387\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3F6FFFFFFFC40000;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"917693\n");
    end
    ReadData1E = 64'h4D60624F59DB184C;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hC1CFFFFFFFFFFDDF;
    ans = 64'h4A20624F59DB184B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"917999\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h4019CB37CAB5578C;
    ReadData3E = 64'h002FFFFDFFFEFFFF;
    ans = 64'h3FF9CB37CAB5578E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"918305\n");
    end
    ReadData1E = 64'hC3D5D738638A1944;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'hC3B5D738638A1944;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"918611\n");
    end
    ReadData1E = 64'hBFC00000FFFFFE00;
    ReadData2E = 64'h30DFFE0000000007;
    ReadData3E = 64'h3FDFFFF00000FFFE;
    ans = 64'h3FDFFFF00000FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"918917\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'hB7E4C284CB2F81D3;
    ans = 64'h3FB0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"919223\n");
    end
    ReadData1E = 64'hB7FEF5B5A6DCC55E;
    ReadData2E = 64'hBFCFFFFFFF80003F;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h3FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"919529\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'h3FAFFFFFFFFFFFF4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"919835\n");
    end
    ReadData1E = 64'h4006E5FE7D1A8DC4;
    ReadData2E = 64'h7FD80000001FFFFF;
    ReadData3E = 64'h3F3000001FEFFFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"920141\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h381FFFE000001000;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"920447\n");
    end
    ReadData1E = 64'h47EFFF7FFFEFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h40AFFFFFFFFFFF08;
    ans = 64'h47DFFF7FFFEFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"920753\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC0A1FFFFFFFFC000;
    ReadData3E = 64'hCF783ACF6657D485;
    ans = 64'hCF783ACF6657D485;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"921059\n");
    end
    ReadData1E = 64'h557BCB0C4F71816F;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h556BCB0C4F71816D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"921365\n");
    end
    ReadData1E = 64'h5920000107FFFFFF;
    ReadData2E = 64'hBBD00000010003FF;
    ReadData3E = 64'h2147FEFFFFFFFFFE;
    ans = 64'hD50000010900040F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"921671\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'hC0355E7698BA0005;
    ans = 64'hC0353E7698BA0005;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"921977\n");
    end
    ReadData1E = 64'h802E001FFFFFFFFF;
    ReadData2E = 64'hB79FFFFDFFDFFFFE;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hBFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"922283\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h3FC0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"922589\n");
    end
    ReadData1E = 64'h43DFFFFFFFA00000;
    ReadData2E = 64'hE0FFFE0000000000;
    ReadData3E = 64'hC3EFF00010000000;
    ans = 64'hE4EFFDFFFFA00600;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"922895\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC1D000002000007F;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC1B0000028000080;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"923201\n");
    end
    ReadData1E = 64'hC027F3B5DBAB58EF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'hC04FFF000001FFFF;
    ans = 64'hC052FDF6BB766B1D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"923507\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h001FFFBDFFFFFFFE;
    ReadData3E = 64'hBFDFFFFC01FFFFFF;
    ans = 64'hBFDFFFFC01FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"923813\n");
    end
    ReadData1E = 64'h3FCDA0B73B3E429D;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h3FCDA0B73B3E429B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"924119\n");
    end
    ReadData1E = 64'h43EFFFFEFFFFFFEF;
    ReadData2E = 64'hB1A7F7D6729904E6;
    ReadData3E = 64'hAD80008000002000;
    ans = 64'hB5A7F7D5B2DA5144;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"924425\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h43F000800007FFFF;
    ans = 64'h43F000800007FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"924731\n");
    end
    ReadData1E = 64'h7CD00004000000FF;
    ReadData2E = 64'h3CA7FFFFBFFFFFFF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h79880005BFFFF17D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"925037\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h3FD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"925343\n");
    end
    ReadData1E = 64'h3D5FEFFFFFDFFFFF;
    ReadData2E = 64'h56B05A2CE13DEC1D;
    ReadData3E = 64'hC1FFFFFFFF400000;
    ans = 64'h542051FFCABCF2FA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"925649\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFCFFFFDBFFFFFFE;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400F800008FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"925955\n");
    end
    ReadData1E = 64'hBFC0000000000018;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'hC3F000000005FFFF;
    ans = 64'hC3F000000005FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"926261\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hFFD26C3811E7D282;
    ReadData3E = 64'h43E6E1E67A741F0C;
    ans = 64'hFFB26C3811E7D283;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"926567\n");
    end
    ReadData1E = 64'hBFC70E2F42B43FA9;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"926873\n");
    end
    ReadData1E = 64'h3850000000000000;
    ReadData2E = 64'h8010000000800080;
    ReadData3E = 64'h3FB0020000000FFF;
    ans = 64'h3FB0020000000FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"927179\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC3C0003F00000000;
    ans = 64'hC3C0003F00000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"927485\n");
    end
    ReadData1E = 64'h400643BE798DFA0B;
    ReadData2E = 64'hBFDB38C21CCBA9FC;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBFF2F0A652A03F0C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"927791\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"928097\n");
    end
    ReadData1E = 64'hC1FFC00000000400;
    ReadData2E = 64'hBFDFFFFEFFFDFFFE;
    ReadData3E = 64'hC1D000200000007F;
    ans = 64'h41E7BFEF01FE07BF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"928403\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hB04ED49F5FEFA1FE;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"928709\n");
    end
    ReadData1E = 64'h3810000040040000;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hC7EFFFFFFFDEFFFE;
    ans = 64'hC7EFFFFFFFDEFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"929015\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC1D0000000080400;
    ReadData3E = 64'h3FBFFFFC07FFFFFF;
    ans = 64'hC1AFFFFFFFD0080A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"929321\n");
    end
    ReadData1E = 64'hC7E008000FFFFFFE;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"929627\n");
    end
    ReadData1E = 64'h8155FDEE412185B1;
    ReadData2E = 64'h3F9FFFFFF8000FFE;
    ReadData3E = 64'hC01000004FFFFFFE;
    ans = 64'hC01000004FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"929933\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'hBFFFFFFFFF0003FF;
    ans = 64'hBFEFFFFFFE0007FD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"930239\n");
    end
    ReadData1E = 64'hB81F0000000007FF;
    ReadData2E = 64'h4021FFFFFFFFFFDF;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h3CA0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"930545\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"930851\n");
    end
    ReadData1E = 64'h0020000040200000;
    ReadData2E = 64'hB808F7D662C77EE8;
    ReadData3E = 64'h7FDC9A96BB0DE738;
    ans = 64'h7FDC9A96BB0DE738;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"931157\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h41D0000001FFFF80;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h41B0000002FFFF81;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"931463\n");
    end
    ReadData1E = 64'h244FFE0000000007;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'h4010000000FFFFFE;
    ans = 64'h4010000000FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"931769\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h7FD000000000401F;
    ReadData3E = 64'h43C6B4670C817DE7;
    ans = 64'h7FB0000000004020;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"932075\n");
    end
    ReadData1E = 64'hC490000000090000;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'hC4AFFFFF80120002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"932381\n");
    end
    ReadData1E = 64'hBCA000000000077F;
    ReadData2E = 64'h3FD000000000003F;
    ReadData3E = 64'h37F4A35F530E6918;
    ans = 64'hBC800000000007BE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"932687\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hBFF30581BE620614;
    ans = 64'h3FE9F4FC833BF3DA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"932993\n");
    end
    ReadData1E = 64'h3FB0000400003FFE;
    ReadData2E = 64'h41E000000007FFFC;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h41A0000400083FFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"933299\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"933605\n");
    end
    ReadData1E = 64'h3FC54E5D41712B9E;
    ReadData2E = 64'h47E000010007FFFF;
    ReadData3E = 64'h33C0002007FFFFFF;
    ans = 64'h47B54E5E9661A6E2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"933911\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFE00001FFF7FFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFF200003FFF0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"934217\n");
    end
    ReadData1E = 64'hBF1FFC0100000000;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'hB7E41A93A1F3F585;
    ans = 64'hC26FFC0100000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"934523\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC7F0007FFFFFFEFF;
    ReadData3E = 64'hB77FFFFFFFF8003F;
    ans = 64'hC7D0007FFFFFFF00;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"934829\n");
    end
    ReadData1E = 64'hBF5FEFFBFFFFFFFE;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hC34007FBFF000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"935135\n");
    end
    ReadData1E = 64'hC14400000003FFFE;
    ReadData2E = 64'h402FFFFFE007FFFE;
    ReadData3E = 64'h40070F2EB2AE69A8;
    ans = 64'hC183FFFFD4F9D14A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"935441\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'hBF30000001FFF7FF;
    ans = 64'h4320000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"935747\n");
    end
    ReadData1E = 64'h7FE3BD3ADDB19B8A;
    ReadData2E = 64'h41E856C1AFCD15DF;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"936053\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h4353FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"936359\n");
    end
    ReadData1E = 64'h8000000004004000;
    ReadData2E = 64'hC7FFEFFFFFDFFFFE;
    ReadData3E = 64'hBFDFFFE000100000;
    ans = 64'hBFDFFFE000100000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"936665\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hF4E2706CF518E262;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'hF4C2706CF518E263;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"936971\n");
    end
    ReadData1E = 64'h82D59DA4597BE47B;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'hC0000000008007FF;
    ans = 64'hC0000000008007FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"937277\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBF40000000003F7E;
    ReadData3E = 64'h036000000000017F;
    ans = 64'hBF20000000003F7F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"937583\n");
    end
    ReadData1E = 64'hBDDBFFFFFEFFFFFF;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h4010000000000001;
    ans = 64'hFDCBFFFFFEFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"937889\n");
    end
    ReadData1E = 64'hB7FFFFFFFFFF7FF8;
    ReadData2E = 64'h41E9D8AF3D98FD59;
    ReadData3E = 64'h3FCCA1065CC75F03;
    ans = 64'h3FCCA1065CC75F03;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"938195\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h400003FBFFFFFFFF;
    ans = 64'h7FC0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"938501\n");
    end
    ReadData1E = 64'h23B930F468025F01;
    ReadData2E = 64'hBFF0AC6DEFA519FB;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"938807\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'hC340000000000000;
    ans = 64'h7FC0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"939113\n");
    end
    ReadData1E = 64'h401A984CEE2CCFDF;
    ReadData2E = 64'h2E8FFFDFFFFFFFBF;
    ReadData3E = 64'hBFBFFFBFFFFFEFFF;
    ans = 64'hBFBFFFBFFFFFEFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"939419\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC3492C4DB0C5D4CF;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hC3292C4DB0C5D4D1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"939725\n");
    end
    ReadData1E = 64'h41EAAC2889278322;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'hBFD20001FFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"940031\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC032A56D96C76E30;
    ReadData3E = 64'hB81000000002FFFF;
    ans = 64'hC012A56D96C76E31;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"940337\n");
    end
    ReadData1E = 64'h3FE0003FFFFFFFBF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'h7FE0003FFFFFFFBE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"940643\n");
    end
    ReadData1E = 64'h936FFFC000000006;
    ReadData2E = 64'h43FF81FFFFFFFFFF;
    ReadData3E = 64'hC3CFFFFFFFFC001E;
    ans = 64'hC3CFFFFFFFFC001E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"940949\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h42DC090B99823732;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"941255\n");
    end
    ReadData1E = 64'h3EE0000FFFFFFDFF;
    ReadData2E = 64'h3800004000002000;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"941561\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"941867\n");
    end
    ReadData1E = 64'hB7F0000000FFFFF8;
    ReadData2E = 64'h47FFFEFFFFFFFFFD;
    ReadData3E = 64'h43D003FDFFFFFFFE;
    ans = 64'h43D003FDFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"942173\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h41D24652A1E9AC61;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h41B24652A1E9AC62;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"942479\n");
    end
    ReadData1E = 64'h3C9003FFF8000000;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h3F2FFFFF000FFFFE;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"942785\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FB3971FA94CB21A;
    ReadData3E = 64'h407FFFFFFFFDFFFF;
    ans = 64'h408000272E3E5299;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"943091\n");
    end
    ReadData1E = 64'h37E800000000001F;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"943397\n");
    end
    ReadData1E = 64'hC1E00001E0000000;
    ReadData2E = 64'h47EFFEFFFFC00000;
    ReadData3E = 64'hBFC0000001BFFFFF;
    ans = 64'hC9DFFF03BFA1FFF9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"943703\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h37F5CAD1D5067C14;
    ans = 64'h37F5CAD1D5067C14;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"944009\n");
    end
    ReadData1E = 64'hB7E000000FFFFFFE;
    ReadData2E = 64'hB3AFFEFFFDFFFFFF;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"944315\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hC00FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"944621\n");
    end
    ReadData1E = 64'h801FFF7FFFFFF7FF;
    ReadData2E = 64'hB52FFFFE00000002;
    ReadData3E = 64'h38700007FFFFFFEF;
    ans = 64'h38700007FFFFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"944927\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h381FFFFFF7FFFF00;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'h37FFFFFFF7FFFF02;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"945233\n");
    end
    ReadData1E = 64'hC0000004003FFFFE;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h41DFFFEFFFDFFFFF;
    ans = 64'h41DFFFEFFFDFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"945539\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC3CFFFFFF0200000;
    ReadData3E = 64'h3FF000200000FFFE;
    ans = 64'hC3AFFFFFF0200002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"945845\n");
    end
    ReadData1E = 64'hC00C000000100000;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"946151\n");
    end
    ReadData1E = 64'h5DA0000203FFFFFE;
    ReadData2E = 64'h3F5FBFFFFFC00000;
    ReadData3E = 64'h3BEFFFFC00000FFF;
    ans = 64'h5D0FC003FFAFFFF4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"946457\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h43E0000FFF000000;
    ans = 64'h43E0000FFF000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"946763\n");
    end
    ReadData1E = 64'h37F54C04C2DD9BA8;
    ReadData2E = 64'hBFB0017FFFFFFFFF;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"947069\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"947375\n");
    end
    ReadData1E = 64'hACDFFF7FFFFFFBFE;
    ReadData2E = 64'h3FD0000FFFBFFFFE;
    ReadData3E = 64'hC01FFFFFFF7FFFDF;
    ans = 64'hC01FFFFFFF7FFFDF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"947681\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC1D00000DFFFFFFE;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'hC1B00000DFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"947987\n");
    end
    ReadData1E = 64'h47E000040000000F;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'hBF50000007FFFBFF;
    ans = 64'hBF50000007FFFBFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"948293\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC80FFFF7FF7FFFFE;
    ReadData3E = 64'h802FFF8000800000;
    ans = 64'hC7EFFFF7FF800000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"948599\n");
    end
    ReadData1E = 64'hB9C01FFFFFFFF800;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"948905\n");
    end
    ReadData1E = 64'hBFC0001007FFFFFE;
    ReadData2E = 64'hBFD10000000001FF;
    ReadData3E = 64'h99700000013FFFFF;
    ans = 64'h3FA10011088001FD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"949211\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'hC3D0000000013FFF;
    ans = 64'hC3D0000000013FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"949517\n");
    end
    ReadData1E = 64'hC38C638EECC31ABD;
    ReadData2E = 64'hC1E0003FFE000000;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h457C640077725BEC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"949823\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hBFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"950129\n");
    end
    ReadData1E = 64'h4024C8265CB4F296;
    ReadData2E = 64'hC033F2685AF47617;
    ReadData3E = 64'h3FF17E93439BFDF1;
    ans = 64'hC069C58B36E690F8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"950435\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBCA001000001FFFF;
    ReadData3E = 64'h8000000000000000;
    ans = 64'hBC80010000020000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"950741\n");
    end
    ReadData1E = 64'h401FFFFFE0000080;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'hC13FFFEFFF800000;
    ans = 64'hC13FFFEFFF800000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"951047\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h407FFFFFFFFFFFFF;
    ReadData3E = 64'hB7FFFFFFFFFBFF7F;
    ans = 64'h4060000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"951353\n");
    end
    ReadData1E = 64'h3FC0000000000BFE;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"951659\n");
    end
    ReadData1E = 64'h20E0000000780000;
    ReadData2E = 64'h3880801FFFFFFFFF;
    ReadData3E = 64'h4020000008080000;
    ans = 64'h4020000008080000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"951965\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h381000000001001E;
    ans = 64'hBC90000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"952271\n");
    end
    ReadData1E = 64'hCBFFFFFFE1FFFFFE;
    ReadData2E = 64'hAA2FFEFFFFFFEFFE;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"952577\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"952883\n");
    end
    ReadData1E = 64'h8023F3F2A8389E11;
    ReadData2E = 64'h066175AB575ABFB2;
    ReadData3E = 64'h3FDFFFFFFFFC01FF;
    ans = 64'h3FDFFFFFFFFC01FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"953189\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h44468C671497C5ED;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"953495\n");
    end
    ReadData1E = 64'h3FC4793378DC78BE;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'hC7FFF08000000000;
    ans = 64'hC7FFF08000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"953801\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FE1638DA0EC77C4;
    ReadData3E = 64'hC1D0000100100000;
    ans = 64'hC1D0000100074E39;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"954107\n");
    end
    ReadData1E = 64'hBB8FBFFEFFFFFFFF;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CB00000FDFFF7FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"954413\n");
    end
    ReadData1E = 64'hBFE78F372393823A;
    ReadData2E = 64'h40284FEE9BB805EE;
    ReadData3E = 64'h3801623519F16C3D;
    ans = 64'hC021E642969DBD5D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"954719\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h7FFEF9A73225A948;
    ans = 64'h7FFEF9A73225A948;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"955025\n");
    end
    ReadData1E = 64'h3FCFF70000000000;
    ReadData2E = 64'hFFE1112C9531B024;
    ReadData3E = 64'h4000000000000001;
    ans = 64'hFFC10C5FC0A7BA2A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"955331\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFE4000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"955637\n");
    end
    ReadData1E = 64'h9B800000200003FF;
    ReadData2E = 64'hBFE014D6B151044F;
    ReadData3E = 64'h3734B9CE039E32CB;
    ans = 64'h3734B9CE039E32CB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"955943\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC7EDFFFF7FFFFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"956249\n");
    end
    ReadData1E = 64'h3E401FFDFFFFFFFE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'hBF8F80000000FFFF;
    ans = 64'hBF8F80008100EFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"956555\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC1DFFEFFFFF7FFFF;
    ReadData3E = 64'hDB300000000FFFBF;
    ans = 64'hDB300000000FFFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"956861\n");
    end
    ReadData1E = 64'h47FE49AB039DFC52;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hC7EE49AB039DFC52;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"957167\n");
    end
    ReadData1E = 64'hC560FFFFFFFFFFBE;
    ReadData2E = 64'hBFFFFFFFFFFFFFE2;
    ReadData3E = 64'hC7FFBED86249373C;
    ans = 64'hC7FFBED86249263C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"957473\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'hBFFD2C90EF0F0A69;
    ans = 64'hBFFF2C90EF0F0A69;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"957779\n");
    end
    ReadData1E = 64'h489CCE60FE90D5B4;
    ReadData2E = 64'hC1FDCD10BFBA3993;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hCAAAD3A117DA0724;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"958085\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'hBFCFFFFFFFFFFFF9;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"958391\n");
    end
    ReadData1E = 64'hB7F0008000000000;
    ReadData2E = 64'hC01FFEFFDFFFFFFF;
    ReadData3E = 64'h7FEFF8000000FFFE;
    ans = 64'h7FEFF8000000FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"958697\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC03BC4DBCEA3AFEF;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC350000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"959003\n");
    end
    ReadData1E = 64'h401FFDFFFFF00000;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'hC3400D15CAF2C5CA;
    ans = 64'hC3400D15CAF2C5CE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"959309\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFCFFFFE00007FFF;
    ReadData3E = 64'hC1F0000003FFBFFF;
    ans = 64'hC1F000000400BFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"959615\n");
    end
    ReadData1E = 64'hC100002000000000;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h4100002000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"959921\n");
    end
    ReadData1E = 64'hFFD800000007FFFE;
    ReadData2E = 64'h3FD08000007FFFFE;
    ReadData3E = 64'hB7F3784914F45A14;
    ans = 64'hFFB8C00000C83FFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"960227\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h002FFFBFFFFFFC00;
    ans = 64'hBFD0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"960533\n");
    end
    ReadData1E = 64'h3FDFDFFFFFFFFFFD;
    ReadData2E = 64'h8017FFFFEFFFFFFF;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"960839\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBFD0000000000004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"961145\n");
    end
    ReadData1E = 64'h3960743905826A36;
    ReadData2E = 64'hC02000001FFFFFEF;
    ReadData3E = 64'hC0365E1AE8AC5C5E;
    ans = 64'hC0365E1AE8AC5C5E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"961451\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC1CFFFFFFBFFFFBF;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h433FFFFFF0000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"961757\n");
    end
    ReadData1E = 64'h001B890A98279E8B;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h408F000000020000;
    ans = 64'h408F000000020000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"962063\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hFFEFFFFFFFFFDFEE;
    ReadData3E = 64'h3CF000000007FFF7;
    ans = 64'hFFCFFFFFFFFFDFF0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"962369\n");
    end
    ReadData1E = 64'hBFCB62E6ED21BD63;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h3FDB62E6ED21BD61;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"962675\n");
    end
    ReadData1E = 64'hC1E000000027FFFE;
    ReadData2E = 64'h3FAFFFFEFFFFFF80;
    ReadData3E = 64'hC02FFFFFFE007FFF;
    ans = 64'hC19FFFFF404FFF76;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"962981\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h4F40000400080000;
    ans = 64'h4F40000400080000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"963287\n");
    end
    ReadData1E = 64'h381FFFFFEDFFFFFF;
    ReadData2E = 64'hFFFFFFFE0000007F;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hFFFFFFFE0000007F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"963593\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h0010000000000001;
    ans = 64'hBFE0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"963899\n");
    end
    ReadData1E = 64'hBCA001FFFFDFFFFF;
    ReadData2E = 64'h3FE2F2314E94F75E;
    ReadData3E = 64'hBAEFFE0000000200;
    ans = 64'hBC92F48F9898A599;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"964205\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h00171492F6962A8C;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"964511\n");
    end
    ReadData1E = 64'h800A1EC2F81DC219;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h37E6F1B7F6356355;
    ans = 64'h37E6F1B7F6356355;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"964817\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h40A0000000800000;
    ReadData3E = 64'hBFFDF4CAB0144EF1;
    ans = 64'h407FE20B364FEBB3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"965123\n");
    end
    ReadData1E = 64'h7FF3BE7F51FBB158;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"965429\n");
    end
    ReadData1E = 64'h37F0000000014000;
    ReadData2E = 64'h89066053EC882A36;
    ReadData3E = 64'hC1B000000011FFFF;
    ans = 64'hC1B000000011FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"965735\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'hC1CFFFFFFFFF8007;
    ans = 64'hC1D00000003FC003;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"966041\n");
    end
    ReadData1E = 64'hC2CFFFFFFFFFDFC0;
    ReadData2E = 64'hC3F01FFFFFFFFFF6;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'h46D01FFFFFFFEFB6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"966347\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'hBFF0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"966653\n");
    end
    ReadData1E = 64'h4E3FFC01FFFFFFFF;
    ReadData2E = 64'h62E1FFFEFFFFFFFE;
    ReadData3E = 64'h3FDFFFFFFC00007F;
    ans = 64'h7131FDC0201FEFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"966959\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBFDFFF8000000080;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h400F0003FFFFFFFB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"967265\n");
    end
    ReadData1E = 64'hC13000001DFFFFFF;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hC000001FFFFFFFF0;
    ans = 64'h414FFFFF3BFE0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"967571\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3253FDFFFFFFFFFF;
    ReadData3E = 64'hBCA1FFFFFFFFE000;
    ans = 64'hBCA1FFFFFFFFE000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"967877\n");
    end
    ReadData1E = 64'h7FD00400000001FF;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"968183\n");
    end
    ReadData1E = 64'h5C2AC90B8977039C;
    ReadData2E = 64'hBCA0000FFFFFFDFF;
    ReadData3E = 64'hC0CFFFFFFFFE7FFF;
    ans = 64'hD8DAC926528289B8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"968489\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hFFDFFFC001FFFFFF;
    ans = 64'hFFDFFFC001FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"968795\n");
    end
    ReadData1E = 64'h47E1FEFFFFFFFFFE;
    ReadData2E = 64'hBFF0041000000000;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hC7E20391BEFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"969101\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"969407\n");
    end
    ReadData1E = 64'hFFF0000400000000;
    ReadData2E = 64'h78400000000FFDFF;
    ReadData3E = 64'hC1F36615D549A0E5;
    ans = 64'hFFF8000400000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"969713\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC006A045CF961FE9;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hC005A81173E587FA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"970019\n");
    end
    ReadData1E = 64'hC1E20FFFFFFFFFFF;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h3FC03FFFFFF80000;
    ans = 64'h4532100000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"970325\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC1FFDFFF7FFFFFFE;
    ReadData3E = 64'hBFDDFFFFEFFFFFFF;
    ans = 64'hC1DFDFFF801E0000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"970631\n");
    end
    ReadData1E = 64'hC80A1EC7C632AD61;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"970937\n");
    end
    ReadData1E = 64'hEEEDFBFFFFFFFFFE;
    ReadData2E = 64'hC3F07FFFBFFFFFFE;
    ReadData3E = 64'h3FFA5202B857DE26;
    ans = 64'h72EEEBDF880FFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"971243\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'hC00FFFFFF8000003;
    ans = 64'hC330000000000004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"971549\n");
    end
    ReadData1E = 64'h41F63343608D9BC4;
    ReadData2E = 64'h51FBCA96E92C466E;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h540347CD24D8C7DD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"971855\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"972161\n");
    end
    ReadData1E = 64'h9AECFD35BA2EEEA1;
    ReadData2E = 64'h45E0000080003FFF;
    ReadData3E = 64'hBFEFFFFFFFBFF7FF;
    ans = 64'hBFEFFFFFFFBFF7FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"972467\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h594FF80000000010;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h592FF80000000012;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"972773\n");
    end
    ReadData1E = 64'h0000008FFFFFFFFE;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'hC055969B9E03C347;
    ans = 64'hC05596A01E03C347;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"973079\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h43DFE000000003FF;
    ReadData3E = 64'h47EFFFFEFFF80000;
    ans = 64'h47EFFFFEFFF80000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"973385\n");
    end
    ReadData1E = 64'h287FEFFEFFFFFFFF;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'hE86FEFFF00000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"973691\n");
    end
    ReadData1E = 64'hC3E800000000003F;
    ReadData2E = 64'h802E7EFB0F9DFB33;
    ReadData3E = 64'hC01FFFFFFFFFFFE7;
    ans = 64'hC01FFFFFFFFFFFE7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"973997\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'hEC7FFFBFFBFFFFFF;
    ans = 64'hFFD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"974303\n");
    end
    ReadData1E = 64'h3CE000003FFEFFFF;
    ReadData2E = 64'hC130040003FFFFFE;
    ReadData3E = 64'h8010000000000000;
    ans = 64'hBE200400440EFFCD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"974609\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"974915\n");
    end
    ReadData1E = 64'hC00DFFBFFFFFFFFF;
    ReadData2E = 64'hC075E9323CB1786B;
    ReadData3E = 64'hC008AD1C6219D7C3;
    ans = 64'h40947E1CB850DA95;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"975221\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBF6FF8000000007F;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'hBFE007FE00000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"975527\n");
    end
    ReadData1E = 64'h69CC4C2584E45E21;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h26CA7C0E0EDC5D70;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"975833\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'h3FE8720FB6B51807;
    ReadData3E = 64'hBFC000007EFFFFFF;
    ans = 64'h3FB0E41E6F6A3013;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"976139\n");
    end
    ReadData1E = 64'hC3F37CDA8E6B566C;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"976445\n");
    end
    ReadData1E = 64'hFFD21FFFFFFFFFFE;
    ReadData2E = 64'hB810000001FEFFFE;
    ReadData3E = 64'hC80FFFFFFFA00000;
    ans = 64'h77F220000242DDFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"976751\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'hC34000000003F000;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"977057\n");
    end
    ReadData1E = 64'h381FFFF7FF7FFFFF;
    ReadData2E = 64'hE60A593DE76F061F;
    ReadData3E = 64'h0000000000000001;
    ans = 64'hDE3A593750B6274B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"977363\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"977669\n");
    end
    ReadData1E = 64'hBFE5C55F9D983738;
    ReadData2E = 64'hBFE0003FFFFDFFFE;
    ReadData3E = 64'h3D4D4C7C8BA68069;
    ans = 64'h3FD5C5B6B3140390;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"977975\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hC3EFFFFFFFFFB7FF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'hC3CFFFFFFFFFB801;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"978281\n");
    end
    ReadData1E = 64'h4023A84702311282;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h4018B473E74D3CF2;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"978587\n");
    end
    ReadData1E = 64'h3FD0000000000001;
    ReadData2E = 64'hBF3000010000007E;
    ReadData3E = 64'hBFFFFFFFFFFFF802;
    ans = 64'hC00000200001FC01;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"978893\n");
    end
    ReadData1E = 64'hFFEFFFFFFEFDFFFF;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"979199\n");
    end
    ReadData1E = 64'h3C300000FFFFF7FF;
    ReadData2E = 64'hC2BFFF0000000004;
    ReadData3E = 64'hBFCBFB075CA1326A;
    ans = 64'hBFCBFC0754B131EA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"979505\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h41AFFFFFFC0001FE;
    ans = 64'h41AFFFFFFC0001FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"979811\n");
    end
    ReadData1E = 64'h40FFFFFFFFFFF03E;
    ReadData2E = 64'hBCAFFFFFF5FFFFFF;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"980117\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hC01FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"980423\n");
    end
    ReadData1E = 64'h401FFFBFFFFFFF7F;
    ReadData2E = 64'hC00C00000000007E;
    ReadData3E = 64'hBEBFFFFF0000FFFE;
    ans = 64'hC03BFFC81FFFFF0D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"980729\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBF8BFFFF80000000;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'hBF7BFFFF800000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"981035\n");
    end
    ReadData1E = 64'h3D10000000004004;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h3CA8781265FFCFE7;
    ans = 64'h3CA8781265FFCFE7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"981341\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FD6B45300124B28;
    ReadData3E = 64'h47E7FFFFF7FFFFFE;
    ans = 64'h47E7FFFFF7FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"981647\n");
    end
    ReadData1E = 64'hB7EFFFEFFDFFFFFE;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"981953\n");
    end
    ReadData1E = 64'hFFFFFFFFFFC3FFFF;
    ReadData2E = 64'h3F900000A0000000;
    ReadData3E = 64'h41E000000083FFFF;
    ans = 64'hFFFFFFFFFFC3FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"982259\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h0010000000000000;
    ReadData3E = 64'h3F8F001000000000;
    ans = 64'h3F8F001000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"982565\n");
    end
    ReadData1E = 64'h8020000000001FFC;
    ReadData2E = 64'hBC0FFFF9FFFFFFFE;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"982871\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'h4010000000000000;
    ans = 64'h4010000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"983177\n");
    end
    ReadData1E = 64'hC34626B788D750F2;
    ReadData2E = 64'h381FFFBFFFFFFDFE;
    ReadData3E = 64'h3FB00080001FFFFF;
    ans = 64'h3FB00080001FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"983483\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC01F780000000000;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'hC00F77FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"983789\n");
    end
    ReadData1E = 64'h121FFFFFFC000002;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h40300003FFFF7FFF;
    ans = 64'h40300003FFFF7FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"984095\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3F400000001FFFDE;
    ReadData3E = 64'h8029A783715CE4CD;
    ans = 64'h3F300000001FFFDD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"984401\n");
    end
    ReadData1E = 64'hFFE000040003FFFF;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'hC00000080007FFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"984707\n");
    end
    ReadData1E = 64'h380FFFFC007FFFFE;
    ReadData2E = 64'h6A9FFFC000000003;
    ReadData3E = 64'h8012C39F78579516;
    ans = 64'h62BFFFBC0087FF01;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"985013\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hBCAFDFFFFFFFFFEE;
    ans = 64'hBCA7DFFFFFFFFFEE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"985319\n");
    end
    ReadData1E = 64'h41AFFFFF8001FFFE;
    ReadData2E = 64'h2C6BBA796AEC3BE5;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"985625\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC000000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"985931\n");
    end
    ReadData1E = 64'hFFD8BB98BFC10007;
    ReadData2E = 64'h41EFFFFEFFFFF800;
    ReadData3E = 64'h37E00000000FFFF8;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"986237\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3ECFFFFFBFFFE000;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h3EBFFFFFBFFFDFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"986543\n");
    end
    ReadData1E = 64'h47E0040FFFFFFFFF;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'hB04007FFFFFFFF7F;
    ans = 64'h4490041000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"986849\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h21000001003FFFFE;
    ReadData3E = 64'h3CA5394998BC9A57;
    ans = 64'h3CA5394998BC9A57;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"987155\n");
    end
    ReadData1E = 64'h31BFFFFFFFC0003F;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hBFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"987461\n");
    end
    ReadData1E = 64'h44A367D1A38248BF;
    ReadData2E = 64'hDD3DC9E8AED9E5B2;
    ReadData3E = 64'hBD000FFFFFFFFEFF;
    ans = 64'hE1F2108733B1A4F0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"987767\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3CAFFFFFFFFFFFFE;
    ReadData3E = 64'hC1F000000080003E;
    ans = 64'hC1F000000080003E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"988073\n");
    end
    ReadData1E = 64'h9C09F2FB6E1B6AC0;
    ReadData2E = 64'hA7D718808707F875;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"988379\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h4000FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"988685\n");
    end
    ReadData1E = 64'hB80FFFFFC007FFFF;
    ReadData2E = 64'hFDFFFFFFE7FFFFFF;
    ReadData3E = 64'hC1EFFFF7FFEFFFFF;
    ans = 64'h761FFFFFA808002E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"988991\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h40100000000FFFFE;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h40000000000FFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"989297\n");
    end
    ReadData1E = 64'hC0F0E9718D9CDCE9;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hC01FFFFFFFFFFEFD;
    ans = 64'hC0D0EB718D9CDCEA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"989603\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FC01FFFFFFFFFEF;
    ReadData3E = 64'hCCC916E161AACEEE;
    ans = 64'hCCC916E161AACEEE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"989909\n");
    end
    ReadData1E = 64'hBF137F0147CBDE6D;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FDFFF6407F5C1A1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"990215\n");
    end
    ReadData1E = 64'hC1F03FFFF7FFFFFF;
    ReadData2E = 64'hC3F0001F7FFFFFFE;
    ReadData3E = 64'hC7E5B399E0BBB6AA;
    ans = 64'hC7E5B399E09B366A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"990521\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h41D200000000FFFF;
    ans = 64'h41D200000010FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"990827\n");
    end
    ReadData1E = 64'hC1B0EE7ED613B6E0;
    ReadData2E = 64'hC3F0504B43F10D03;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'h45B14376F61A6E2B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"991133\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hBFE7FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"991439\n");
    end
    ReadData1E = 64'h4068B9229919C3B2;
    ReadData2E = 64'hC1F9CD735BDB5221;
    ReadData3E = 64'h3BF8F8569B788D11;
    ans = 64'hC273EF5E2D8D3101;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"991745\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h43EFE0000FFFFFFF;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"992051\n");
    end
    ReadData1E = 64'hC3C0000000003FBF;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'hBD0FFFFFFFFFC001;
    ans = 64'hC3B0000000003FBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"992357\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFF00000001FFFFD;
    ReadData3E = 64'h43400000000FDFFF;
    ans = 64'h43400000000FDFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"992663\n");
    end
    ReadData1E = 64'h4021FFFEFFFFFFFF;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'h4010FFFF00000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"992969\n");
    end
    ReadData1E = 64'h3FD43A64D1B392BB;
    ReadData2E = 64'h40A2000000400000;
    ReadData3E = 64'h3FCFF00FFFFFFFFE;
    ans = 64'h4086C3B06D3AEEA6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"993275\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h3FB0008000002000;
    ans = 64'h3FE20010000003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"993581\n");
    end
    ReadData1E = 64'hBFEFFFFEFFFFFFDF;
    ReadData2E = 64'h47F0000001C00000;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hC7EFFFFF037FFFC3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"993887\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"994193\n");
    end
    ReadData1E = 64'h47F400000000001E;
    ReadData2E = 64'h403D60E42520C790;
    ReadData3E = 64'h5FFFFFFFBFFFFFEE;
    ans = 64'h5FFFFFFFBFFFFFEE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"994499\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h8D2FFFFFFFFF7F00;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"994805\n");
    end
    ReadData1E = 64'hBE7FFFFFF0000080;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hBFEFD6F62F0019CE;
    ans = 64'hBFEFD6F66F0019AE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"995111\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h406F7FFFFFDFFFFE;
    ReadData3E = 64'h3FF91C0E0D8CFD19;
    ans = 64'h405FE470381633F1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"995417\n");
    end
    ReadData1E = 64'hC7EFFC001FFFFFFE;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'hC7EFFC0020000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"995723\n");
    end
    ReadData1E = 64'h409FFEFBFFFFFFFF;
    ReadData2E = 64'h47E0003FFF7FFFFE;
    ReadData3E = 64'h0020000FFFFFFFFB;
    ans = 64'h488FFF7BFAF0081B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"996029\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'hB9500000002007FF;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"996335\n");
    end
    ReadData1E = 64'h0021FBFFFFFFFFFE;
    ReadData2E = 64'hD2657E7C804AB0C5;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"996641\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'h3FE7FFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"996947\n");
    end
    ReadData1E = 64'hC3E020001FFFFFFE;
    ReadData2E = 64'hC01800000007FFFF;
    ReadData3E = 64'hC1FFFFF7FFFFFFDF;
    ans = 64'h440830002FF81000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"997253\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBA4FFFFFF0003FFF;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"997559\n");
    end
    ReadData1E = 64'h3FC23813383D12F2;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'hC01000001FF00000;
    ans = 64'hC00DB8FDD8D85DA2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"997865\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hB810DF99665D6F69;
    ReadData3E = 64'h43F0000800100000;
    ans = 64'h43F0000800100000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"998171\n");
    end
    ReadData1E = 64'h3FF1A6FD72CBC7F4;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'h4001A6FD72CBC7F4;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"998477\n");
    end
    ReadData1E = 64'h40040000003FFFFF;
    ReadData2E = 64'h3FDFFFBFFFFE0000;
    ReadData3E = 64'h3FFFFFFFFC000001;
    ans = 64'h4009FFEBFE1F5FC0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"998783\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hBFCFFFFEFF000000;
    ans = 64'h3FE8000040400001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"999089\n");
    end
    ReadData1E = 64'hC34F85E5C3567603;
    ReadData2E = 64'hC1D000000080FFFF;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'h452F85E5C4549D8D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"999395\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"999701\n");
    end
    ReadData1E = 64'h37EFFFFF00000003;
    ReadData2E = 64'h5E90000001BFFFFE;
    ReadData3E = 64'h47F0000020000002;
    ans = 64'h568FFFFF037FFFE3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1000007\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h41C609CDCAAE88A3;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFF4FB191A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1000313\n");
    end
    ReadData1E = 64'hC060008000FFFFFF;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'h25C2C89884490777;
    ans = 64'hC080008000FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1000619\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC3DC007FFFFFFFFF;
    ReadData3E = 64'h00032CA7CDBDDF36;
    ans = 64'hC3CC007FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1000925\n");
    end
    ReadData1E = 64'h802FFFFFE0040000;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h804DFFFFE0040000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1001231\n");
    end
    ReadData1E = 64'hB80C07FFFFFFFFFE;
    ReadData2E = 64'hBFE800000000003F;
    ReadData3E = 64'h40241C223A4D9841;
    ans = 64'h40241C223A4D9841;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1001537\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'h40000007FFFFC000;
    ans = 64'h40100003FFFFE000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1001843\n");
    end
    ReadData1E = 64'hC000000000BFFFFE;
    ReadData2E = 64'hB7FFFFDFFFFBFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1002149\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'h4000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1002455\n");
    end
    ReadData1E = 64'hC03FFFFF7FFC0000;
    ReadData2E = 64'h380FFFFFFFFFF807;
    ReadData3E = 64'h40479C03BA46E5FF;
    ans = 64'h40479C03BA46E5FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1002761\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC0105FFFFFFFFFFF;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h4017CFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1003067\n");
    end
    ReadData1E = 64'h32A659D12B3A57E2;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hC0501FFE00000000;
    ans = 64'hC0501FFE00000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1003373\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC3E0207FFFFFFFFF;
    ReadData3E = 64'h402003FFFFFFFFDE;
    ans = 64'hC3D0207FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1003679\n");
    end
    ReadData1E = 64'h381FFFFBFFEFFFFF;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h384FFFFBFFEFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1003985\n");
    end
    ReadData1E = 64'h3B50000080000040;
    ReadData2E = 64'h43CFFFE003FFFFFF;
    ReadData3E = 64'hC1DFFFFFDFFFFEFF;
    ans = 64'hC1DFFFFFDFFFFAFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1004291\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'hC0210341F15BAF08;
    ans = 64'h432FFFFFFFFFFFEE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1004597\n");
    end
    ReadData1E = 64'hFFD00007FFFFE000;
    ReadData2E = 64'h4040010000200000;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1004903\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h4330000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1005209\n");
    end
    ReadData1E = 64'h115FFBFFFFFC0000;
    ReadData2E = 64'h3F3FFFFFBFFF8000;
    ReadData3E = 64'hFFDBFFFFEFFFFFFF;
    ans = 64'hFFDBFFFFEFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1005515\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC035395F6673F88F;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC02D395F6673F88E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1005821\n");
    end
    ReadData1E = 64'h002FFC00FFFFFFFF;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'h00198167B6DD2E02;
    ans = 64'h038FFC00FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1006127\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFFE0B95ECD9ACF7;
    ReadData3E = 64'h3FD0000080000007;
    ans = 64'hBFE60B95ACD9ACF3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1006433\n");
    end
    ReadData1E = 64'h4806C87C6E9AD566;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1006739\n");
    end
    ReadData1E = 64'hC019E1FC6C077592;
    ReadData2E = 64'h3CAC9006ED57E4C3;
    ReadData3E = 64'h40100007FFFFBFFE;
    ans = 64'h40100007FFFFBFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1007045\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'hBFEFFFFFEFFEFFFF;
    ans = 64'h433FFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1007351\n");
    end
    ReadData1E = 64'h3D2666AE90703854;
    ReadData2E = 64'hBC30FFFFFFFF7FFE;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h3FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1007657\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'h8000000000000001;
    ans = 64'h7FCFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1007963\n");
    end
    ReadData1E = 64'h7FF78D4C8A6F6B33;
    ReadData2E = 64'h1EEF6DBA3B66A8C5;
    ReadData3E = 64'hC12FFDFBFFFFFFFF;
    ans = 64'h7FFF8D4C8A6F6B33;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1008269\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h4340000010000020;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h4330000010000021;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1008575\n");
    end
    ReadData1E = 64'h3FC07FFF7FFFFFFE;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'hC3E631875CDB7E45;
    ans = 64'h7FB07FFF7FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1008881\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBF2000001FFBFFFE;
    ReadData3E = 64'h43E0000012000000;
    ans = 64'h43E0000012000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1009187\n");
    end
    ReadData1E = 64'h3FD45A9AEEFFF3C6;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1009493\n");
    end
    ReadData1E = 64'hBED003FFC0000000;
    ReadData2E = 64'hC00FFDFFFFFEFFFE;
    ReadData3E = 64'hBF4FFFFFFBFFFFF7;
    ans = 64'hBF4F7FE7FFFFE3F8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1009799\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'h3FB7FFE000000000;
    ans = 64'h7FDFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1010105\n");
    end
    ReadData1E = 64'h3F3FFFFFFDFFFFFE;
    ReadData2E = 64'hB81FFF0000FFFFFE;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hBCA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1010411\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1010717\n");
    end
    ReadData1E = 64'h8FFC99FE1EB38741;
    ReadData2E = 64'hC1EFFEBFFFFFFFFF;
    ReadData3E = 64'h43C000200003FFFE;
    ans = 64'h43C000200003FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1011023\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFEFFFFFC2000000;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hC003FFFFF83FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1011329\n");
    end
    ReadData1E = 64'hC070000000FFBFFF;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'hC1C6E2D6423AD7C0;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1011635\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h4060020007FFFFFF;
    ReadData3E = 64'hBFCFEFBFFFFFFFFF;
    ans = 64'h404FE4104FFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1011941\n");
    end
    ReadData1E = 64'h80222D75B3139904;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1012247\n");
    end
    ReadData1E = 64'hC7F478EFD91BB2AE;
    ReadData2E = 64'h24B0000000080006;
    ReadData3E = 64'h7FDFFFFFFFFDFFEE;
    ans = 64'h7FDFFFFFFFFDFFEE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1012553\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h2B8FFFFFFFFF83FE;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1012859\n");
    end
    ReadData1E = 64'h7FDFFFFFFFFF5FFF;
    ReadData2E = 64'h401FFF8000000001;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1013165\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1013471\n");
    end
    ReadData1E = 64'hBFC92B62A147BA62;
    ReadData2E = 64'hC10FF800000000FF;
    ReadData3E = 64'h3FCFFDFFFFFC0000;
    ans = 64'h40E9251FC81F693B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1013777\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h434FFFEFFFFBFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h433FFFEFFFFBFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1014083\n");
    end
    ReadData1E = 64'hBF7FEFFFEFFFFFFE;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'hC0EFFFFFFFF9FFFF;
    ans = 64'hC0EFFFFFFFF9FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1014389\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFE581C023E170BC;
    ReadData3E = 64'hFFF000003FBFFFFF;
    ans = 64'hFFF800003FBFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1014695\n");
    end
    ReadData1E = 64'hC00DC00000000000;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1015001\n");
    end
    ReadData1E = 64'h4800000080000000;
    ReadData2E = 64'h381C5A57C936F7B7;
    ReadData3E = 64'h4808000000000FFE;
    ans = 64'h4808000000000FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1015307\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'h3CA0000FFFF7FFFF;
    ans = 64'h3CA0000FFFF7FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1015613\n");
    end
    ReadData1E = 64'hBFEFFFFFBFFFDFFF;
    ReadData2E = 64'hBFEFFF000FFFFFFF;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'h3FEFFEFFD001DFDF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1015919\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1016225\n");
    end
    ReadData1E = 64'h3940000014000000;
    ReadData2E = 64'h001FFFBFFFFFFFFF;
    ReadData3E = 64'hB95FFFFFC003FFFF;
    ans = 64'hB95FFFFFC003FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1016531\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h8ADFFFFFF000000F;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hBFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1016837\n");
    end
    ReadData1E = 64'h406DF9FFE3D019C9;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'h41C0000000007EFF;
    ans = 64'h41C0000000007EFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1017143\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h43CFFFFFFFFFFBFF;
    ReadData3E = 64'h4000040001FFFFFE;
    ans = 64'h43BFFFFFFFFFFBFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1017449\n");
    end
    ReadData1E = 64'hCE8FFFFFFFFFFFE8;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1017755\n");
    end
    ReadData1E = 64'h0960000800100000;
    ReadData2E = 64'h403FFC0001000000;
    ReadData3E = 64'h41D7FFFFFFFFFFC0;
    ans = 64'h41D7FFFFFFFFFFC0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1018061\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'hC1FFFFFFE0003FFF;
    ans = 64'hC1FFFFFFE0003FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1018367\n");
    end
    ReadData1E = 64'h410FFFF7FDFFFFFE;
    ReadData2E = 64'h480BFFFFC0000000;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h492BFFF8BE401002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1018673\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1018979\n");
    end
    ReadData1E = 64'h43EFDF7FFFFFFFFF;
    ReadData2E = 64'hB7E0001007FFFFFF;
    ReadData3E = 64'hB4C000201FFFFFFF;
    ans = 64'hBBDFDF9FEF6FBFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1019285\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC1DFFFFFFFFEFDFE;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'hC1CFFFFFFFDEFDFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1019591\n");
    end
    ReadData1E = 64'hC080000000201000;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'h41CFFDFFFFFFFFFE;
    ans = 64'h41CFFDFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1019897\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hFFF0100000000002;
    ReadData3E = 64'hBF9FDFFFEFFFFFFF;
    ans = 64'hFFF8100000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1020203\n");
    end
    ReadData1E = 64'h480FFFFFE00003FF;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'hC4BFFFFFE00003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1020509\n");
    end
    ReadData1E = 64'h43DD2781CF9BF717;
    ReadData2E = 64'hBFB1000000000006;
    ReadData3E = 64'hB7E0000000001FDF;
    ans = 64'hC39EF9F9EC95B693;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1020815\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h37E0041FFFFFFFFF;
    ans = 64'hBC90000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1021121\n");
    end
    ReadData1E = 64'h4AA000000003FFF0;
    ReadData2E = 64'hBFFFFFFFFFEFFC00;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1021427\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1021733\n");
    end
    ReadData1E = 64'h4010000008000001;
    ReadData2E = 64'h3FD00FFF00000000;
    ReadData3E = 64'hBFC000400000007E;
    ans = 64'h3FEC1FEE100FFEE3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1022039\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h472FC0003FFFFFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'h471FC0003FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1022345\n");
    end
    ReadData1E = 64'hC80FFFFFBFFFFFDF;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h798FFF0040000000;
    ans = 64'h798FFF0040000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1022651\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hB400E44229C276B0;
    ReadData3E = 64'h403020001FFFFFFF;
    ans = 64'h403020001FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1022957\n");
    end
    ReadData1E = 64'hB48784AE13E55D35;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1023263\n");
    end
    ReadData1E = 64'h4340000001FFFFFE;
    ReadData2E = 64'hC1F07FFFFFFF7FFE;
    ReadData3E = 64'hC3F1FFFFFFFEFFFF;
    ans = 64'hC5408000920F7FFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1023569\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFD0000000000000;
    ReadData3E = 64'h42B0000002100000;
    ans = 64'h42B00000020FFFE0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1023875\n");
    end
    ReadData1E = 64'hCE71FFFFFFFFFFFB;
    ReadData2E = 64'h9B2FFFFFFFFFFF6E;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1024181\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'h400EFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1024487\n");
    end
    ReadData1E = 64'hC194000000001000;
    ReadData2E = 64'h43B496260249CFD3;
    ReadData3E = 64'h400FEFFFFDFFFFFE;
    ans = 64'hC559BBAF82DC585E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1024793\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h5C6B01BA6B0ECB17;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h5C5B01BA6B0ECB16;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1025099\n");
    end
    ReadData1E = 64'hBFF0000000001002;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'h30A0000FFFFFFF7F;
    ans = 64'h3FE0000000001001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1025405\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h4080000000FFFFC0;
    ReadData3E = 64'h37EFE0007FFFFFFF;
    ans = 64'h4070000000FFFFBF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1025711\n");
    end
    ReadData1E = 64'hBFC4195BFCEF85E3;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FF14195BFCEF85F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1026017\n");
    end
    ReadData1E = 64'hFFDFFFFFFFFFF7DE;
    ReadData2E = 64'hC3CFFFFFFFBFF7FE;
    ReadData3E = 64'hC03FFFFF003FFFFE;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1026323\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'h434D28185A46B2EB;
    ans = 64'h434D28185A46B2EB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1026629\n");
    end
    ReadData1E = 64'hB8DFFFFFFFF00010;
    ReadData2E = 64'h3C8FEFDFFFFFFFFE;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h434FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1026935\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC002000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1027241\n");
    end
    ReadData1E = 64'h8D1CB6E31C42D6D0;
    ReadData2E = 64'h4026B2B4559344C6;
    ReadData3E = 64'hC7EFFCFFFFFFFFFF;
    ans = 64'hC7EFFCFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1027547\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h3FDFFFF801FFFFFF;
    ReadData3E = 64'h8010000000000000;
    ans = 64'h3FCFFFF801FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1027853\n");
    end
    ReadData1E = 64'hC028000001FFFFFF;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h41EFC00020000000;
    ans = 64'h41EFC00020C00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1028159\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC3EFFFFFFBFFFFFF;
    ReadData3E = 64'hC3D8697088B915AB;
    ans = 64'hC3EC34B8425C8AD5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1028465\n");
    end
    ReadData1E = 64'h407C66F7D3FB5A4F;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hC07C76F7D3FB5A4E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1028771\n");
    end
    ReadData1E = 64'hBFC937040830C79E;
    ReadData2E = 64'hBFBFFE0100000000;
    ReadData3E = 64'hBE3FFFFF7FFF7FFF;
    ans = 64'h3F993570E1A866D3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1029077\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFEFFFFFFFFFFFFE;
    ReadData3E = 64'hBF4FFFFFFFFC03FF;
    ans = 64'hBFE007FFFFFFFEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1029383\n");
    end
    ReadData1E = 64'h416FFFFEFFF7FFFE;
    ReadData2E = 64'h3FE0000008000002;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hC33FFFFFFF800004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1029689\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'h3FF0000000000001;
    ans = 64'h3FE0000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1029995\n");
    end
    ReadData1E = 64'h7FFFFFF0FFFFFFFF;
    ReadData2E = 64'hBF6000000FFFFFFE;
    ReadData3E = 64'h3E9F8F47F01A6A7B;
    ans = 64'h7FFFFFF0FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1030301\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h380DACDF6E0D7A8D;
    ReadData3E = 64'h0000000000000001;
    ans = 64'h37FDACDF6E0D7A8C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1030607\n");
    end
    ReadData1E = 64'h401FE148734D4B74;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h281FFFF00000000E;
    ans = 64'hC01FE148734D4B76;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1030913\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h41E0000FFFFFFFBF;
    ReadData3E = 64'hFFD88149181B8BAA;
    ans = 64'hFFD88149181B8BAA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1031219\n");
    end
    ReadData1E = 64'hC08250FEC0C6C7E6;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'h409252FEC0C6C7E5;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1031525\n");
    end
    ReadData1E = 64'h432EE7DCB359DC6B;
    ReadData2E = 64'h40B074A36C4386CA;
    ReadData3E = 64'hBFFFFFE0007FFFFE;
    ans = 64'h43EFC9295EA60C0F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1031831\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h400FFEFFFFFEFFFE;
    ans = 64'h4007FEFFFFFEFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1032137\n");
    end
    ReadData1E = 64'hC343B1E8463B2B26;
    ReadData2E = 64'h7FF007FFFFFF7FFF;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h7FF807FFFFFF7FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1032443\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'hBFEFFFFFFFFFFFFF;
    ans = 64'hBFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1032749\n");
    end
    ReadData1E = 64'h40190979C4E0CA3F;
    ReadData2E = 64'hBFE000001001FFFE;
    ReadData3E = 64'hBCBBFFFFFFFFFFFC;
    ans = 64'hC0090979DDED6531;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1033055\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h825FFFFFFFFFFFFE;
    ReadData3E = 64'h7FFFFFFFFFFFFFFF;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1033361\n");
    end
    ReadData1E = 64'hC7F07F7FFFFFFFFF;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h41EFFFFFEFFFFFF8;
    ans = 64'h48007F7FFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1033667\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC0A000000002001E;
    ReadData3E = 64'hFFFC4C6E938BCF63;
    ans = 64'hFFFC4C6E938BCF63;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1033973\n");
    end
    ReadData1E = 64'h3EF4000000001FFF;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hBFD000A000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1034279\n");
    end
    ReadData1E = 64'h001FFF01FFFFFFFF;
    ReadData2E = 64'h7FF020000001FFFF;
    ReadData3E = 64'hC00FFFFFFF002000;
    ans = 64'h7FF820000001FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1034585\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h41FFFF7FFFFFFFFD;
    ans = 64'h41FFFF7FFFDFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1034891\n");
    end
    ReadData1E = 64'h40600000002007FF;
    ReadData2E = 64'h404FFFFFFFFEFFBF;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'h40BFFC00003F0FBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1035197\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'h3FDFFFFFFFFFFFFE;
    ans = 64'hBFF7FFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1035503\n");
    end
    ReadData1E = 64'h7FEFFFFFFFFFF01F;
    ReadData2E = 64'hBFC00000003FFFBF;
    ReadData3E = 64'h41684E7E00A71AE6;
    ans = 64'hFFC00000003FF7CE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1035809\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hCDC6B77D8EB7FA00;
    ReadData3E = 64'hFFEFFFFFFFFFFFFE;
    ans = 64'hFFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1036115\n");
    end
    ReadData1E = 64'hBAB0000FFFFFFFEF;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h939FF7FFFFFE0000;
    ans = 64'h3AD0000FFFFFFFEF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1036421\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hFFF0001FFFFF8000;
    ReadData3E = 64'h402FFF7000000000;
    ans = 64'hFFF8001FFFFF8000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1036727\n");
    end
    ReadData1E = 64'h380FFFFBFFFFFFFF;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'h3CA0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1037033\n");
    end
    ReadData1E = 64'h801DFFFFFFFDFFFF;
    ReadData2E = 64'hC0800001FFFFFFC0;
    ReadData3E = 64'hBFD00101FFFFFFFE;
    ans = 64'hBFD00101FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1037339\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'hCA404000000007FE;
    ans = 64'hCA404000000007FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1037645\n");
    end
    ReadData1E = 64'h801FFFD800000000;
    ReadData2E = 64'h3FCFFFFFFDFFFFFD;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'h3FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1037951\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'hBFD0000000000000;
    ans = 64'hC010FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1038257\n");
    end
    ReadData1E = 64'h9DE0000007FEFFFF;
    ReadData2E = 64'hC3E001FFFFFF8000;
    ReadData3E = 64'h400FFFFFAFFFFFFE;
    ans = 64'h400FFFFFAFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1038563\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h37F0000017FFFFFF;
    ReadData3E = 64'h7FE0000000000000;
    ans = 64'h7FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1038869\n");
    end
    ReadData1E = 64'h3FB1FFFFFFFFEFFE;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'h40CF7FFFC0000000;
    ans = 64'h40CF7FB7C0000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1039175\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hFFDE8D5A73D26464;
    ReadData3E = 64'h37F9F76C367CE95E;
    ans = 64'hFFCE8D5A73D26463;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1039481\n");
    end
    ReadData1E = 64'h3E0FA37875781127;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'hC15FA37875781127;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1039787\n");
    end
    ReadData1E = 64'hB81F7FFFFFFFDFFF;
    ReadData2E = 64'hC800000100000000;
    ReadData3E = 64'h6CB1FFFFFFBFFFFE;
    ans = 64'h6CB1FFFFFFBFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1040093\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h3C50003FFFFFC000;
    ans = 64'hC330000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1040399\n");
    end
    ReadData1E = 64'hC3E000000FFFFFFC;
    ReadData2E = 64'hBF8FEFFFFFFFEFFF;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'h437FF0001FEFEFF7;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1040705\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'h3CA0000000000001;
    ans = 64'hC33FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1041011\n");
    end
    ReadData1E = 64'h3FD00FFFFC000000;
    ReadData2E = 64'hE76F7B616143C433;
    ReadData3E = 64'h47F0004000080000;
    ans = 64'hE74F9ADCBAC62F9F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1041317\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h33E000FFFFFFFDFF;
    ReadData3E = 64'hC340000000000001;
    ans = 64'hC340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1041623\n");
    end
    ReadData1E = 64'hFFFFFFEFFFFF7FFF;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'hC03001FFFFFFFFFE;
    ans = 64'hFFFFFFEFFFFF7FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1041929\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBACFFBFDFFFFFFFF;
    ReadData3E = 64'hC00007FFFFF7FFFE;
    ans = 64'hC00007FFFFF7FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1042235\n");
    end
    ReadData1E = 64'h4060803FFFFFFFFF;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1042541\n");
    end
    ReadData1E = 64'h3FF040000000000F;
    ReadData2E = 64'hC1FFFFFF7FFFFFF0;
    ReadData3E = 64'h570000007FC00000;
    ans = 64'h570000007FC00000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1042847\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'hC7EFFFFFFEFFFDFF;
    ans = 64'hFFCFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1043153\n");
    end
    ReadData1E = 64'hBFEFEFFFBFFFFFFE;
    ReadData2E = 64'hA15FF00000000100;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1043459\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'h801FFFFFFFFFFFFF;
    ans = 64'hFFD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1043765\n");
    end
    ReadData1E = 64'hBDE0A431843E6962;
    ReadData2E = 64'hC007BF7F4F6FE319;
    ReadData3E = 64'h7FD0000000002200;
    ans = 64'h7FD0000000002200;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1044071\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC7E311181221223D;
    ReadData3E = 64'h401FFFFFFFFFFFFF;
    ans = 64'hC7D311181221223C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1044377\n");
    end
    ReadData1E = 64'h60CC008000000000;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'hBFDAB06782893A16;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1044683\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hC00FFFE000000002;
    ReadData3E = 64'hBFDFFFFFFFFFF00F;
    ans = 64'hC003FFEFFFFFFE02;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1044989\n");
    end
    ReadData1E = 64'h4040000000101FFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'h8000000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1045295\n");
    end
    ReadData1E = 64'h41C0EE6217580667;
    ReadData2E = 64'hFFEFFFFFF7FFFFFC;
    ReadData3E = 64'h3FF2E298683C3F2D;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1045601\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h450FFFFFFFFFFFFF;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1045907\n");
    end
    ReadData1E = 64'h380FFDFFFFFDFFFF;
    ReadData2E = 64'hC3F0FFFFFFFFFEFE;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hBFDFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1046213\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'h000FFFFFFFFFFFFE;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1046519\n");
    end
    ReadData1E = 64'hBFDFFEFFEFFFFFFE;
    ReadData2E = 64'hBFDCAB6630300A64;
    ReadData3E = 64'hC1DFFBFFFFFFE000;
    ans = 64'hC1DFFBFFFFF18AC0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1046825\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hBFFFFF3FFFFFFFFF;
    ReadData3E = 64'hC00FFFFFFFFFFFFE;
    ans = 64'hC013FFE7FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1047131\n");
    end
    ReadData1E = 64'hFFD0080000007FFE;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h480ECB8DF4D61707;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1047437\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'h002FFFFFFFFFFFFF;
    ReadData3E = 64'h4944C0B502DF4D9E;
    ans = 64'h4944C0B502DF4D9E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1047743\n");
    end
    ReadData1E = 64'hBCBFC01000000000;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1048049\n");
    end
    ReadData1E = 64'hC2AFFFFFFFFFFFFF;
    ReadData2E = 64'hBDFBFFFFFFFFFF7F;
    ReadData3E = 64'hBF83698EAB72255D;
    ans = 64'h40BBFFFD92CE2A10;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1048355\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFF;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h801A50D4FC9D3CC9;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1048661\n");
    end
    ReadData1E = 64'hBF94D8272D1C6FD0;
    ReadData2E = 64'hBF8FFFFFFFFFF0FF;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3F34D8272D1C760A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1048967\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'h8000000000000000;
    ans = 64'h0000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1049273\n");
    end
    ReadData1E = 64'h3F00000000100010;
    ReadData2E = 64'h5E5BFFFFFFFFEFFF;
    ReadData3E = 64'hBFDFF00007FFFFFE;
    ans = 64'h5D6C0000001BF01B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1049579\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFC0000009FFFFFF;
    ReadData3E = 64'h4000000000000000;
    ans = 64'h3FFEFFFFFF600000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1049885\n");
    end
    ReadData1E = 64'hB7FDFBFFFFFFFFFE;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h7FD000000000203F;
    ans = 64'h7FD000000000203F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1050191\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h434FFFFFFFE0001F;
    ReadData3E = 64'hBFFFFEFFFF7FFFFF;
    ans = 64'h433FFFFFFFE0001B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1050497\n");
    end
    ReadData1E = 64'hC0C00000400001FF;
    ReadData2E = 64'h000FFFFFFFFFFFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1050803\n");
    end
    ReadData1E = 64'h4BC0040000003FFF;
    ReadData2E = 64'h41DFE0000000FFFF;
    ReadData3E = 64'h3FC0000FFFFFFFFF;
    ans = 64'h4DAFE7F800017FBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1051109\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'h7FDFFFFFF000001E;
    ans = 64'h7FDFFFFFF000001E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1051415\n");
    end
    ReadData1E = 64'h3FC0000001000010;
    ReadData2E = 64'hBFF5BD425B6267A3;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBFC5BD425CBE3BE2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1051721\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h000FFFFFFFFFFFFE;
    ReadData3E = 64'hFFF0000000000001;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1052027\n");
    end
    ReadData1E = 64'h52FFFFF801FFFFFF;
    ReadData2E = 64'h3D30000000007FFF;
    ReadData3E = 64'hBFE1B8A4DD23031A;
    ans = 64'h503FFFF80200FFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1052333\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFE14F5BF1948F88;
    ReadData3E = 64'hBFF0000000000001;
    ans = 64'hBFF453D6FC6523E3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1052639\n");
    end
    ReadData1E = 64'h40096DFEA6B9CF4D;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'hC3C2000000000040;
    ans = 64'hC3C2000000000040;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1052945\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3F00000020080000;
    ReadData3E = 64'h000BF714D245ED69;
    ans = 64'h3EF000002007FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1053251\n");
    end
    ReadData1E = 64'hC93FFFFFFBFFEFFF;
    ReadData2E = 64'h0010000000000001;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1053557\n");
    end
    ReadData1E = 64'h5ED64804F957BD63;
    ReadData2E = 64'h7FEA233C26339EFB;
    ReadData3E = 64'hC7E79C84C2D4C3CC;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1053863\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h001FFFFFFFFFFFFF;
    ReadData3E = 64'h353FDF8000000000;
    ans = 64'h353FDF8000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1054169\n");
    end
    ReadData1E = 64'hC1EFFFFFFBF80000;
    ReadData2E = 64'h4050000000000000;
    ReadData3E = 64'h0010000000000001;
    ans = 64'hC24FFFFFFBF80000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1054475\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h001FFFFFFFFFFFFE;
    ReadData3E = 64'h7FEFFFFFFFFFFFFF;
    ans = 64'h7FEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1054781\n");
    end
    ReadData1E = 64'h41D4C6B47A165676;
    ReadData2E = 64'hBEEFFFFFFFFFFEFE;
    ReadData3E = 64'hC027ABDD993EECCF;
    ans = 64'hC0D4C9A9F5C97DAC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1055087\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFFAB60866FB2019;
    ReadData3E = 64'h3FEFFFFFFFFFFFFF;
    ans = 64'h3FC527DE64137F9F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1055393\n");
    end
    ReadData1E = 64'h3FDEFFF7FFFFFFFE;
    ReadData2E = 64'h3CA0000000000000;
    ReadData3E = 64'h3FBFF7FFFFFC0000;
    ans = 64'h3FBFF7FFFFFC0004;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1055699\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h38103FFFFFFFFFDF;
    ReadData3E = 64'hC800020000000000;
    ans = 64'hC800020000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1056005\n");
    end
    ReadData1E = 64'h3FB000000010007F;
    ReadData2E = 64'h3CA0000000000001;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1056311\n");
    end
    ReadData1E = 64'hC06FFFFFFEF00000;
    ReadData2E = 64'hBCA3F00000000000;
    ReadData3E = 64'hC1D659022B8CF3A6;
    ans = 64'hC1D659022B8CF3A6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1056617\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hFFE80000000000FF;
    ans = 64'hFFE80000000000FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1056923\n");
    end
    ReadData1E = 64'hBE5FFFFE0000007E;
    ReadData2E = 64'hBFFFF8000007FFFE;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'h3E6FF7FE0088007B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1057229\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3CAFFFFFFFFFFFFF;
    ReadData3E = 64'hC34FFFFFFFFFFFFE;
    ans = 64'hC34FFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1057535\n");
    end
    ReadData1E = 64'hBFFFFE000000003F;
    ReadData2E = 64'hE5CC78B61B1FB3DF;
    ReadData3E = 64'h7FEFFFFE00000001;
    ans = 64'h7FEFFFFE00000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1057841\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hB52FFFFFFF0FFFFF;
    ReadData3E = 64'hBFDFFFFFFFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1058147\n");
    end
    ReadData1E = 64'hC3DFFFFFA0000000;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'h40B00003FFFFFFE0;
    ans = 64'hC3BFFFFF9FFFFFF0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1058453\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC030D5C41923E3C4;
    ReadData3E = 64'h3FE080FFFFFFFFFF;
    ans = 64'hC01F9B683247C786;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1058759\n");
    end
    ReadData1E = 64'hC18FFFBFE0000000;
    ReadData2E = 64'h3FD0000000000000;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC16FFFC060000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1059065\n");
    end
    ReadData1E = 64'hC05F800007FFFFFE;
    ReadData2E = 64'hBCB0000002200000;
    ReadData3E = 64'h7FD000FFFFFFFFFF;
    ans = 64'h7FD000FFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1059371\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3FD0000000000001;
    ReadData3E = 64'hC3CFFFFFEFFFFFFC;
    ans = 64'hC3CFFFFFEFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1059677\n");
    end
    ReadData1E = 64'hC34FFDFFFFFFF000;
    ReadData2E = 64'hC0621FFFFFFFFFFF;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1059983\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3FDFFFFFFFFFFFFF;
    ReadData3E = 64'h4340000000000000;
    ans = 64'h4340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1060289\n");
    end
    ReadData1E = 64'h3B20000207FFFFFF;
    ReadData2E = 64'h802000800000001F;
    ReadData3E = 64'h3E804000000000FE;
    ans = 64'h3E804000000000FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1060595\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h43403FFFFFFFDFFE;
    ReadData3E = 64'h3FD0000000000000;
    ans = 64'h43303FFFFFFFDFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1060901\n");
    end
    ReadData1E = 64'hBBAFFFFC7FFFFFFE;
    ReadData2E = 64'h3FDFFFFFFFFFFFFE;
    ReadData3E = 64'h7FFF800007FFFFFE;
    ans = 64'h7FFF800007FFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1061207\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h37BFE0003FFFFFFF;
    ReadData3E = 64'h3F700000200007FE;
    ans = 64'h3F700000200007FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1061513\n");
    end
    ReadData1E = 64'hBFCFFFEDFFFFFFFF;
    ReadData2E = 64'h3FE0000000000000;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h400F00008FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1061819\n");
    end
    ReadData1E = 64'h3FA000001FFFFFE0;
    ReadData2E = 64'hBEA0020000001FFF;
    ReadData3E = 64'h40EFFFFFFFFFFFE7;
    ans = 64'h40EFFFFFFFFFF7E6;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1062125\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'hC02FFFFF000007FE;
    ans = 64'hC02F7FFF000007FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1062431\n");
    end
    ReadData1E = 64'hC03BE7FA8F940E1A;
    ReadData2E = 64'hC7FFBFFFFFFFDFFF;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1062737\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3FE0000000000001;
    ReadData3E = 64'hC010000000000001;
    ans = 64'hC00E000000000002;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1063043\n");
    end
    ReadData1E = 64'h7FF6F884134124A3;
    ReadData2E = 64'h1124345EB2F9D7C2;
    ReadData3E = 64'h3FF093DCB192A945;
    ans = 64'h7FFEF884134124A3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1063349\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC44FFFEFFFFBFFFE;
    ReadData3E = 64'hBCA0000000000001;
    ans = 64'hC43FFFEFFFFBFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1063655\n");
    end
    ReadData1E = 64'h43F000FFFF7FFFFF;
    ReadData2E = 64'h3FEFFFFFFFFFFFFF;
    ReadData3E = 64'h3FC08000007FFFFF;
    ans = 64'h43F000FFFF7FFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1063961\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3D7FFFC001FFFFFE;
    ReadData3E = 64'h7FE2742ED7A37446;
    ans = 64'h7FE2742ED7A37446;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1064267\n");
    end
    ReadData1E = 64'hC02003FFE0000000;
    ReadData2E = 64'h3FEFFFFFFFFFFFFE;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hC02403FFDFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1064573\n");
    end
    ReadData1E = 64'h432A1E70319EA668;
    ReadData2E = 64'hC3C00000000000EF;
    ReadData3E = 64'hFFDFFFFBFFFFFFFC;
    ans = 64'hFFDFFFFBFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1064879\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3FF0000000000000;
    ReadData3E = 64'hC08EAC5CED022B82;
    ans = 64'hC08EA85CED022B82;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1065185\n");
    end
    ReadData1E = 64'hBFC01AA92E0FD347;
    ReadData2E = 64'h43DBFFDFFFFFFFFF;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1065491\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3FF0000000000001;
    ReadData3E = 64'h400FFFFFFFFFFFFF;
    ans = 64'h4011FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1065797\n");
    end
    ReadData1E = 64'hC0C08000000003FF;
    ReadData2E = 64'hC1DFFFF3FFFFFFFE;
    ReadData3E = 64'h41F9C64E718456C2;
    ans = 64'h42B0819634E71C43;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1066103\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h4465F3D65FF408DD;
    ReadData3E = 64'h001FFFFFFFFFFFFF;
    ans = 64'h4455F3D65FF408DC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1066409\n");
    end
    ReadData1E = 64'hB58FEFFC00000000;
    ReadData2E = 64'h3FFFFFFFFFFFFFFF;
    ReadData3E = 64'hBE600FFFF8000000;
    ans = 64'hBE600FFFF8000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1066715\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h403000000010000F;
    ReadData3E = 64'h7FD0001FDFFFFFFF;
    ans = 64'h7FD0001FDFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1067021\n");
    end
    ReadData1E = 64'hB810002000000000;
    ReadData2E = 64'h3FFFFFFFFFFFFFFE;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h3FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1067327\n");
    end
    ReadData1E = 64'h443FFFF800020000;
    ReadData2E = 64'h232BFFFFFFFFF7FF;
    ReadData3E = 64'hF7EFFFFFFF87FFFF;
    ans = 64'hF7EFFFFFFF87FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1067633\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'hC01806C8304C3CA3;
    ans = 64'hC01406C8304C3CA3;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1067939\n");
    end
    ReadData1E = 64'h40200007FFFFFFBF;
    ReadData2E = 64'h43FFFFFFFFFFFF5F;
    ReadData3E = 64'h434FFFFFFFFFFFFF;
    ans = 64'h44300047FFFFFF6E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1068245\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h4000000000000000;
    ReadData3E = 64'hBFFFFFFFFFFFFFFE;
    ans = 64'hBFEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1068551\n");
    end
    ReadData1E = 64'h417000000001001E;
    ReadData2E = 64'hB4F1BB57032E8670;
    ReadData3E = 64'hBFEFFFFFFFFF9FFF;
    ans = 64'hBFEFFFFFFFFF9FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1068857\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC8048B8D53C7F5BB;
    ReadData3E = 64'h800FFFFFFFFFFFFE;
    ans = 64'hC7F48B8D53C7F5BA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1069163\n");
    end
    ReadData1E = 64'hC01FFFFFFFFEFFF7;
    ReadData2E = 64'h4000000000000001;
    ReadData3E = 64'hBE4E000FFFFFFFFF;
    ans = 64'hC0300000003B801C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1069469\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h7FE1000000007FFF;
    ReadData3E = 64'hC01F8000000001FE;
    ans = 64'h7FD1000000007FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1069775\n");
    end
    ReadData1E = 64'h41EFFFFFFEFFFFBE;
    ReadData2E = 64'h400FFFFFFFFFFFFF;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'h420FFFFFFEFBFFBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1070081\n");
    end
    ReadData1E = 64'hC18000000F000000;
    ReadData2E = 64'h488000000000FFFE;
    ReadData3E = 64'h41C010003FFFFFFE;
    ans = 64'hCA1000000F00FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1070387\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h400FFFFFFFFFFFFE;
    ReadData3E = 64'hB7E9007C3CE5D513;
    ans = 64'h3FFFFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1070693\n");
    end
    ReadData1E = 64'h002FFFFFFFFFF006;
    ReadData2E = 64'hFFF8A6F40A830B75;
    ReadData3E = 64'hC01FFFFFFFFFFFFE;
    ans = 64'hFFF8A6F40A830B75;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1070999\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h4010000000000000;
    ReadData3E = 64'h3FF0000000000000;
    ans = 64'h4007FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1071305\n");
    end
    ReadData1E = 64'hC03000100FFFFFFF;
    ReadData2E = 64'h381FFFC00000003F;
    ReadData3E = 64'hBCA0200000004000;
    ans = 64'hBCA0200000004000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1071611\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h710FF1FFFFFFFFFE;
    ReadData3E = 64'h0000000000000000;
    ans = 64'h70FFF1FFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1071917\n");
    end
    ReadData1E = 64'h47F0FFF7FFFFFFFE;
    ReadData2E = 64'h4010000000000001;
    ReadData3E = 64'hB7E0000000000060;
    ans = 64'h4810FFF7FFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1072223\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC3D0000001FFDFFF;
    ReadData3E = 64'h400FFBFFFF7FFFFF;
    ans = 64'hC3C0000001FFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1072529\n");
    end
    ReadData1E = 64'hC4B53D377B80B5C0;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'hC4E53D377B80B5BF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1072835\n");
    end
    ReadData1E = 64'hBFAB3C37F3B44C72;
    ReadData2E = 64'h3801000004000000;
    ReadData3E = 64'hC00FFFFF6FFFFFFE;
    ans = 64'hC00FFFFF6FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1073141\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h401FFFFFFFFFFFFF;
    ReadData3E = 64'hBE3F0000000FFFFF;
    ans = 64'h400FFFFFFF07FFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1073447\n");
    end
    ReadData1E = 64'hC39FFFFFFBFFFDFF;
    ReadData2E = 64'h4770000002080000;
    ReadData3E = 64'h4010000000000000;
    ans = 64'hCB2000000007FEFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1073753\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h401FFFFFFFFFFFFE;
    ReadData3E = 64'hBFE0000000000001;
    ans = 64'h400BFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1074059\n");
    end
    ReadData1E = 64'h419000001FFFFEFF;
    ReadData2E = 64'h471FFC95BD9A3D25;
    ReadData3E = 64'h37EF600000000000;
    ans = 64'h48BFFC95FD93669E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1074365\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFD000009FFFFFFF;
    ReadData3E = 64'h7FF0000000000001;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1074671\n");
    end
    ReadData1E = 64'h47EFE1FFFFFFFFFE;
    ReadData2E = 64'h4340000000000000;
    ReadData3E = 64'hC00FFFFFFF03FFFE;
    ans = 64'h4B3FE1FFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1074977\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hB77565980E71074C;
    ReadData3E = 64'hC07AEC57F394DD51;
    ans = 64'hC07AEC57F394DD51;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1075283\n");
    end
    ReadData1E = 64'h3FB7EA8EA0DB9F94;
    ReadData2E = 64'h4340000000000001;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'h4307EA8EA0DB9F95;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1075589\n");
    end
    ReadData1E = 64'h400FFFFFFE007FFE;
    ReadData2E = 64'h711400000FFFFFFF;
    ReadData3E = 64'h43E57BE412E34D74;
    ans = 64'h713400000EC04FFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1075895\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h434FFFFFFFFFFFFF;
    ReadData3E = 64'hC1CEFFFEFFFFFFFF;
    ans = 64'h433FFFFFC20001FD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1076201\n");
    end
    ReadData1E = 64'h3FFFFFFFFFFFBFFC;
    ReadData2E = 64'hC3C60616C794210B;
    ReadData3E = 64'hC000000000000001;
    ans = 64'hC3D60616C793F4FC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1076507\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h434FFFFFFFFFFFFE;
    ReadData3E = 64'h3FDFFFFFFFFFFFFF;
    ans = 64'h433FFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1076813\n");
    end
    ReadData1E = 64'hC1EDFFFFF7FFFFFE;
    ReadData2E = 64'h540FFFFFFC003FFF;
    ReadData3E = 64'h3E5FEFFBFFFFFFFF;
    ans = 64'hD60DFFFFF4403BFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1077119\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC9700003FFDFFFFF;
    ReadData3E = 64'hFFEFFFFFFFFFFFFF;
    ans = 64'hFFEFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1077425\n");
    end
    ReadData1E = 64'h3FBFFFFFFFFFF01F;
    ReadData2E = 64'h7FE0000000000000;
    ReadData3E = 64'hC026694A871241A8;
    ans = 64'h7FAFFFFFFFFFF01F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1077731\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3FCBFFFFBFFFFFFF;
    ReadData3E = 64'h40AFF7FFFEFFFFFF;
    ans = 64'h40AFF837FEFF7FFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1078037\n");
    end
    ReadData1E = 64'hC00FFFFFFEE00000;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1078343\n");
    end
    ReadData1E = 64'h3FB0000001EFFFFE;
    ReadData2E = 64'h80100000007FFFFE;
    ReadData3E = 64'h41E0FF7FFFFFFFFE;
    ans = 64'h41E0FF7FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1078649\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h7FE0000000000001;
    ReadData3E = 64'h41D007FFFFDFFFFF;
    ans = 64'h7FD0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1078955\n");
    end
    ReadData1E = 64'h17C00000000800FE;
    ReadData2E = 64'h3FEFFFFFFFFFFE00;
    ReadData3E = 64'h3FFFFFFFFFFFFFFF;
    ans = 64'h3FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1079261\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h7FEFFFFFFFFFFFFF;
    ReadData3E = 64'hBCAFFFFFFFFFFFFE;
    ans = 64'h7FDFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1079567\n");
    end
    ReadData1E = 64'h43CF7FFFFFFF8000;
    ReadData2E = 64'hBFBB9EDE467471F8;
    ReadData3E = 64'h4893FFFFFFFFFFFF;
    ans = 64'h4893FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1079873\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h43C2B9FEFACE0877;
    ReadData3E = 64'h434FFFFFFFFFFFFE;
    ans = 64'h43B2F9FEFACE0876;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1080179\n");
    end
    ReadData1E = 64'h43EDD3E4302B9937;
    ReadData2E = 64'h7FEFFFFFFFFFFFFE;
    ReadData3E = 64'hB7E8380CF656C23A;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1080485\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBCA7BFFFFFFFFFFF;
    ReadData3E = 64'h399FFFF7FFFFF7FF;
    ans = 64'hBC97BFFFFFFFFFDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1080791\n");
    end
    ReadData1E = 64'hBFCFFFFFFFF00000;
    ReadData2E = 64'h7FF0000000000000;
    ReadData3E = 64'h8010000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1081097\n");
    end
    ReadData1E = 64'hBF872FE583CF2298;
    ReadData2E = 64'h3CA2BBA100BAE285;
    ReadData3E = 64'hC35FF7FFFFFEFFFE;
    ans = 64'hC35FF7FFFFFEFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1081403\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h7FF0000000000001;
    ReadData3E = 64'h429B13F3D64F1831;
    ans = 64'h7FF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1081709\n");
    end
    ReadData1E = 64'hBFB00001003FFFFE;
    ReadData2E = 64'h41CFFFE000000001;
    ReadData3E = 64'hBFEFFFFFFFFFFFFE;
    ans = 64'hC18FFFE2087DFF7D;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1082015\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h7FFFFFFFFFFFFFFF;
    ReadData3E = 64'h3CA0000000000000;
    ans = 64'h7FFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1082321\n");
    end
    ReadData1E = 64'h200DA806045D23E7;
    ReadData2E = 64'hBFDFFE0010000000;
    ReadData3E = 64'hC1500001BFFFFFFF;
    ans = 64'hC1500001BFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1082627\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3FD03FFFFFFFFFFF;
    ReadData3E = 64'hC340000000000000;
    ans = 64'hC340000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1082933\n");
    end
    ReadData1E = 64'h3800000008003FFF;
    ReadData2E = 64'h7FFFFFFFFFFFFFFE;
    ReadData3E = 64'h41D00008000001FF;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1083239\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h37E000000007BFFE;
    ReadData3E = 64'h3CAFF00003FFFFFF;
    ans = 64'h3CAFF00003FFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1083545\n");
    end
    ReadData1E = 64'hBEAF80000000001E;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h000FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1083851\n");
    end
    ReadData1E = 64'h3805456CE15D5053;
    ReadData2E = 64'h5E8000020001FFFF;
    ReadData3E = 64'h7470000000FFEFFF;
    ans = 64'h7470000000FFEFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1084157\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h8000000000000000;
    ReadData3E = 64'h40300000002003FF;
    ans = 64'h40300000002003FF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1084463\n");
    end
    ReadData1E = 64'hABEBA5B21AB91F03;
    ReadData2E = 64'hC3A8E57EF02FF6AE;
    ReadData3E = 64'h3FE0000000000000;
    ans = 64'h3FE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1084769\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h8000000000000001;
    ReadData3E = 64'h8010000000000001;
    ans = 64'h8010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1085075\n");
    end
    ReadData1E = 64'h7FF1C68B26BDE7D1;
    ReadData2E = 64'hC80FBFFFFFFFF7FF;
    ReadData3E = 64'h3E2040000000FFFF;
    ans = 64'h7FF9C68B26BDE7D1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1085381\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hB7FF800000000FFE;
    ReadData3E = 64'h4010000000000001;
    ans = 64'h4010000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1085687\n");
    end
    ReadData1E = 64'hC00FC577FE8BADBD;
    ReadData2E = 64'h800FFFFFFFFFFFFF;
    ReadData3E = 64'hBF10000080007FFE;
    ans = 64'hBF10000080007FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1085993\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC7F8000000000007;
    ReadData3E = 64'h41E1CD72352B6440;
    ans = 64'hC7E8000000000005;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1086299\n");
    end
    ReadData1E = 64'h4030000000000FFF;
    ReadData2E = 64'h800FFFFFFFFFFFFE;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1086605\n");
    end
    ReadData1E = 64'hC03FFFF80FFFFFFF;
    ReadData2E = 64'hFFD122B8091B2689;
    ReadData3E = 64'h2220000000000210;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1086911\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h8010000000000000;
    ReadData3E = 64'hC1ED21FF51F1AB58;
    ans = 64'hC1ED21FF51F1AB58;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1087217\n");
    end
    ReadData1E = 64'hC34FFFFF80000007;
    ReadData2E = 64'h3FCA54333558DA5E;
    ReadData3E = 64'hBFD0000000000001;
    ans = 64'hC32A5432CC080D8F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1087523\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h8010000000000001;
    ReadData3E = 64'h000FFFFFFFFFFFFF;
    ans = 64'h0007FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1087829\n");
    end
    ReadData1E = 64'hC07DFFFF7FFFFFFF;
    ReadData2E = 64'h40DFFF8000003FFE;
    ReadData3E = 64'hC3E40B9AC28BB34B;
    ans = 64'hC3E40B9AC28BD14B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1088135\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hE869B71D58044CB4;
    ReadData3E = 64'hC00FFFFFFFFFFFFF;
    ans = 64'hE859B71D58044CB2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1088441\n");
    end
    ReadData1E = 64'hC01FFFFFFEDFFFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'h452192C043B9A0C1;
    ans = 64'h452192C043B9A0C1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1088747\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h430800FFFFFFFFFF;
    ReadData3E = 64'hBFFFFF7FFFFFFBFF;
    ans = 64'h42F800FFFFFFFFDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1089053\n");
    end
    ReadData1E = 64'h7D3FF0007FFFFFFF;
    ReadData2E = 64'h801FFFFFFFFFFFFF;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1089359\n");
    end
    ReadData1E = 64'h2A4200003FFFFFFF;
    ReadData2E = 64'hC01FFFBFFBFFFFFF;
    ReadData3E = 64'h48B000000000BFFF;
    ans = 64'h48B000000000BFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1089665\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h801FFFFFFFFFFFFE;
    ReadData3E = 64'hC03FFFDFFFFFFFDE;
    ans = 64'hC03FFFDFFFFFFFDE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1089971\n");
    end
    ReadData1E = 64'h88FFFFFFFF5FFFFF;
    ReadData2E = 64'hA454024CF36DECDB;
    ReadData3E = 64'h3CAFFFFFFFFFFFFF;
    ans = 64'h3CAFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1090277\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBCA0000000000000;
    ReadData3E = 64'h7FFFFFFFFFFFFFFE;
    ans = 64'h7FFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1090583\n");
    end
    ReadData1E = 64'h3F4FFFFFFBFFFFFF;
    ReadData2E = 64'hBFCF7FFFFFFF8000;
    ReadData3E = 64'hC01FDFFFFFFFFFFF;
    ans = 64'hC01FE03EFFFFF81E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1090889\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC80040000007FFFE;
    ReadData3E = 64'h3FFFFFFFFFFFFFFE;
    ans = 64'hC7F040000007FFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1091195\n");
    end
    ReadData1E = 64'hBFBB61DA36F5D5A6;
    ReadData2E = 64'hBCA0000000000001;
    ReadData3E = 64'h3FE0007FFFFBFFFE;
    ans = 64'h3FE0007FFFFBFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1091501\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC7EFFFFFFFE0007E;
    ReadData3E = 64'hFFF000001FFF7FFE;
    ans = 64'hFFF800001FFF7FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1091807\n");
    end
    ReadData1E = 64'hB7F893A187691B38;
    ReadData2E = 64'hBCAFFFFFFFFFFFFF;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1092113\n");
    end
    ReadData1E = 64'h37F004003FFFFFFE;
    ReadData2E = 64'h21A00008000007FF;
    ReadData3E = 64'hB7EFFFF800007FFE;
    ans = 64'hB7EFFFF800007FFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1092419\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'h47E059952FFAECBD;
    ans = 64'h47E059952FFAECBD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1092725\n");
    end
    ReadData1E = 64'hBFB000001003FFFE;
    ReadData2E = 64'h41D800000007FFFF;
    ReadData3E = 64'h801FFFFFFFFFFFFE;
    ans = 64'hC1980000180DFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1093031\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBCAFFFFFFFFFFFFE;
    ReadData3E = 64'hFFF0000000000000;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1093337\n");
    end
    ReadData1E = 64'h95BFFEFFFFFEFFFF;
    ReadData2E = 64'h41E0000000000300;
    ReadData3E = 64'h402FFFFFBFFFFFBE;
    ans = 64'h402FFFFFBFFFFFBE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1093643\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC89FE00000000003;
    ReadData3E = 64'hBFF0000000000000;
    ans = 64'hC88FE00000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1093949\n");
    end
    ReadData1E = 64'h498FFFFFFFFFBFFE;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'hC1E0100000002000;
    ans = 64'hC96FFFFFFFFFC000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1094255\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h556E8DC5F6D2D345;
    ReadData3E = 64'hC340B053BA11A120;
    ans = 64'h555E8DC5F6D2D343;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1094561\n");
    end
    ReadData1E = 64'hC3CFFFFFFFFFDFDF;
    ReadData2E = 64'hBFD0000000000001;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'h43AF7FFFFFFFDFE1;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1094867\n");
    end
    ReadData1E = 64'h404EFFF7FFFFFFFE;
    ReadData2E = 64'hFFDFFFFFEFFFFBFE;
    ReadData3E = 64'h36201FFDFFFFFFFE;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1095173\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFF;
    ReadData3E = 64'hBF81687F93BD4C1C;
    ans = 64'hBFD08B43FC9DEA5F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1095479\n");
    end
    ReadData1E = 64'h3FEBFFFFFFFFFF80;
    ReadData2E = 64'h380F7FFFFFDFFFFF;
    ReadData3E = 64'h0010000000000000;
    ans = 64'h380B8FFFFFE3FF81;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1095785\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFDFFFFFFFFFFFFE;
    ReadData3E = 64'h7FE0000000000001;
    ans = 64'h7FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1096091\n");
    end
    ReadData1E = 64'h434ACAA091937E4F;
    ReadData2E = 64'h41E7FDFFFFFFFFFF;
    ReadData3E = 64'h40E5EEF38E6512EB;
    ans = 64'h4544164BC3258583;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1096397\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h001CE5742BF6BCA4;
    ReadData3E = 64'h3FE0000000000001;
    ans = 64'h3FE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1096703\n");
    end
    ReadData1E = 64'h3E800010000001FF;
    ReadData2E = 64'hBFE0000000000000;
    ReadData3E = 64'hBFE63277397ECFAD;
    ans = 64'hBFE63277597EEFAD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1097009\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC03037FFFFFFFFFF;
    ReadData3E = 64'hC27087FFFFFFFFFE;
    ans = 64'hC2708800000081BE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1097315\n");
    end
    ReadData1E = 64'h401EFEFFFFFFFFFF;
    ReadData2E = 64'hBFE0000000000001;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h4010807FFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1097621\n");
    end
    ReadData1E = 64'hB7E5E9AED5C5CD08;
    ReadData2E = 64'h40100010000000FF;
    ReadData3E = 64'h00077CD5AB42C193;
    ans = 64'hB805E9C4BF74A42B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1097927\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'h801FFFE7FFFFFFFE;
    ans = 64'hBFDFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1098233\n");
    end
    ReadData1E = 64'hBFCC7D5BA9493FB9;
    ReadData2E = 64'h33679E2657415F04;
    ReadData3E = 64'h8000000000000001;
    ans = 64'hB34506E6F8CC5CED;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1098539\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFEFFFFFFFFFFFFF;
    ReadData3E = 64'hC34FFFFFFFFFFFFF;
    ans = 64'hC34FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1098845\n");
    end
    ReadData1E = 64'h3800000000000010;
    ReadData2E = 64'hC1FFFE00001FFFFE;
    ReadData3E = 64'hD6EFFFFFFFFFFFEB;
    ans = 64'hD6EFFFFFFFFFFFEB;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1099151\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC66302F95CD69379;
    ReadData3E = 64'hBFDFFFFFFFFFFFFF;
    ans = 64'hC65302F95CD69378;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1099457\n");
    end
    ReadData1E = 64'hB7FFFFFFFFFFFC3E;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'hE1507FFF7FFFFFFF;
    ans = 64'hE1507FFF7FFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1099763\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hA2FE0000000000FF;
    ReadData3E = 64'h3801FFFFFFFFFFFF;
    ans = 64'h3801FFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1100069\n");
    end
    ReadData1E = 64'hC77BC12E7CB1723C;
    ReadData2E = 64'hBFF0000000000000;
    ReadData3E = 64'hC010000000000000;
    ans = 64'h477BC12E7CB1723C;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1100375\n");
    end
    ReadData1E = 64'hBCE000000001FC00;
    ReadData2E = 64'h3FF80CABB0668B82;
    ReadData3E = 64'hB9FFFFFFFCFFFFFF;
    ans = 64'hBCE80CABB0698754;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1100681\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFF0000000000001;
    ReadData3E = 64'h3FBFFFFFFFFFFF83;
    ans = 64'hBFD800000000001F;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1100987\n");
    end
    ReadData1E = 64'hBD8184351B39575C;
    ReadData2E = 64'hA0AFFFFBFFFFFFF8;
    ReadData3E = 64'hFFFFFFFFFFFFFFFF;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1101293\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFFFFFFFFFFFFFFF;
    ReadData3E = 64'h401FFFFFFFFFFFFE;
    ans = 64'h401BFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1101599\n");
    end
    ReadData1E = 64'h3F9FFFDFFEFFFFFE;
    ReadData2E = 64'hB7E00007EFFFFFFF;
    ReadData3E = 64'hC1EFFFFFFFDFF000;
    ans = 64'hC1EFFFFFFFDFF000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1101905\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h301FFDFFFDFFFFFE;
    ReadData3E = 64'h3CAFFFFFFFFFFFFE;
    ans = 64'h3CAFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1102211\n");
    end
    ReadData1E = 64'hC06FD8A0BAEFFBA0;
    ReadData2E = 64'hBFFFFFFFFFFFFFFE;
    ReadData3E = 64'h4000020000000200;
    ans = 64'h407FF8A4BAEFFBA2;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1102517\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h32F0007FFFFEFFFF;
    ReadData3E = 64'hBD81FFFFFF7FFFFF;
    ans = 64'hBD81FFFFFF7FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1102823\n");
    end
    ReadData1E = 64'h691FFFDFF0000000;
    ReadData2E = 64'hC000000000000000;
    ReadData3E = 64'h4000000000000001;
    ans = 64'hE92FFFDFF0000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1103129\n");
    end
    ReadData1E = 64'h7FF000004FFFFFFE;
    ReadData2E = 64'hC039E997F6D37059;
    ReadData3E = 64'h331C0003FFFFFFFE;
    ans = 64'h7FF800004FFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1103435\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'h3E9000100FFFFFFF;
    ans = 64'hBFEFFFFF7FFF7F80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1103741\n");
    end
    ReadData1E = 64'hB7E9521D61269217;
    ReadData2E = 64'h3FCFFFFFBFFFBFFF;
    ReadData3E = 64'h7FEFFFFFFFFFFFFE;
    ans = 64'h7FEFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1104047\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC000000000000001;
    ReadData3E = 64'hC010000000000000;
    ans = 64'hC014000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1104353\n");
    end
    ReadData1E = 64'hCD1F485CE2E7C107;
    ReadData2E = 64'hC00F000003FFFFFF;
    ReadData3E = 64'hBF7407FFFFFFFFFE;
    ans = 64'h4D3E4E19FFB98E9A;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1104659\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBF04000000007FFF;
    ReadData3E = 64'hBCA0000000000000;
    ans = 64'hBEF400000000FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1104965\n");
    end
    ReadData1E = 64'h3FC000003FFFFBFF;
    ReadData2E = 64'hC00FFFFFFFFFFFFF;
    ReadData3E = 64'h001000001FFFFFFE;
    ans = 64'hBFE000003FFFFBFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1105271\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hCC0F7F226DFC3A30;
    ReadData3E = 64'h3810800000000080;
    ans = 64'hCBFF7F226DFC3A2E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1105577\n");
    end
    ReadData1E = 64'h4011FFFFFF800000;
    ReadData2E = 64'hC00FFFFFFFFFFFFE;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hC033FFFFFF7FFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1105883\n");
    end
    ReadData1E = 64'h0120000000020080;
    ReadData2E = 64'hC02FFBFFBFFFFFFF;
    ReadData3E = 64'hBFE7FFFFFFFEFFFF;
    ans = 64'hBFE7FFFFFFFEFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1106189\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC010000000000000;
    ReadData3E = 64'h3FDFFC0000007FFF;
    ans = 64'hBFF800FFFFFFDFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1106495\n");
    end
    ReadData1E = 64'hC7E000000401FFFF;
    ReadData2E = 64'h3FD000000200001E;
    ReadData3E = 64'hFFE0000000000000;
    ans = 64'hFFE0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1106801\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC010000000000001;
    ReadData3E = 64'h4000000000000001;
    ans = 64'h3CC0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1107107\n");
    end
    ReadData1E = 64'hC07400000000000F;
    ReadData2E = 64'hC22F7EFFFFFFFFFF;
    ReadData3E = 64'h402E0FDFBF9D095F;
    ans = 64'h42B3AF6000000F16;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1107413\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h41C0020000040000;
    ReadData3E = 64'h0010000000000001;
    ans = 64'h41B002000003FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1107719\n");
    end
    ReadData1E = 64'hAD8FFFFFFFF80007;
    ReadData2E = 64'hC01FFFFFFFFFFFFF;
    ReadData3E = 64'h42960AE5D04FDC6B;
    ans = 64'h42960AE5D04FDC6B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1108025\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC0251BE94A80FD14;
    ReadData3E = 64'hBFEFFFFFFFDFFFC0;
    ans = 64'hC0191BE94A7CFD0B;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1108331\n");
    end
    ReadData1E = 64'hBE700000081FFFFE;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'h3FF0000080000040;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1108637\n");
    end
    ReadData1E = 64'h001FFE0000001FFF;
    ReadData2E = 64'h3F40B59FC2D8D6D9;
    ReadData3E = 64'hD4800001FFFFFFFE;
    ans = 64'hD4800001FFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1108943\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC01FFFFFFFFFFFFE;
    ReadData3E = 64'h43EFFC000000FFFF;
    ans = 64'h43EFFC000000FFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1109249\n");
    end
    ReadData1E = 64'h3F50004000001FFF;
    ReadData2E = 64'h4040000003FBFFFF;
    ReadData3E = 64'h4340000000000001;
    ans = 64'h4340000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1109555\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC340000000000000;
    ReadData3E = 64'hBFFFFFFFFFFFFFFF;
    ans = 64'hC330000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1109861\n");
    end
    ReadData1E = 64'hC9A000000FFFFFC0;
    ReadData2E = 64'h401FC0000000001F;
    ReadData3E = 64'h41E62D1014622F6F;
    ans = 64'hC9CFC0001FBFFFA0;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1110167\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC08000000004003F;
    ReadData3E = 64'h800FFFFFFFFFFFFF;
    ans = 64'hC07000000004003E;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1110473\n");
    end
    ReadData1E = 64'hC7E00020000007FF;
    ReadData2E = 64'hC340000000000001;
    ReadData3E = 64'h3D8081FFFFFFFFFF;
    ans = 64'h4B30002000000800;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1110779\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3096C2873C3B7AB7;
    ReadData3E = 64'h3CCFFFFDFFFFBFFF;
    ans = 64'h3CCFFFFDFFFFBFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1111085\n");
    end
    ReadData1E = 64'h403FFFFEFFFFFFFB;
    ReadData2E = 64'hC34FFFFFFFFFFFFF;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hC39FFFFEFFFFFFFA;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1111391\n");
    end
    ReadData1E = 64'hC1F2000FFFFFFFFF;
    ReadData2E = 64'h4F3003FFFFFEFFFE;
    ReadData3E = 64'h4BDFFFFFFF8001FE;
    ans = 64'hD132049003FEDFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1111697\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hC34FFFFFFFFFFFFE;
    ReadData3E = 64'hB8101000003FFFFF;
    ans = 64'hC33FFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1112003\n");
    end
    ReadData1E = 64'h3FF719E56C2EBF0F;
    ReadData2E = 64'hB802BAD5445737FA;
    ReadData3E = 64'hC01FFFFFFFFFFFFF;
    ans = 64'hC01FFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1112309\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hFFE0000000000000;
    ReadData3E = 64'h3FEFFFFFFFFFFFFE;
    ans = 64'hFFCFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1112615\n");
    end
    ReadData1E = 64'h37FF9FFFFFFFFFFF;
    ReadData2E = 64'hC85FFFFFFFFFF07E;
    ReadData3E = 64'h40010000000FFFFF;
    ans = 64'hC06F5BFFFFFFB0AC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1112921\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hBFFFFFF8000003FE;
    ReadData3E = 64'hFFFFFFFFFFFFFFFE;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1113227\n");
    end
    ReadData1E = 64'hBFCEFFFFFFFFEFFF;
    ReadData2E = 64'hFFE0000000000001;
    ReadData3E = 64'hC030000000001200;
    ans = 64'h7FBEFFFFFFFFF001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1113533\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hFFDFFFDFFFFFFFFE;
    ReadData3E = 64'hBFD0100007FFFFFF;
    ans = 64'hFFCFFFDFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1113839\n");
    end
    ReadData1E = 64'h401000000007FBFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1114145\n");
    end
    ReadData1E = 64'hC1CAE55EE9B7731D;
    ReadData2E = 64'hC1DBFFFFFFFFFFFE;
    ReadData3E = 64'hC1E05FFFFFFFFFFE;
    ans = 64'h43B788B30BFD84B8;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1114451\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFF;
    ReadData3E = 64'h34FFFFFEFFFF0000;
    ans = 64'hFFDFFFFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1114757\n");
    end
    ReadData1E = 64'h3FBFFBFFFFFFFFFE;
    ReadData2E = 64'hFFEFFFFBFFBFFFFF;
    ReadData3E = 64'h400FFFFFFFFFFFFE;
    ans = 64'hFFBFFBFC004007FD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1115063\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hFFEFFFFFFFFFFFFE;
    ReadData3E = 64'hBFE0000000000000;
    ans = 64'hFFDFFFFFFFFFFFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1115369\n");
    end
    ReadData1E = 64'h002007FFFFFBFFFF;
    ReadData2E = 64'hBFC926A50C7B04BB;
    ReadData3E = 64'h3DCFFFFFFFFFFFFE;
    ans = 64'h3DCFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1115675\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hB707FFFFFFEFFFFF;
    ReadData3E = 64'h7FF0000000000000;
    ans = 64'h7FF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1115981\n");
    end
    ReadData1E = 64'h381000007F000000;
    ReadData2E = 64'hFFF0000000000000;
    ReadData3E = 64'h7FDFFFFF00000020;
    ans = 64'hFFF0000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1116287\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3CAFFDFFFFFFFF00;
    ReadData3E = 64'h39CFDFFFFDFFFFFE;
    ans = 64'h3C9FFDFFFFFFFFFD;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1116593\n");
    end
    ReadData1E = 64'h3FFFFFFFFFFFF7EE;
    ReadData2E = 64'hFFF0000000000001;
    ReadData3E = 64'hBCAFFFFFFFFFFFFF;
    ans = 64'hFFF8000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1116899\n");
    end
    ReadData1E = 64'h41503FFFFFFFEFFE;
    ReadData2E = 64'hB67FFFF7FDFFFFFF;
    ReadData3E = 64'hBFD7FFFFFFFEFFFF;
    ans = 64'hBFD7FFFFFFFEFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1117205\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hFFFFFFFFFFFFFFFF;
    ReadData3E = 64'h366FFE0000000007;
    ans = 64'hFFFFFFFFFFFFFFFF;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1117511\n");
    end
    ReadData1E = 64'hB7F0001FFFDFFFFF;
    ReadData2E = 64'hBF5FFFFFF07FFFFF;
    ReadData3E = 64'hC000000000000000;
    ans = 64'hC000000000000000;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1117817\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'hFFFFFFFFFFFFFFFE;
    ReadData3E = 64'h3FD0000000000001;
    ans = 64'hFFFFFFFFFFFFFFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1118123\n");
    end
    ReadData1E = 64'h47E0000000003EFE;
    ReadData2E = 64'hB72000013FFFFFFF;
    ReadData3E = 64'h403FFFFEFFFFF7FE;
    ans = 64'h403FFFFAFFFFA7FE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1118429\n");
    end
    ReadData1E = 64'h3FDFFFFFFFFFFFFE;
    ReadData2E = 64'h3F0FFFFFFFF7FFC0;
    ReadData3E = 64'hFFE0000000000001;
    ans = 64'hFFE0000000000001;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1118735\n");
    end
    ReadData1E = 64'h9280000001FFFFFE;
    ReadData2E = 64'h0000000000000000;
    ReadData3E = 64'hBF5008000001FFFE;
    ans = 64'hBF5008000001FFFE;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1119041\n");
    end
    ReadData1E = 64'h3FE0000000000000;
    ReadData2E = 64'h41E0000000003F80;
    ReadData3E = 64'h3FFFBFFFFFFFF000;
    ans = 64'h41D00000007F3F80;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1119347\n");
    end
    ReadData1E = 64'hC3F000200003FFFE;
    ReadData2E = 64'h0000000000000001;
    ReadData3E = 64'h001FFFFFFFFFFFFE;
    ans = 64'h80CFFC400007FFFC;
    FrmE = 3'b000;
#10
    // IEEE 754-2008 section 6.3 states: "When ether an input or result is NaN, this
    //                                     standard does not interpret the sign of a NaN."
	wnan = &FmaResultM[62:52] && |FmaResultM[51:0]; 
	xnan = &ReadData1E[62:52] && |ReadData1E[51:0]; 
	ynan = &ReadData2E[62:52] && |ReadData2E[51:0]; 
	znan = &ReadData3E[62:52] && |ReadData3E[51:0]; 
	ansnan = &ans[62:52] && |ans[51:0]; 
	xnorm = ~(|ReadData1E[62:52]) && |ReadData1E[51:0] ? {ReadData1E[50:0], 1'b0} : ReadData1E; 
	ynorm = ~(|ReadData2E[62:52]) && |ReadData2E[51:0] ? {ReadData2E[50:0], 1'b0} : ReadData2E;
	if((!wnan && (FmaResultM != ans)) || (wnan && ansnan && ~(((xnan && (FmaResultM[62:0] == {ReadData1E[62:52],1'b1,ReadData1E[50:0]})) || (ynan && (FmaResultM[62:0] == {ReadData2E[62:52],1'b1,ReadData2E[50:0]}))  || (znan && (FmaResultM[62:0] == {ReadData3E[62:52],1'b1,ReadData3E[50:0]})) || (FmaResultM[62:0] == ans[62:0])) ))) begin
		$fwrite(fp, "%h %h %h %h %h  Wrong ",ReadData1E,ReadData2E, ReadData3E, FmaResultM, ans);
		$fwrite(fp, "FmaResultM=%d ",$signed(aligncntE));
		if(FmaResultM == 64'h8000000000000000) $fwrite(fp, "FmaResultM=-zero ");
		if(~(|ReadData1E[62:52]) && |ReadData1E[51:0]) $fwrite(fp, "xdenorm ");
		if(~(|ReadData2E[62:52]) && |ReadData2E[51:0]) $fwrite(fp, "ydenorm ");
		if(~(|ReadData3E[62:52]) && |ReadData3E[51:0]) $fwrite(fp, "zdenorm ");
		if(FmaFlagsM[4] != 0) $fwrite(fp, "invld ");
		if(FmaFlagsM[2] != 0) $fwrite(fp, "ovrflw ");
		if(FmaFlagsM[1] != 0) $fwrite(fp, "unflw ");
		if(FmaResultM == 64'hFFF0000000000000) $fwrite(fp, "FmaResultM=-inf ");
		if(FmaResultM == 64'h7FF0000000000000) $fwrite(fp, "FmaResultM=+inf ");
		if(FmaResultM >  64'h7FF0000000000000 && FmaResultM <  64'h7FF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >  64'hFFF8000000000000 && FmaResultM <  64'hFFF8000000000000 ) $fwrite(fp, "FmaResultM=sigNaN ");
		if(FmaResultM >= 64'h7FF8000000000000 && FmaResultM <= 64'h7FFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(FmaResultM >= 64'hFFF8000000000000 && FmaResultM <= 64'hFFFfffffffffffff ) $fwrite(fp, "FmaResultM=qutNaN ");
		if(ans == 64'hFFF0000000000000) $fwrite(fp, "ans=-inf ");
		if(ans == 64'h7FF0000000000000) $fwrite(fp, "ans=+inf ");
		if(ans >  64'h7FF0000000000000 && ans <  64'h7FF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >  64'hFFF8000000000000 && ans <  64'hFFF8000000000000 ) $fwrite(fp, "ans=sigNaN ");
		if(ans >= 64'h7FF8000000000000 && ans <= 64'h7FFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
		if(ans >= 64'hFFF8000000000000 && ans <= 64'hFFFfffffffffffff ) $fwrite(fp, "ans=qutNaN ");
    	$fwrite(fp,"1119653\n");
    end
	$stop;
	end
endmodule